/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.controller;

import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.RGB;

import com.haundrix.animation.FadeCallback;
import com.haundrix.animation.MoveCallback;
import com.haundrix.animation.Mover;
import com.haundrix.games.chess.view.ChessPieceType;
import com.haundrix.games.view.BoardAdapter;
import com.haundrix.games.view.BoardPiece;
import com.haundrix.games.view.BoardSquare;
import com.haundrix.games.view.BoardSquareEvent;
import com.haundrix.games.view.IDraggingPieceHandler;
import com.haundrix.games.view.IPieceConfigurator;
import com.haundrix.games.view.PieceDragEvent;
import com.haundrix.games.view.PieceGrabEvent;
import com.haundrix.games.view.RectangularBoardControl;
import com.haundrix.games.view.SquareClickEvent;
import com.haundrix.games.view.SquareClickEvent.MouseButton;
import com.haundrix.graphics.GraphicsUtils;
import com.haundrix.graphics.Highlight;
import com.haundrix.utils.IDisposeListener;
import com.haundrix.utils.IDisposable;

public class TestBoardController extends BoardAdapter<ChessPieceType> {
    
    private Highlight pieceHighlight = new Highlight(new RGB(217.0F, 1.0F, 0.5F), 0.75F, 0.15F);
    private float draggingPieceAlpha = 0.9F;
    private float draggingPieceShadowIntensity = 0.15F;
    private Highlight anchorHighlight = new Highlight(new RGB(217.0F, 1.0F, 1.0F), 0.75F, 0.15F);

    private Highlight checkHighlight = new Highlight(new RGB(255, 0, 0), 0.6F, 0.30F);
    private int checkHighlightDuration = 300;
    private int nCheckHighlights = 6;

    private int pieceFadeDuration = 500;
    private int promotionDuration = 1500;
    private int minMovementDuration = 300;
    private int maxMovementDuration = 1500;
    
    private RectangularBoardControl<ChessPieceType> board;

    private BoardSquare currentSquare;

    private BoardSquare anchoredSquare = null;
    private BoardPiece<ChessPieceType> anchoredPiece = null;
    private IDisposeListener anchoredPieceDisposeListener = new IDisposeListener() {
        @Override
        public void disposed(IDisposable disposable) {
            anchoredPiece = null;
        }
    };
    
    private ChessPieceType previousDeletedPiece = null;
    
    private int nDrops = 0;

    public TestBoardController(RectangularBoardControl<ChessPieceType> board) {
        this.board = board;
    }

    public Highlight getPieceHighlight() {
        return pieceHighlight;
    }

    public void setPieceHighlight(Highlight pieceHighlight) {
        this.pieceHighlight = pieceHighlight;
    }

    public float getDraggingPieceAlpha() {
        return draggingPieceAlpha;
    }

    public void setDraggingPieceAlpha(float draggingPieceAlpha) {
        if((draggingPieceAlpha < 0.0F) || (draggingPieceAlpha > 1.0F)) {
            throw new IllegalArgumentException("Dragging piece alpha must be a number within 0 and 1.");
        }
        
        this.draggingPieceAlpha = draggingPieceAlpha;
    }

    public float getDraggingPieceShadowIntensity() {
        return draggingPieceShadowIntensity;
    }

    public void setDraggingPieceShadowIntensity(float draggingPieceShadowIntensity) {
        if((draggingPieceShadowIntensity < 0.0F) || (draggingPieceShadowIntensity > 1.0F)) {
            throw new IllegalArgumentException("Dragging piece shadow intensity must be a number within 0 and 1.");
        }
        
        this.draggingPieceShadowIntensity = draggingPieceShadowIntensity;
    }

    public Highlight getAnchorHighlight() {
        return anchorHighlight;
    }

    public void setAnchorHighlight(Highlight anchorHighlight) {
        this.anchorHighlight = anchorHighlight;
    }

    public Highlight getCheckHighlight() {
        return checkHighlight;
    }

    public void setCheckHighlight(Highlight checkHighlight) {
        this.checkHighlight = checkHighlight;
    }

    public int getCheckHighlightDuration() {
        return checkHighlightDuration;
    }

    public void setCheckHighlightDuration(int checkHighlightDuration) {
        this.checkHighlightDuration = checkHighlightDuration;
    }

    public int getNCheckHighlights() {
        return nCheckHighlights;
    }

    public void setNCheckHighlights(int checkHighlights) {
        nCheckHighlights = checkHighlights;
    }

    public int getPieceFadeDuration() {
        return pieceFadeDuration;
    }

    public void setPieceFadeDuration(int pieceFadeDuration) {
        if(pieceFadeDuration < 0) {
            throw new IllegalArgumentException("Piece fade duration can't be negative.");
        }
        
        this.pieceFadeDuration = pieceFadeDuration;
    }

    public int getPromotionDuration() {
        return promotionDuration;
    }

    public void setPromotionDuration(int promotionDuration) {
        if(promotionDuration < 0) {
            throw new IllegalArgumentException("Promotion duration can't be negative.");
        }
        
        this.promotionDuration = promotionDuration;
    }

    public int getMinMovementDuration() {
        return minMovementDuration;
    }

    public void setMinMovementDuration(int minMovementDuration) {
        if(minMovementDuration < 0) {
            throw new IllegalArgumentException("Minimum movement duration can't be negative.");
        }
        if(minMovementDuration > maxMovementDuration) {
            throw new IllegalArgumentException("Minimum movement duration can't exceed maximum movement duration.");
        }
        
        this.minMovementDuration = minMovementDuration;
    }

    public int getMaxMovementDuration() {
        return maxMovementDuration;
    }

    public void setMaxMovementDuration(int maxMovementDuration) {
        if(maxMovementDuration < 0) {
            throw new IllegalArgumentException("Maximum movement duration can't be negative.");
        }
        if(maxMovementDuration < minMovementDuration) {
            throw new IllegalArgumentException("Maximum movement duration can't be less than the minimum movement duration.");
        }
        
        this.maxMovementDuration = maxMovementDuration;
    }
    
    public void setMovementDurationBounds(int minMovementDuration, int maxMovementDuration) {
        if(minMovementDuration < 0) {
            throw new IllegalArgumentException("Minimum movement duration can't be negative.");
        }
        if(maxMovementDuration < 0) {
            throw new IllegalArgumentException("Maximum movement duration can't be negative.");
        }
        if(minMovementDuration > maxMovementDuration) {
            throw new IllegalArgumentException("Maximum movement duration must be higher than the minimum movement duration.");
        }
        
        this.minMovementDuration = minMovementDuration;
        this.maxMovementDuration = maxMovementDuration;
    }

    @Override
    public void squareLeft(BoardSquareEvent<ChessPieceType> event) {
        currentSquare = null;
        if(anchoredPiece == null) {
            BoardPiece<ChessPieceType> piece = event.getPiece();
            if(piece != null) {
                piece.setHighlight(null);
            }
        } else {
            board.setSquareHighlight(event.getSquare(), null);
        }
    }

    @Override
    public void squareEntered(BoardSquareEvent<ChessPieceType> event) {
        currentSquare = event.getSquare();
        BoardPiece<ChessPieceType> piece = event.getPiece();
        if(anchoredPiece == null) {
            if(piece != null) {
                piece.setHighlight(pieceHighlight);
            }
        } else {
            if((piece == null) || (piece != anchoredPiece)) {
                board.setSquareHighlight(event.getSquare(), pieceHighlight);
            }
        }
    }

    @Override
    public void squareClicked(SquareClickEvent<ChessPieceType> event) {
        BoardPiece<ChessPieceType> targetPiece = event.getPiece();
        final BoardSquare targetSquare = event.getSquare();
        if(event.getClickedButton() == MouseButton.LEFT_BUTTON) {
            if(anchoredPiece == null) {
                if(targetPiece != null) {
                    targetPiece.setHighlight(anchorHighlight);
                    anchoredSquare = targetSquare;
                    anchoredPiece = targetPiece;
                    anchoredPiece.addDisposeListener(anchoredPieceDisposeListener);
                }
            } else {
                board.setSquareHighlight(targetSquare, null);
                anchoredPiece.setHighlight(null);
                anchoredPiece.removeDisposeListener(anchoredPieceDisposeListener);
                if(!targetSquare.equals(anchoredSquare)) {
                    Point origin = board.getSquareLocation(anchoredSquare);
                    Point destination = board.getSquareLocation(targetSquare);
                    int movementDuration = getMovementDuration(origin, destination);
                    // stores the anchored piece in a local variable in order
                    // to prevent NPEs later in the move callbacks below
                    final BoardPiece<ChessPieceType> theAnchoredPiece = anchoredPiece;
                    if(targetPiece == null) {
                        board.movePiece(anchoredSquare, targetSquare, movementDuration, null,
                                new MoveCallback() {
                                    @Override
                                    public void ended(Mover mover) {
                                        handleMoveEnded(theAnchoredPiece, targetSquare);
                                    }
                                });
                    } else {
                        board.movePiece(anchoredSquare, targetSquare, movementDuration, null,
                                new MoveCallback() {
                            
                                    private boolean fadeOutTriggered = false;
                                    
                                    @Override
                                    public void moved(Mover mover) {
                                        if(!fadeOutTriggered) {
                                            if(mover.getRemainingTime() <= pieceFadeDuration) {
                                                board.fadeOutPiece(targetSquare, pieceFadeDuration);
                                                fadeOutTriggered = true;
                                            }
                                        }
                                    }
                                    
                                    @Override
                                    public void ended(Mover mover) {
                                        handleMoveEnded(theAnchoredPiece, targetSquare);
                                    }

                                });
                    }
                }
                anchoredSquare = null;
                anchoredPiece = null;
            }
        } else if(event.getClickedButton() == MouseButton.RIGHT_BUTTON) {
            if(targetPiece == null) {
                if(previousDeletedPiece != null) {
                    board.fadeInPiece(previousDeletedPiece, targetSquare, pieceFadeDuration, new IPieceConfigurator<ChessPieceType>() {
                        @Override
                        public void configure(BoardPiece<ChessPieceType> piece) {
                            piece.setHighlight(pieceHighlight);
                        }
                    });
                }
            } else {
                board.fadeOutPiece(targetSquare, pieceFadeDuration);
                previousDeletedPiece = targetPiece.getPiece();
            }
        }
    }

    @Override
    public void pieceGrabbed(PieceGrabEvent<ChessPieceType> event) {
        if(anchoredPiece == null) {
            BoardPiece<ChessPieceType> grabbedPiece = event.getGrabbedPiece();
            grabbedPiece.setAlpha(draggingPieceAlpha);
            grabbedPiece.setShadowIntensity(draggingPieceShadowIntensity);
        } else {
            event.allow = false;
        }
    }
    
    @Override
    public void pieceDropped(final PieceDragEvent<ChessPieceType> event, final IDraggingPieceHandler draggingPieceHandler) {
        final BoardPiece<ChessPieceType> draggedPiece = event.getDraggedPiece();
        final BoardSquare targetSquare = event.getTargetSquare();
        if(targetSquare == null) {
            if(nDrops++ % 2 == 0) {
                Point origin = draggedPiece.getLocation();
                Point destination = board.getSquareLocation(event.getDraggingStartSquare());
                int movementDuration = getMovementDuration(origin, destination);
                draggingPieceHandler.moveBackToOriginalSquare(movementDuration, null, new MoveCallback() {
                    @Override
                    public void ended(Mover mover) {
                        TestBoardController.clearPiece(draggedPiece);
                        handleHighlightPiece(draggedPiece, event.getDraggingStartSquare());
                    }
                });
            } else {
                draggingPieceHandler.fadeOut(pieceFadeDuration);
            }
        } else {
            if(event.getTargetPiece() == null) {
                draggingPieceHandler.moveToTargetSquare(minMovementDuration, null, new MoveCallback() {
                    @Override
                    public void ended(Mover mover) {
                        handleMoveEnded(draggedPiece, targetSquare);
                    }
                });
            } else {
                draggingPieceHandler.moveToTargetSquare(minMovementDuration, null, new MoveCallback() {
                    
                    @Override
                    public void started(final Mover mover) {
                        mover.pause();
                        board.fadeOutPiece(targetSquare, pieceFadeDuration, new FadeCallback() {
                            @Override
                            public void ended() {
                                mover.resume();
                            }
                        });
                    }

                    @Override
                    public void ended(Mover mover) {
                        handleMoveEnded(draggedPiece, targetSquare);
                    }

                });
            }
        }
    }

    private int getMovementDuration(Point origin, Point destination) {
        double distance = GraphicsUtils.getDistance(origin, destination);
        int squareSize = board.getSquareSize();
        if(distance <= squareSize) {
            return minMovementDuration;
        }
        
        Point topLeftSquareLocation = board.getSquareLocation(new BoardSquare(0, 0));
        Point bottomRightSquareLocation = board.getSquareLocation(new BoardSquare(board.getRowCount() - 1, board.getColumnCount() - 1));
        double maxDistance = GraphicsUtils.getDistance(topLeftSquareLocation, bottomRightSquareLocation);
        
        if(distance >= maxDistance) {
            return maxMovementDuration;
        }
        
        return (int) (minMovementDuration + (maxMovementDuration - minMovementDuration) /
                (maxDistance - squareSize) * (distance - squareSize));
    }

    private void handleMoveEnded(BoardPiece<ChessPieceType> movedPiece, BoardSquare targetSquare) {
        TestBoardController.clearPiece(movedPiece);
        handleHighlightPiece(movedPiece, targetSquare);
        handleCheck(movedPiece);
        handlePromotion(movedPiece, targetSquare);
    }

    private static void clearPiece(final BoardPiece<ChessPieceType> piece) {
        piece.setHighlight(null);
        piece.setAlpha(1.0F);
        piece.setShadowIntensity(0.0F);
    }
    
    private void handleHighlightPiece(BoardPiece<ChessPieceType> movedPiece, BoardSquare targetSquare) {
        if((currentSquare != null) && currentSquare.equals(targetSquare)) {
            movedPiece.setHighlight(pieceHighlight);
        }
    }

    private void handleCheck(BoardPiece<ChessPieceType> movedPiece) {
        ChessPieceType kingToHighlight = null;
        if(movedPiece.getPiece() == ChessPieceType.BLACK_QUEEN) {
            kingToHighlight = ChessPieceType.WHITE_KING;
        } else if(movedPiece.getPiece() == ChessPieceType.WHITE_QUEEN) {
            kingToHighlight = ChessPieceType.BLACK_KING;
        }
        
        if(kingToHighlight != null) {
            for(BoardSquare boardSquare : board) {
                if(board.getPiece(boardSquare) == kingToHighlight) {
                    board.getBoardPiece(boardSquare).pulse(checkHighlight, checkHighlightDuration, nCheckHighlights);
                    break;
                }
            }
        }
    }
    
    private void handlePromotion(BoardPiece<ChessPieceType> piece, BoardSquare targetSquare) {
        boolean promoteBlackPawn = (piece.getPiece() == ChessPieceType.BLACK_PAWN) &&
            (targetSquare.getRow() == board.getRowCount() - 1);
        if(promoteBlackPawn) {
            piece.crossfadeTo(ChessPieceType.BLACK_QUEEN, promotionDuration);
        } else {
            boolean promoteWhitePawn = (piece.getPiece() == ChessPieceType.WHITE_PAWN) &&
                (targetSquare.getRow() == 0);
            if(promoteWhitePawn) {
                piece.crossfadeTo(ChessPieceType.WHITE_QUEEN, promotionDuration);
            }
        }
    }

}
