/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.png;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;

import org.joda.time.LocalDate;
import org.testng.Assert;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import com.haundrix.games.chess.model.ChessGame;
import com.haundrix.games.chess.model.ChessGameResult;
import com.haundrix.games.chess.model.PromotionMove;
import com.haundrix.games.chess.model.Queen;
import com.haundrix.games.chess.pgn.InvalidPgnFileException;
import com.haundrix.games.chess.pgn.PgnExporter;
import com.haundrix.games.chess.pgn.PgnImporter;
import com.haundrix.games.model.GameInfo;
import com.haundrix.games.model.GameOutcome;
import com.haundrix.games.model.Move;
import com.haundrix.games.model.Square;
import com.haundrix.tests.HaundrixAssert;
import com.haundrix.tests.TestUtils;

public class PgnGenerationTest {

    @Test
    public void testChessGameExportImport() throws IOException {
        Move[] moves = {
                new Move(new Square(6, 4), new Square(4, 4)),  // e4
                new Move(new Square(1, 3), new Square(3, 3)),  //            d5
                new Move(new Square(4, 4), new Square(3, 4)),  // e5
                new Move(new Square(1, 5), new Square(3, 5)),  //            f5
                new Move(new Square(3, 4), new Square(2, 5)),  // exf6 e.p.
                new Move(new Square(0, 1), new Square(2, 2)),  //            Nc6
                new Move(new Square(7, 6), new Square(5, 5)),  // Nf3
                new Move(new Square(0, 2), new Square(4, 6)),  //            Bg4
                new Move(new Square(7, 5), new Square(4, 2)),  // Bc4
                new Move(new Square(0, 3), new Square(2, 3)),  //            Qd6
                new Move(new Square(7, 4), new Square(7, 6)),  // 0-0
                new Move(new Square(0, 4), new Square(0, 2)),  //            0-0-0
                new Move(new Square(2, 5), new Square(1, 6)),  // fxg7
                new Move(new Square(3, 3), new Square(4, 2)),  //            dxc4
                new Move(new Square(6, 1), new Square(4, 1)),  // b4
                new Move(new Square(0, 6), new Square(2, 5)),  //            Nf6
                new PromotionMove(new Square(1, 6), new Square(0, 5), Queen.class),  // gxf8=Q
                new Move(new Square(0, 7), new Square(0, 5)),  //            Rhxf8
                new Move(new Square(6, 3), new Square(5, 3)),  // d3
                new Move(new Square(2, 3), new Square(5, 3)),  //            Qxd3
                new Move(new Square(7, 1), new Square(6, 3)),  // Nbd2
                new Move(new Square(4, 6), new Square(5, 5)),  //            Bxf3
                new Move(new Square(7, 3), new Square(5, 5)),  // Qxf3
                new Move(new Square(5, 3), new Square(7, 5)),  //            Qxf1+
                new Move(new Square(6, 3), new Square(7, 5)),  // Nxf1
                new Move(new Square(2, 2), new Square(4, 1)),  //            Nxb4
                new Move(new Square(5, 5), new Square(3, 5)),  // Qf5+
                new Move(new Square(0, 2), new Square(0, 1)),  //            Kb8
                new Move(new Square(7, 5), new Square(5, 6)),  // Ng3
                new Move(new Square(0, 5), new Square(0, 4)),  //            Rfe8
                new Move(new Square(5, 6), new Square(4, 4)),  // Ne4
                new Move(new Square(0, 3), new Square(7, 3)),  //            Rd1++
            };

        GameInfo gameInfo = new GameInfo();
        gameInfo.setWhitePlayer("Gary Kasparov");
        gameInfo.setWhiteElo(2900);
        gameInfo.setBlackPlayer("Anatoly Karpov");
        gameInfo.setBlackElo(2899);
        gameInfo.setEvent("Chess World Championship Match");
        gameInfo.setSite("London, United Kingdom");
        gameInfo.setDate(new LocalDate(1989, 12, 24));
        gameInfo.setRound(7);
        
        ChessGame game = new ChessGame();
        game.setInfo(gameInfo);
        for(Move move : moves) {
            game.move(move);
        }

        File tempFile = new File(TestUtils.getTempDir(PgnGenerationTest.class), "game.pgn");
        Writer writer = new FileWriter(tempFile);
        new PgnExporter(writer).exportGame(game);
        writer.close();
        
        File benchmarkFile = new File(TestUtils.getDataDir(PgnGenerationTest.class), "game.pgn");
        HaundrixAssert.contentEqualsIgnoringLineTerminator(tempFile, benchmarkFile);
        
        Reader reader = new FileReader(tempFile);
        ChessGame importedGame = new PgnImporter(reader).importGame();
        reader.close();
        Assert.assertEquals(importedGame, game);
    }
    
    @DataProvider(name="pgnFiles")
    public Object[][] getTestPgnFiles() {
        String[][] fileNames = {
            {"game1.pgn", "game1.pgn"},
            {"game2.pgn", "game2Benchmark.pgn"},
            {"game3.pgn", "game3Benchmark.pgn"},
            {"brokenGame.pgn", "brokenGameBenchmark.pgn"},
            {"complexGame.pgn", "complexGameBenchmark.pgn"},
            {"setupPositionGame.pgn", "setupPositionGameBenchmark.pgn"},
            {"noMovesGame.pgn", "noMovesGameBenchmark.pgn"},
            {"whiteWinsGame.pgn", "whiteWinsGame.pgn"},
            {"blackWinsGame.pgn", "blackWinsGame.pgn"},
            {"drawGame.pgn", "drawGame.pgn"},
            {"undefinedResultGame.pgn", "undefinedResultGame.pgn"}
        };
        
        File dataDir = TestUtils.getDataDir(PgnGenerationTest.class);
        File[][] files = new File[fileNames.length][2];
        for(int i = 0; i < fileNames.length; i++) {
            for(int j = 0; j < 2; j++) {
                files[i][j] = new File(dataDir, fileNames[i][j]);
            }
        }
        
        return files;
    }

    @Test(dataProvider="pgnFiles")
    public void testPgnFileExportImport(File originalFile, File benchmarkExportedFile) throws IOException {
        // should import the original file with no exceptions
        Reader reader = new FileReader(originalFile);
        ChessGame importedGame = new PgnImporter(reader).importGame();
        reader.close();
        
        // should export the imported game with no exceptions
        File tempDir = TestUtils.getTempDir(PgnGenerationTest.class);
        File tempFile = new File(tempDir, originalFile.getName());
        Writer writer = new FileWriter(tempFile);
        new PgnExporter(writer).exportGame(importedGame);
        writer.close();
        
        // compares the exported file with the benchmark file
        HaundrixAssert.contentEqualsIgnoringLineTerminator(tempFile, benchmarkExportedFile);
        
        // should be able to import the exported game with no exceptions
        reader = new FileReader(tempFile);
        ChessGame reimportedGame = new PgnImporter(reader).importGame();
        reader.close();
        
        // compares the imported games
        Assert.assertEquals(reimportedGame, importedGame);
    }
    
    @Test
    public void testGameResultPgnImport() throws IOException {
        File dataDir = TestUtils.getDataDir(PgnGenerationTest.class);
        
        File whiteWinsFile = new File(dataDir, "whiteWinsGame.pgn");
        ChessGame whiteWinsGame = new PgnImporter(new FileReader(whiteWinsFile)).importGame();
        Assert.assertEquals(whiteWinsGame.getResult(), new ChessGameResult(GameOutcome.WHITE_WINS));
        
        File blackWinsFile = new File(dataDir, "blackWinsGame.pgn");
        ChessGame blackWinsGame = new PgnImporter(new FileReader(blackWinsFile)).importGame();
        Assert.assertEquals(blackWinsGame.getResult(), new ChessGameResult(GameOutcome.BLACK_WINS));
        
        File drawFile = new File(dataDir, "drawGame.pgn");
        ChessGame drawGame = new PgnImporter(new FileReader(drawFile)).importGame();
        Assert.assertEquals(drawGame.getResult(), new ChessGameResult(GameOutcome.DRAW));
        
        File undefinedResultFile = new File(dataDir, "undefinedResultGame.pgn");
        ChessGame undefinedResultGame = new PgnImporter(new FileReader(undefinedResultFile)).importGame();
        Assert.assertNull(undefinedResultGame.getResult());
        
        File resultErrorFile = new File(dataDir, "resultErrorGame.pgn");
        try {
            new PgnImporter(new FileReader(resultErrorFile)).importGame();
            Assert.fail("Error! Failed to detect inconsistent game result when importing PGN file.");
        } catch(InvalidPgnFileException e) {
            // it works!
        }
    }
    
}
