/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
 */
package com.haundrix.games.model;

import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import com.haundrix.games.model.Piece.PieceColor;
import com.haundrix.tests.HaundrixAssert;

public class RectangularBoardTest {

    private class P extends Piece {

        public P(PieceColor color) {
            super(color);
        }
        
    }

    private P black = new P(PieceColor.BLACK);
    private P white = new P(PieceColor.WHITE);
    
    private RectangularBoard board;

    @BeforeMethod
    public void setUp() {
        board = new RectangularBoard(new P[][] {
    // columns :   0,     1,     2,     3,     4
    /* row 0 */ {null , black, null , null , null }, // row 0
    /* row 1 */ {null , null , white, null , black}, // row 1
    /* row 2 */ {white, null , null , null , null }, // row 2
    /* row 3 */ {null , null , null , null , null }, // row 3
    /* row 4 */ {null , null , black, white, null }, // row 4
    /* row 5 */ {black, null , null , white, null }, // row 5
    // columns :   0,     1,     2,     3,     4
        });
    }
    
    @AfterMethod
    public void tearDown() {
        board = null;
    }
    
    @Test
    public void testGetSquareColor() {
        Assert.assertEquals(board.getSquareColor(new Square(0, 0)), SquareColor.LIGHT);
        Assert.assertEquals(board.getSquareColor(new Square(4, 1)), SquareColor.DARK);
        Assert.assertEquals(board.getSquareColor(2, 4), SquareColor.LIGHT);
        Assert.assertEquals(board.getSquareColor(5, 4), SquareColor.DARK);
        
        board.setUpperLeftSquareColor(SquareColor.DARK);
        Assert.assertEquals(board.getSquareColor(new Square(0, 0)), SquareColor.DARK);
        Assert.assertEquals(board.getSquareColor(new Square(4, 1)), SquareColor.LIGHT);
        Assert.assertEquals(board.getSquareColor(2, 4), SquareColor.DARK);
        Assert.assertEquals(board.getSquareColor(5, 4), SquareColor.LIGHT);
    }

    @Test
    public void testGetLShapedPathSquaresFrom() {
        Square origin = new Square(0, 4);
        Square[] expectedLShapedPathSquares = {new Square(1, 2), new Square(2, 3)};
        Square[] actualLShapedPathSquares = board.getLShapedPathSquaresFrom(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualLShapedPathSquares, expectedLShapedPathSquares);
        
        origin = new Square(5, 2);
        expectedLShapedPathSquares = new Square[] {new Square(4, 0), new Square(3, 1), new Square(3, 3),
                new Square(4, 4)};
        actualLShapedPathSquares = board.getLShapedPathSquaresFrom(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualLShapedPathSquares, expectedLShapedPathSquares);
        
        origin = new Square(2, 2);
        expectedLShapedPathSquares = new Square[] {new Square(1, 0), new Square(0, 1), new Square(0, 3),
                new Square(1, 4), new Square(3, 0), new Square(4, 1), new Square(4, 3), new Square(3, 4)};
        actualLShapedPathSquares = board.getLShapedPathSquaresFrom(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualLShapedPathSquares, expectedLShapedPathSquares);
    }
    
    @Test
    public void testIsNeighbouringSquare() {
        Assert.assertTrue(board.isNeighbouringSquare(new Square(2, 2), new Square(1, 2)));
        Assert.assertTrue(board.isNeighbouringSquare(new Square(3, 1), new Square(2, 1)));
        Assert.assertTrue(board.isNeighbouringSquare(new Square(5, 4), new Square(5, 3)));
        Assert.assertTrue(board.isNeighbouringSquare(new Square(0, 1), new Square(1, 2)));
        Assert.assertTrue(board.isNeighbouringSquare(new Square(3, 0), new Square(2, 1)));
        
        Assert.assertFalse(board.isNeighbouringSquare(new Square(3, 3), new Square(3, 3)));
        Assert.assertFalse(board.isNeighbouringSquare(new Square(0, 0), new Square(5, 4)));
        Assert.assertFalse(board.isNeighbouringSquare(new Square(1, 1), new Square(3, 2)));
        Assert.assertFalse(board.isNeighbouringSquare(new Square(0, 4), new Square(2, 4)));
        Assert.assertFalse(board.isNeighbouringSquare(new Square(5, 0), new Square(3, 2)));
        
        try {
            board.isNeighbouringSquare(new Square(0, 0), new Square(-1, 0));
            Assert.fail();
        } catch(IllegalArgumentException e) {
            // that's fine
        }
    }
    
    @Test
    public void testIsColumn() {
        Assert.assertTrue(board.isColumn(new Square(0, 1), new Square(5, 1)));
        Assert.assertTrue(board.isColumn(new Square(5, 3), new Square(0, 3)));
        Assert.assertTrue(board.isColumn(new Square(3, 2), new Square(1, 2)));
        Assert.assertTrue(board.isColumn(new Square(4, 0), new Square(1, 0)));
        Assert.assertTrue(board.isColumn(new Square(1, 3), new Square(4, 3)));

        Assert.assertFalse(board.isColumn(new Square(3, 4), new Square(3, 4)));
        Assert.assertFalse(board.isColumn(new Square(5, 0), new Square(5, 4)));
        Assert.assertFalse(board.isColumn(new Square(3, 2), new Square(3, 0)));
        Assert.assertFalse(board.isColumn(new Square(1, 3), new Square(0, 4)));
        Assert.assertFalse(board.isColumn(new Square(4, 1), new Square(1, 4)));
        
        try {
            board.isColumn(new Square(0, 0), new Square(6, 0));
            Assert.fail();
        } catch(IllegalArgumentException e) {
            // that's fine
        }
    }
    
    @Test
    public void testIsFreeColumn() {
        Assert.assertTrue(board.isFreeColumn(new Square(5, 1), new Square(1, 1)));
        Assert.assertTrue(board.isFreeColumn(new Square(0, 3), new Square(3, 3)));
        Assert.assertTrue(board.isFreeColumn(new Square(5, 0), new Square(2, 0)));
        Assert.assertTrue(board.isFreeColumn(new Square(5, 3), new Square(4, 3)));
        Assert.assertTrue(board.isFreeColumn(new Square(0, 4), new Square(1, 4)));

        Assert.assertFalse(board.isFreeColumn(new Square(3, 3), new Square(3, 3)));
        Assert.assertFalse(board.isFreeColumn(new Square(0, 4), new Square(5, 4)));
        Assert.assertFalse(board.isFreeColumn(new Square(5, 4), new Square(0, 4)));
        Assert.assertFalse(board.isFreeColumn(new Square(0, 0), new Square(4, 0)));
        Assert.assertFalse(board.isFreeColumn(new Square(4, 0), new Square(0, 0)));
        Assert.assertFalse(board.isFreeColumn(new Square(3, 0), new Square(3, 4)));
        Assert.assertFalse(board.isFreeColumn(new Square(5, 1), new Square(4, 2)));
        
        try {
            board.isFreeColumn(new Square(0, 0), new Square(20, 0));
            Assert.fail();
        } catch(IllegalArgumentException e) {
            // that's fine
        }
    }
    
    @Test
    public void testIsRank() {
        Assert.assertTrue(board.isRank(new Square(4, 0), new Square(4, 4)));
        Assert.assertTrue(board.isRank(new Square(0, 1), new Square(0, 3)));
        Assert.assertTrue(board.isRank(new Square(3, 4), new Square(3, 1)));
        Assert.assertTrue(board.isRank(new Square(5, 2), new Square(5, 3)));
        Assert.assertTrue(board.isRank(new Square(2, 1), new Square(2, 0)));
        
        Assert.assertFalse(board.isRank(new Square(3, 1), new Square(3, 1)));
        Assert.assertFalse(board.isRank(new Square(4, 2), new Square(0, 2)));
        Assert.assertFalse(board.isRank(new Square(5, 4), new Square(4, 3)));
        Assert.assertFalse(board.isRank(new Square(5, 1), new Square(4, 0)));
        Assert.assertFalse(board.isRank(new Square(0, 3), new Square(2, 3)));
        
        try {
            board.isRank(new Square(1, 2), null);
            Assert.fail();
        } catch(IllegalArgumentException e) {
            // that's fine
        }
    }

    @Test
    public void testIsFreeRank() {
        Assert.assertTrue(board.isFreeRank(new Square(3, 0), new Square(3, 4)));
        Assert.assertTrue(board.isFreeRank(new Square(0, 0), new Square(0, 1)));
        Assert.assertTrue(board.isFreeRank(new Square(1, 2), new Square(1, 4)));
        Assert.assertTrue(board.isFreeRank(new Square(4, 2), new Square(4, 3)));
        Assert.assertTrue(board.isFreeRank(new Square(5, 0), new Square(5, 3)));

        Assert.assertFalse(board.isFreeRank(new Square(4, 1), new Square(4, 1)));
        Assert.assertFalse(board.isFreeRank(new Square(0, 0), new Square(0, 4)));
        Assert.assertFalse(board.isFreeRank(new Square(0, 4), new Square(0, 0)));
        Assert.assertFalse(board.isFreeRank(new Square(1, 1), new Square(1, 3)));
        Assert.assertFalse(board.isFreeRank(new Square(1, 3), new Square(1, 1)));
        Assert.assertFalse(board.isFreeRank(new Square(4, 3), new Square(5, 3)));
        Assert.assertFalse(board.isFreeRank(new Square(4, 1), new Square(3, 2)));
        
        try {
            board.isFreeRank(null, new Square(1, 2));
            Assert.fail();
        } catch(IllegalArgumentException e) {
            // that's fine
        }
    }
    
    @Test
    public void testIsDiagonal() {
        Assert.assertTrue(board.isDiagonal(new Square(0, 0), new Square(4, 4)));
        Assert.assertTrue(board.isDiagonal(new Square(4, 0), new Square(0, 4)));
        Assert.assertTrue(board.isDiagonal(new Square(2, 2), new Square(3, 3)));
        Assert.assertTrue(board.isDiagonal(new Square(4, 1), new Square(3, 0)));
        Assert.assertTrue(board.isDiagonal(new Square(5, 2), new Square(3, 4)));

        Assert.assertFalse(board.isDiagonal(new Square(4, 1), new Square(4, 1)));
        Assert.assertFalse(board.isDiagonal(new Square(0, 1), new Square(0, 3)));
        Assert.assertFalse(board.isDiagonal(new Square(1, 2), new Square(5, 2)));
        Assert.assertFalse(board.isDiagonal(new Square(2, 2), new Square(3, 2)));
        Assert.assertFalse(board.isDiagonal(new Square(5, 1), new Square(0, 0)));
        
        try {
            board.isDiagonal(new Square(3, 3), new Square(5, 8));
            Assert.fail();
        } catch(IllegalArgumentException e) {
            // that's fine
        }
    }
    
    @Test
    public void testIsFreeDiagonal() {
        Assert.assertTrue(board.isFreeDiagonal(new Square(0, 0), new Square(4, 4)));
        Assert.assertTrue(board.isFreeDiagonal(new Square(4, 0), new Square(0, 4)));
        Assert.assertTrue(board.isFreeDiagonal(new Square(5, 2), new Square(3, 0)));
        Assert.assertTrue(board.isFreeDiagonal(new Square(5, 0), new Square(1, 4)));
        Assert.assertTrue(board.isFreeDiagonal(new Square(0, 1), new Square(1, 2)));

        Assert.assertFalse(board.isFreeDiagonal(new Square(1, 1), new Square(1, 1)));
        Assert.assertFalse(board.isFreeDiagonal(new Square(5, 4), new Square(1, 0)));
        Assert.assertFalse(board.isFreeDiagonal(new Square(3, 4), new Square(5, 2)));
        Assert.assertFalse(board.isFreeDiagonal(new Square(0, 3), new Square(3, 0)));
        Assert.assertFalse(board.isFreeDiagonal(new Square(0, 1), new Square(2, 3)));
        
        try {
            board.isFreeDiagonal(new Square(-3, 3), new Square(5, 1));
            Assert.fail();
        } catch(IllegalArgumentException e) {
            // that's fine
        }
    }
    
    @Test
    public void testIsLShapedPath() {
        Assert.assertTrue(board.isLShapedPath(new Square(0, 0), new Square(1, 2)));
        Assert.assertTrue(board.isLShapedPath(new Square(5, 3), new Square(3, 4)));
        Assert.assertTrue(board.isLShapedPath(new Square(2, 4), new Square(1, 2)));
        Assert.assertTrue(board.isLShapedPath(new Square(2, 1), new Square(4, 0)));
        Assert.assertTrue(board.isLShapedPath(new Square(5, 0), new Square(3, 1)));

        Assert.assertFalse(board.isLShapedPath(new Square(1, 1), new Square(1, 1)));
        Assert.assertFalse(board.isLShapedPath(new Square(0, 0), new Square(3, 1)));
        Assert.assertFalse(board.isLShapedPath(new Square(5, 4), new Square(4, 1)));
        Assert.assertFalse(board.isLShapedPath(new Square(5, 2), new Square(2, 2)));
        Assert.assertFalse(board.isLShapedPath(new Square(3, 2), new Square(1, 4)));
        
        try {
            board.isLShapedPath(new Square(3, 100), new Square(5, 1));
            Assert.fail();
        } catch(IllegalArgumentException e) {
            // that's fine
        }
    }
    
    @Test
    public void testGetNeighbouringSquares() {
        Square origin = new Square(0, 0);
        Square[] expectedNeighbouringSquares = {new Square(0, 1), new Square(1, 0), new Square(1, 1)};
        Square[] actualNeighbouringSquares = board.getNeighbouringSquares(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualNeighbouringSquares, expectedNeighbouringSquares);
        
        origin = new Square(5, 3);
        expectedNeighbouringSquares = new Square[] {new Square(5, 2), new Square(5, 4), new Square(4, 2),
                new Square(4, 3), new Square(4, 4)};
        actualNeighbouringSquares = board.getNeighbouringSquares(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualNeighbouringSquares, expectedNeighbouringSquares);

        origin = new Square(2, 2);
        expectedNeighbouringSquares = new Square[] {new Square(1, 1), new Square(1, 2), new Square(1, 3),
                new Square(2, 1), new Square(2, 3), new Square(3, 1), new Square(3, 2), new Square(3, 3)};
        actualNeighbouringSquares = board.getNeighbouringSquares(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualNeighbouringSquares, expectedNeighbouringSquares);
    }
    
    @Test
    public void testGetFreeColumnSquaresFrom() {
        Square origin = new Square(5, 4);
        Square[] expectedFreeColumnSquares = {new Square(4, 4), new Square(3, 4), new Square(2, 4),
                new Square(1, 4)};
        Square[] actualFreeColumnSquares = board.getFreeColumnSquaresFrom(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualFreeColumnSquares, expectedFreeColumnSquares);
        
        origin = new Square(3, 0);
        expectedFreeColumnSquares = new Square[] {new Square(2, 0), new Square(4, 0), new Square(5, 0)};
        actualFreeColumnSquares = board.getFreeColumnSquaresFrom(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualFreeColumnSquares, expectedFreeColumnSquares);
        
        origin = new Square(1, 1);
        expectedFreeColumnSquares = new Square[] {new Square(0, 1), new Square(2, 1), new Square(3, 1),
                new Square(4, 1), new Square(5, 1)};
        actualFreeColumnSquares = board.getFreeColumnSquaresFrom(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualFreeColumnSquares, expectedFreeColumnSquares);
    }
    
    @Test
    public void testGetFreeRankSquaresFrom() {
        Square origin = new Square(5, 0);
        Square[] expectedFreeRankSquares = {new Square(5, 1), new Square(5, 2), new Square(5, 3)};
        Square[] actualFreeRankSquares = board.getFreeRankSquaresFrom(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualFreeRankSquares, expectedFreeRankSquares);
        
        origin = new Square(0, 3);
        expectedFreeRankSquares = new Square[] {new Square(0, 1), new Square(0, 2), new Square(0, 4)};
        actualFreeRankSquares = board.getFreeRankSquaresFrom(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualFreeRankSquares, expectedFreeRankSquares);
        
        origin = new Square(3, 2);
        expectedFreeRankSquares = new Square[] {new Square(3, 0), new Square(3, 1), new Square(3, 3),
                new Square(3, 4)};
        actualFreeRankSquares = board.getFreeRankSquaresFrom(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualFreeRankSquares, expectedFreeRankSquares);
    }
    
    @Test
    public void testGetFreeDiagonalSquaresFrom() {
        Square origin = new Square(5, 0);
        Square[] expectedFreeDiagonalSquares = {new Square(4, 1), new Square(3, 2), new Square(2, 3),
                new Square(1, 4)};
        Square[] actualFreeDiagonalSquares = board.getFreeDiagonalSquaresFrom(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualFreeDiagonalSquares, expectedFreeDiagonalSquares);
        
        origin = new Square(2, 4);
        expectedFreeDiagonalSquares = new Square[] {new Square(1, 3), new Square(0, 2), new Square(3, 3),
                new Square(4, 2)};
        actualFreeDiagonalSquares = board.getFreeDiagonalSquaresFrom(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualFreeDiagonalSquares, expectedFreeDiagonalSquares);
        
        origin = new Square(3, 2);
        expectedFreeDiagonalSquares = new Square[] {new Square(1, 0), new Square(2, 1), new Square(4, 3),
                new Square(5, 0), new Square(4, 1), new Square(2, 3), new Square(1, 4)};
        actualFreeDiagonalSquares = board.getFreeDiagonalSquaresFrom(origin).toArray(new Square[0]);
        HaundrixAssert.areEqualIgnoringOrder(actualFreeDiagonalSquares, expectedFreeDiagonalSquares);
    }
    
}
