/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.tests;

import java.util.Arrays;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.PaletteData;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.widgets.Display;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

/**
 * Tests whether SWT composes transparent images correctly.
 */
@Test(groups=TestGroups.GUI_TESTS)
public class SwtCompositeTest {

    private static final int WIDTH = 200;
    private static final int HEIGHT = 100;
    
    private static final int PIXEL_VALUE = 212;
    private static final int ALPHA_VALUE = 70;
    
    private Display display;
    private Image image;
    private GC gc;
    
    @BeforeMethod
    public void setUp() {
        display = new Display();
        image = new Image(display, WIDTH, HEIGHT);
        gc = new GC(image);
        gc.setBackground(display.getSystemColor(SWT.COLOR_WHITE));
        gc.fillRectangle(0, 0, WIDTH, HEIGHT);
    }
    
    @AfterMethod
    public void tearDown() {
        gc.dispose();
        image.dispose();
        display.dispose();
    }
    
    @Test
    public void testGlobalAlphaComposite() {
        ImageData imageData = new ImageData(WIDTH, HEIGHT, 24, new PaletteData(0xFF0000, 0xFF00, 0xFF));
        Arrays.fill(imageData.data, (byte) PIXEL_VALUE);
        imageData.alpha = ALPHA_VALUE;
        Image transparentImage = new Image(display, imageData);
        gc.drawImage(transparentImage, 0, 0);
        transparentImage.dispose();
        
        checkImagePixels();
    }

    @Test
    public void testCompositeByPixel() {
        ImageData imageData = new ImageData(WIDTH, HEIGHT, 24, new PaletteData(0xFF0000, 0xFF00, 0xFF));
        Arrays.fill(imageData.data, (byte) PIXEL_VALUE);
        byte[] alphas = new byte[WIDTH * HEIGHT];
        Arrays.fill(alphas, (byte) ALPHA_VALUE);
        imageData.alphaData = alphas;
        Image transparentImage = new Image(display, imageData);
        gc.drawImage(transparentImage, 0, 0);
        transparentImage.dispose();
        
        checkImagePixels();
    }
    
    @Test
    public void testGcAlpha() {
        Color backgroundColor = new Color(display, PIXEL_VALUE, PIXEL_VALUE, PIXEL_VALUE);
        gc.setBackground(backgroundColor);
        gc.setAlpha(ALPHA_VALUE);
        gc.fillRectangle(image.getBounds());
        backgroundColor.dispose();
        
        checkImagePixels();
    }

    private void checkImagePixels() {
        int expectedPixelValue = (int) (255.0 * (1.0 - ALPHA_VALUE / 255.0) + PIXEL_VALUE * (ALPHA_VALUE / 255.0));
        
        int[] pixels = new int[WIDTH * HEIGHT];
        ImageData imageData = image.getImageData();
        imageData.getPixels(0, 0, pixels.length, pixels, 0);
        for(int pixel : pixels) {
            RGB rgb = imageData.palette.getRGB(pixel);
            
            Assert.assertEquals(rgb.red, expectedPixelValue);
            Assert.assertEquals(rgb.green, expectedPixelValue);
            Assert.assertEquals(rgb.blue, expectedPixelValue);
        }
    }
    
}
