/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.animation;

import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Display;
import org.jdesktop.animation.timing.TimingTarget;
import org.jdesktop.animation.timing.TimingTargetAdapter;
import org.jdesktop.animation.timing.interpolation.Interpolator;
import org.jdesktop.animation.timing.interpolation.PropertySetter;

import com.haundrix.utils.IDisposeListener;
import com.haundrix.utils.IDisposable;

public class Mover {

    private static Interpolator DEFAULT_INTERPOLATOR = CompressedSpringInterpolator.getInstance();

    private IMovable movable;
    private Point startLocation;
    private Point endLocation;
    private MoveCallback moveCallback;
    
    private boolean moving = false;
    private boolean paused = false;
    private boolean aboutToEnd = false;
    
    private SwtAnimator animator;
    
    private IDisposeListener disposeListener;
    private boolean isMovableDisposed = false;

    public Mover(Display display, IMovable movable, Point endLocation, int duration, MoveCallback moveCallback) {
        this.movable = movable;
        this.endLocation = endLocation;
        this.moveCallback = moveCallback != null ? moveCallback : new MoveCallback();
        
        PropertySetter propertySetter = new PropertySetter(movable, "location", endLocation);
        animator = new SwtAnimator(display, duration, propertySetter);
        
        if(movable instanceof IDisposable) {
            disposeListener = new IDisposeListener() {
                @Override
                public void disposed(IDisposable disposable) {
                    isMovableDisposed = true;
                    animator.cancel();
                    Mover.this.moveCallback.movableDisposed(Mover.this, disposable);
                }
            };
            ((IDisposable) movable).addDisposeListener(disposeListener);
        }

        animator.addTarget(new TimingTargetAdapter() {

            @Override
            public void begin() {
                Mover.this.moveCallback.started(Mover.this);
            }

            @Override
            public void end() {
                aboutToEnd = true;
                Mover.this.moveCallback.aboutToEnd(Mover.this);
                if(!paused) {
                    endMovement();
                }
            }

            @Override
            public void timingEvent(float fraction) {
                Mover.this.moveCallback.moved(Mover.this);
            }
            
        });

        animator.setInterpolator(DEFAULT_INTERPOLATOR);
    }

    public static TimingTarget createTimingTarget(IMovable movable, Point endLocation) {
        PropertySetter propertySetter = new PropertySetter(movable, "location", endLocation);
        TimingTarget timingTarget = new InterpolatorTimingTarget(DEFAULT_INTERPOLATOR, propertySetter);
        return movable instanceof IDisposable ?
                new DisposableAwareTimingTarget((IDisposable) movable, timingTarget) :
                timingTarget;
    }
    
    public Point getStartLocation() {
        return moving ? startLocation : movable.getLocation();
    }
    
    public Point getEndLocation() {
        return endLocation;
    }
    
    public Point getCurrentLocation() {
        if(!moving) {
            throw new IllegalStateException("Mover is not moving.");
        }
        
        return movable.getLocation();
    }
    
    public int getDuration() {
        return animator.getDuration();
    }
    
    public int getElapsedTime() {
        if(!moving) {
            throw new IllegalStateException("Mover is not moving.");
        }
        if(paused) {
            throw new IllegalStateException("Mover is paused.");
        }
        
        return (int) animator.getTotalElapsedTime();
    }
    
    public int getRemainingTime() {
        if(!moving) {
            throw new IllegalStateException("Mover is not moving.");
        }
        if(paused) {
            throw new IllegalStateException("Mover is paused.");
        }
        
        return getDuration() - getElapsedTime();
    }
    
    public Interpolator getInterpolator() {
        return animator.getInterpolator();
    }
    
    public void setInterpolator(Interpolator interpolator) {
        animator.setInterpolator(interpolator);
    }
    
    public void move() {
        if(moving) {
            throw new IllegalStateException("Mover is already moving.");
        }
        
        animator.start();
        startLocation = movable.getLocation();
        moving = true;
    }
    
    public void pause() {
        if(!moving) {
            throw new IllegalStateException("Mover is not moving.");
        }
        
        if(!paused) {
            if(!aboutToEnd) {
                animator.pause();
            }
            paused = true;
        }
    }
    
    public void resume() {
        if(!moving) {
            throw new IllegalStateException("Mover is not moving.");
        }
        
        if(paused) {
            if(!aboutToEnd) {
                animator.resume();
            } else {
                if(!isMovableDisposed) {
                    endMovement();
                }
            }
            
            paused = false;
        }
    }
    
    private void endMovement() {
        moving = false;
        paused = false;
        aboutToEnd = false;
        startLocation = null;
        if(movable instanceof IDisposable) {
            ((IDisposable) movable).removeDisposeListener(disposeListener);
        }
        
        moveCallback.ended(this);
    }

}
