/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.animation;

import org.eclipse.swt.widgets.Display;
import org.jdesktop.animation.timing.TimingSource;

public class SwtTimingSource extends TimingSource {
    
    private Display display;
    
    private int startDelay = 0;
    private int resolution = 20;
    
    private volatile Tick nextTick = null;

    public SwtTimingSource(Display display) {
        if(display == null) {
            throw new IllegalArgumentException("Error! Display cannot be null.");
        }
        
        this.display = display;
    }

    public SwtTimingSource(Display display, int resolution) {
        if(display == null) {
            throw new IllegalArgumentException("Error! Display cannot be null.");
        }
        
        this.display = display;
        if(resolution >= 0) {
            this.resolution = resolution;
        }
    }

    public SwtTimingSource(Display display, int startDelay, int resolution) {
        if(display == null) {
            throw new IllegalArgumentException("Error! Display cannot be null.");
        }
        
        this.display = display;
        if(startDelay >= 0) {
            this.startDelay = startDelay;
        }
        if(resolution >= 0) {
            this.resolution = resolution;
        }
    }

    @Override
    public void setStartDelay(int startDelay) {
        checkIsNotRunning();
        this.startDelay = startDelay;
    }
    
    @Override
    public void setResolution(int resolution) {
        checkIsNotRunning();
        this.resolution = resolution;
    }

    @Override
    public void start() {
        checkIsNotRunning();
        if(Display.getCurrent() != display) {
            display.syncExec(new Runnable() {
                @Override
                public void run() {
                    doStart();
                }
            });
        } else {
            doStart();
        }
    }

    @Override
    public synchronized void stop() {
        if(nextTick != null) {
            nextTick.cancel();
            nextTick = null;
        }
    }

    public boolean isRunning() {
        return nextTick != null;
    }
    
    private void checkIsNotRunning() {
        if(isRunning()) {
            throw new IllegalStateException("Cannot perform this operation while TimingSource is running.");
        }
    }

    private void doStart() {
        long expectedStartTime = System.nanoTime() / 1000000 + startDelay;
        nextTick = new Tick(expectedStartTime);
        display.timerExec(startDelay, nextTick);
    }
    
    private class Tick implements Runnable {
        
        private long expectedTime;
        private volatile boolean cancel = false;
        
        public Tick(long expectedTime) {
            this.expectedTime = expectedTime;
        }
        
        public void cancel() {
            cancel = true;
        }

        @Override
        public void run() {
            synchronized(SwtTimingSource.this) {
                if(!cancel) {
                    timingEvent();
                    // the callbacks triggered by the call to timingEvent() above
                    // may have requested this timing source to stop, so let's check it
                    // again and not schedule the next tick if this is the case
                    if(!cancel) {
                        int delay;
                        if(resolution == 0) {
                            // special case, schedule the tick event ASAP
                            delay = 0;
                            nextTick = new Tick(0L);
                        } else {
                            long currentTime = System.nanoTime() / 1000000;
                            delay =  currentTime < expectedTime ?
                                ((int) (expectedTime - currentTime)) + resolution :
                                resolution - (((int) (currentTime - expectedTime)) % resolution);
                            long newExpectedTime = currentTime + delay;
                            nextTick = new Tick(newExpectedTime);
                        }
                        display.timerExec(delay, nextTick);
                    }
                }
            }
        }
        
    }

}
