/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.controller;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.swt.widgets.Composite;

import com.haundrix.games.chess.model.Bishop;
import com.haundrix.games.chess.model.ChessGame;
import com.haundrix.games.chess.model.King;
import com.haundrix.games.chess.model.Knight;
import com.haundrix.games.chess.model.Pawn;
import com.haundrix.games.chess.model.PromoteablePiece;
import com.haundrix.games.chess.model.PromotionMove;
import com.haundrix.games.chess.model.Queen;
import com.haundrix.games.chess.model.Rook;
import com.haundrix.games.chess.view.ChessPieceType;
import com.haundrix.games.controller.GameController;
import com.haundrix.games.controller.MoveAnimationEvent;
import com.haundrix.games.model.GameAdapter;
import com.haundrix.games.model.GameResult;
import com.haundrix.games.model.Move;
import com.haundrix.games.model.Piece;
import com.haundrix.games.model.Square;
import com.haundrix.games.model.Piece.PieceColor;
import com.haundrix.games.view.BoardPiece;
import com.haundrix.games.view.BoardSquare;
import com.haundrix.games.view.PieceSelectorDialog;
import com.haundrix.games.view.PieceSet;
import com.haundrix.utils.BidirectionalMap;

public class ChessGameController extends GameController<ChessPieceType> {
    
    private static final BidirectionalMap<Class<? extends Piece>, ChessPieceType> BLACK_PIECE_TYPES;
    private static final BidirectionalMap<Class<? extends Piece>, ChessPieceType> WHITE_PIECE_TYPES;
    static {
        BLACK_PIECE_TYPES = new BidirectionalMap<Class<? extends Piece>, ChessPieceType>();
        BLACK_PIECE_TYPES.put(Rook.class, ChessPieceType.BLACK_ROOK);
        BLACK_PIECE_TYPES.put(Knight.class, ChessPieceType.BLACK_KNIGHT);
        BLACK_PIECE_TYPES.put(Bishop.class, ChessPieceType.BLACK_BISHOP);
        BLACK_PIECE_TYPES.put(Queen.class, ChessPieceType.BLACK_QUEEN);
        BLACK_PIECE_TYPES.put(King.class, ChessPieceType.BLACK_KING);
        BLACK_PIECE_TYPES.put(Pawn.class, ChessPieceType.BLACK_PAWN);
        
        WHITE_PIECE_TYPES = new BidirectionalMap<Class<? extends Piece>, ChessPieceType>();
        WHITE_PIECE_TYPES.put(Rook.class, ChessPieceType.WHITE_ROOK);
        WHITE_PIECE_TYPES.put(Knight.class, ChessPieceType.WHITE_KNIGHT);
        WHITE_PIECE_TYPES.put(Bishop.class, ChessPieceType.WHITE_BISHOP);
        WHITE_PIECE_TYPES.put(Queen.class, ChessPieceType.WHITE_QUEEN);
        WHITE_PIECE_TYPES.put(King.class, ChessPieceType.WHITE_KING);
        WHITE_PIECE_TYPES.put(Pawn.class, ChessPieceType.WHITE_PAWN);
    }
    
    private static final int PIECE_SIZE_IN_PROMOTION_DIALOG = 48;

    private PromotionMove promotionMove;
    
    private int checkHighlightDuration = 250;
    private int nCheckHighlights = 4;

    private int promotionDuration = 1500;
    
    private List<IChessGameControllerListener> chessGameControllerListeners =
        new ArrayList<IChessGameControllerListener>();

    private PieceSelectorDialog<ChessPieceType> pieceSelectorDialog = null;

    public ChessGameController(ChessGame chessGame, Composite parent, PieceSet<ChessPieceType> pieceSet) {
        super(chessGame, parent, pieceSet, BLACK_PIECE_TYPES, WHITE_PIECE_TYPES);
        
        chessGame.addGameListener(new GameAdapter() {
            @Override
            public void ended(GameResult result) {
                if(pieceSelectorDialog != null) {
                    pieceSelectorDialog.dispose();
                }
            }
        });
    }
    
    @Override
    public ChessGame getGame() {
        return (ChessGame) super.getGame();
    }

    public int getCheckHighlightDuration() {
        return checkHighlightDuration;
    }

    public void setCheckHighlightDuration(int checkHighlightDuration) {
        this.checkHighlightDuration = checkHighlightDuration;
    }

    public int getNCheckHighlights() {
        return nCheckHighlights;
    }

    public void setNCheckHighlights(int checkHighlights) {
        nCheckHighlights = checkHighlights;
    }

    public int getPromotionDuration() {
        return promotionDuration;
    }

    public void setPromotionDuration(int promotionDuration) {
        this.promotionDuration = promotionDuration;
    }

    public void addChessGameControllerListener(IChessGameControllerListener chessGameControllerListener) {
        synchronized(gameControllerListeners) {
            synchronized(chessGameControllerListeners) {
                addGameControllerListener(chessGameControllerListener);
                chessGameControllerListeners.add(chessGameControllerListener);
            }
        }
    }
    
    public void removeChessGameControllerListener(IChessGameControllerListener chessGameControllerListener) {
        synchronized(gameControllerListeners) {
            synchronized(chessGameControllerListeners) {
                removeGameControllerListener(chessGameControllerListener);
                chessGameControllerListeners.remove(chessGameControllerListener);
            }
        }
    }

    @Override
    protected void completePlayingMove(Move move, boolean isNewMove, final Runnable endPlayCallback) {
        promotionMove = null;
        
        boolean isCastle = isNewMove ? getGame().isCastle(move) : getGame().previousMoveWasCastle();
        boolean isPromotion = isNewMove ? (move instanceof PromotionMove) || getGame().isPromotion(move) :
                getGame().previousMoveWasPromotion();
        if(isCastle) {
            completePlayingCastleMove(move, isNewMove, endPlayCallback);
        } else if(isPromotion) {
            completePlayingPromotionMove(move, isNewMove, endPlayCallback);
        } else {
            endPlayCallback.run();
        }
    }

    @Override
    protected void startUnplayingMove(Move move, final Runnable endUnplayCallback) {
        if(getGame().isCastle(move)) {
            Move rookMove = getGame().getRookCastleMove(move);
            int movementDuration = getMovementDuration(rookMove);
            BoardSquare boardOrigin = GameController.getBoardSquare(rookMove.getOrigin());
            BoardSquare boardDestination = GameController.getBoardSquare(rookMove.getDestination());
            boardControl.movePiece(boardDestination, boardOrigin, movementDuration, endUnplayCallback);
        } else if(getGame().isPromotion(move)) {
            Pawn promotingPawn = (Pawn) getGame().getBoard().getPiece(move.getOrigin());
            BoardSquare boardDestination = GameController.getBoardSquare(move.getDestination());
            BoardPiece<ChessPieceType> promotedBoardPiece = boardControl.getBoardPiece(boardDestination);
            promotedBoardPiece.crossfadeTo(getPieceType(promotingPawn), promotionDuration, endUnplayCallback);
        } else {
            endUnplayCallback.run();
        }
    }

    @Override
    protected void afterFinishedPlaying() {
        if(getGame().isCheck()) {
            Square kingSquare = getGame().getCheckedKingSquare();
            BoardSquare kingBoardSquare = GameController.getBoardSquare(kingSquare);
            boardControl.getBoardPiece(kingBoardSquare).pulse(highlightScheme.getCapturingPieceStrongHighlight(),
                    checkHighlightDuration, nCheckHighlights);
        }
    }

    @Override
    protected Move getMove(Square origin, Square destination) {
        Move move = new Move(origin, destination);
        return getGame().isPromotion(move) ? promotionMove : move;
    }

    private void completePlayingCastleMove(Move move, boolean isNewMove, final Runnable endPlayCallback) {
        Move rookMove = isNewMove ? getGame().getRookCastleMove(move) : getGame().getPreviousRookCastleMove();
        int movementDuration = getMovementDuration(rookMove);
        final MoveAnimationEvent event = new MoveAnimationEvent(movementDuration, move, isNewMove);
        BoardSquare boardOrigin = GameController.getBoardSquare(rookMove.getOrigin());
        BoardSquare boardDestination = GameController.getBoardSquare(rookMove.getDestination());
        final BoardPiece<ChessPieceType> boardRook = boardControl.getBoardPiece(boardOrigin);
        notifyBeginCastleRookMoveAnimation(event);
        boardControl.movePiece(boardOrigin, boardDestination, movementDuration, new Runnable() {
            @Override
            public void run() {
                notifyEndCastleRookMoveAnimation(event);
                boardRook.setHighlight(null);
                endPlayCallback.run();
            }
        });
    }

    private ChessPieceType[] getValidPromotionPieces(PieceColor promotingColor) {
        ChessPieceType[] validPromotionPieces = new ChessPieceType[ChessGame.PROMOTEABLE_PIECES.length];
        for(int i = 0; i < validPromotionPieces.length; i++) {
            validPromotionPieces[i] = getPieceType(ChessGame.PROMOTEABLE_PIECES[i], promotingColor);
        }
        
        return validPromotionPieces;
    }

    private void completePlayingPromotionMove(Move move, boolean isNewMove, final Runnable endPlayCallback) {
        ChessPieceType promotionPiece;
        if(isNewMove) {
            if(move instanceof PromotionMove) {
                promotionMove = (PromotionMove) move;
                PieceColor promotingPieceColor = getGame().getState().getColorToMove();
                promotionPiece = getPieceType(promotionMove.getPromotedPiece(), promotingPieceColor);
            } else {
                PieceColor promotingColor = getGame().getState().getColorToMove();
                ChessPieceType[] promotionPieces = getValidPromotionPieces(promotingColor);
                pieceSelectorDialog = new PieceSelectorDialog<ChessPieceType>(boardControl, promotionPieces, PIECE_SIZE_IN_PROMOTION_DIALOG);
                pieceSelectorDialog.setText("Choose the promotion piece:");  // TODO externalize this
                promotionPiece = pieceSelectorDialog.getPiece();
                pieceSelectorDialog = null;
                
                if(promotionPiece != null) {
                    @SuppressWarnings("unchecked")
                    Class<? extends PromoteablePiece> promotionPieceClass =
                        (Class<? extends PromoteablePiece>) getPieceClass(promotionPiece);
                    promotionMove = new PromotionMove(move.getOrigin(), move.getDestination(), promotionPieceClass);
                } else {
                    promotionMove = null;
                }
            }
        } else {
            promotionMove = (PromotionMove) getGame().getPreviousMove();
            PieceColor promotingPieceColor = getGame().getState().getColorToMove().getOppositeColor();
            promotionPiece = getPieceType(promotionMove.getPromotedPiece(), promotingPieceColor);
        }
        
        if(promotionPiece != null) {
            BoardSquare boardDestination = GameController.getBoardSquare(move.getDestination());
            BoardPiece<ChessPieceType> promotingBoardPiece = boardControl.getBoardPiece(boardDestination);
            final MoveAnimationEvent event = new MoveAnimationEvent(promotionDuration, promotionMove, isNewMove);
            notifyBeginPromotionMoveAnimation(event);
            promotingBoardPiece.crossfadeTo(promotionPiece, promotionDuration, new Runnable() {
                @Override
                public void run() {
                    notifyEndPromotionMoveAnimation(event);
                    endPlayCallback.run();
                }
            });
        } else {
            endPlayCallback.run();
        }
    }

    private void notifyBeginCastleRookMoveAnimation(MoveAnimationEvent event) {
        synchronized(chessGameControllerListeners) {
            for(IChessGameControllerListener chessGameControllerListener : chessGameControllerListeners) {
                chessGameControllerListener.beginCastleRookMoveAnimation(event);
            }
        }
    }
    
    private void notifyEndCastleRookMoveAnimation(MoveAnimationEvent event) {
        synchronized(chessGameControllerListeners) {
            for(IChessGameControllerListener chessGameControllerListener : chessGameControllerListeners) {
                chessGameControllerListener.endCastleRookMoveAnimation(event);
            }
        }
    }

    private void notifyBeginPromotionMoveAnimation(MoveAnimationEvent event) {
        synchronized(chessGameControllerListeners) {
            for(IChessGameControllerListener chessGameControllerListener : chessGameControllerListeners) {
                chessGameControllerListener.beginPromotionMoveAnimation(event);
            }
        }
    }
    
    private void notifyEndPromotionMoveAnimation(MoveAnimationEvent event) {
        synchronized(chessGameControllerListeners) {
            for(IChessGameControllerListener chessGameControllerListener : chessGameControllerListeners) {
                chessGameControllerListener.endPromotionMoveAnimation(event);
            }
        }
    }

}
