/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.controller;

import com.haundrix.games.chess.model.ChessGame;
import com.haundrix.games.chess.model.ChessGameResult;
import com.haundrix.games.chess.model.ChessGameResultReason;
import com.haundrix.games.chess.model.Knight;
import com.haundrix.games.chess.model.PromoteablePiece;
import com.haundrix.games.chess.model.PromotionMove;
import com.haundrix.games.controller.AnimationEvent;
import com.haundrix.games.controller.MoveAnimationEvent;
import com.haundrix.games.model.GameOutcome;
import com.haundrix.games.model.Piece.PieceColor;
import com.haundrix.utils.SoundFile;
import com.haundrix.utils.SoundPlayer;

public class ChessGameSoundController {

    private ChessGame game;
    private IChessSoundTheme soundTheme;
    private SoundPlayer soundPlayer;

    public ChessGameSoundController(ChessGameController gameController, IChessSoundTheme soundTheme,
            SoundPlayer soundPlayer) {
        game = gameController.getGame();
        this.soundTheme = soundTheme;
        this.soundPlayer = soundPlayer;
        
        gameController.addChessGameControllerListener(new InternalChessGameControllerListener());
    }
    
    public void playNewGameSound() {
        soundPlayer.play(soundTheme.getNewGameSound());
    }
    
    public void playGameEndingSound(ChessGameResult result) {
        playGameEndingSound(result, 0);
    }

    private void playGameEndingSound(ChessGameResult result, int interval) {
        if(result != null) {
            SoundFile sound;
            GameOutcome outcome = result.getOutcome();
            if((outcome == GameOutcome.WHITE_WINS) || (outcome == GameOutcome.BLACK_WINS)) {
                if(result.getResultReason() == ChessGameResultReason.TIME_FORFEIT) {
                    sound = soundTheme.getTimeoutSound();
                } else {
                    sound = soundTheme.getWinSound();
                }
            } else {
                sound = soundTheme.getDrawSound();
            }
            
            soundPlayer.playAtBegin(sound, interval);
        }
    }

    private class InternalChessGameControllerListener extends ChessGameControllerAdapter {

        @Override
        public void pieceDraggedOffBoard(AnimationEvent event) {
            handleIllegalMove(event);
        }

        @Override
        public void illegalMoveAttempted(AnimationEvent event) {
            handleIllegalMove(event);
        }

        @Override
        public void beginMoveAnimation(MoveAnimationEvent event) {
            if(isPromotion(event)) {
                // the initial pawn move is always played in promotions
                if(isCaptureUnchecked(event)) {
                    handleCaptureMove(event);
                } else {
                    handleNormalMove(event);
                }
            } else if(isGameEnding(event)) {
                if(isCastle(event)) {
                    // the game ending sound will be played only after the rook move
                    return;
                } else {
                    handleGameEndingMove(event);
                }
            } else if(isKnightAttack(event)) {
                handleKnightAttackMove();
            } else if(isCheck(event)) {
                if(isCastle(event)) {
                    // the check sound will be played only after the rook move
                    return;
                } else {
                    handleCheckMove(event);
                }
            } else if(isKnightCapture(event)) {
                handleKnightCaptureMove(event);
            } else if(isCapture(event)) {
                handleCaptureMove(event);
            } else {
                handleNormalMove(event);
            }
        }

        @Override
        public void beginTransitionAnimation(AnimationEvent event) {
            if(game.isFinished()) {
                playGameEndingSound(game.getResult(), event.getDuration());
            } else {
                soundPlayer.playAtFinish(soundTheme.getTransitionSound(), event.getDuration());
            }
        }

        @Override
        public void beginCastleRookMoveAnimation(MoveAnimationEvent event) {
            if(isCheck(event)) {
                handleCheckMove(event);
            } else {
                handleNormalMove(event);
            }
        }

        @Override
        public void beginPromotionMoveAnimation(MoveAnimationEvent event) {
            Class<? extends PromoteablePiece> promotedPiece = ((PromotionMove) event.getMove()).getPromotedPiece();
            if(promotedPiece == Knight.class) {
                handleKnightAttackMove();
            } else if(isCheck(event)) {
                handleCheckMove(event);
            } else {
                soundPlayer.play(soundTheme.getPromotionSound());
            }
        }

        private boolean isCapture(MoveAnimationEvent event) {
            return event.isNewMove() ? game.isCaptureMove(event.getMove()) : game.previousMoveWasCapture();
        }

        private boolean isCaptureUnchecked(MoveAnimationEvent event) {
            return event.isNewMove() ? game.isCapture(event.getMove()) : game.previousMoveWasCapture();
        }

        private boolean isKnightCapture(MoveAnimationEvent event) {
            return event.isNewMove() ? game.isKnightCaptureMove(event.getMove()) : game.previousMoveWasKnightCapture();
        }

        private boolean isKnightAttack(MoveAnimationEvent event) {
            return event.isNewMove() ? game.isKnightAttackMove(event.getMove()) : game.previousMoveWasKnightAttack();
        }

        private boolean isCheck(MoveAnimationEvent event) {
            return event.isNewMove() ? game.isCheckMove(event.getMove()) : game.previousMoveWasCheck();
        }

        private boolean isCastle(MoveAnimationEvent event) {
            return event.isNewMove() ? game.isCastleMove(event.getMove()) : game.previousMoveWasCastle();
        }

        private boolean isPromotion(MoveAnimationEvent event) {
            return event.isNewMove() ? game.isPromotion(event.getMove()) : game.previousMoveWasPromotion();
        }

        private boolean isGameEnding(MoveAnimationEvent event) {
            return event.isNewMove() ? game.isGameEndingMove(event.getMove()) : game.isFinished();
        }

        private void handleIllegalMove(AnimationEvent event) {
            if(event.getDuration() <= soundTheme.getShortIllegalMoveDurationThreshold()) {
                soundPlayer.play(soundTheme.getIllegalMoveShortSound());
            } else {
                soundPlayer.play(soundTheme.getIllegalMoveLongSound());
            }
        }

        private void handleNormalMove(MoveAnimationEvent event) {
            PieceColor movedColor = event.isNewMove() ? game.getCurrentColorToMove() :
                game.getCurrentColorToMove().getOppositeColor();
            if(movedColor == PieceColor.WHITE) {
                soundPlayer.playAtFinish(soundTheme.getWhiteMoveSound(), event.getDuration());
            } else {
                soundPlayer.playAtFinish(soundTheme.getBlackMoveSound(), event.getDuration());
            }
        }

        private void handleCaptureMove(MoveAnimationEvent event) {
            soundPlayer.playAtFinish(soundTheme.getCaptureSound(), event.getDuration());
        }

        private void handleKnightCaptureMove(MoveAnimationEvent event) {
            soundPlayer.playAtFinish(soundTheme.getKnightCaptureSound(), event.getDuration());
        }

        private void handleKnightAttackMove() {
            soundPlayer.play(soundTheme.getKnightAttackSound());
        }

        private void handleCheckMove(MoveAnimationEvent event) {
            soundPlayer.playAtBegin(soundTheme.getCheckSound(), event.getDuration());
        }

        private void handleGameEndingMove(MoveAnimationEvent event) {
            ChessGameResult result = event.isNewMove() ? game.getResult(event.getMove()) : game.getResult();
            playGameEndingSound(result, event.getDuration());
        }

    }

}
