/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.model;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.haundrix.games.chess.model.ChessGameState.BoardSide;
import com.haundrix.games.model.Game;
import com.haundrix.games.model.IMoveNotation;
import com.haundrix.games.model.IllegalMoveException;
import com.haundrix.games.model.Move;
import com.haundrix.games.model.RectangularBoard;
import com.haundrix.games.model.Square;
import com.haundrix.games.model.Piece.PieceColor;
import com.haundrix.utils.BidirectionalMap;

// TODO Some other non official options this class may support:
//          - use ":" symbol instead of "x" to indicate captures,
//          - suppress capture, check and checkmate signs.
//
// This class allows for some customization of the notation, such
// as indicating or not en passant captures, using the # or ++ sign
// for annotating checkmates, etc. The default values for these options
// correspond to the official FIDE rules, except for using the = sign
// to annotate promotions, which should have its default value set
// to false, as in the official FIDE rules the = sign is used to indicate
// the offer of a draw.
//
// A few notes about disambiguation:
// When a move may be disambiguated by both the originating file and the
// originating rank of the moving piece, the FIDE rules SUGGEST that the
// originating file be used, whereas the PGN specification MANDATES that
// the originating file be used. Therefore, this implementation always
// uses the originating file for converting moves to descriptions and
// the other way around.
// Also, the FIDE rules do not consider the special case when the moving piece
// can not be disambiguated by its originating file or rank alone (and therefore
// the full square coordinate must be supplied), whereas the PGN specification
// does account for it.
public class ChessAlgebraicMoveNotation extends ChessNotation implements IMoveNotation {
    
    private static final String EN_PASSANT_SIGN = " e.p.";
    private static final String ESCAPED_EN_PASSANT_SIGN = " e\\.p\\.";
    
    private static final String DIGIT_CASTLE_SIGN = "0";
    private static final String ESCAPED_DIGIT_CASTLE_SIGN = "0";
    private static final String LETTER_CASTLE_SIGN = "O";
    private static final String ESCAPED_LETTER_CASTLE_SIGN = "O";
    
    private static final String CAPTURE_SIGN = "x";
    private static final String ESCAPED_CAPTURE_SIGN = "x";
    
    private static final String EQUALS_SIGN = "=";
    private static final String ESCAPED_EQUALS_SIGN = "=";
    
    private static final String CHECK_SIGN = "+";
    private static final String ESCAPED_CHECK_SIGN = "\\+";
    
    private static final String HASH_MATE_SIGN = "#";
    private static final String ESCAPED_HASH_MATE_SIGN = "#";
    private static final String PLUSPLUS_MATE_SIGN = "++";
    private static final String ESCAPED_PLUSPLUS_MATE_SIGN = "\\+\\+";
    
    private boolean indicateEnPassant = true;
    
    private String castleSign = DIGIT_CASTLE_SIGN;
    private String escapedCastleSign = ESCAPED_DIGIT_CASTLE_SIGN;
    
    private boolean useEqualsForPromotion = true;
    
    private String checkmateSign = PLUSPLUS_MATE_SIGN;
    private String escapedCheckmateSign = ESCAPED_PLUSPLUS_MATE_SIGN;
    
    private boolean patternsAreDirty = true;  // flags that the patterns below need to be (re)calculated
    private Pattern queensideCastleMovePattern;
    private Pattern kingsideCastleMovePattern;
    private Pattern pieceMovePattern;
    private Pattern pawnMovePattern;

    public ChessAlgebraicMoveNotation() {
    }
    
    public ChessAlgebraicMoveNotation(BidirectionalMap<Class<? extends ChessPiece>, Character> pieceCharacters) {
        super(pieceCharacters);
    }
    
    public boolean getIndicateEnPassant() {
        return indicateEnPassant;
    }

    public void setIndicateEnPassant(boolean indicateEnPassant) {
        this.indicateEnPassant = indicateEnPassant;
        patternsAreDirty = true;
    }

    public boolean getUseLetterOForCastle() {
        return castleSign.equals(LETTER_CASTLE_SIGN);
    }
    
    public void setUseLetterOForCastle(boolean useLetterOForCastle) {
        if(useLetterOForCastle) {
            castleSign = LETTER_CASTLE_SIGN;
            escapedCastleSign = ESCAPED_LETTER_CASTLE_SIGN;
        } else {
            castleSign = DIGIT_CASTLE_SIGN;
            escapedCastleSign = ESCAPED_DIGIT_CASTLE_SIGN;
        }
        patternsAreDirty = true;
    }

    public boolean getUseEqualsForPromotion() {
        return useEqualsForPromotion;
    }

    public void setUseEqualsForPromotion(boolean useEqualsForPromotion) {
        this.useEqualsForPromotion = useEqualsForPromotion;
        patternsAreDirty = true;
    }

    public boolean getUseHashForCheckmate() {
        return checkmateSign.equals(HASH_MATE_SIGN);
    }
    
    public void setUseHashForCheckmate(boolean useHashForCheckmate) {
        if(useHashForCheckmate) {
            checkmateSign = HASH_MATE_SIGN;
            escapedCheckmateSign = ESCAPED_HASH_MATE_SIGN;
        } else {
            checkmateSign = PLUSPLUS_MATE_SIGN;
            escapedCheckmateSign = ESCAPED_PLUSPLUS_MATE_SIGN;
        }
        patternsAreDirty = true;
    }
    
    @Override
    public String getMoveDescription(Move move, Game game) {
        if(!(game instanceof ChessGame)) {
            throw new IllegalArgumentException("Error! Must supply an instance of ChessGame.");
        }
        
        ChessGame chessGame = (ChessGame) game;
        
        if(!chessGame.isLegal(move)) {
            throw new IllegalMoveException();
        }
        
        String description;
        
        if(chessGame.isQueensideCastleMove(move)) {
            description = castleSign + "-" + castleSign + "-" + castleSign;
        } else if(chessGame.isKingsideCastleMove(move)) {
            description = castleSign + "-" + castleSign;
        } else {
            Square origin = move.getOrigin();
            Square destination = move.getDestination();

            ChessPiece piece = (ChessPiece) chessGame.getBoard().getPiece(origin);
            if(piece.getClass() == Pawn.class) {
                if(chessGame.isCaptureMove(move)) {
                    description = getColumnDescription(origin);
                } else {
                    description = "";
                }
            } else {
                description = getPieceDescription(piece.getClass()).toString();
                description += getDisanbiguationString(chessGame, piece, origin, destination);
            }
            
            if(chessGame.isCaptureMove(move)) {
                description += CAPTURE_SIGN;
            }
            
            description += getSquareDescription(destination);
            
            if(indicateEnPassant && chessGame.isEnPassantMove(move)) {
                description += EN_PASSANT_SIGN;
            }
            if(move instanceof PromotionMove) {
                if(useEqualsForPromotion) {
                    description += EQUALS_SIGN;
                }
                Class<? extends PromoteablePiece> promotedPiece = ((PromotionMove) move).getPromotedPiece();
                description += getPieceDescription(promotedPiece);
            }
        }
        
        if(chessGame.isCheckmateMove(move)) {
            description += checkmateSign;
        } else if(chessGame.isCheckMove(move)) {
            description += CHECK_SIGN;
        }
        
        return description;
    }

    @Override
    public Move getMove(String moveDescription, Game game) {
        if(!(game instanceof ChessGame)) {
            throw new IllegalArgumentException("Error! Must supply an instance of ChessGame.");
        }

        ChessGame chessGame = (ChessGame) game;
        
        if(patternsAreDirty) {
            calculatePatterns();
        }
        
        if(queensideCastleMovePattern.matcher(moveDescription).matches()) {
            return parseCastleMove(moveDescription, chessGame, BoardSide.QUEEN_SIDE);
        } else if(kingsideCastleMovePattern.matcher(moveDescription).matches()) {
            return parseCastleMove(moveDescription, chessGame, BoardSide.KING_SIDE);
        } else if(pieceMovePattern.matcher(moveDescription).matches()) {
            return parsePieceMove(moveDescription, chessGame);
        } else if(pawnMovePattern.matcher(moveDescription).matches()) {
            return parsePawnMove(moveDescription, chessGame);
        } else {
            throw new IllegalArgumentException("Error! Incorrect move description: " + moveDescription + ".");
        }
    }

    private String getDisanbiguationString(ChessGame game, ChessPiece piece, Square origin, Square destination) {
        Square[] origins = game.getLegalOriginsWithPieceTo(piece.getClass(), destination);
        if(origins.length > 1) {
            RectangularBoard board = game.getBoard();
            if(!board.hasSameColumn(origins)) {
                return getColumnDescription(origin);
            }
            if(!board.hasSameRow(origins)) {
                return getRowDescription(origin);
            }
            return getSquareDescription(origin);
        } else {
            return "";
        }
    }
    
    private void calculatePatterns() {
        String checkPattern = "((" + ESCAPED_CHECK_SIGN + ")|(" + escapedCheckmateSign + "))?";
        
        queensideCastleMovePattern = Pattern.compile(escapedCastleSign + "-" + escapedCastleSign + "-" +
                escapedCastleSign + checkPattern);
        kingsideCastleMovePattern = Pattern.compile(escapedCastleSign + "-" + escapedCastleSign + checkPattern);
        
        String columnPattern = "[a-h]";
        String rowPattern = "[1-8]";
        String squarePattern = "(" + columnPattern + rowPattern + ")";
        String piecePattern = "([A-Z])";
        
        String disambiguationPattern = "((" + rowPattern + ")|(" + columnPattern + ")|" + squarePattern + ")?";
        String capturePattern = "(" + ESCAPED_CAPTURE_SIGN + ")?";
        
        pieceMovePattern = Pattern.compile(piecePattern + disambiguationPattern + capturePattern +
                squarePattern + checkPattern);
        
        String pawnCapturePattern = "(?:(" + columnPattern + ")(?:" + ESCAPED_CAPTURE_SIGN + ")?)?";
        
        String enPassantPattern = "(";
        if(indicateEnPassant) {
            enPassantPattern += ESCAPED_EN_PASSANT_SIGN;
        }
        enPassantPattern += ")?";
        
        // Note that the = sign is made optional regardless of useEqualsForPromotion is set to true or false.
        // This is to allow parsing of the move from broken PGN games.
        String promotionPattern = "(?:(?:" + ESCAPED_EQUALS_SIGN + ")?" + piecePattern + ")?";
        
        pawnMovePattern = Pattern.compile(pawnCapturePattern + squarePattern + enPassantPattern +
                promotionPattern + checkPattern);
        
        patternsAreDirty = false;
    }

    // Check indication characters are made optional, so that moves that miss them in broken
    // PGN files can still be parsed. However, if provided they must be correct.
    private void verifyCheckIndication(Matcher matcher, int checkPatternGroup, Move move, ChessGame game,
            String errorMessage) {
        String checkIndication = matcher.group(checkPatternGroup);
        if((checkIndication != null) && !checkIndication.equals("")) {
            boolean indicatesCheck = matcher.group(checkPatternGroup + 1) != null;
            boolean isCheck = game.isCheckMove(move) && !game.isCheckmateMove(move);
            if(indicatesCheck && !isCheck) {
                throw new IllegalArgumentException(errorMessage + "this move is not check.");
            } else if(isCheck && !indicatesCheck) {
                throw new IllegalArgumentException(errorMessage + "this move is check and must be indicated as such.");
            }
            
            boolean indicatesCheckmate = matcher.group(checkPatternGroup + 2) != null;
            boolean isCheckmate = game.isCheckmateMove(move);
            if(indicatesCheckmate && !isCheckmate) {
                throw new IllegalArgumentException(errorMessage + "this move is not checkmate.");
            } else if(isCheckmate && !indicatesCheckmate) {
                throw new IllegalArgumentException(errorMessage + "this move is checkmate and must be indicated as such.");
            }
        }
    }
  
    private Move parseCastleMove(String moveDescription, ChessGame game, BoardSide boardSide) {
        String errorMessage = "Error parsing move " + moveDescription + ": ";
        
        Move move = game.getCastleMove(boardSide);
        if(move == null) {
            throw new IllegalArgumentException(errorMessage + "can't castle in this position.");
        }
        
        Pattern castlePattern = boardSide == BoardSide.QUEEN_SIDE ? queensideCastleMovePattern :
                kingsideCastleMovePattern;
        Matcher matcher = castlePattern.matcher(moveDescription);
        matcher.matches();
        
        verifyCheckIndication(matcher, 1, move, game, errorMessage);
        
        return move;
    }

    private Move parsePieceMove(String moveDescription, ChessGame game) {
        String errorMessage = "Error parsing move " + moveDescription + ": ";
        
        Matcher matcher = pieceMovePattern.matcher(moveDescription);
        matcher.matches();
        
        String pieceDescription = matcher.group(1);
        Class<? extends ChessPiece> pieceType = getPieceClass(pieceDescription.charAt(0));
        if(pieceType == null) {
            throw new IllegalArgumentException(errorMessage + "unknown piece type.");
        } else if(pieceType == Pawn.class) {
            throw new IllegalArgumentException(errorMessage + "can't indicate the moving piece in pawn moves.");
        }
        
        String destinationDescription = matcher.group(7);
        Square destination = getSquare(destinationDescription);
        
        Square origin = parsePieceMoveOrigin(matcher, game, pieceType, destination, errorMessage);
        
        Move move = new Move(origin, destination);
        
        // The capture indication character is made optional, in order to allow the move to be parsed
        // even in broken PGN files. However, if indicated then the move must really be a capture.
        boolean indicatesCapture = (matcher.group(6) != null) && !matcher.group(6).equals("");
        if(indicatesCapture && !game.isCaptureMove(move)) {
            throw new IllegalArgumentException(errorMessage + "this move is not a capture.");
        }
        
        verifyCheckIndication(matcher, 8, move, game, errorMessage);
        
        return move;
    }
    
    private Square parsePieceMoveOrigin(Matcher matcher, ChessGame game, Class<? extends ChessPiece> pieceType,
            Square destination, String errorMessage) {
        Square[] candidateOrigins = game.getLegalOriginsWithPieceTo(pieceType, destination);
        if(candidateOrigins.length == 0) {
            errorMessage += "no piece of the specified type can move to the specified destination.";
            throw new IllegalArgumentException(errorMessage);
        } else if(candidateOrigins.length == 1) {
            if((matcher.group(2) != null) && !matcher.group(2).equals("")) {
                errorMessage += "can't disambiguate move since only one piece of the specified type can ";
                errorMessage += "move to the specified destination.";
                throw new IllegalArgumentException(errorMessage);
            }
            return candidateOrigins[0];
        } else {
            if((matcher.group(2) == null ) || matcher.group(2).equals("")) {
                errorMessage += "the move must be disambiguated, since more than one piece of the specified ";
                errorMessage += "type can move to the specified destination.";
                throw new IllegalArgumentException(errorMessage);
            }
            RectangularBoard board = game.getBoard();
            if(!board.hasSameColumn(candidateOrigins)) {
                String columnDescription = matcher.group(4);
                if(columnDescription == null) {
                    errorMessage += "the move must be disambiguated by file (only), since no pieces of the ";
                    errorMessage += "specified type that can move to the specified destination are on the same ";
                    errorMessage += "file.";
                    throw new IllegalArgumentException(errorMessage);
                }
                int column = getColumn(columnDescription);
                return board.getSquareWithColumn(column, candidateOrigins);
            } else if(!board.hasSameRow(candidateOrigins)) {
                String rowDescription = matcher.group(3);
                if(rowDescription == null) {
                    errorMessage += "the move must be disambiguated by rank (only), since no pieces of the ";
                    errorMessage += "specified type that can move to the specified destination are on the same ";
                    errorMessage += "rank.";
                    throw new IllegalArgumentException(errorMessage);
                }
                int row = getRow(rowDescription);
                return board.getSquareWithRow(row, candidateOrigins);
            } else {
                String originDescription = matcher.group(5);
                if(originDescription == null) {
                    errorMessage += "the move must be disambiguated by the full originating square description, ";
                    errorMessage += "since there are some pieces of the selected type that can move to the ";
                    errorMessage += "selected destination that are placed on the same row and same column.";
                    throw new IllegalArgumentException(errorMessage);
                }
                return getSquare(originDescription);
            }
        }
    }

    @SuppressWarnings("unchecked")
    private Move parsePawnMove(String moveDescription, ChessGame game) {
        String errorMessage = "Error parsing move: " + moveDescription + ": ";
        
        Matcher matcher = pawnMovePattern.matcher(moveDescription);
        matcher.matches();
        
        String destinationDescription = matcher.group(2);
        Square destination = getSquare(destinationDescription);

        Square origin = null;
        RectangularBoard board = game.getBoard();
        int rowDifference = game.getState().getColorToMove() == PieceColor.BLACK ? 1 : -1;
        int originRow = destination.getRow() - rowDifference;
        if((originRow >= 0) && (originRow < board.getRowCount())) {
            String originColumnDescription = matcher.group(1);
            if((originColumnDescription == null) || originColumnDescription.equals("")) {
                origin = new Square(originRow, destination.getColumn());
                ChessPiece piece = (ChessPiece) board.getPiece(origin);
                if((piece == null) || !(piece instanceof Pawn)) {
                    originRow = destination.getRow() - 2 * rowDifference;
                    if((originRow >= 0) && (originRow < board.getRowCount())) {
                        origin = new Square(originRow, destination.getColumn());
                    } else {
                        origin = null;
                    }
                }
            } else {
                int originColumn = getColumn(originColumnDescription);
                if(originColumn == destination.getColumn()) {
                    throw new IllegalArgumentException(errorMessage + "this move is not a capture!");
                }
                origin = new Square(originRow, originColumn);
            }
        }
        ChessPiece piece = origin != null ? (ChessPiece) board.getPiece(origin) : null;
        if((origin == null) || (piece == null) || !(piece instanceof Pawn)) {
            errorMessage += "no pawn can move to the specified destination.";
            throw new IllegalArgumentException(errorMessage);
        }
        
        Move move = new Move(origin, destination);
        String promotionPieceDescription = matcher.group(4);
        if(game.isPromotion(move)) {
            if((promotionPieceDescription == null) || (promotionPieceDescription.equals(""))) {
                throw new IllegalArgumentException(errorMessage + "missing promotion piece.");
            }
            Class<? extends ChessPiece> promotionPieceType = getPieceClass(promotionPieceDescription.charAt(0));
            if((promotionPieceType == null) || !(PromoteablePiece.class.isAssignableFrom(promotionPieceType))) {
                throw new IllegalArgumentException(errorMessage + "invalid promotion piece.");
            }
            move = new PromotionMove(move.getOrigin(), move.getDestination(), (Class<? extends PromoteablePiece>) promotionPieceType);
        } else {
            if((promotionPieceDescription != null) && !promotionPieceDescription.equals("")) {
                throw new IllegalArgumentException(errorMessage + "this move is not a promotion move.");
            }
        }
        if(!game.isLegal(move)) {
            errorMessage += "the specified pawn can not move to the specified destination.";
            throw new IllegalArgumentException(errorMessage);
        }
        
        if(indicateEnPassant) {
            boolean indicatesEnPassant = (matcher.group(3) != null) && !matcher.group(3).equals("");
            boolean isEnPassant = game.isEnPassantMove(move);
            if(isEnPassant && !indicatesEnPassant) {
                errorMessage += "the specified move is an en passant capture and must be indicated as such.";
                throw new IllegalArgumentException(errorMessage);
            } else if(indicatesEnPassant && !isEnPassant) {
                errorMessage += "the specified move is not an en passant capture.";
                throw new IllegalArgumentException(errorMessage);
            }
        }
        
        verifyCheckIndication(matcher, 5, move, game, errorMessage);
        
        return move;
    }

}
