/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.model;

import java.util.ArrayList;
import java.util.List;

import com.haundrix.games.chess.model.ChessGameState.BoardSide;
import com.haundrix.games.model.Game;
import com.haundrix.games.model.GameOutcome;
import com.haundrix.games.model.GameResult;
import com.haundrix.games.model.Move;
import com.haundrix.games.model.Square;
import com.haundrix.games.model.SquareColor;
import com.haundrix.games.model.Piece.PieceColor;
import com.haundrix.utils.Utils;

public class ChessGame extends Game {
    
    public static final int ROW_COUNT = 8;
    public static final int COLUMN_COUNT = 8;
    
    private final int initialPlyClock;
    
    private final boolean autoClaimDrawByPositionRepetition;
    private final boolean autoClaimDrawBy50MoveRule;
    
    @SuppressWarnings("unchecked")
    public static final Class<? extends PromoteablePiece>[] PROMOTEABLE_PIECES = new Class[] {
        Queen.class, Rook.class, Bishop.class, Knight.class};
    
    public ChessGame() {
        this(false, false);
    }
    
    public ChessGame(boolean autoClaimDrawByPositionRepetition, boolean autoClaimDrawBy50MoveRule) {
        super(new ChessGameState(), 0);
        initialPlyClock = 0;
        
        this.autoClaimDrawByPositionRepetition = autoClaimDrawByPositionRepetition;
        this.autoClaimDrawBy50MoveRule = autoClaimDrawBy50MoveRule;
        
        ((ChessGameState) gameState).setGame(this);
    }
    
    // TODO Clone the supplied game state in order to cut the links to the outside world.
    public ChessGame(ChessGameState gameState, int initialPlyCount, int initialPlyClock) {
        this(gameState, initialPlyCount, initialPlyClock, false, false);
    }

    // TODO Clone the supplied game state in order to cut the links to the outside world.
    public ChessGame(ChessGameState gameState, int initialPlyCount, int initialPlyClock,
            boolean autoClaimDrawByPositionRepetition, boolean autoClaimDrawBy50MoveRule) {
        super(gameState, initialPlyCount);
        this.initialPlyClock = initialPlyClock;
        
        this.autoClaimDrawByPositionRepetition = autoClaimDrawByPositionRepetition;
        this.autoClaimDrawBy50MoveRule = autoClaimDrawBy50MoveRule;

        ((ChessGameState) gameState).setGame(this);
        
        result = evaluateGameResult();
    }

    public int getInitialPlyClock() {
        return initialPlyClock;
    }

    public boolean isAutoClaimDrawByPositionRepetition() {
        return autoClaimDrawByPositionRepetition;
    }

    public boolean isAutoClaimDrawBy50MoveRule() {
        return autoClaimDrawBy50MoveRule;
    }

    // TODO Document this API so that clients get warned that the returned object is just a clone.
    //      Tell them the valid ways to change the game's internal state (game.move() or the navigation API).
    @Override
    public ChessGameState getState() {
        return ((ChessGameState) gameState).clone();
    }

    @Override
    public ChessGameResult getResult() {
        return (ChessGameResult) super.getResult();
    }

    @Override
    public ChessGameResult getResult(Move move) {
        return (ChessGameResult) super.getResult(move);
    }
    
    @Override
    public void setResult(GameResult result) {
        if(result instanceof ChessGameResult) {
            super.setResult(result);
        } else if((result.getClass() == GameResult.class) && (result.getResultReason() == null)) {
            ChessGameResult chessResult = new ChessGameResult(result.getOutcome());
            super.setResult(chessResult);
        } else {
            throw new IllegalArgumentException("Error! Invalid result argument.");
        }
    }

    @Override
    public boolean equals(Object otherObject) {
        if(!super.equals(otherObject)) {
            return false;
        }
        
        if(!(otherObject instanceof ChessGame)) {
            return false;
        }
        
        return initialPlyClock == ((ChessGame) otherObject).getInitialPlyClock();
    }

    @Override
    public ChessGame clone() {
        ChessGame clone = (ChessGame) super.clone();
        ((ChessGameState) clone.gameState).setGame(clone);
        
        return clone;
    }

    // TODO return a clone?
    public ChessGameState getInitialPosition() {
        int currentPosition = getCurrentPosition();
        doGoToBegin();
        ChessGameState gameState = getState();
        doGoTo(currentPosition);
        return gameState;
    }
    
    public void resign(PieceColor resigningColor) {
        if(isOver()) {
            throw new IllegalStateException("Error! The game is already ended.");
        }
        
        GameOutcome outcome = resigningColor == PieceColor.WHITE ? GameOutcome.BLACK_WINS : GameOutcome.WHITE_WINS;
        setResult(new ChessGameResult(outcome, ChessGameResultReason.RESIGNATION));
    }

    public void callTimeFlag(PieceColor timeForfeitingColor) {
        if(isOver()) {
            throw new IllegalStateException("Error! The game is already ended.");
        }
        
        GameOutcome outcome = timeForfeitingColor == PieceColor.WHITE ? GameOutcome.BLACK_WINS : GameOutcome.WHITE_WINS;
        setResult(new ChessGameResult(outcome, ChessGameResultReason.TIME_FORFEIT));
    }
    
    public boolean claimDrawByPositionRepetition() {
        if(isOver()) {
            throw new IllegalStateException("Error! The game is already ended.");
        }

        int currentPosition = getCurrentPosition();
        doGoToEnd();
        
        if(isDrawByPositionRepetition()) {
            doGoTo(currentPosition);
            setResult(new ChessGameResult(GameOutcome.DRAW, ChessGameResultReason.POSITION_REPETITION));
            return true;
        } else {
            doGoTo(currentPosition);
            return false;
        }
    }
    
    public boolean claimDrawBy50MoveRule() {
        if(isOver()) {
            throw new IllegalStateException("Error! The game is already ended.");
        }
        
        int currentPosition = getCurrentPosition();
        doGoToEnd();
        
        if(isDrawBy50MoveRule()) {
            doGoTo(currentPosition);
            setResult(new ChessGameResult(GameOutcome.DRAW, ChessGameResultReason.FIFTY_MOVE_RULE));
            return true;
        } else {
            doGoTo(currentPosition);
            return false;
        }
    }

    public void setDrawByCommonAgreement() {
        if(isOver()) {
            throw new IllegalStateException("Error! The game is already ended.");
        }
        
        setResult(new ChessGameResult(GameOutcome.DRAW, ChessGameResultReason.COMMON_AGREEMENT));
    }

    @Override
    public boolean isLegal(Move move) {
        Square origin = move.getOrigin();
        Square destination = move.getDestination();
        
        getBoard().checkBounds(origin, destination);
        
        if(origin.equals(destination)) {
            return false;
        }
        if(!isValidMoveOrigin(origin)) {
            return false;
        }
        if(!isValidMoveDestination(destination)) {
            return false;
        }
        
        ChessPiece movingPiece = (ChessPiece) getBoard().getPiece(origin);
        if(!movingPiece.canMove(move)) {
            return false;
        }
        if(moveExposesKing(move)) {
            return false;
        }
        
        return true;
    }

    @Override
    public Square[] getLegalDestinationsFrom(Square origin) {
        getBoard().checkBounds(origin);
        
        if(!isValidMoveOrigin(origin)) {
            return new Square[0];
        }
        
        ChessPiece piece = (ChessPiece) getBoard().getPiece(origin);
        List<Square> candidateDestinations = piece.getDestinationsFrom(origin);
        List<Square> validatedDestinations = new ArrayList<Square>();
        for(Square candidateDestination : candidateDestinations) {
            if(!isValidMoveDestination(candidateDestination)) {
                continue;
            }
            Move candidateMove = new Move(origin, candidateDestination);
            if(piece instanceof Pawn) {
                candidateMove = ((Pawn) piece).getDefaultPromotionMove(candidateMove);
            }
            if(moveExposesKing(candidateMove)) {
                continue;
            }
            validatedDestinations.add(candidateDestination);
        }
        
        return validatedDestinations.toArray(new Square[0]);
    }

    @Override
    public Square[] getLegalOriginsTo(Square destination) {
        getBoard().checkBounds(destination);
        
        if(!isValidMoveDestination(destination)) {
            return new Square[0];
        }
        
        List<Square> origins = new ArrayList<Square>();
        for(Square candidateOrigin : getBoard()) {
            if(!candidateOrigin.equals(destination) && isValidMoveOrigin(candidateOrigin)) {
                ChessPiece piece = (ChessPiece) getBoard().getPiece(candidateOrigin);
                Move candidateMove = new Move(candidateOrigin, destination);
                if(piece instanceof Pawn) {
                    candidateMove = ((Pawn) piece).getDefaultPromotionMove(candidateMove);
                }
                if(piece.canMove(candidateMove) && !moveExposesKing(candidateMove)) {
                    origins.add(candidateOrigin);
                }
            }
        }
        
        return origins.toArray(new Square[0]);
    }
    
    @Override
    public Move[] getAllLegalMoves() {
        List<Move> legalMoves = new ArrayList<Move>();
        for(Move move : super.getAllLegalMoves()) {
            if(isPromotion(move)) {
                Pawn pawn = (Pawn) getBoard().getPiece(move.getOrigin());
                legalMoves.add(pawn.getDefaultPromotionMove(move));
            } else {
                legalMoves.add(move);
            }
        }
        
        return legalMoves.toArray(new Move[0]);
    }

    @Override
    public Square getAttackedSquare(Move move) {
        ChessPiece piece = (ChessPiece) getBoard().getPiece(move.getOrigin());
        if(piece == null) {
            return null;
        }
        
        if(piece instanceof Pawn) {
            move = ((Pawn) piece).getDefaultPromotionMove(move);
        }
        if(!isLegal(move)) {
            return null;
        }
        
        return piece.getAttackedSquare(move);
    }

    public Square[] getLegalOriginsWithPieceTo(Class<? extends ChessPiece> pieceType, Square destination) {
        getBoard().checkBounds(destination);
        
        if(!isValidMoveDestination(destination)) {
            return new Square[0];
        }
        
        Square[] candidateOrigins = getBoard().findPieces(pieceType, gameState.getColorToMove());
        List<Square> legalOrigins = new ArrayList<Square>();
        for(Square candidateOrigin : candidateOrigins) {
            ChessPiece piece = (ChessPiece) getBoard().getPiece(candidateOrigin);
            Move candidateMove = new Move(candidateOrigin, destination);
            if(pieceType == Pawn.class) {
                candidateMove = ((Pawn) piece).getDefaultPromotionMove(candidateMove); 
            }
            if(piece.canMove(candidateMove) && !moveExposesKing(candidateMove)) {
                legalOrigins.add(candidateOrigin);
            }
        }
        
        return legalOrigins.toArray(new Square[0]);
    }

    // This method will return true even in the special case when it's check mate.
    // Clients who care about this should call isCheckMate() first prior to calling this method.
    public boolean isCheck() {
        return isKingAttacked(gameState.getColorToMove().getOppositeColor());
    }
    
    // This method will return true even in the special case when it's a check mate move.
    public boolean isCheck(Move move) {
        doMakeMove(move);
        boolean isCheckMove = isCheck();
        doUnmakePreviousMove();
        
        return isCheckMove;
    }

    // This method will return true even in the special case when it's a check mate move.
    // Clients who care about this should call isCheckMateMove(Move) first prior to calling this method.
    public boolean isCheckMove(Move move) {
        checkLegal(move);
        return isCheck(move);
    }
    
    public boolean previousMoveWasCheck() {
        return isCheck();
    }

    public boolean isCheckmate() {
        return isCheck() && !existLegalMove();
    }

    public boolean isCheckmateMove(Move move) {
        checkLegal(move);
        
        doMakeMove(move);
        boolean isCheckmateMove = isCheckmate();
        doUnmakePreviousMove();
        
        return isCheckmateMove;
    }

    public boolean isGameEndingMove(Move move) {
        checkLegal(move);
        
        doMakeMove(move);
        boolean isGameEndingMove = result != null;
        doUnmakePreviousMove();
        
        return isGameEndingMove;
    }

    public Square getCheckedKingSquare() {
        if(!isCheck()) {
            return null;
        }
        
        return getKingSquare(gameState.getColorToMove());
    }
    
    public boolean isEnPassantMove(Move move) {
        checkLegal(move);
        
        ChessPiece piece = (ChessPiece) getBoard().getPiece(move.getOrigin());
        if(!(piece instanceof Pawn)) {
            return false;
        }
        
        return ((Pawn) piece).isEnPassantMove(move);
    }

    public boolean isCastle(Move move) {
        ChessPiece piece = (ChessPiece) getBoard().getPiece(move.getOrigin());
        if(!(piece instanceof King)) {
            return false;
        }
        
        return ((King) piece).isCastleMove(move);
    }

    public boolean isCastleMove(Move move) {
        return isQueensideCastleMove(move) || isKingsideCastleMove(move);
    }

    public boolean previousMoveWasCastle() {
        doGoToPrevious();
        Move previousMove = getNextMove();
        boolean previousMoveWasCastle = isCastle(previousMove);
        doGoToNext();
        
        return previousMoveWasCastle;
    }

    public boolean isQueensideCastleMove(Move move) {
        checkLegal(move);
        
        ChessPiece piece = (ChessPiece) getBoard().getPiece(move.getOrigin());
        if(!(piece instanceof King)) {
            return false;
        }
        
        return ((King) piece).isQueensideCastleMove(move);
    }

    public boolean isKingsideCastleMove(Move move) {
        checkLegal(move);
        
        ChessPiece piece = (ChessPiece) getBoard().getPiece(move.getOrigin());
        if(!(piece instanceof King)) {
            return false;
        }
        
        return ((King) piece).isKingsideCastleMove(move);
    }

    public Move getCastleMove(BoardSide boardSide) {
        Square origin = ChessGameState.getKingStartSquare(gameState.getColorToMove());
        int destinationColumn = boardSide == BoardSide.QUEEN_SIDE ? origin.getColumn() - 2 :
            origin.getColumn() + 2;
        Square destination = new Square(origin.getRow(), destinationColumn);
        Move move = new Move(origin, destination);
        return isLegal(move) ? move : null;
    }

    public Move getRookCastleMove(Move move) {
        if(!isCastleMove(move)) {
            throw new IllegalArgumentException("Error! The supplied move is not a castle move.");
        }
        
        King king = (King) getBoard().getPiece(move.getOrigin());
        return king.getRookCastleMove(move);
    }

    public Move getPreviousRookCastleMove() {
        doGoToPrevious();
        Move previousMove = getNextMove();
        Move previousRookCastleMove = getRookCastleMove(previousMove);
        doGoToNext();
        
        return previousRookCastleMove;
    }

    public boolean isPromotion(Move move) {
        ChessPiece piece = (ChessPiece) getBoard().getPiece(move.getOrigin());
        if(!(piece instanceof Pawn)) {
            return false;
        }
        
        move = ((Pawn) piece).getDefaultPromotionMove(move);
        return move instanceof PromotionMove;
    }

    public boolean previousMoveWasPromotion() {
        return getPreviousMove() instanceof PromotionMove;
    }
    
    public boolean isKnightCapture(Move move) {
        if(!movingPieceIsKnight(move)) {
            return false;
        }
        
        return isCapture(move);
    }
    
    public boolean isKnightCaptureMove(Move move) {
        checkLegal(move);
        return isKnightCapture(move);
    }

    public boolean previousMoveWasKnightCapture() {
        doGoToPrevious();
        Move previousMove = getNextMove();
        boolean previousMoveWasKnightCapture = isKnightCapture(previousMove);
        doGoToNext();
        
        return previousMoveWasKnightCapture;
    }

    // Returns true if the move is a knight check or multiple attack.
    public boolean isKnightAttack(Move move) {
        if(!movingPieceIsKnight(move)) {
            return false;
        }
        
        if(isCheck(move)) {
            return true;
        }
        
        // let's temporarily make the move and then pretend it's the same color to move again!
        doMakeMove(move);
        gameState.switchColorToMove();
        int attackedPiecesCount = 0;
        Square knightOrigin = move.getDestination();
        for(Square knightDestination : getLegalDestinationsFrom(knightOrigin)) {
            Square attackedSquare = getAttackedSquare(new Move(knightOrigin, knightDestination));
            ChessPiece attackedPiece = (ChessPiece) gameState.getBoard().getPiece(attackedSquare);
            if((attackedPiece != null) && ((attackedPiece instanceof Rook) || (attackedPiece instanceof Queen))) {
                attackedPiecesCount++;
            }
        }
        gameState.switchColorToMove();
        doUnmakePreviousMove();
        
        return attackedPiecesCount > 1;
    }

    // Returns true if the move is a knight check or multiple attack.
    public boolean isKnightAttackMove(Move move) {
        checkLegal(move);
        return isKnightAttack(move);
    }

    public boolean previousMoveWasKnightAttack() {
        doGoToPrevious();
        Move previousMove = getNextMove();
        boolean previousMoveWasKnightAttack = isKnightAttack(previousMove);
        doGoToNext();
        
        return previousMoveWasKnightAttack;
    }

    public int getCurrentPlyClock() {
        int plyClock = 0;
        
        int currentPosition = getCurrentPosition();
        boolean moveIsPlyClock = true;
        while(!isAtBegin()) {
            doGoToPrevious();
            if(isPlyClockMove(getNextMove())) {
                plyClock++;
            } else {
                moveIsPlyClock = false;
                break;
            }
        }
        if(isAtBegin() && moveIsPlyClock) {
            plyClock += initialPlyClock;
        }
        doGoTo(currentPosition);
        
        return plyClock;
    }
    
    @Override
    protected void doMove(Move move) {
        ChessPiece piece = (ChessPiece) getBoard().getPiece(move.getOrigin());
        piece.move(move);
        
        gameState.switchColorToMove();
    }
    
    @Override
    protected void doUnmove(Move move) {
        ChessPiece movedPiece = (ChessPiece) getBoard().getPiece(move.getDestination());
        movedPiece.unmove(move);
        
        gameState.switchColorToMove();
    }
    
    @Override
    protected ChessGameResult evaluateGameResult() {
        if(!existLegalMove()) {
            if(isCheck()) {
                GameOutcome outcome = gameState.getColorToMove() == PieceColor.BLACK ?
                        GameOutcome.WHITE_WINS : GameOutcome.BLACK_WINS;
                return new ChessGameResult(outcome, ChessGameResultReason.CHECKMATE);
            } else {
                return new ChessGameResult(GameOutcome.DRAW, ChessGameResultReason.STALEMATE);
            }
        } else if(isDeadPosition()) {
            return new ChessGameResult(GameOutcome.DRAW, ChessGameResultReason.DEAD_POSITION);
        } else if(autoClaimDrawByPositionRepetition && isDrawByPositionRepetition()) {
            return new ChessGameResult(GameOutcome.DRAW, ChessGameResultReason.POSITION_REPETITION);
        } else if(autoClaimDrawBy50MoveRule && isDrawBy50MoveRule()) {
            return new ChessGameResult(GameOutcome.DRAW, ChessGameResultReason.FIFTY_MOVE_RULE);
        } else {
            return null;
        }
    }

    boolean isSquareAttacked(Square attackedSquare, PieceColor attackingPieceColor) {
        for(Square attackingPieceSquare : getSquaresOccupiedByPiecesOfColor(attackingPieceColor)) {
            if(!attackingPieceSquare.equals(attackedSquare)) {
                ChessPiece attackingPiece = (ChessPiece) getBoard().getPiece(attackingPieceSquare);
                if(attackingPiece.attacksSquare(attackingPieceSquare, attackedSquare)) {
                    return true;
                }
            }
        }
        
        return false;
    }

    private boolean isValidMoveOrigin(Square origin) {
        ChessPiece movingPiece = (ChessPiece) getBoard().getPiece(origin);
        if(movingPiece == null) {
            return false;
        }
        if(movingPiece.getColor() != gameState.getColorToMove()) {
            return false;
        }
        return true;
    }

    private boolean isValidMoveDestination(Square destination) {
        ChessPiece targetPiece = (ChessPiece) getBoard().getPiece(destination);
        if((targetPiece == null) || (targetPiece.getColor() != gameState.getColorToMove())) {
            return true;
        } else {
            return false;
        }
    }

    private Square getKingSquare(PieceColor kingColor) {
        return getBoard().findPiece(King.class, kingColor);
    }

    private boolean isKingAttacked(PieceColor attackingPieceColor) {
        Square kingSquare = getKingSquare(attackingPieceColor.getOppositeColor());
        
        if(kingSquare == null) {
            throw new AssertionError("Error! King not found on board.");
        }
        
        return isSquareAttacked(kingSquare, attackingPieceColor);
    }

    private boolean moveExposesKing(Move move) {
        doMove(move);
        boolean exposesKing = isKingAttacked(gameState.getColorToMove());
        doUnmove(move);
        
        return exposesKing;
    }

    private boolean movingPieceIsKnight(Move move) {
        ChessPiece movingPiece = (ChessPiece) gameState.getBoard().getPiece(move.getOrigin());
        return (movingPiece != null) && (movingPiece instanceof Knight);
    }

    private boolean isPlyClockMove(Move move) {
        ChessPiece movingPiece = (ChessPiece) getBoard().getPiece(move.getOrigin());
        boolean isPawnMove = (movingPiece != null) && (movingPiece.getClass() == Pawn.class);
        
        return !isPawnMove && !isCaptureMove(move);
    }

    /*
     * A dead position is reached when neither player can give checkmate,
     * no matter how unskillfully his opponent plays.
     * 
     * This happens with the following pieces on board:
     * - king versus king,
     * - king and knight versus king,
     * - king and bishop versus king,
     *   - king and several bishops of the same color versus king,
     *   - king and several bishops versus king and several bishops, all bishops having the same color.
     */
    private boolean isDeadPosition() {
        Square[] whitePieceSquares = getSquaresOccupiedByPiecesOfColor(PieceColor.WHITE);
        Square[] blackPieceSquares = getSquaresOccupiedByPiecesOfColor(PieceColor.BLACK);
        
        if(isKingVersusKingPosition(whitePieceSquares, blackPieceSquares)) {
            return true;
        } else if(isKingAndKnightVersusKingPosition(whitePieceSquares, blackPieceSquares)) {
            return true;
        } else if(isSameColorBishopsPosition(whitePieceSquares, blackPieceSquares)) {
            return true;
        } else {
            return false;
        }
    }

    private boolean isKingVersusKingPosition(Square[] aColorPieceSquares, Square[] otherColorPieceSquares) {
        return (aColorPieceSquares.length == 1) && (otherColorPieceSquares.length == 1);
    }

    private boolean isKingAndKnightVersusKingPosition(Square[] aColorPieceSquares, Square[] otherColorPieceSquares) {
        Square[] kingAndKnightSideSquares;
        if((aColorPieceSquares.length == 1) && (otherColorPieceSquares.length == 2)) {
            kingAndKnightSideSquares = otherColorPieceSquares;
        } else if((aColorPieceSquares.length == 2) && (otherColorPieceSquares.length == 1)) {
            kingAndKnightSideSquares = aColorPieceSquares;
        } else {
            return false;
        }
        
        boolean firstPieceIsKnight = getBoard().getPiece(kingAndKnightSideSquares[0]).getClass() == Knight.class;
        boolean secondPieceIsKnight = getBoard().getPiece(kingAndKnightSideSquares[1]).getClass() == Knight.class;
        return firstPieceIsKnight || secondPieceIsKnight;
    }

    private boolean isSameColorBishopsPosition(Square[] aColorPieceSquares, Square[] otherColorPieceSquares) {
        SquareColor bishopSquareColor = null;
        Square[] allSquares = Utils.addAll(aColorPieceSquares, otherColorPieceSquares);
        for(Square square : allSquares) {
            Class<?> pieceClass = getBoard().getPiece(square).getClass();
            if(pieceClass == King.class) {
                continue;
            } else if(pieceClass == Bishop.class) {
                if(bishopSquareColor == null) {
                    bishopSquareColor = getBoard().getSquareColor(square);
                } else {
                    if(getBoard().getSquareColor(square) != bishopSquareColor) {
                        return false;
                    }
                }
            } else {
                return false;
            }
        }
        
        if(bishopSquareColor == null) {
            return false;
        }
        
        return true;
    }

    private boolean isDrawByPositionRepetition() {
        ChessGameState currentGameState = ((ChessGameState) gameState).clone();
        int repetitionCount = 1;
        
        int currentPlyClock = getCurrentPlyClock();
        for(int i = 0; i < currentPlyClock; i++) {
            if(isAtBegin()) {
                doGoToEnd();
                return false;
            }
            
            doGoToPrevious();
            if(gameState.equals(currentGameState)) {
                if(++repetitionCount == 3) {
                    doGoToEnd();
                    return true;
                }
            }
        }
        
        doGoToEnd();
        return false;
    }

    private boolean isDrawBy50MoveRule() {
        return getCurrentPlyClock() >= 100;
    }

}
