/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.model;

import com.haundrix.games.model.GameState;
import com.haundrix.games.model.Piece;
import com.haundrix.games.model.RectangularBoard;
import com.haundrix.games.model.Square;
import com.haundrix.games.model.Piece.PieceColor;
import com.haundrix.utils.Utils;

public class ChessGameState extends GameState {
    
    public enum BoardSide {
        QUEEN_SIDE,
        KING_SIDE
    }

    private Square enPassantSquare;
    
    private ChessGame game;
    
    public ChessGameState() {
        super(new RectangularBoard(ChessGame.ROW_COUNT, ChessGame.COLUMN_COUNT), PieceColor.WHITE);
        
        board.setPiece(new Rook(PieceColor.BLACK, this, false), 0, 0);
        board.setPiece(new Knight(PieceColor.BLACK, this), 0, 1);
        board.setPiece(new Bishop(PieceColor.BLACK, this), 0, 2);
        board.setPiece(new Queen(PieceColor.BLACK, this), 0, 3);
        board.setPiece(new King(PieceColor.BLACK, this, false), 0, 4);
        board.setPiece(new Bishop(PieceColor.BLACK, this), 0, 5);
        board.setPiece(new Knight(PieceColor.BLACK, this), 0, 6);
        board.setPiece(new Rook(PieceColor.BLACK, this, false), 0, 7);
        for(int column = 0; column < 8; column++) {
            board.setPiece(new Pawn(PieceColor.BLACK, this), 1, column);
        }
        board.setPiece(new Rook(PieceColor.WHITE, this, false), 7, 0);
        board.setPiece(new Knight(PieceColor.WHITE, this), 7, 1);
        board.setPiece(new Bishop(PieceColor.WHITE, this), 7, 2);
        board.setPiece(new Queen(PieceColor.WHITE, this), 7, 3);
        board.setPiece(new King(PieceColor.WHITE, this, false), 7, 4);
        board.setPiece(new Bishop(PieceColor.WHITE, this), 7, 5);
        board.setPiece(new Knight(PieceColor.WHITE, this), 7, 6);
        board.setPiece(new Rook(PieceColor.WHITE, this, false), 7, 7);
        for(int column = 0; column < 8; column++) {
            board.setPiece(new Pawn(PieceColor.WHITE, this), 6, column);
        }

        enPassantSquare = null;
    }
    
    // TODO For the moment no check is made that the supplied board position is legal!
    public ChessGameState(RectangularBoard board, PieceColor colorToMove, Square enPassantSquare) {
        super(board, colorToMove);

        if((board.getRowCount() != ChessGame.ROW_COUNT) || (board.getColumnCount() != ChessGame.COLUMN_COUNT)) {
            throw new IllegalArgumentException("Error! The supplied board has an incorrect number of rows or columns.");
        }
        
        this.enPassantSquare = enPassantSquare;
    }

    public Square getEnPassantSquare() {
        return enPassantSquare;
    }

    public void setEnPassantSquare(Square enPassantSquare) {
        this.enPassantSquare = enPassantSquare;
    }

    public ChessGame getGame() {
        return game;
    }

    public void setGame(ChessGame game) {
        this.game = game;
    }

    @Override
    public boolean equals(Object otherObject) {
        if(!super.equals(otherObject)) {
            return false;
        }
        
        if(!(otherObject instanceof ChessGameState)) {
            return false;
        }
        
        ChessGameState otherGameState = (ChessGameState) otherObject;
        return Utils.nullableEquals(enPassantSquare, otherGameState.getEnPassantSquare());
    }

    public boolean equalsIgnoreMoved(ChessGameState otherGameState) {
        RectangularBoard otherBoard = otherGameState.getBoard();
        
        if(board.getRowCount() != otherBoard.getRowCount()) {
            return false;
        }
        if(board.getColumnCount() != otherBoard.getColumnCount()) {
            return false;
        }
        
        for(int row = 0; row < board.getRowCount(); row++) {
            for(int column = 0; column < board.getColumnCount(); column++) {
                ChessPiece piece = (ChessPiece) board.getPiece(row, column);
                ChessPiece otherPiece = (ChessPiece) otherBoard.getPiece(row, column);
                boolean samePiece;
                if(piece == null) {
                    samePiece = otherPiece == null;
                } else {
                    if(piece instanceof Rook) {
                        samePiece = ((Rook) piece).equalsIgnoreMoved(otherPiece);
                    } else if(piece instanceof King) {
                        samePiece = ((King) piece).equalsIgnoreMoved(otherPiece);
                    } else {
                        samePiece = piece.equals(otherPiece);
                    }
                }
                if(!samePiece) {
                    return false;
                }
            }
        }
        
        if(!Utils.nullableEquals(enPassantSquare, otherGameState.getEnPassantSquare())) {
            return false;
        }
        
        return isSameCastlePossibility(PieceColor.BLACK, BoardSide.QUEEN_SIDE, otherGameState) &&
            isSameCastlePossibility(PieceColor.BLACK, BoardSide.KING_SIDE, otherGameState) &&
            isSameCastlePossibility(PieceColor.WHITE, BoardSide.QUEEN_SIDE, otherGameState) &&
            isSameCastlePossibility(PieceColor.WHITE, BoardSide.KING_SIDE, otherGameState);
    }

    @Override
    public ChessGameState clone() {
        ChessGameState clone;
        clone = (ChessGameState) super.clone();
        
        RectangularBoard clonedBoard = clone.board;
        for(Square square : clonedBoard) {
            ChessPiece piece = (ChessPiece) clonedBoard.getPiece(square);
            if(piece != null) {
                piece.setGameState(clone);
            }
        }
        
        clone.game = new ChessGame(clone, game.getCurrentPlyCount(), game.getCurrentPlyClock());
        
        return clone;
    }
    
    public static Square getKingStartSquare(PieceColor kingColor) {
        return new Square(ChessGameState.getPiecesStartRow(kingColor), 4);
    }
    
    public static Square getRookStartSquare(PieceColor rookColor, BoardSide boardSide) {
        int rookRow = ChessGameState.getPiecesStartRow(rookColor);
        int rookColumn = boardSide == BoardSide.QUEEN_SIDE ? 0 : ChessGame.COLUMN_COUNT - 1;
        return new Square(rookRow, rookColumn);
    }
    
    public boolean isCastlePossible(PieceColor castlingSideColor, BoardSide boardSide) {
        Piece piece = board.getPiece(ChessGameState.getKingStartSquare(castlingSideColor));
        if((piece == null) || !(piece instanceof King) || ((King) piece).isMoved()) {
            return false;
        }
        
        piece = board.getPiece(ChessGameState.getRookStartSquare(castlingSideColor, boardSide));
        return (piece != null) && (piece instanceof Rook) && !((Rook) piece).isMoved();
    }

    // Warning: this method only sets the castling availability if the rook is still in its starting square.
    //          It is an error to call RectangularBoard.setPiece(Piece, Square) passing a non-moved King or Rook
    //          and a square other than their starting square.
    public void setCastlePossible(PieceColor castlingSideColor, BoardSide boardSide, boolean isCastlePossible) {
        Piece piece = board.getPiece(ChessGameState.getRookStartSquare(castlingSideColor, boardSide));
        boolean noRookThere = (piece == null) || !(piece instanceof Rook);
        
        if(noRookThere) {
            if(isCastlePossible) {
                throw new IllegalArgumentException("Error! The rook has already moved.");
            }
        } else {
            ((Rook) piece).setMoved(!isCastlePossible);
            if(isCastlePossible) {
                piece = board.getPiece(ChessGameState.getKingStartSquare(castlingSideColor));
                boolean noKingThere = (piece == null) || !(piece instanceof King);
                if(noKingThere) {
                    throw new IllegalArgumentException("Error! The king has already moved.");
                }
                ((King) piece).setMoved(false);
            }
        }
    }

    private static int getPiecesStartRow(PieceColor color) {
        return color == PieceColor.BLACK ? 0 : 7;
    }

    private boolean isSameCastlePossibility(PieceColor castlingSideColor, BoardSide boardSide,
            ChessGameState otherGameState) {
        return isCastlePossible(castlingSideColor, boardSide) ==
            otherGameState.isCastlePossible(castlingSideColor, boardSide); 
    }

}
