/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.model;

import java.util.List;

import com.haundrix.games.model.Piece;
import com.haundrix.games.model.Move;
import com.haundrix.games.model.RectangularBoard;
import com.haundrix.games.model.Square;

public abstract class ChessPiece extends Piece {
    
    protected enum UnmoveInfoProperties {
        CAPTURED_PIECE,
        EN_PASSANT_SQUARE
    }
    
    protected ChessGameState gameState;
    
    public ChessPiece(PieceColor color, ChessGameState gameState) {
        super(color);
        this.gameState = gameState;
    }
    
    public abstract boolean canMove(Move move);
    
    public abstract List<Square> getDestinationsFrom(Square origin);

    @Override
    public boolean equals(Object object) {
        if(!super.equals(object)) {
            return false;
        }
        
        return object.getClass() == getClass();
    }

    @Override
    public ChessPiece clone() {
        return (ChessPiece) super.clone();
    }
    
    protected RectangularBoard getBoard() {
        return gameState.getBoard();
    }
    
    // Implementations of this method may assume the move is legal.
    protected void move(Move move) {
        checkMove(move);

        Square origin = move.getOrigin();
        Square destination = move.getDestination();

        ChessPiece capturedPiece = (ChessPiece) getBoard().getPiece(destination);
        move.addUnmoveInfo(UnmoveInfoProperties.CAPTURED_PIECE, capturedPiece);
        move.addUnmoveInfo(UnmoveInfoProperties.EN_PASSANT_SQUARE, gameState.getEnPassantSquare());
        
        getBoard().setPiece(null, origin);
        getBoard().setPiece(this, destination);
        gameState.setEnPassantSquare(null);
    }

    // sanity check utility method
    protected void checkMove(Move move) {
        if(getBoard().getPiece(move.getOrigin()) != this) {
            throw new IllegalArgumentException("Error moving piece: the origin square is not occupied by this piece.");
        }
    }
    
    // Implementations of this method may assume the move is legal.
    // The supplied instance of Move must be the same one passed to a previous call to move(Move).
    protected void unmove(Move move) {
        checkUnmove(move);

        ChessPiece capturedPiece = (ChessPiece) move.getUnmoveInfo(UnmoveInfoProperties.CAPTURED_PIECE);
        getBoard().setPiece(capturedPiece, move.getDestination());
        getBoard().setPiece(this, move.getOrigin());
        
        Square enPassantSquare = (Square) move.getUnmoveInfo(UnmoveInfoProperties.EN_PASSANT_SQUARE);
        gameState.setEnPassantSquare(enPassantSquare);
        
        move.clearUnmoveInfos();
    }

    // sanity check utility method
    protected void checkUnmove(Move move) {
        if(getBoard().getPiece(move.getDestination()) != this) {
            throw new IllegalArgumentException("Error unmoving piece: the destination square is not occupied by this piece.");
        }
        if(getBoard().getPiece(move.getOrigin()) != null) {
            throw new IllegalArgumentException("Error unmoving piece: the origin square is occupied.");
        }
    }
    
    protected boolean attacksSquare(Square pieceSquare, Square attackedSquare) {
        return canMove(new Move(pieceSquare, attackedSquare));
    }

    // This method may assume the move is legal.
    protected Square getAttackedSquare(Move move) {
        return move.getDestination();
    }
    
    // for using when cloning the piece
    void setGameState(ChessGameState gameState) {
        this.gameState = gameState;
    }
    
}
