/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.model;

import java.util.regex.Pattern;

import org.apache.commons.lang.CharUtils;

import com.haundrix.games.chess.model.ChessGameState.BoardSide;
import com.haundrix.games.model.RectangularBoard;
import com.haundrix.games.model.Square;
import com.haundrix.games.model.Piece.PieceColor;

public class ForsythEdwardsNotation extends ChessNotation {

    private static final String NULL_DESCRIPTION = "-";
    
    private static final String ROW_DESCRIPTION_SEPARATOR = "/";
    
    private static final String BLACK_DESCRIPTION = "b";
    private static final String WHITE_DESCRIPTION = "w";

    private static final String WHITE_KINGSIDE_CASTLE_DESCRIPTION = "K";
    private static final String WHITE_QUEENSIDE_CASTLE_DESCRIPTION = "Q";
    private static final String BLACK_KINGSIDE_CASTLE_DESCRIPTION = "k";
    private static final String BLACK_QUEENSIDE_CASTLE_DESCRIPTION = "q";
    
    private static ForsythEdwardsNotation instance;
    
    private ForsythEdwardsNotation() {
    }
    
    public static ForsythEdwardsNotation getInstance() {
        if(instance == null) {
            instance = new ForsythEdwardsNotation();
        }
        
        return instance;
    }

    public String getPositionDescription(ChessGame game) {
        return getPositionDescription(game.getState(), game.getNextMoveNumber(),
                game.getCurrentPlyClock());
    }

    public String getPositionDescription(ChessGameState gameState, int nextMoveNumber, int plyClock) {
        String positionDescription = getPiecePlacementDescription(gameState);
        positionDescription += " " + getColorToMoveDescription(gameState);
        positionDescription += " " + getCastlingAvailabilityDescription(gameState);
        positionDescription += " " + getEnPassantSquareDescription(gameState);
        positionDescription += " " + plyClock;
        positionDescription += " " + nextMoveNumber;
        
        return positionDescription;
    }

    public ChessGame getGame(String positionDescription) {
        String[] tokens = positionDescription.split(" ");
        
        ChessGameState gameState = processPiecePlacementDescription(tokens[0]);
        processColorToMoveDescription(tokens[1], gameState);
        processCastlingAvailabilityDescription(tokens[2], gameState);
        processEnPassantSquareDescription(tokens[3], gameState);
        
        int initialPlyClock = Integer.parseInt(tokens[4]);
        
        int nextMoveNumber = Integer.parseInt(tokens[5]);
        int initialPlyCount = (nextMoveNumber - 1) * 2;
        if(gameState.getColorToMove() == PieceColor.BLACK) {
            initialPlyCount++;
        }
        
        return new ChessGame(gameState, initialPlyCount, initialPlyClock);
    }

    private String getPiecePlacementDescription(ChessGameState gameState) {
        StringBuilder piecePlacementDescription = new StringBuilder();
        
        RectangularBoard board = gameState.getBoard();
        
        for(int row = 0; row < board.getRowCount(); row++) {
            int nEmptySquares = 0;
            for(int column = 0; column < board.getColumnCount(); column++) {
                ChessPiece piece = (ChessPiece) board.getPiece(row, column);
                if(piece == null) {
                    nEmptySquares++;
                } else {
                    if(nEmptySquares != 0) {
                        piecePlacementDescription.append(nEmptySquares);
                        nEmptySquares = 0;
                    }
                    String pieceDescription = getPieceDescription(piece.getClass()).toString();
                    if(piece.getColor() == PieceColor.BLACK) {
                        pieceDescription = pieceDescription.toLowerCase();
                    }
                    piecePlacementDescription.append(pieceDescription);
                }
            }
            if(nEmptySquares != 0) {
                piecePlacementDescription.append(nEmptySquares);
            }
            if(row != board.getRowCount() - 1) {
                piecePlacementDescription.append(ROW_DESCRIPTION_SEPARATOR);
            }
        }
        
        return piecePlacementDescription.toString();
    }

    private ChessGameState processPiecePlacementDescription(String piecePlacementDescription) {
        RectangularBoard board = new RectangularBoard(ChessGame.ROW_COUNT, ChessGame.COLUMN_COUNT);
        ChessGameState gameState = new ChessGameState(board, PieceColor.BLACK, null);
        
        String[] rowDescriptions = piecePlacementDescription.split(ROW_DESCRIPTION_SEPARATOR);
        if(rowDescriptions.length != ChessGame.ROW_COUNT) {
            throw new IllegalArgumentException("Error! Incorrect number of rows when processing piece placement data string.");
        }
        
        for(int row = 0; row < ChessGame.ROW_COUNT; row++) {
            String errorMessage = "Error processing piece placement description for row " + row + " :\n";
            String rowDescription = rowDescriptions[row];
            int column = 0;
            for(int i = 0; i < rowDescription.length(); i++) {
                if(column >= ChessGame.COLUMN_COUNT) {
                    errorMessage += "this row specifies more than " + ChessGame.COLUMN_COUNT + " columns.";
                    throw new IllegalArgumentException(errorMessage);
                }
                char character = rowDescription.charAt(i);
                if(CharUtils.isAsciiNumeric(character)) {
                    column += CharUtils.toIntValue(character);
                } else {
                    PieceColor color = PieceColor.WHITE;
                    if(Character.isLowerCase(character)) {
                        color = PieceColor.BLACK;
                        character = Character.toUpperCase(character);
                    }
                    Class<? extends ChessPiece> pieceClass = getPieceClass(character);
                    if(pieceClass == null) {
                        errorMessage += "unknown piece character: " + character;
                        throw new IllegalArgumentException(errorMessage);
                    }
                    ChessPiece piece = createPiece(pieceClass, color, gameState);
                    board.setPiece(piece, row, column);
                    column++;
                }
            }
            if(column != ChessGame.COLUMN_COUNT) {
                errorMessage += "this row does not specify exactly " + ChessGame.COLUMN_COUNT + " columns.";
                throw new IllegalArgumentException(errorMessage);
            }
        }
        
        return gameState;
    }

    private String getColorToMoveDescription(ChessGameState gameState) {
        return gameState.getColorToMove() == PieceColor.BLACK ? BLACK_DESCRIPTION : WHITE_DESCRIPTION;
    }

    private void processColorToMoveDescription(String colorToMoveDescription, ChessGameState gameState) {
        if(colorToMoveDescription.equals(BLACK_DESCRIPTION)) {
            gameState.setColorToMove(PieceColor.BLACK);
        } else if(colorToMoveDescription.equals(WHITE_DESCRIPTION)) {
            gameState.setColorToMove(PieceColor.WHITE);
        } else {
            throw new IllegalArgumentException("Error! Invalid color to move description.");
        }
    }

    private String getCastlingAvailabilityDescription(ChessGameState gameState) {
        StringBuilder castlingAvailabilityDescription = new StringBuilder();
        
        if(gameState.isCastlePossible(PieceColor.WHITE, BoardSide.KING_SIDE)) {
            castlingAvailabilityDescription.append(WHITE_KINGSIDE_CASTLE_DESCRIPTION);
        }
        if(gameState.isCastlePossible(PieceColor.WHITE, BoardSide.QUEEN_SIDE)) {
            castlingAvailabilityDescription.append(WHITE_QUEENSIDE_CASTLE_DESCRIPTION);
        }
        if(gameState.isCastlePossible(PieceColor.BLACK, BoardSide.KING_SIDE)) {
            castlingAvailabilityDescription.append(BLACK_KINGSIDE_CASTLE_DESCRIPTION);
        }
        if(gameState.isCastlePossible(PieceColor.BLACK, BoardSide.QUEEN_SIDE)) {
            castlingAvailabilityDescription.append(BLACK_QUEENSIDE_CASTLE_DESCRIPTION);
        }
        
        if(castlingAvailabilityDescription.length() == 0) {
            castlingAvailabilityDescription.append(NULL_DESCRIPTION);
        }
        
        return castlingAvailabilityDescription.toString();
    }

    private void processCastlingAvailabilityDescription(String castlingAvailabilityDescription,
            ChessGameState gameState) {
        String pattern = "(" + WHITE_KINGSIDE_CASTLE_DESCRIPTION + ")?(" +
            WHITE_QUEENSIDE_CASTLE_DESCRIPTION + ")?(" +
            BLACK_KINGSIDE_CASTLE_DESCRIPTION + ")?(" +
            BLACK_QUEENSIDE_CASTLE_DESCRIPTION + ")?" ;
        boolean isValidDescription = (Pattern.matches(pattern, castlingAvailabilityDescription) &&
                !castlingAvailabilityDescription.equals("")) ||
                castlingAvailabilityDescription.equals(NULL_DESCRIPTION);
        if(!isValidDescription) {
            throw new IllegalArgumentException("Error! Invalid castling availability description.");
        }
        
        boolean isWhiteKingsideCastlePossible =
            castlingAvailabilityDescription.contains(WHITE_KINGSIDE_CASTLE_DESCRIPTION); 
        gameState.setCastlePossible(PieceColor.WHITE, BoardSide.KING_SIDE, isWhiteKingsideCastlePossible);
        
        boolean isWhiteQueensideCastlePossible =
            castlingAvailabilityDescription.contains(WHITE_QUEENSIDE_CASTLE_DESCRIPTION); 
        gameState.setCastlePossible(PieceColor.WHITE, BoardSide.QUEEN_SIDE, isWhiteQueensideCastlePossible);

        boolean isBlackKingsideCastlePossible =
            castlingAvailabilityDescription.contains(BLACK_KINGSIDE_CASTLE_DESCRIPTION); 
        gameState.setCastlePossible(PieceColor.BLACK, BoardSide.KING_SIDE, isBlackKingsideCastlePossible);

        boolean isBlackQueensideCastlePossible =
            castlingAvailabilityDescription.contains(BLACK_QUEENSIDE_CASTLE_DESCRIPTION); 
        gameState.setCastlePossible(PieceColor.BLACK, BoardSide.QUEEN_SIDE, isBlackQueensideCastlePossible);
    }

    private String getEnPassantSquareDescription(ChessGameState gameState) {
        Square enPassantSquare = gameState.getEnPassantSquare();
        return enPassantSquare != null ? getSquareDescription(enPassantSquare) : NULL_DESCRIPTION;
    }

    private void processEnPassantSquareDescription(String enPassantSquareDescription, ChessGameState gameState) {
        if(!enPassantSquareDescription.equals(NULL_DESCRIPTION)) {
            Square enPassantSquare = getSquare(enPassantSquareDescription);
            gameState.setEnPassantSquare(enPassantSquare);
        }
    }

    private ChessPiece createPiece(Class<? extends ChessPiece> pieceClass, PieceColor color,
            ChessGameState gameState) {
        if(pieceClass == Rook.class) {
            return new Rook(color, gameState, true);
        } else if(pieceClass == Knight.class) {
            return new Knight(color, gameState);
        } else if(pieceClass == Bishop.class) {
            return new Bishop(color, gameState);
        } else if(pieceClass == Queen.class) {
            return new Queen(color, gameState);
        } else if(pieceClass == King.class) {
            return new King(color, gameState, true);
        } else if(pieceClass == Pawn.class) {
            return new Pawn(color, gameState);
        } else {
            throw new IllegalArgumentException("Error! Unknown chess piece class: " + pieceClass + ".");
        }
    }

}
