/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.model;

import java.util.ArrayList;
import java.util.List;

import com.haundrix.games.model.Move;
import com.haundrix.games.model.Square;

public class King extends ChessPiece {
    
    private enum KingUnmoveInfoProperties {
        FIRST_MOVE,
        ROOK_MOVE
    }
    
    private boolean moved;

    public King(PieceColor color, ChessGameState gameState) {
        super(color, gameState);
        moved = true;
    }

    public King(PieceColor color, ChessGameState gameState, boolean moved) {
        super(color, gameState);
        this.moved = moved;
    }

    public boolean isMoved() {
        return moved;
    }
    
    public void setMoved(boolean moved) {
        this.moved = moved;
    }

    @Override
    public boolean canMove(Move move) {
        Square origin = move.getOrigin();
        Square destination = move.getDestination();
        
        if(getBoard().isNeighbouringSquare(origin, destination)) {
            return true;
        }
        
        return canCastle(origin, destination);
    }

    @Override
    public List<Square> getDestinationsFrom(Square origin) {
        List<Square> destinations = new ArrayList<Square>();
        
        destinations.addAll(getBoard().getNeighbouringSquares(origin));
        
        Square destination = new Square(origin.getRow(), origin.getColumn() - 2);
        addCastleDestination(origin, destination, destinations);
        destination = new Square(origin.getRow(), origin.getColumn() + 2);
        addCastleDestination(origin, destination, destinations);
        
        return destinations;
    }

    @Override
    public boolean equals(Object object) {
        if(!super.equals(object)) {
            return false;
        }
        
        return ((King) object).moved == moved;
    }

    public boolean equalsIgnoreMoved(ChessPiece otherPiece) {
        return super.equals(otherPiece);
    }

    @Override
    protected void move(Move move) {
        checkMove(move);
        
        if(isCastleMove(move)) {
            Move rookMove = getRookCastleMove(move);
            Rook rook = (Rook) getBoard().getPiece(rookMove.getOrigin());  // we're asserting there's a rook there
            rook.move(rookMove);  // the en passant square unmove property is tracked by the rook
            move.addUnmoveInfo(KingUnmoveInfoProperties.ROOK_MOVE, rookMove);
            getBoard().setPiece(null, move.getOrigin());
            getBoard().setPiece(this, move.getDestination());
        } else {
            super.move(move);
            move.addUnmoveInfo(KingUnmoveInfoProperties.FIRST_MOVE, !moved);
        }
        
        moved = true;
    }

    @Override
    protected void unmove(Move move) {
        checkUnmove(move);
        
        if(isCastleMove(move)) {
            getBoard().setPiece(null, move.getDestination());
            getBoard().setPiece(this, move.getOrigin());
            moved = false;

            Move rookMove = (Move) move.getUnmoveInfo(KingUnmoveInfoProperties.ROOK_MOVE);
            Rook rook = (Rook) getBoard().getPiece(rookMove.getDestination());  // we're asserting there's a rook there
            rook.unmove(rookMove);
            
            move.clearUnmoveInfos();
        } else {
            moved = ! (Boolean) move.getUnmoveInfo(KingUnmoveInfoProperties.FIRST_MOVE);
            super.unmove(move);
        }
    }

    // The implementation of this method may assume the supplied move is legal.
    boolean isCastleMove(Move move) {
        return isQueensideCastleMove(move) || isKingsideCastleMove(move);
    }

    // The implementation of this method may assume the supplied move is legal.
    boolean isQueensideCastleMove(Move move) {
        return move.getOrigin().getColumn() - move.getDestination().getColumn() == 2;
    }

    // The implementation of this method may assume the supplied move is legal.
    boolean isKingsideCastleMove(Move move) {
        return move.getOrigin().getColumn() - move.getDestination().getColumn() == -2;
    }

    // The implementation of this method may assume that the supplied move is indeed a legal castle move.
    Move getRookCastleMove(Move move) {
        Square origin = move.getOrigin();
        Square destination = move.getDestination();
        
        int columnDifference = destination.getColumn() - origin.getColumn();
        Square intermediateSquare;
        Square rookSquare;
        if(columnDifference == -2) {
            // queenside castle
            intermediateSquare = new Square(origin.getRow(), origin.getColumn() - 1);
            rookSquare = new Square(origin.getRow(), 0);
        } else {
            // kingside castle
            intermediateSquare = new Square(origin.getRow(), origin.getColumn() + 1);
            rookSquare = new Square(origin.getRow(), getBoard().getColumnCount() - 1);
        }

        return new Move(rookSquare, intermediateSquare);
    }

    private boolean canCastle(Square origin, Square destination) {
        if(moved) {
            return false;
        }
        
        int castleRow = color == PieceColor.BLACK ? 0 : getBoard().getRowCount() - 1;
        if((origin.getRow() != castleRow) || (destination.getRow() != castleRow)) {
            return false;
        }
        int columnDifference = destination.getColumn() - origin.getColumn();
        if(Math.abs(columnDifference) != 2) {
            return false;
        }
        
        ChessGame game = gameState.getGame();
        if(game.isSquareAttacked(origin, color.getOppositeColor())) {
            return false;
        }

        Square intermediateSquare;
        Square farAwaySquare = null;
        ChessPiece rook;
        if(columnDifference == -2) {
            // queenside castle
            intermediateSquare = new Square(origin.getRow(), origin.getColumn() - 1);
            farAwaySquare = new Square(origin.getRow(), 1);
            rook = (ChessPiece) getBoard().getPiece(origin.getRow(), 0);
        } else {
            // kingside castle
            intermediateSquare = new Square(origin.getRow(), origin.getColumn() + 1);
            rook = (ChessPiece) getBoard().getPiece(origin.getRow(), getBoard().getColumnCount() - 1);
        }
        if((rook == null) || !(rook instanceof Rook)) {
            return false;
        }
        if(((Rook) rook).isMoved()) {
            return false;
        }
        if(getBoard().getPiece(intermediateSquare) != null) {
            return false;
        }
        if(getBoard().getPiece(destination) != null) {
            return false;
        }
        if((farAwaySquare != null) && (getBoard().getPiece(farAwaySquare) != null)) {
            return false;
        }
        if(game.isSquareAttacked(intermediateSquare, color.getOppositeColor())) {
            return false;
        }
        
        return true;
    }

    private void addCastleDestination(Square origin, Square destination, List<Square> destinations) {
        int destinationColumn = destination.getColumn();
        if((destinationColumn >= 0) && (destinationColumn < getBoard().getColumnCount())) {
            if(canCastle(origin, destination)) {
                destinations.add(destination);
            }
        }
    }

}
