/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.model;

import java.util.ArrayList;
import java.util.List;

import com.haundrix.games.chess.model.PromoteablePiece.PromoteablePieceUnmoveInfoProperties;
import com.haundrix.games.model.Piece;
import com.haundrix.games.model.Move;
import com.haundrix.games.model.Square;

public class Pawn extends ChessPiece {
    
    public Pawn(PieceColor color, ChessGameState gameState) {
        super(color, gameState);
    }

    @Override
    public boolean canMove(Move move) {
        Square origin = move.getOrigin();
        Square destination = move.getDestination();
        
        int rowDifference = destination.getRow() - origin.getRow();
        if(color == PieceColor.WHITE) {
            // dirty trick to avoid duplicating too much code for black and white pawns
            rowDifference = -rowDifference;
        }
        int absColumnDifference = Math.abs(destination.getColumn() - origin.getColumn());

        if((rowDifference <= 0) || (rowDifference > 2)) {
            return false;
        } else if(rowDifference == 1) {
            if(absColumnDifference > 1) {
                return false;
            } else if(absColumnDifference == 0) {
                if(getBoard().getPiece(destination) != null) {
                    return false;
                }
            } else if(absColumnDifference == 1) {
                if((getBoard().getPiece(destination) == null) && !destination.equals(gameState.getEnPassantSquare())) {
                    return false;
                }
            }
        } else if(rowDifference == 2) {
            if((origin.getRow() != getStartRow()) || (absColumnDifference != 0)) {
                return false;
            }
            int intermediateSquareRow = color == PieceColor.BLACK ? origin.getRow() + 1 : origin.getRow() - 1;
            Square intermediateSquare = new Square(intermediateSquareRow, origin.getColumn());
            if((getBoard().getPiece(intermediateSquare) != null) || (getBoard().getPiece(destination) != null)) {
                return false;
            }
        }
        
        if((destination.getRow() == getPromotionRow()) && !(move instanceof PromotionMove)) {
            return false;
        }
        
        return true;
    }

    @Override
    public List<Square> getDestinationsFrom(Square origin) {
        List<Square> destinations = new ArrayList<Square>();
        
        int increment = color == PieceColor.BLACK ? 1 : -1;
        
        Square frontSquare = new Square(origin.getRow() + increment, origin.getColumn());
        if(getBoard().getPiece(frontSquare) == null) {
            destinations.add(frontSquare);
            if(origin.getRow() == getStartRow()) {
                Square frontFrontSquare = new Square(origin.getRow() + 2 * increment, origin.getColumn());
                if(getBoard().getPiece(frontFrontSquare) == null) {
                    destinations.add(frontFrontSquare);
                }
            }
        }
        
        Square diagonalSquare = new Square(origin.getRow() + increment, origin.getColumn() - 1);
        addCaptureDestination(origin, diagonalSquare, destinations);
        diagonalSquare = new Square(origin.getRow() + increment, origin.getColumn() + 1);
        addCaptureDestination(origin, diagonalSquare, destinations);
        
        return destinations;
    }

    public int getPromotionRow() {
        return color == PieceColor.BLACK ? getBoard().getRowCount() - 1 : 0;
    }

    public Move getDefaultPromotionMove(Move move) {
        if(move.getDestination().getRow() == getPromotionRow()) {
            return new PromotionMove(move.getOrigin(), move.getDestination(), Queen.class);
        } else {
            return move;
        }
    }

    @Override
    protected void move(Move move) {
        checkMove(move);
        
        Square origin = move.getOrigin();
        Square destination = move.getDestination();
        
        getBoard().setPiece(null, origin);
        
        int absColumnDifference = Math.abs(origin.getColumn() - destination.getColumn());
        boolean isEnPassantCapture = destination.equals(gameState.getEnPassantSquare());
        if((absColumnDifference == 1) && isEnPassantCapture) {
            Square capturedPawnSquare = new Square(origin.getRow(), destination.getColumn());
            Pawn capturedPawn = (Pawn) getBoard().getPiece(capturedPawnSquare);  // we're also asserting that the captured piece is indeed a pawn!
            move.addUnmoveInfo(UnmoveInfoProperties.CAPTURED_PIECE, capturedPawn);
            getBoard().setPiece(null, capturedPawnSquare);
        }
        
        move.addUnmoveInfo(UnmoveInfoProperties.EN_PASSANT_SQUARE, gameState.getEnPassantSquare());
        int absRowDifference = Math.abs(origin.getRow() - destination.getRow());
        if(absRowDifference == 2) {
            int enPassantSquareRow = color == PieceColor.BLACK ?
                    destination.getRow() - 1 :
                    destination.getRow() + 1;
            gameState.setEnPassantSquare(new Square(enPassantSquareRow, destination.getColumn()));
        } else {
            gameState.setEnPassantSquare(null);
        }
        
        if(!isEnPassantCapture) {
            move.addUnmoveInfo(UnmoveInfoProperties.CAPTURED_PIECE, getBoard().getPiece(destination));
        }
        if(move instanceof PromotionMove) {
            Class<? extends PromoteablePiece> promotedPiece = ((PromotionMove) move).getPromotedPiece();
            move.addUnmoveInfo(PromoteablePieceUnmoveInfoProperties.PROMOTED_PAWN, this);
            getBoard().setPiece(createNewPiece(promotedPiece), destination);
        } else {
            getBoard().setPiece(this, destination);
        }
    }

    // This method does not need to take promotion moves into account, since after
    // a promotion an instance of PromoteablePiece replaces the instance of Pawn on the board.
    // Promotion moves will be handled by the implementation of PromoteablePiece.unmove() . 
    @Override
    protected void unmove(Move move) {
        checkUnmove(move);
        
        Square origin = move.getOrigin();
        Square destination = move.getDestination();

        Square enPassantSquare = (Square) move.getUnmoveInfo(UnmoveInfoProperties.EN_PASSANT_SQUARE);
        if((enPassantSquare != null) && (destination.equals(enPassantSquare))) {
            getBoard().setPiece(null, destination);
            getBoard().setPiece(this, origin);
            Pawn capturedPawn = (Pawn) move.getUnmoveInfo(UnmoveInfoProperties.CAPTURED_PIECE);  // we're also asserting that the captured piece was indeed a pawn!
            Square capturedPawnSquare = new Square(origin.getRow(), destination.getColumn());
            getBoard().setPiece(capturedPawn, capturedPawnSquare);
            gameState.setEnPassantSquare(enPassantSquare);
            move.clearUnmoveInfos();
        } else {
            super.unmove(move);
        }
    }

    @Override
    protected boolean attacksSquare(Square pieceSquare, Square attackedSquare) {
        int rowDifference = attackedSquare.getRow() - pieceSquare.getRow();
        int expectedRowDifference = color == PieceColor.BLACK ? 1 : -1;
        if(rowDifference != expectedRowDifference) {
            return false;
        }
        
        int absColumnDifference = Math.abs(attackedSquare.getColumn() - pieceSquare.getColumn());
        return absColumnDifference == 1;
    }

    @Override
    protected Square getAttackedSquare(Move move) {
        if(move.getDestination().equals(gameState.getEnPassantSquare())) {
            return new Square(move.getOrigin().getRow(), move.getDestination().getColumn());
        } else {
            return super.getAttackedSquare(move);
        }
    }

    // The implementation of this method may assume the supplied move is legal.
    boolean isEnPassantMove(Move move) {
        return move.getDestination().equals(gameState.getEnPassantSquare());
    }

    private int getStartRow() {
        return color == PieceColor.BLACK ? 1 : getBoard().getRowCount() - 2;
    }

    private Piece createNewPiece(Class<? extends PromoteablePiece> promotedPiece) {
        if(promotedPiece == Queen.class) {
            return new Queen(color, gameState);
        } else if(promotedPiece == Rook.class) {
            return new Rook(color, gameState);
        } else if(promotedPiece == Bishop.class) {
            return new Bishop(color, gameState);
        } else if(promotedPiece == Knight.class) {
            return new Knight(color, gameState);
        } else {
            throw new IllegalArgumentException("Invalid promoting piece: " + promotedPiece + ".");
        }
    }

    private void addCaptureDestination(Square origin, Square diagonalSquare, List<Square> destinations) {
        int destinationColumn = diagonalSquare.getColumn();
        if((destinationColumn < 0) || (destinationColumn >= getBoard().getColumnCount())) {
            return;
        }
        
        if(getBoard().getPiece(diagonalSquare) != null) {
            destinations.add(diagonalSquare);
        } else if(diagonalSquare.equals(gameState.getEnPassantSquare())) {
            destinations.add(diagonalSquare);
        }
    }

}
