/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.model.playing;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.haundrix.games.chess.model.ChessAlgebraicMoveNotation;
import com.haundrix.games.chess.model.ChessGame;
import com.haundrix.games.chess.model.ChessGameResult;
import com.haundrix.games.chess.model.ChessGameResultReason;
import com.haundrix.games.chess.model.ChessSimpleMoveNotation;
import com.haundrix.games.model.GameOutcome;
import com.haundrix.games.model.GameResult;
import com.haundrix.games.model.IMoveNotation;
import com.haundrix.games.model.Move;
import com.haundrix.games.model.Piece.PieceColor;
import com.haundrix.games.model.playing.ClockTime;
import com.haundrix.games.model.playing.Engine;
import com.haundrix.games.model.playing.EngineConfig;
import com.haundrix.games.model.playing.EngineLog;
import com.haundrix.games.model.playing.FischerClock;
import com.haundrix.games.model.playing.IClock;
import com.haundrix.games.model.playing.ICommandLine;
import com.haundrix.games.model.playing.IGameProxy;
import com.haundrix.games.model.playing.IPlayer;
import com.haundrix.games.model.playing.MovesPerTimeClock;
import com.haundrix.games.model.playing.TimePerMoveClock;
import com.haundrix.utils.IExceptionHandler;
import com.haundrix.utils.ProcessConfig;

/*
 * This class interfaces with a chess engine running as an external process through
 * the Chess Engine Communication Protocol (a.k.a. Winboard 2 protocol).
 * 
 * This implementation currently supports only games at the starting position. 
 */
public class ComputerPlayer implements IPlayer {
    
    private static final int BUFFER_SIZE = 1024;
    
    // Note:
    // using "\r\n" causes problems with some engines even on Windows
    // (happened when sending out moves to Sjeng on Windows).
    private static final String LINE_DELIMITER = "\n";
    
    private static final long FEATURE_PROCESSING_TIMEOUT_PERIOD = 2000L;  // in milliseconds
    
    private static int engineListeningThreadNumber = 0;
    
    private PieceColor myColor;
    private Engine engine;
    
    private IGameProxy gameProxy;
    private IExceptionHandler exceptionHandler;

    private EngineLog log = new EngineLog();
    
    // These objects are accessed by both threads, but need not be synchronized
    // since a single thread at at time will be accessing them.
    // When it's the opponent's turn to move they are accessed by the GUI thread,
    // whereas when it's the engine's turn they are accessed by the engine listening thread.
    private ChessGame internalGame;
    private IMoveNotation notation;
    
    private boolean isListeningToEngine = false;
    
    private volatile boolean allFeaturesInputed = false;
    private LinkedHashMap<String, Object> featureValues = new LinkedHashMap<String, Object>();

    private BufferedReader in;
    private OutputStreamWriter out;

    private boolean useLongMoveCommand = false;

    private volatile boolean waitingForMove;
    private volatile boolean isDrawOffered = false;
    private volatile boolean isDisposed = false;

    public ComputerPlayer(PieceColor engineColor, IClock clock, Engine engine, EngineConfig engineConfig,
            boolean opponentIsComputer, IGameProxy gameProxy, IExceptionHandler exceptionHandler) throws IOException {
        myColor = engineColor;
        this.engine = engine;
        
        this.gameProxy = gameProxy;
        this.exceptionHandler = exceptionHandler;
        
        internalGame = new ChessGame();
        notation = new ChessSimpleMoveNotation();
        
        startProcess(engine, engineConfig);
        
        processFeatures();
        
        output("post");
        
        if(engineConfig.isPonder()) {
            output("hard");
        } else {
            output("easy");
        }
        
        if(opponentIsComputer) {
            output("computer");
        }
        
        waitingForMove = true;
        output("new");
        if(myColor == PieceColor.WHITE) {
            waitingForMove = false;
            output("force");
        }

        configureClock(clock);
    }
    
    public EngineLog getLog() {
        return log;
    }

    @Override
    public void go() {
        if(myColor == PieceColor.WHITE) {
            output("go");
        }
    }

    @Override
    public void move(Move move) {
        if(!waitingForMove) {
            String errorMessage = "Error! A move was sent to the engine when it was the engine's turn to play.";
            throw new IllegalArgumentException(errorMessage);
        }
        
        internalGame.move(move.clone());
        String moveDescription = internalGame.getLastMoveDescription(notation);
        try {
            waitingForMove = false;
            if(useLongMoveCommand) {
                output("usermove " + moveDescription);
            } else {
                output(moveDescription);
            }
        } catch(Exception e) {
            exceptionHandler.handle(e);
        }
    }
    
    @Override
    public void setTime(ClockTime time) {
        long myTime, opponentTime;
        if(myColor == PieceColor.WHITE) {
            myTime = time.getWhiteTime();
            opponentTime = time.getBlackTime();
        } else {
            myTime = time.getBlackTime();
            opponentTime = time.getWhiteTime();
        }
        
        try {
            output("time " + myTime / 10L);
            output("otim " + opponentTime / 10L);
        } catch(Exception e) {
            exceptionHandler.handle(e);
        }
    }

    @Override
    public long calculateTimeThreshold(long availableTime) {
        return engine.calculateTimeThreshold(availableTime);
    }

    @Override
    public void moveNow() {
        try {
            output("?");
        } catch(Exception e) {
            exceptionHandler.handle(e);
        }
    }
    
    @Override
    public void setResult(GameResult result) {
        GameOutcome outcome = result.getOutcome();
        ChessGameResultReason resultReason = ((ChessGameResult) result).getResultReason();
        
        String outcomeDescription;
        String resultReasonDescription;
        if(outcome == GameOutcome.WHITE_WINS) {
            outcomeDescription = "1-0";
            if(resultReason == ChessGameResultReason.CHECKMATE) {
                resultReasonDescription = "White mates";
            } else if(resultReason == ChessGameResultReason.RESIGNATION) {
                resultReasonDescription = "Black resigns";
            } else if(resultReason == ChessGameResultReason.TIME_FORFEIT) {
                resultReasonDescription = "Black forfeits on time";
            } else {
                resultReasonDescription = "White wins";
            }
        } else if(outcome == GameOutcome.BLACK_WINS) {
            outcomeDescription = "0-1";
            if(resultReason == ChessGameResultReason.CHECKMATE) {
                resultReasonDescription = "Black mates";
            } else if(resultReason == ChessGameResultReason.RESIGNATION) {
                resultReasonDescription = "White resigns";
            } else if(resultReason == ChessGameResultReason.TIME_FORFEIT) {
                resultReasonDescription = "White forfeits on time";
            } else {
                resultReasonDescription = "Black wins";
            }
        } else {
            outcomeDescription = "1/2-1/2";
            if(resultReason == ChessGameResultReason.STALEMATE) {
                resultReasonDescription = "Stalemate";
            } else if(resultReason == ChessGameResultReason.DEAD_POSITION) {
                resultReasonDescription = "Insufficient material";
            } else if(resultReason == ChessGameResultReason.POSITION_REPETITION) {
                resultReasonDescription = "Draw by 3-fold repetition";
            } else if(resultReason == ChessGameResultReason.FIFTY_MOVE_RULE) {
                resultReasonDescription = "Draw by 50 move rule";
            } else if(resultReason == ChessGameResultReason.COMMON_AGREEMENT) {
                resultReasonDescription = "Draw by common agreement";
            } else {
                resultReasonDescription = "Draw";
            }
        }
        
        try {
            output("result " + outcomeDescription + " {" + resultReasonDescription + "}");
        } catch(Exception e) {
            exceptionHandler.handle(e);
        }
    }

    @Override
    public boolean isDisposed() {
        return isDisposed;
    }

    @Override
    public void dispose() {
        try {
            output("quit");
            isDisposed = true;
        } catch(Exception e) {
            // never mind
        }
        
        // TODO Ensure that the external process really ended.
        // Check that the external process really ended and
        // implement a timeout mechanism to kill misbehaving processes.
        // The Watchdog class in the Apache Commons Exec library might be useful.
    }

    public void offerDraw() {
        try {
            isDrawOffered = true;
            output("draw");
        } catch(Exception e) {
            exceptionHandler.handle(e);
        }
    }

    private void outputAtPrompt(String promptString, String command) {
        try {
            StringBuilder input = new StringBuilder();
            
            char[] buffer = new char[BUFFER_SIZE];
            while(!input.toString().endsWith(promptString)) {
                int charsRead = in.read(buffer);
                input.append(buffer, 0, charsRead);
            }
            
            String[] lines = input.toString().split(LINE_DELIMITER, -1);
            for(int i = 0; i < lines.length - 1; i++) {
                log.log(lines[i]);
            }
            
            log.log(promptString + command);
            
            out.write(command + LINE_DELIMITER);
            out.flush();
        } catch(IOException e) {
            throw new RuntimeException(e);
        }
    }

    private void output(String command) {
        try {
            log.log(command);
            
            out.write(command + LINE_DELIMITER);
            out.flush();
        } catch(IOException e) {
            throw new RuntimeException(e);
        }
    }
    
    private void startProcess(Engine engine, EngineConfig engineConfig) throws IOException {
        ProcessConfig processConfig = engine.preConfigure(engineConfig);
        
        String executable = processConfig.getExecutable();
        String[] parameters = processConfig.getParameters();
        String[] command = new String[parameters.length + 1];
        command[0] = executable;
        System.arraycopy(parameters, 0, command, 1, parameters.length);
        ProcessBuilder processBuilder = new ProcessBuilder(command);
        
        Map<String, String> environment = processBuilder.environment();
        for(String variableName : processConfig.getEnvironment().keySet()) {
            String variableValue = processConfig.getEnvironment().get(variableName);
            environment.put(variableName, variableValue);
        }
        
        processBuilder.directory(new File(executable).getParentFile());
        processBuilder.redirectErrorStream(true);
        
        Process process = processBuilder.start();
        
        in = new BufferedReader(new InputStreamReader(new BufferedInputStream(process.getInputStream())));
        out = new OutputStreamWriter(new BufferedOutputStream(process.getOutputStream()));
        
        // after the first call to output(String command) the communication will 
        // be handled asynchronously and the input will be given in complete lines
        engine.postConfigure(engineConfig, new ICommandLine() {
            
            @Override
            public void outputAtPrompt(String promptString, String command) {
                if(isListeningToEngine) {
                    throw new IllegalStateException("Error! Communication with the engine is being handled asynchronously.");
                }
                
                ComputerPlayer.this.outputAtPrompt(promptString, command);
            }
            
            @Override
            public void output(String command) {
                startListeningToEngine();
                ComputerPlayer.this.output(command);
            }
        });
        startListeningToEngine();
    }
    
    private void startListeningToEngine() {
        if(!isListeningToEngine) {
            new EngineListeningThread().start();
            isListeningToEngine = true;
        }
    }

    private void processFeatures() throws IOException {
        output("protover 2");
        
        final long startTime = System.currentTimeMillis();
        while(!allFeaturesInputed) {
            synchronized(this) {
                long now = System.currentTimeMillis();
                if(now - startTime > FEATURE_PROCESSING_TIMEOUT_PERIOD) {
                    allFeaturesInputed = true;
                    break;
                }
            }
            
            Thread.yield();
        }
        
        for(String featureName : featureValues.keySet()) {
            processFeature(featureName, featureValues.get(featureName));
        }
    }

    private void processFeature(String featureName, Object featureValue) {
        CECPFeature feature = CECPFeature.getFeature(featureName);
        
        if(feature == null) {
            output("rejected " + featureName);
            return;
        }
        
        if(!feature.validate(featureValue)) {
            output("Error! Invalid feature value for " + featureName + ": " + featureValue + " (" +
                    feature.getFriendlyTypeName() + " value expected).");
            output("rejected " + featureName);
            return;
        }
        
        if(featureName.equals(CECPFeature.SAN.getName())) {
            if(((Integer) featureValue) == 1) {
                ChessAlgebraicMoveNotation algebraicNotation = new ChessAlgebraicMoveNotation();
                algebraicNotation.setIndicateEnPassant(false);
                algebraicNotation.setUseLetterOForCastle(true);
                algebraicNotation.setUseEqualsForPromotion(true);
                algebraicNotation.setUseHashForCheckmate(true);
                
                notation = algebraicNotation;
            }
        } else if(featureName.equals(CECPFeature.USER_MOVE.getName())) {
            if(((Integer) featureValue) == 1) {
                useLongMoveCommand = true;
            }
        }
        output("accepted " + featureName);
    }

    private void configureClock(IClock clock) {
        if(clock != null) {
            String clockCommand;
            if(clock instanceof MovesPerTimeClock) {
                MovesPerTimeClock movesPerTimeClock = (MovesPerTimeClock) clock;
                
                clockCommand = "level " + movesPerTimeClock.getMovesCount() + " ";
                clockCommand += movesPerTimeClock.getClockTime() + " 0";
            } else if(clock instanceof FischerClock) {
                FischerClock fischerClock = (FischerClock) clock;
                
                clockCommand = "level 0 " + fischerClock.getFixedTime() + " ";
                clockCommand += fischerClock.getIncrement();
            } else if(clock instanceof TimePerMoveClock) {
                clockCommand = "st " + ((TimePerMoveClock) clock).getTimePerMove();
            } else {
                throw new IllegalArgumentException("Error! Invalid clock.");
            }
            
            output(clockCommand);
        }
    }

    private class EngineListeningThread extends Thread {
        
        private final Pattern FEATURE_PATTERN = Pattern.compile("(\\w+)=(?:(\\d+)|(?:\"(.*?)\"))");
        private final Pattern MOVE_COMMAND_PATTERN = Pattern.compile("move\\s+(\\S+)");
        
        private final String DRAW_COMMAND = "offer draw";
        
        private Pattern resignCommandPattern;
        
        public EngineListeningThread() {
            super("Engine Listening Thread " + ComputerPlayer.engineListeningThreadNumber++);
            
            String resignOutcomeDescription = myColor == PieceColor.WHITE ? "0-1" : "1-0";
            String resignCommandPatternString = "(?:(?:resign)|(?:result " + resignOutcomeDescription + " \\{.*\\}))";
            resignCommandPattern = Pattern.compile(resignCommandPatternString);
        }
        
        @Override
        public void run() {
            while(true) {
                try {
                    // blocks waiting for commands from the engine's external process
                    String input = in.readLine();
                    if(isDisposed || (input == null)) {
                        break;  // (input == null) means the external process has ended
                    }
                    log.log(input);
                    input = input.trim();
                    
                    synchronized(ComputerPlayer.this) {
                        if(!allFeaturesInputed) {
                            if(input.startsWith("feature")) {
                                Matcher matcher = FEATURE_PATTERN.matcher(input);
                                while(matcher.find()) {
                                    String featureName = matcher.group(1);
                                    Object featureValue = matcher.group(2) != null ?
                                            new Integer(matcher.group(2)) :
                                            matcher.group(3);
                                    featureValues.put(featureName, featureValue);
                                    if(matcher.group().equals("done=1")) {
                                        allFeaturesInputed = true;
                                        break;
                                    }
                                }
                            }
                        } else {
                            Matcher matcher = MOVE_COMMAND_PATTERN.matcher(input);
                            if(matcher.matches()) {
                                if(!internalGame.isOver()) {
                                    String moveDescription = matcher.group(1);
                                    Move move = notation.getMove(moveDescription, internalGame);
                                    internalGame.move(move);
                                    waitingForMove = true;
                                    isDrawOffered = false;
                                    gameProxy.makeMove(move.clone());
                                }
                            } else if(DRAW_COMMAND.equals(input)) {
                                if(isDrawOffered) {
                                    isDrawOffered = false;
                                    gameProxy.communicateDrawAcceptance();
                                } else {
                                    gameProxy.makeDrawOffer(myColor);
                                }
                            } else if(resignCommandPattern.matcher(input).matches()) {
                                gameProxy.communicateResignation(myColor);
                            }
                        }
                    }
                } catch(IOException e) {
                    throw new RuntimeException(e);
                }
            }
        }
        
    }

}
