/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.pgn;

import java.io.Writer;

import com.haundrix.games.chess.model.ChessAlgebraicMoveNotation;
import com.haundrix.games.chess.model.ChessGame;
import com.haundrix.games.chess.model.ChessGameState;
import com.haundrix.games.chess.model.ForsythEdwardsNotation;
import com.haundrix.games.model.GameInfo;
import com.haundrix.games.model.GameOutcome;
import com.haundrix.games.model.Piece.PieceColor;

public class PgnExporter {
    
    private PgnWriter pgnWriter;
    
    public PgnExporter(Writer writer) {
        pgnWriter = new PgnWriter(writer);
    }
    
    public void exportGame(ChessGame game) {
        exportInfo(game);
        exportMoves(game);
    }

    private void exportInfo(ChessGame game) {
        GameInfo gameInfo = game.getInfo();
        
        // Writes the seven tag roaster:
        
        String event = gameInfo.getEvent();
        String eventTagValue = (event != null) && !event.trim().equals("") ? event : PgnConstants.UNKNOWN_TAG_VALUE;
        pgnWriter.writeTagPair(PgnConstants.EVENT_TAG_NAME, eventTagValue);
        
        String site = gameInfo.getSite();
        String siteTagValue = (site != null) && !site.trim().equals("") ? site : PgnConstants.UNKNOWN_TAG_VALUE;
        pgnWriter.writeTagPair(PgnConstants.SITE_TAG_NAME, siteTagValue);
        
        pgnWriter.writeTagPair(PgnConstants.DATE_TAG_NAME, PgnFormat.format(gameInfo.getDate()));
        
        String roundTagValue = gameInfo.getRound() != -1 ? Integer.toString(gameInfo.getRound())
                : PgnConstants.UNKNOWN_TAG_VALUE;
        pgnWriter.writeTagPair(PgnConstants.ROUND_TAG_NAME, roundTagValue);
        
        String whitePlayer = gameInfo.getWhitePlayer();
        String whiteTagValue = (whitePlayer != null) && !whitePlayer.trim().equals("") ? whitePlayer :
            PgnConstants.UNKNOWN_TAG_VALUE;
        pgnWriter.writeTagPair(PgnConstants.WHITE_PLAYER_TAG_NAME, whiteTagValue);
        
        String blackPlayer = gameInfo.getBlackPlayer();
        String blackTagValue = (blackPlayer != null) && !blackPlayer.trim().equals("") ? blackPlayer :
            PgnConstants.UNKNOWN_TAG_VALUE;
        pgnWriter.writeTagPair(PgnConstants.BLACK_PLAYER_TAG_NAME, blackTagValue);
        
        GameOutcome outcome = game.getResult() != null ? game.getResult().getOutcome() : null;
        PgnGameOutcome pgnGameOutcome = PgnGameOutcome.fromGameOutcome(outcome);
        pgnWriter.writeTagPair(PgnConstants.RESULT_TAG_NAME, PgnFormat.format(pgnGameOutcome));
        
        // Writes additional tag pairs:
        
        if(gameInfo.getWhiteElo() != -1) {
            pgnWriter.writeTagPair(PgnConstants.WHITE_ELO_RATING_TAG_NAME,
                    PgnFormat.format(gameInfo.getWhiteElo()));
        }
        if(gameInfo.getBlackElo() != -1) {
            pgnWriter.writeTagPair(PgnConstants.BLACK_ELO_RATING_TAG_NAME,
                    PgnFormat.format(gameInfo.getBlackElo()));
        }
        
        if(game.getInitialPlyCount() > 0) {
            pgnWriter.writeTagPair(PgnConstants.IS_SETUP_POSITION_TAG_NAME, PgnFormat.format(true));
            
            ChessGameState gameState = game.getInitialPosition();
            int initialNextMoveNumber = game.getInitialNextMoveNumber();
            int initialPlyClock = game.getInitialPlyClock();
            
            String initialPosition = ForsythEdwardsNotation.getInstance().getPositionDescription(gameState,
                    initialNextMoveNumber, initialPlyClock);
            pgnWriter.writeTagPair(PgnConstants.SETUP_POSITION_TAG_NAME, initialPosition);
        }
        
        pgnWriter.closeTagPairSection();
    }
    
    private void exportMoves(ChessGame game) {
        ChessAlgebraicMoveNotation notation = new ChessAlgebraicMoveNotation();
        notation.setIndicateEnPassant(false);
        notation.setUseLetterOForCastle(true);
        notation.setUseEqualsForPromotion(true);
        notation.setUseHashForCheckmate(true);
        
        int moveNumber = game.getInitialNextMoveNumber();
        PieceColor colorToMove = game.getInitialColorToMove();
        pgnWriter.setInitialMove(moveNumber, colorToMove);
        
        for(String moveDescription : game.getMoveDescriptions(notation)) {
            pgnWriter.writeMove(moveDescription);
        }
        
        GameOutcome outcome = game.getResult() != null ? game.getResult().getOutcome() : null;
        pgnWriter.closeMovetextSection(PgnGameOutcome.fromGameOutcome(outcome));
    }

}
