/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.pgn;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.joda.time.DateTime;
import org.joda.time.LocalDate;
import org.joda.time.LocalTime;
import org.joda.time.format.DateTimeFormat;

//TODO Add support for writing the time control tag value (create a new class hierarchy to represent the
//possible time control value types). 
//TODO Add support for writing list tag values (separated by : according to the PGN specification). 
class PgnFormat {

    private static final String UNKNOWN_DATE_VALUE = "????.??.??";
    private static final String DATE_FORMAT = "YYYY.MM.dd";
    private static final Pattern datePattern;
    static {
        String yearPattern = "(?:(\\d{4})|(?:\\?{4}))";
        String monthPattern = "(?:(\\d{2})|(?:\\?{2}))";
        String dayPattern = monthPattern;
        String separatorPattern = "\\.";
        
        datePattern = Pattern.compile(yearPattern + separatorPattern + monthPattern + separatorPattern +
                dayPattern);
    }
    
    private static final String TIME_FORMAT = "HH:mm:ss";
    
    public static String format(boolean tagBooleanValue) {
        return tagBooleanValue ? "1" : "0";
    }
    
    public static boolean parseBoolean(String tagValueDescription) {
        if(tagValueDescription.equals("1")) {
            return true;
        } else if(tagValueDescription.equals("0")) {
            return false;
        } else {
            throw new IllegalArgumentException("Error parsing PGN tag boolean value:" + tagValueDescription);
        }
    }
    
    public static String format(int tagIntegerValue) {
        return Integer.toString(tagIntegerValue);
    }
    
    public static int parseInteger(String tagValueDescription) {
        return Integer.parseInt(tagValueDescription);
    }
    
    public static String format(LocalDate tagDateValue) {
        return tagDateValue == null ? UNKNOWN_DATE_VALUE : tagDateValue.toString(DATE_FORMAT);
    }
    
    public static LocalDate parseDate(String tagValueDescription) {
        Matcher matcher = PgnFormat.datePattern.matcher(tagValueDescription);
        if(matcher.matches()) {
            if(matcher.group(1) == null) {
                return null;
            }
            int year = Integer.parseInt(matcher.group(1));
            int month, day;
            if((matcher.group(2) == null) || (matcher.group(3) == null)) {
                month = day = 1;
            } else {
                month = Integer.parseInt(matcher.group(2));
                day = Integer.parseInt(matcher.group(3));
            }
            return new LocalDate(year, month, day);
        } else {
            String errorMessage = "Error parsing PGN tag date value: " + tagValueDescription + ".";
            errorMessage += "\nThe date must be formatted like YYYY.MM.dd, with optional ? for unknown values.";
            throw new IllegalArgumentException(errorMessage);
        }
    }
    
    public static String format(LocalTime tagTimeValue) {
        return tagTimeValue.toString(TIME_FORMAT);
    }
    
    public static LocalTime parseTime(String tagValueDescription) {
        DateTime dateTime = DateTimeFormat.forPattern(TIME_FORMAT).parseDateTime(tagValueDescription);
        return new LocalTime(dateTime);
    }

    public static String format(PgnGameOutcome tagGameResultValue) {
        return tagGameResultValue.getDescription();
    }
    
    public static PgnGameOutcome parseGameResult(String tagValueDescription) {
        return PgnGameOutcome.parse(tagValueDescription);
    }
    
}
