/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.pgn;

import java.io.IOException;
import java.io.Reader;

import com.haundrix.games.chess.model.ChessAlgebraicMoveNotation;
import com.haundrix.games.chess.model.ChessGame;
import com.haundrix.games.chess.model.ChessGameResult;
import com.haundrix.games.chess.model.ForsythEdwardsNotation;
import com.haundrix.games.chess.pgn.PgnReader.PgnProcessor;
import com.haundrix.games.model.GameInfo;
import com.haundrix.games.model.GameOutcome;
import com.haundrix.games.model.Piece.PieceColor;

public class PgnImporter {

    private PgnReader pgnReader;
    
    private ChessAlgebraicMoveNotation notation;
    
    private GameInfo gameInfo = new GameInfo();
    private ChessGame game;
    
    private PgnGameOutcome expectedOutcome;
    
    public PgnImporter(Reader reader) {
        pgnReader = new PgnReader(reader);
        
        notation = new ChessAlgebraicMoveNotation();
        notation.setIndicateEnPassant(false);
        notation.setUseLetterOForCastle(true);
        notation.setUseEqualsForPromotion(true);
        notation.setUseHashForCheckmate(true);
    }
    
    public ChessGame importGame() throws IOException {
        pgnReader.read(new InternalPgnProcessor());
        
        return game;
    }
    
    private class InternalPgnProcessor implements PgnProcessor {

        @Override
        public void processTagPair(String tagName, String tagValue) {
            if(tagName.equals(PgnConstants.EVENT_TAG_NAME) &&
                    (!tagValue.equals(PgnConstants.UNKNOWN_TAG_VALUE))) {
                gameInfo.setEvent(tagValue);
            } else if(tagName.equals(PgnConstants.SITE_TAG_NAME) &&
                    (!tagValue.equals(PgnConstants.UNKNOWN_TAG_VALUE))) {
                gameInfo.setSite(tagValue);
            } else if(tagName.equals(PgnConstants.DATE_TAG_NAME)) {
                gameInfo.setDate(PgnFormat.parseDate(tagValue));
            } else if(tagName.equals(PgnConstants.ROUND_TAG_NAME) &&
                    (!tagValue.equals(PgnConstants.UNKNOWN_TAG_VALUE))) {
                gameInfo.setRound(PgnFormat.parseInteger(tagValue));
            } else if(tagName.equals(PgnConstants.WHITE_PLAYER_TAG_NAME) &&
                    (!tagValue.equals(PgnConstants.UNKNOWN_TAG_VALUE))) {
                gameInfo.setWhitePlayer(tagValue);
            } else if(tagName.equals(PgnConstants.BLACK_PLAYER_TAG_NAME) &&
                    (!tagValue.equals(PgnConstants.UNKNOWN_TAG_VALUE))) {
                gameInfo.setBlackPlayer(tagValue);
            } else if(tagName.equals(PgnConstants.RESULT_TAG_NAME)) {
                expectedOutcome = PgnFormat.parseGameResult(tagValue);
            } else if(tagName.equals(PgnConstants.WHITE_ELO_RATING_TAG_NAME) &&
                    (!tagValue.equals(PgnConstants.UNKNOWN_TAG_VALUE))) {
                gameInfo.setWhiteElo(PgnFormat.parseInteger(tagValue));
            } else if(tagName.equals(PgnConstants.BLACK_ELO_RATING_TAG_NAME) &&
                    (!tagValue.equals(PgnConstants.UNKNOWN_TAG_VALUE))) {
                gameInfo.setBlackElo(PgnFormat.parseInteger(tagValue));
            } else if(tagName.equals(PgnConstants.SETUP_POSITION_TAG_NAME)) {
                game = ForsythEdwardsNotation.getInstance().getGame(tagValue);
            }
        }

        @Override
        public void endTagPairSection() {
            if(game == null) {
                game = new ChessGame();
            }
            game.setInfo(gameInfo);
        }

        @Override
        public void setInitialMove(int moveNumber, PieceColor colorToMove) {
            String errorMessage = "Error processing first move: ";
            
            int expectedMoveNumber = game.getInitialNextMoveNumber();
            if(moveNumber != expectedMoveNumber) {
                errorMessage += "incorrect move number.";
                errorMessage += "\nExpected " + expectedMoveNumber + ", but found " + moveNumber + ".";
                throw new InvalidPgnFileException(errorMessage);
            }
            
            if(colorToMove != game.getInitialColorToMove()) {
                throw new InvalidPgnFileException(errorMessage + "incorrect move color.");
            }
        }

        @Override
        public void processMove(String moveDescription) {
            game.move(notation.getMove(moveDescription, game));
        }

        @Override
        public void endGame(PgnGameOutcome pgnGameOutcome) {
            String errorMessage = "Error processing game termination marker.";
            
            if((expectedOutcome != null) && (pgnGameOutcome != expectedOutcome)) {
                errorMessage += "\nThe tag pair section specifies " + expectedOutcome.getDescription();
                errorMessage += ", but the game termination marker is " + pgnGameOutcome.getDescription() + ".";
                throw new InvalidPgnFileException(errorMessage);
            }
            
            if(game.getResult() != null) {
                String gameTerminationMarker = pgnGameOutcome.getDescription();
                String gameOutcomeDescription = PgnGameOutcome.fromGameOutcome(game.getResult().getOutcome()).getDescription();
                if(!gameTerminationMarker.equals(gameOutcomeDescription)) {
                    errorMessage += "\nThe game termination marker is " + gameTerminationMarker;
                    errorMessage += ", whereas the actual game result is " + gameOutcomeDescription + ".";
                    throw new InvalidPgnFileException(errorMessage);
                }
            } else {
                GameOutcome gameOutcome = pgnGameOutcome.getGameOutcome();
                if(gameOutcome != null) {
                    game.setResult(new ChessGameResult(gameOutcome));
                }
            }
        }

    }

}
