/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.pgn;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.haundrix.games.model.Piece.PieceColor;

class PgnReader {
    
    private static Pattern TAG_PAIR_PATTERN = Pattern.compile("\\[\\s*([A-Z]\\w*)\\s*\"(.*)\"\\s*\\]");
    private static Pattern WHITE_MOVE_NUMBER_PATTERN = Pattern.compile("(\\d+)\\.");
    private static Pattern BLACK_MOVE_NUMBER_PATTERN = Pattern.compile("(\\d+)\\.\\.\\.");
    private static Pattern START_COMMENT_PATTERN = Pattern.compile("\\{.*");
    private static Pattern END_COMMENT_PATTERN = Pattern.compile(".*\\}");
    private static Pattern START_VARIATION_PATTERN = Pattern.compile("\\(.*");
    private static Pattern END_VARIATION_PATTERN = Pattern.compile(".*\\)");
    private static Pattern ANNOTATION_GLYPH_PATTERN = Pattern.compile("\\$\\d+");
    private static Pattern GAME_RESULT_PATTERN = Pattern.compile("(?:0-1|1-0|1/2-1/2|\\*)");

    public interface PgnProcessor {

        void processTagPair(String tagName, String tagValue);
        void endTagPairSection();
        void setInitialMove(int moveNumber, PieceColor colorToMove);
        void processMove(String moveDescription);
        void endGame(PgnGameOutcome gameResult);

    }

    private BufferedReader reader;
    
    private boolean isFirstToken = true;
    private boolean readMoveNumber = false;
    private int moveNumber;
    private PieceColor colorToMove;
    private boolean isInsideComment = false;
    private int variationNestLevel = 0;
    private boolean justReadMoveNumber = false;
    private boolean gameEnded = false;
    
    public PgnReader(Reader reader) {
        this.reader = new BufferedReader(reader);
    }
    
    public void read(PgnProcessor processor) throws IOException {
        readTagPairSection(processor);
        readMovetextSection(processor);
    }

    private void readTagPairSection(PgnProcessor processor) throws IOException {
        String line = reader.readLine();
        if(line == null) {
            throw new InvalidPgnFileException("Error! The file is empty.");
        }
        
        while((line != null) && !line.trim().equals("")) {
            Matcher matcher = TAG_PAIR_PATTERN.matcher(line);
            if(!matcher.matches()) {
                throw new InvalidPgnFileException("Error! Invalid tag pair data: " + line + ".");
            }
            
            String tagName = matcher.group(1);
            String tagValue = matcher.group(2);
            processor.processTagPair(tagName, tagValue);
            
            line = reader.readLine();
        }
        if(line == null) {
            throw new InvalidPgnFileException("Error! Unexpected end of file: missing movetext section.");
        }
        
        processor.endTagPairSection();
    }
    
    private void readMovetextSection(PgnProcessor processor) throws IOException {
        outer: while(!gameEnded) {
            String line = reader.readLine();
            if((line == null) || line.trim().equals("")) {
                break;
            }
            
            String[] tokens = line.trim().split("\\s+");
            for(int i = 0; i < tokens.length; i++) {
                readToken(tokens[i], processor);
                if(gameEnded) {
                    if(i != tokens.length - 1) {
                        String errorMessage = "Error processing the following line:\n";
                        errorMessage += line + "\n";
                        errorMessage += "The game termination marker must be the last token " +
                        		"in the movetext section.";
                        throw new InvalidPgnFileException(errorMessage);
                    }
                    break outer;
                }
            }
        }
        
        if(!gameEnded) {
            throw new InvalidPgnFileException("Error! No game termination marker found.");
        }
    }

    private void readToken(String token, PgnProcessor processor) {
        String errorMessage = "Error processing token " + token + ": ";
        if(isFirstToken) {
            readFirstToken(token, processor, errorMessage);
        } else {
            if(isInsideComment) {
                if(END_COMMENT_PATTERN.matcher(token).matches()) {
                    isInsideComment = false;
                }
                return;
            }
            if(variationNestLevel > 0) {
                if(START_VARIATION_PATTERN.matcher(token).matches()) {
                    variationNestLevel++;
                } else if(END_VARIATION_PATTERN.matcher(token).matches()) {
                    variationNestLevel--;
                }
                return;
            }
            
            if(END_COMMENT_PATTERN.matcher(token).matches()) {
                throw new InvalidPgnFileException("Error! Unmatched end of comment token: " + token + ".");
            }
            if(END_VARIATION_PATTERN.matcher(token).matches()) {
                throw new InvalidPgnFileException("Error! Unmatched end of variation token: " + token + ".");
            }

            if(START_COMMENT_PATTERN.matcher(token).matches()) {
                readStartCommentToken(errorMessage);
            } else if(START_VARIATION_PATTERN.matcher(token).matches()) {
                readStartVariationToken(errorMessage);
            } else if(ANNOTATION_GLYPH_PATTERN.matcher(token).matches()) {
                // ignores numerical annotation glyphs for now
            } else if(WHITE_MOVE_NUMBER_PATTERN.matcher(token).matches()) {
                readMoveNumberToken(token, PieceColor.WHITE, errorMessage);
            } else if(BLACK_MOVE_NUMBER_PATTERN.matcher(token).matches()) {
                readMoveNumberToken(token, PieceColor.BLACK, errorMessage);
            } else if(GAME_RESULT_PATTERN.matcher(token).matches()){
                readGameResultToken(token, processor, errorMessage);
            } else {
                readMoveToken(token, processor, errorMessage);
            }
            
            justReadMoveNumber = false;
        }
    }

    private void readFirstToken(String token, PgnProcessor processor, String errorMessage) {
        if(GAME_RESULT_PATTERN.matcher(token).matches()) {
            // special case: there are no moves in the movetext section!
            processor.setInitialMove(1, PieceColor.WHITE);
            readGameResultToken(token, processor, errorMessage);
            return;
        }
        
        // Apart from the special case above, the first token
        // must be either a white move number or a black move number.
        Matcher matcher = WHITE_MOVE_NUMBER_PATTERN.matcher(token);
        if(matcher.matches()) {
            moveNumber = Integer.parseInt(matcher.group(1));
            colorToMove = PieceColor.WHITE;
        } else {
            matcher = BLACK_MOVE_NUMBER_PATTERN.matcher(token);
            if(matcher.matches()) {
                moveNumber = Integer.parseInt(matcher.group(1));
                colorToMove = PieceColor.BLACK;
            } else {
                errorMessage += "this can't be the first token in the movetext section;";
                throw new InvalidPgnFileException(errorMessage);
            }
        }
        
        processor.setInitialMove(moveNumber, colorToMove);
        readMoveNumber = true;
        justReadMoveNumber = true;
        isFirstToken = false;
    }
    
    private void readStartCommentToken(String errorMessage) {
        if(justReadMoveNumber) {
            errorMessage += "a comment can't be specified at this location.";
            throw new InvalidPgnFileException(errorMessage);
        }
        isInsideComment = true;
        readMoveNumber = false;
    }

    private void readStartVariationToken(String errorMessage) {
        if(justReadMoveNumber) {
            errorMessage += "a variation can't be specified at this location.";
            throw new InvalidPgnFileException(errorMessage);
        }
        variationNestLevel++;
        readMoveNumber = false;
    }

    private void readMoveNumberToken(String token, PieceColor moveColor, String errorMessage) {
        if(readMoveNumber || (colorToMove != moveColor)) {
            errorMessage += "this move number can't be specified at this location.";
            throw new InvalidPgnFileException(errorMessage);
        }
        Pattern moveNumberPattern = moveColor == PieceColor.BLACK ? BLACK_MOVE_NUMBER_PATTERN :
                WHITE_MOVE_NUMBER_PATTERN;
        Matcher matcher = moveNumberPattern.matcher(token);
        matcher.matches();
        int moveNumber = Integer.parseInt(matcher.group(1));
        if(moveNumber != this.moveNumber) {
            errorMessage += "invalid move number (expected " + this.moveNumber + ").";
            throw new InvalidPgnFileException(errorMessage);
        }
        readMoveNumber = true;
        justReadMoveNumber = true;
    }

    private void readGameResultToken(String token, PgnProcessor processor, String errorMessage) {
        if(justReadMoveNumber) {
            errorMessage += "a game termination marker can't be specified at this location.";
            throw new InvalidPgnFileException(errorMessage);
        }
        processor.endGame(PgnGameOutcome.parse(token));
        gameEnded = true;
    }

    private void readMoveToken(String token, PgnProcessor processor, String errorMessage) {
        if(!readMoveNumber) {
            throw new InvalidPgnFileException(errorMessage + "missing move number specifier.");
        }
        processor.processMove(token);
        colorToMove = colorToMove.getOppositeColor();
        if(colorToMove == PieceColor.WHITE) {
            moveNumber++;
            readMoveNumber = false;
        }
    }

}
