/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.chess.pgn;

import java.io.BufferedWriter;
import java.io.PrintWriter;
import java.io.Writer;
import java.util.regex.Pattern;

import com.haundrix.games.model.Piece.PieceColor;

class PgnWriter {
    
    private static final Pattern TAG_NAME_PATTERN = Pattern.compile("[A-Z]\\w*");
    
    private static final int MAX_MOVETEXT_LINE_LENGTH = 80;

    private PrintWriter writer;
    
    private boolean isWritingTagPairSection = true;
    private boolean wroteFirstMove = false;
    private int currentLineLength = 0;
    
    private int moveNumber = 1;
    private PieceColor colorToMove = PieceColor.WHITE;
    
    public PgnWriter(Writer writer) {
        this.writer = new PrintWriter(new BufferedWriter(writer));
    }

    public boolean isWritingTagPairSection() {
        return isWritingTagPairSection;
    }
    
    public void writeTagPair(String tagName, String tagValue) {
        checkValidStateMethod(true);
        
        if(!TAG_NAME_PATTERN.matcher(tagName).matches()) {
            String errorMessage = "Error! The tag name must start with an uppercase letter ";
            errorMessage += "and contain only alphanumeric characters or the underscore.";
            throw new IllegalArgumentException(errorMessage);
        }
        
        writer.println("[" + tagName + " \"" + tagValue + "\"]");
    }
    
    public void closeTagPairSection() {
        checkValidStateMethod(true);
        
        writer.println();
        isWritingTagPairSection = false;
    }
    
    public void setInitialMove(int moveNumber, PieceColor colorToMove) {
        checkValidStateMethod(false);

        if(wroteFirstMove) {
            String errorMessage = "Error! Cannot set the initial move number after the move list ";
            errorMessage += "has already started being written.";
            throw new IllegalStateException(errorMessage);
        }
        
        this.moveNumber = moveNumber;
        this.colorToMove = colorToMove;
    }
    
    public void writeMove(String moveDescription) {
        if(colorToMove == PieceColor.WHITE) {
            writeMovetextToken(moveNumber + ".", true);
        } else {
            if(!wroteFirstMove) {
                writeMovetextToken(moveNumber + "...", true);
            }
        }
        
        writeMovetextToken(moveDescription, true);
        
        colorToMove = colorToMove.getOppositeColor();
        if(colorToMove == PieceColor.WHITE) {
            moveNumber++;
        }
        
        wroteFirstMove = true;
    }
    
    public void closeMovetextSection(PgnGameOutcome gameResult) {
        writeMovetextToken(gameResult.getDescription(), false);
        writer.println();
        writer.flush();
    }

    private void checkValidStateMethod(boolean isTagPairSectionMethod) {
        if(isTagPairSectionMethod && !isWritingTagPairSection) {
            throw new IllegalStateException("Error! This method may only be called when writing the tag pair section.");
        } else if(!isTagPairSectionMethod && isWritingTagPairSection) {
            throw new IllegalStateException("Error! This method may only be called when writing the movetext section.");
        }
    }
    
    private void writeMovetextToken(String token, boolean addSpace) {
        checkValidStateMethod(false);
        
        int tokenLength = token.length();
        if(addSpace) {
            tokenLength++;
        }
        if(currentLineLength + tokenLength > MAX_MOVETEXT_LINE_LENGTH) {
            writer.println();
            currentLineLength = 0;
        }
        
        writer.print(token);
        if(addSpace) {
            writer.print(" ");
        }
        currentLineLength += tokenLength;
    }
    
}
