/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.controller;

import java.io.InputStream;

import org.apache.commons.io.IOUtils;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.graphics.Device;
import org.eclipse.swt.graphics.ImageData;

import com.haundrix.swt.SwtUtils;

// For the moment the cursor theme is hardcoded!
public class CursorTheme {
    
    private static final String CURSORS_DIR = "/rsc/cursors/";
    
    private static final int WAITING_CURSOR_FRAME_COUNT = 36;

    private Cursor defaultCursor;
    private Cursor ownPieceHoverCursor;
    private Cursor enemyPieceHoverCursor;
    private Cursor destinationSquareCursor;
    private Cursor pieceGrabCursor;
    private Cursor forbiddenCursor;
    private Cursor gameOverCursor;
    
    private Cursor[] waitingCursors;
    
    public CursorTheme(Device device) {
        // Chooses the appropriate image format based on the platform.
        // On Windows cursors created from transparent PNG images are rendered incorrectly,
        // whereas on Linux cursor images from BMP files aren't loaded.
        // See this bug:
        // https://bugs.eclipse.org/bugs/show_bug.cgi?id=281695
        // It has already been fixed in recent builds, and once it gets through a production
        // release we may use the same PNG images both on Linux and Windows.
        String extension = SwtUtils.isWindows() ? ".bmp" : ".png";
        
        InputStream defaultImage = CursorTheme.class.getResourceAsStream(CURSORS_DIR + "default" + extension);
        defaultCursor = new Cursor(device, new ImageData(defaultImage), 8, 5);
        IOUtils.closeQuietly(defaultImage);
        
        InputStream ownPieceHoverImage = CursorTheme.class.getResourceAsStream(CURSORS_DIR + "ownPieceHover"
                + extension);
        ownPieceHoverCursor = new Cursor(device, new ImageData(ownPieceHoverImage), 15, 7);
        IOUtils.closeQuietly(ownPieceHoverImage);
        
        InputStream enemyPieceHoverImage = CursorTheme.class.getResourceAsStream(CURSORS_DIR +
                "enemyPieceHover" + extension);
        enemyPieceHoverCursor = new Cursor(device, new ImageData(enemyPieceHoverImage), 14, 5);
        IOUtils.closeQuietly(enemyPieceHoverImage);
        
        InputStream destinationSquareImage = CursorTheme.class.getResourceAsStream(CURSORS_DIR +
                "destinationSquare" + extension);
        destinationSquareCursor = new Cursor(device, new ImageData(destinationSquareImage), 14, 5);
        IOUtils.closeQuietly(destinationSquareImage);
        
        InputStream pieceGrabImage = CursorTheme.class.getResourceAsStream(CURSORS_DIR + "pieceGrab" + extension);
        pieceGrabCursor = new Cursor(device, new ImageData(pieceGrabImage), 16, 13);
        IOUtils.closeQuietly(pieceGrabImage);
        
        InputStream forbiddenImage = CursorTheme.class.getResourceAsStream(CURSORS_DIR + "forbidden" + extension);
        forbiddenCursor = new Cursor(device, new ImageData(forbiddenImage), 20, 14);
        IOUtils.closeQuietly(forbiddenImage);
        
        InputStream gameOverImage = CursorTheme.class.getResourceAsStream(CURSORS_DIR + "gameOver" + extension);
        gameOverCursor = new Cursor(device, new ImageData(gameOverImage), 20, 12);
        IOUtils.closeQuietly(gameOverImage);

        waitingCursors = new Cursor[WAITING_CURSOR_FRAME_COUNT];
        for(int i = 0; i < WAITING_CURSOR_FRAME_COUNT; i++) {
            String fileName = CURSORS_DIR + "waiting" + i + extension;
            InputStream waitingImage = CursorTheme.class.getResourceAsStream(fileName);
            waitingCursors[i] = new Cursor(device, new ImageData(waitingImage), 20, 14);
            IOUtils.closeQuietly(waitingImage);
        }
    }

    public Cursor getDefaultCursor() {
        return defaultCursor;
    }

    public Cursor getOwnPieceHoverCursor() {
        return ownPieceHoverCursor;
    }

    public Cursor getEnemyPieceHoverCursor() {
        return enemyPieceHoverCursor;
    }

    public Cursor getDestinationSquareCursor() {
        return destinationSquareCursor;
    }

    public Cursor getPieceGrabCursor() {
        return pieceGrabCursor;
    }

    public Cursor getForbiddenCursor() {
        return forbiddenCursor;
    }
    
    public Cursor getGameOverCursor() {
        return gameOverCursor;
    }
    
    public int getWaitingCursorFrameCount() {
        return WAITING_CURSOR_FRAME_COUNT;
    }

    public Cursor getWaitingCursor(int frameNumber) {
        return waitingCursors[frameNumber];
    }

    public void dispose() {
        defaultCursor.dispose();
        ownPieceHoverCursor.dispose();
        enemyPieceHoverCursor.dispose();
        destinationSquareCursor.dispose();
        pieceGrabCursor.dispose();
        forbiddenCursor.dispose();
        gameOverCursor.dispose();
        
        for(Cursor waitingCursor : waitingCursors) {
            waitingCursor.dispose();
        }
    }
}
