/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.controller;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.apache.commons.lang.ArrayUtils;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseWheelListener;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.widgets.Composite;

import com.haundrix.animation.FadeCallback;
import com.haundrix.animation.MoveCallback;
import com.haundrix.animation.Mover;
import com.haundrix.games.model.Game;
import com.haundrix.games.model.GameAdapter;
import com.haundrix.games.model.GameResult;
import com.haundrix.games.model.IllegalMoveException;
import com.haundrix.games.model.Move;
import com.haundrix.games.model.Piece;
import com.haundrix.games.model.RectangularBoard;
import com.haundrix.games.model.Square;
import com.haundrix.games.model.Piece.PieceColor;
import com.haundrix.games.view.BoardAdapter;
import com.haundrix.games.view.BoardPiece;
import com.haundrix.games.view.BoardSquare;
import com.haundrix.games.view.BoardSquareEvent;
import com.haundrix.games.view.CrossfadeAnimation;
import com.haundrix.games.view.FadeInAnimation;
import com.haundrix.games.view.FadeOutAnimation;
import com.haundrix.games.view.IBoardAnimation;
import com.haundrix.games.view.IDraggingPieceHandler;
import com.haundrix.games.view.IPieceType;
import com.haundrix.games.view.MoveAnimation;
import com.haundrix.games.view.PieceDragEvent;
import com.haundrix.games.view.PieceGrabEvent;
import com.haundrix.games.view.PieceSet;
import com.haundrix.games.view.RectangularBoardControl;
import com.haundrix.games.view.SquareClickEvent;
import com.haundrix.games.view.SquareClickEvent.MouseButton;
import com.haundrix.graphics.GraphicsUtils;
import com.haundrix.graphics.Highlight;
import com.haundrix.utils.BidirectionalMap;

public class GameController<P extends IPieceType> {
    
    protected Game game;
    protected HighlightScheme highlightScheme;
    private BidirectionalMap<Class<? extends Piece>, P> blackPieceTypes;
    private BidirectionalMap<Class<? extends Piece>, P> whitePieceTypes;
    
    protected RectangularBoardControl<P> boardControl;
    
    private boolean allowUserWhiteMoves = true;
    private boolean allowUserBlackMoves = true;

    private int minMovementDuration = 300;
    private int maxMovementDuration = 1500;
    private int droppedPieceMovementDuration = minMovementDuration / 2;
    private int pieceFadeDuration = 500;
    private int boardTransitionDuration = 700;
    
    private float draggingPieceAlpha = 0.9F;
    private float draggingPieceShadowIntensity = 0.15F;

    private BoardSquare currentBoardSquare;
    private BoardPiece<P> highlightedBoardPiece;
    
    private Square[] allowedDestinations = null;
    private Square[] allowedOrigins = null;
    private BoardPiece<P>[] attackedBoardPieces = null;
    private int selectedDestination = -1;
    private int selectedOrigin = -1;
    private boolean isShowingAllowedMoves = false;

    private CursorTheme cursorTheme;
    private WaitingCursorAnimator waitingCursorAnimator;
    
    private boolean isAnimating = false;
    private boolean isDraggingPiece = false;
    private boolean isMakingInvisibleMove = false;

    private int boardControlPosition;

    protected List<IGameControllerListener> gameControllerListeners = new ArrayList<IGameControllerListener>();

    public GameController(Game aGame, Composite parent, PieceSet<P> pieceSet,
            BidirectionalMap<Class<? extends Piece>, P> blackPieceTypes,
            BidirectionalMap<Class<? extends Piece>, P> whitePieceTypes) {
        game = aGame;
        
        Highlight squareWeakHighlight = new Highlight(new RGB(217.0F, 1.0F, 0.5F), 0.75F, 0.15F);
        Highlight squareStrongHighlight = new Highlight(new RGB(217.0F, 1.0F, 1.0F), 0.75F, 0.15F);
        Highlight movingPieceWeakHighlight = new Highlight(new RGB(217.0F, 1.0F, 0.5F), 0.75F, 0.15F);
        Highlight movingPieceStrongHighlight = new Highlight(new RGB(217.0F, 1.0F, 1.0F), 0.75F, 0.15F);
        Highlight capturingPieceWeakHighlight = new Highlight(new RGB(0.0F, 1.0F, 0.75F), 0.75F, 0.30F);
        Highlight capturingPieceStrongHighlight = new Highlight(new RGB(0.0F, 1.0F, 1.0F), 0.75F, 0.30F);
        highlightScheme = new HighlightScheme(squareWeakHighlight, squareStrongHighlight, movingPieceWeakHighlight,
            movingPieceStrongHighlight, capturingPieceWeakHighlight, capturingPieceStrongHighlight);
        
        RectangularBoard board = game.getBoard();
        boardControl = new RectangularBoardControl<P>(parent, board.getColumnCount(), board.getRowCount(), pieceSet);

        this.blackPieceTypes = blackPieceTypes;
        this.whitePieceTypes = whitePieceTypes;
        
        game.addGameListener(new InternalGameListener());
        
        cursorTheme = new CursorTheme(boardControl.getDisplay());
        boardControl.addDisposeListener(new DisposeListener() {
            @Override
            public void widgetDisposed(DisposeEvent e) {
                cursorTheme.dispose();
            }
        });
        waitingCursorAnimator = new WaitingCursorAnimator(cursorTheme, boardControl);
        
        boardControl.addBoardListener(new InternalBoardListener());
        boardControl.addMouseWheelListener(new MouseWheelListener() {
            @Override
            public void mouseScrolled(MouseEvent e) {
                onMouseScrolled(e);
            }
        });
        
        synchronize();
        boardControlPosition = game.getCurrentPosition();
        
        updateCursor();
    }
    
    public Game getGame() {
        return game;
    }
    
    public HighlightScheme getHighlightScheme() {
        return highlightScheme;
    }

    public void setHighlightScheme(HighlightScheme highlightScheme) {
        this.highlightScheme = highlightScheme;
    }

    public RectangularBoardControl<P> getBoardControl() {
        return boardControl;
    }

    public boolean isAllowUserWhiteMoves() {
        return allowUserWhiteMoves;
    }

    public void setAllowUserWhiteMoves(boolean allowUserWhiteMoves) {
        this.allowUserWhiteMoves = allowUserWhiteMoves;
        updateCursor();
    }

    public boolean isAllowUserBlackMoves() {
        return allowUserBlackMoves;
    }

    public void setAllowUserBlackMoves(boolean allowUserBlackMoves) {
        this.allowUserBlackMoves = allowUserBlackMoves;
        updateCursor();
    }

    public int getMinMovementDuration() {
        return minMovementDuration;
    }

    public void setMinMovementDuration(int minMovementDuration) {
        if(minMovementDuration < 0) {
            throw new IllegalArgumentException("Minimum movement duration cannot be negative.");
        }
        if(minMovementDuration > maxMovementDuration) {
            throw new IllegalArgumentException("Minimum movement duration cannot exceed maximum movement duration.");
        }
        
        this.minMovementDuration = minMovementDuration;
    }

    public int getMaxMovementDuration() {
        return maxMovementDuration;
    }

    public void setMaxMovementDuration(int maxMovementDuration) {
        if(maxMovementDuration < 0) {
            throw new IllegalArgumentException("Maximum movement duration cannot be negative.");
        }
        if(maxMovementDuration < minMovementDuration) {
            throw new IllegalArgumentException("Maximum movement duration cannot be less than the minimum movement duration.");
        }
        
        this.maxMovementDuration = maxMovementDuration;
    }
    
    public void setMovementDurationBounds(int minMovementDuration, int maxMovementDuration) {
        if(minMovementDuration < 0) {
            throw new IllegalArgumentException("Minimum movement duration cannot be negative.");
        }
        if(maxMovementDuration < 0) {
            throw new IllegalArgumentException("Maximum movement duration cannot be negative.");
        }
        if(minMovementDuration > maxMovementDuration) {
            throw new IllegalArgumentException("Maximum movement duration must be higher than the minimum movement duration.");
        }
        
        this.minMovementDuration = minMovementDuration;
        this.maxMovementDuration = maxMovementDuration;
    }

    public int getDroppedPieceMovementDuration() {
        return droppedPieceMovementDuration;
    }

    public void setDroppedPieceMovementDuration(int droppedPieceMovementDuration) {
        this.droppedPieceMovementDuration = droppedPieceMovementDuration;
    }

    public int getPieceFadeDuration() {
        return pieceFadeDuration;
    }

    public void setPieceFadeDuration(int pieceFadeDuration) {
        if(pieceFadeDuration < 0) {
            throw new IllegalArgumentException("Piece fade duration cannot be negative.");
        }
        
        this.pieceFadeDuration = pieceFadeDuration;
    }

    public int getBoardTransitionDuration() {
        return boardTransitionDuration;
    }

    public void setBoardTransitionDuration(int boardTransitionDuration) {
        if(boardTransitionDuration < 0) {
            throw new IllegalArgumentException("Board transition duration cannot be negative.");
        }
        
        this.boardTransitionDuration = boardTransitionDuration;
    }

    public float getDraggingPieceAlpha() {
        return draggingPieceAlpha;
    }

    public void setDraggingPieceAlpha(float draggingPieceAlpha) {
        if((draggingPieceAlpha < 0.0F) || (draggingPieceAlpha > 1.0F)) {
            throw new IllegalArgumentException("Dragging piece alpha must be a number within 0 and 1.");
        }
        
        this.draggingPieceAlpha = draggingPieceAlpha;
    }

    public float getDraggingPieceShadowIntensity() {
        return draggingPieceShadowIntensity;
    }

    public void setDraggingPieceShadowIntensity(float draggingPieceShadowIntensity) {
        if((draggingPieceShadowIntensity < 0.0F) || (draggingPieceShadowIntensity > 1.0F)) {
            throw new IllegalArgumentException("Dragging piece shadow intensity must be a number within 0 and 1.");
        }
        
        this.draggingPieceShadowIntensity = draggingPieceShadowIntensity;
    }

    public void addGameControllerListener(IGameControllerListener gameControllerListener) {
        synchronized(gameControllerListeners) {
            gameControllerListeners.add(gameControllerListener);
        }
    }
    
    public void removeGameControllerListener(IGameControllerListener gameControllerListener) {
        synchronized(gameControllerListeners) {
            gameControllerListeners.remove(gameControllerListener);
        }
    }

    public void makeMove(Move move) {
        if(game.isOver()) {
            throw new IllegalStateException("Error! The game is already ended.");
        }
        if((game.getLastColorToMove() == PieceColor.WHITE) && allowUserWhiteMoves) {
            String errorMessage = "Error! Trying to programatically enter a white move which should be input by the user.";
            throw new IllegalArgumentException(errorMessage);
        }
        if((game.getLastColorToMove() == PieceColor.BLACK) && allowUserBlackMoves) {
            String errorMessage = "Error! Trying to programatically enter a black move which should be input by the user.";
            throw new IllegalArgumentException(errorMessage);
        }
        
        final Move usedMove = move.clone();

        if(!game.isAtEnd()) {
            // no animation to play
            game.move(usedMove);
            return;
        }
        
        if(!game.isLegal(usedMove)) {
            throw new IllegalMoveException();
        }
        
        if(!boardControl.isVisible()) {
            // no need to animate the move if the user won't see it!
            isMakingInvisibleMove = true;
            game.move(usedMove);
            synchronize();
            isMakingInvisibleMove = false;
            updateCursor();
            return;
        }
        
        BoardSquare boardOrigin = GameController.getBoardSquare(move.getOrigin());
        final BoardPiece<P> movingPiece = boardControl.getBoardPiece(boardOrigin);
        movingPiece.setHighlight(highlightScheme.getMovingPieceStrongHighlight());
        
        Square attackedSquare = game.getAttackedSquare(move);
        BoardSquare attackedBoardSquare = GameController.getBoardSquare(attackedSquare);
        BoardPiece<P> attackedPiece = boardControl.getBoardPiece(attackedBoardSquare);
        if(attackedPiece != null) {
            attackedPiece.setHighlight(highlightScheme.getCapturingPieceStrongHighlight());
        }
        
        isAnimating = true;
        game.setLocked(true);
        playMove(usedMove, true, new Runnable() {
            @Override
            public void run() {
                movingPiece.setHighlight(null);
                
                doAfterMoveInput(usedMove.getOrigin(), usedMove.getDestination());
            }
        });
    }

    protected static Square getSquare(BoardSquare boardSquare) {
        if(boardSquare == null) {
            return null;
        }
        
        return new Square(boardSquare.getRow(), boardSquare.getColumn());
    }

    protected static BoardSquare getBoardSquare(Square square) {
        if(square == null) {
            return null;
        }
        
        return new BoardSquare(square.getRow(), square.getColumn());
    }
    
    protected P getPieceType(Class<? extends Piece> pieceClass, PieceColor pieceColor) {
        return pieceColor == PieceColor.BLACK ? blackPieceTypes.getValue(pieceClass) :
            whitePieceTypes.getValue(pieceClass);
    }
    
    protected P getPieceType(Piece piece) {
        if(piece == null) {
            return null;
        }
        
        return getPieceType(piece.getClass(), piece.getColor());
    }

    protected Class<? extends Piece> getPieceClass(P pieceType) {
        Class<? extends Piece> pieceClass = blackPieceTypes.getKey(pieceType);
        return pieceClass != null ? pieceClass : whitePieceTypes.getKey(pieceType);
    }

    protected int getMovementDuration(Move move) {
        BoardSquare boardOrigin = GameController.getBoardSquare(move.getOrigin());
        BoardSquare boardDestination = GameController.getBoardSquare(move.getDestination());
        return getMovementDuration(boardOrigin, boardDestination);
    }

    protected int getMovementDuration(BoardSquare boardOrigin, BoardSquare boardDestination) {
        Point origin = boardControl.getSquareLocation(boardOrigin);
        Point destination = boardControl.getSquareLocation(boardDestination);
        return getMovementDuration(origin, destination);
    }

    protected Move getMove(Square origin, Square destination) {
        return new Move(origin, destination);
    }

    // Subclasses willing to complement the default move playing animation with some
    // custom animations may override this method.
    // This hook method is called after the standard move animation has already played.
    // The supplied end play callback must be called after all animations have ended.
    protected void completePlayingMove(Move move, boolean isNewMove, Runnable endPlayCallback) {
        endPlayCallback.run();
    }

    // Subclasses that override the above method must also override this one in order
    // to undo the custom animations that were performed after the standard move animations.
    // This hook method is called before the standard move animation unplays.
    // The supplied end unplay callback must be called after the custom animations,
    // which will then cause the standard unplay move animations to run and finishing the unmove.
    protected void startUnplayingMove(Move move, Runnable endUnplayCallback) {
        endUnplayCallback.run();
    }

    // This hook method is called at the end of playing/unplaying animations.
    protected void afterFinishedPlaying() {
    }

    private void onMouseScrolled(MouseEvent e) {
        if(!isLocked() && isShowingAllowedMoves) {
            if(e.count > 0) {
                // mouse scroll up
                if(allowedDestinations.length > 0) {
                    selectedDestination++;
                    if(selectedDestination >= allowedDestinations.length) {
                        selectedDestination = 0;
                    }
                    updateSelectedDestinations();
                } else if(allowedOrigins.length > 0) {
                    selectedOrigin++;
                    if(selectedOrigin >= allowedOrigins.length) {
                        selectedOrigin = 0;
                    }
                    updateSelectedOrigins();
                }
            } else if(e.count < 0) {
                // mouse scroll down
                if(allowedDestinations.length > 0) {
                    selectedDestination--;
                    if(selectedDestination < 0) {
                        selectedDestination = allowedDestinations.length - 1;
                    }
                    updateSelectedDestinations();
                } else if(allowedOrigins.length > 0) {
                    selectedOrigin--;
                    if(selectedOrigin < 0) {
                        selectedOrigin = allowedOrigins.length - 1;
                    }
                    updateSelectedOrigins();
                }
            }
        }
    }

    private int getMovementDuration(Point origin, Point destination) {
        double distance = GraphicsUtils.getDistance(origin, destination);
        int squareSize = boardControl.getSquareSize();
        if(distance <= squareSize) {
            return minMovementDuration;
        }
        
        Point topLeftSquareLocation = boardControl.getSquareLocation(new BoardSquare(0, 0));
        Point bottomRightSquareLocation = boardControl.getSquareLocation(new BoardSquare(boardControl.getRowCount() - 1, boardControl.getColumnCount() - 1));
        double maxDistance = GraphicsUtils.getDistance(topLeftSquareLocation, bottomRightSquareLocation);
        
        if(distance >= maxDistance) {
            return maxMovementDuration;
        }
        
        return (int) (minMovementDuration + (maxMovementDuration - minMovementDuration) /
                (maxDistance - squareSize) * (distance - squareSize));
    }

    private boolean isWaiting() {
        if(!game.isAtEnd() || game.isFinished()) {
            return false;
        }
        
        PieceColor colorToMove = game.getState().getColorToMove();
        return colorToMove == PieceColor.WHITE ? !allowUserWhiteMoves : !allowUserBlackMoves;
    }

    private boolean isLocked() {
        return isAnimating || !game.isAtEnd() || game.isFinished() || isWaiting();
    }

    private void updateAllowedMoves() {
        clearAllowedMoves();
        setUpAllowedMoves();
    }

    private void clearAllowedMoves() {
        if(isShowingAllowedMoves) {
            if(allowedDestinations.length > 0) {
                for(int i = 0; i < allowedDestinations.length; i++) {
                    BoardSquare boardSquare = GameController.getBoardSquare(allowedDestinations[i]);
                    boardControl.setSquareHighlight(boardSquare, null);
                    if((attackedBoardPieces[i] != null) && !attackedBoardPieces[i].isDisposed()) {
                        attackedBoardPieces[i].setHighlight(null);
                    }
                }
            } else if(allowedOrigins.length > 0) {
                for(int i = 0; i < allowedOrigins.length; i++) {
                    BoardSquare boardSquare = GameController.getBoardSquare(allowedOrigins[i]);
                    BoardPiece<P> boardPiece = boardControl.getBoardPiece(boardSquare);
                    // this piece will no longer be in its origin if it was the moving piece
                    if(boardPiece != null) {
                        boardPiece.setHighlight(null);
                    }
                    if((attackedBoardPieces[i] != null) && !attackedBoardPieces[i].isDisposed()) {
                        attackedBoardPieces[i].setHighlight(null);
                    }
                }
            }
            isShowingAllowedMoves = false;
        }
        if((highlightedBoardPiece != null) && !highlightedBoardPiece.isDisposed()) {
            highlightedBoardPiece.setHighlight(null);
        }
        highlightedBoardPiece = null;
        
        allowedDestinations = null;
        allowedOrigins = null;
        attackedBoardPieces = null;
        selectedDestination = -1;
        selectedOrigin = -1;
    }

    private void setUpAllowedMoves() {
        if(!isLocked() && (currentBoardSquare != null)) {
            Square currentSquare = getSquare(currentBoardSquare);
            Highlight pieceHighlight = null;
            allowedDestinations = game.getLegalDestinationsFrom(currentSquare);
            if(allowedDestinations.length > 0) {
                Arrays.sort(allowedDestinations, new SquareComparator(currentSquare));
                initAttackedBoardPieces(allowedDestinations, true);
                selectedDestination = 0;
                pieceHighlight = highlightScheme.getMovingPieceStrongHighlight();
            } else {
                allowedOrigins = game.getLegalOriginsTo(currentSquare);
                if(allowedOrigins.length > 0) {
                    Arrays.sort(allowedOrigins, new SquareComparator(currentSquare));
                    initAttackedBoardPieces(allowedOrigins, false);
                    selectedOrigin = 0;
                    pieceHighlight = highlightScheme.getCapturingPieceStrongHighlight();
                }
            }
            
            highlightedBoardPiece = boardControl.getBoardPiece(currentBoardSquare);
            if(highlightedBoardPiece != null) {
                highlightedBoardPiece.setHighlight(pieceHighlight);
            }
        }
    }

    @SuppressWarnings("unchecked")
    private void initAttackedBoardPieces(Square[] squares, boolean areDestinationSquares) {
        Square currentSquare = getSquare(currentBoardSquare);
        attackedBoardPieces = new BoardPiece[squares.length];
        for(int i = 0; i < squares.length; i++) {
            Move move = areDestinationSquares ? new Move(currentSquare, squares[i]) :
                new Move(squares[i], currentSquare);
            Square attackedPieceSquare = game.getAttackedSquare(move);
            BoardSquare attackedPieceBoardSquare = GameController.getBoardSquare(attackedPieceSquare);
            attackedBoardPieces[i] = boardControl.getBoardPiece(attackedPieceBoardSquare);
        }
    }
    
    private void updateCursor() {
        if(isWaiting()) {
            if(!waitingCursorAnimator.isRunning()) {
                waitingCursorAnimator.start();
            }
            return;
        } else {
            if(waitingCursorAnimator.isRunning()) {
                waitingCursorAnimator.stop();
            }
        }
        
        if(game.isFinished()) {
            boardControl.setCursor(cursorTheme.getGameOverCursor());
            return;
        }
        
        if((currentBoardSquare == null) || !game.isAtEnd()) {
            boardControl.setCursor(cursorTheme.getDefaultCursor());
        } else {
            if(isDraggingPiece) {
                boardControl.setCursor(cursorTheme.getPieceGrabCursor());
            } else {
                if((allowedDestinations != null) && (allowedDestinations.length > 0)) {
                    boardControl.setCursor(cursorTheme.getOwnPieceHoverCursor());
                } else if((allowedOrigins != null) && (allowedOrigins.length > 0)) {
                    if(boardControl.getPiece(currentBoardSquare) != null) {
                        boardControl.setCursor(cursorTheme.getEnemyPieceHoverCursor());
                    } else {
                        boardControl.setCursor(cursorTheme.getDestinationSquareCursor());
                    }
                } else {
                    boardControl.setCursor(cursorTheme.getForbiddenCursor());
                }
            }
        }
    }

    private void updateSelectedDestinations() {
        Highlight squareHighlight;
        Highlight pieceHighlight;
        for(int i = 0; i < allowedDestinations.length; i++) {
            if(i == selectedDestination) {
                squareHighlight = highlightScheme.getSquareStrongHighlight();
                pieceHighlight = highlightScheme.getCapturingPieceStrongHighlight();
            } else {
                squareHighlight = highlightScheme.getSquareWeakHighlight();
                pieceHighlight = highlightScheme.getCapturingPieceWeakHighlight();
            }
            BoardSquare boardSquare = GameController.getBoardSquare(allowedDestinations[i]);
            boardControl.setSquareHighlight(boardSquare, squareHighlight);
            if(attackedBoardPieces[i] != null) {
                attackedBoardPieces[i].setHighlight(pieceHighlight);
            }
        }
        isShowingAllowedMoves = true;
    }

    private void updateSelectedOrigins() {
        Highlight movingPieceHighlight;
        Highlight capturingPieceHighlight;
        for(int i = 0; i < allowedOrigins.length; i++) {
            if(i == selectedOrigin) {
                movingPieceHighlight = highlightScheme.getMovingPieceStrongHighlight();
                capturingPieceHighlight = highlightScheme.getCapturingPieceStrongHighlight();
            } else {
                movingPieceHighlight = highlightScheme.getMovingPieceWeakHighlight();
                capturingPieceHighlight = null;
            }
            BoardSquare boardSquare = GameController.getBoardSquare(allowedOrigins[i]);
            boardControl.getBoardPiece(boardSquare).setHighlight(movingPieceHighlight);
            BoardPiece<P> attackedBoardPiece = attackedBoardPieces[i];
            if((attackedBoardPiece != null) && (attackedBoardPiece != highlightedBoardPiece)) {
                attackedBoardPiece.setHighlight(capturingPieceHighlight);
            }
        }
        isShowingAllowedMoves = true;
    }

    private void synchronize() {
        RectangularBoard board = game.getBoard();
        for(Square square : board) {
            P pieceType = getPieceType(board.getPiece(square));
            BoardSquare boardSquare = GameController.getBoardSquare(square);
            if(boardControl.getPiece(boardSquare) != pieceType) {
                boardControl.setPiece(pieceType, boardSquare);
            }
        }
        
        boardControlPosition = game.getCurrentPosition();
    }

    private void finishMoving() {
        isAnimating = false;
        updateAllowedMoves();
        game.setLocked(false);
        if(waitingCursorAnimator.isRunning()) {
            waitingCursorAnimator.stop();
        }
        updateCursor();
    }
    
    private void doAfterMoveInput(Square origin, Square destination) {
        game.setLocked(false);
        if(!game.isOver()) {
            boardControlPosition++;
            game.move(getMove(origin, destination));
        }
        synchronize();
        finishMoving();
    }

    private void playMove(final Move move, final boolean isNewMove, final Runnable endPlayCallback) {
        final Square origin = move.getOrigin();
        final Square destination = move.getDestination();
        BoardSquare boardOrigin = GameController.getBoardSquare(origin);
        final BoardSquare boardDestination = GameController.getBoardSquare(destination);
        final int movementDuration = getMovementDuration(boardOrigin, boardDestination);
        Square attackedSquare = isNewMove ? game.getAttackedSquare(move) :
            game.getPreviousAttackedSquare();
        final BoardSquare attackedBoardSquare = GameController.getBoardSquare(attackedSquare);
        final BoardPiece<P> capturedBoardPiece = boardControl.getBoardPiece(attackedBoardSquare);
        
        final Runnable usedEndPlayCallback = new Runnable() {
            @Override
            public void run() {
                if(!(game.isOver() && isNewMove)) {
                    completePlayingMove(move, isNewMove, new Runnable() {
                        @Override
                        public void run() {
                            endPlayCallback.run();
                            afterFinishedPlaying();
                        }
                    });
                } else {
                    endPlayCallback.run();
                }
            }
        };
        
        if(capturedBoardPiece == null) {
            final MoveAnimationEvent event = new MoveAnimationEvent(movementDuration, move, isNewMove);
            notifyBeginMoveAnimation(event);
            boardControl.movePiece(boardOrigin, boardDestination, movementDuration, new Runnable() {
                @Override
                public void run() {
                    notifyEndMoveAnimation(event);
                    usedEndPlayCallback.run();
                }
            });
        } else {
            if(attackedSquare.equals(destination)) {
                final MoveAnimationEvent event = new MoveAnimationEvent(movementDuration, move, isNewMove);
                notifyBeginMoveAnimation(event);
                boardControl.movePiece(boardOrigin, boardDestination, movementDuration,
                    new MoveCallback() {
                        
                        private boolean fadeOutTriggered = false;
                        
                        private int fadeOutDuration = Math.min(pieceFadeDuration, movementDuration);

                        @Override
                        public void moved(Mover mover) {
                            if(!fadeOutTriggered && (mover.getRemainingTime() < fadeOutDuration)) {
                                boardControl.fadeOutPiece(boardDestination, fadeOutDuration);
                                fadeOutTriggered = true;
                            }
                        }

                        @Override
                        public void ended(Mover mover) {
                            notifyEndMoveAnimation(event);
                            usedEndPlayCallback.run();
                        }
                        
                    });
            } else {
                final MoveAnimationEvent event = new MoveAnimationEvent(movementDuration + pieceFadeDuration, move,
                        isNewMove);
                notifyBeginMoveAnimation(event);
                boardControl.movePiece(boardOrigin, boardDestination, movementDuration, new Runnable() {
                    @Override
                    public void run() {
                        if(!(game.isOver() && isNewMove)) {
                            boardControl.fadeOutPiece(attackedBoardSquare, pieceFadeDuration, new Runnable() {
                                @Override
                                public void run() {
                                    notifyEndMoveAnimation(event);
                                    usedEndPlayCallback.run();
                                }
                            });
                        } else {
                            usedEndPlayCallback.run();
                        }
                    }
                });
            }
        }
    }

    private void unplayMove(final Move move, final Runnable endUnplayCallback) {
        startUnplayingMove(move, new Runnable() {
            @Override
            public void run() {
                final Runnable usedEndUnplayCallback = new Runnable() {
                    @Override
                    public void run() {
                        endUnplayCallback.run();
                        afterFinishedPlaying();
                    }
                };
                
                Square destination = move.getDestination();
                final BoardSquare boardOrigin = GameController.getBoardSquare(move.getOrigin());
                final BoardSquare boardDestination = GameController.getBoardSquare(destination);
                final int movementDuration = getMovementDuration(boardOrigin, boardDestination);
                Square attackedSquare = game.getAttackedSquare(move);
                final Piece attackedPiece = game.getBoard().getPiece(attackedSquare);
                if(attackedPiece == null) {
                    boardControl.movePiece(boardDestination, boardOrigin, movementDuration, usedEndUnplayCallback);
                } else {
                    final BoardSquare attackedBoardSquare =
                        GameController.getBoardSquare(attackedSquare);
                    final P attackedPieceType = getPieceType(attackedPiece);
                    if(attackedSquare.equals(destination)) {
                        boardControl.movePiece(boardDestination, boardOrigin, movementDuration,
                            new MoveCallback(usedEndUnplayCallback) {
                                @Override
                                public void started(Mover mover) {
                                    int fadeInDuration = Math.min(pieceFadeDuration, movementDuration);
                                    boardControl.fadeInPiece(attackedPieceType, attackedBoardSquare,
                                        fadeInDuration);
                                }
                            });
                    } else {
                        boardControl.fadeInPiece(attackedPieceType, attackedBoardSquare, pieceFadeDuration,
                            null, new Runnable() {
                                @Override
                                public void run() {
                                    boardControl.movePiece(boardDestination, boardOrigin, movementDuration,
                                        usedEndUnplayCallback);
                                }
                            });
                    }
                }
            }
        });
    }
    
    private void notifyPieceDraggedOffBoard(AnimationEvent event) {
        synchronized(gameControllerListeners) {
            for(IGameControllerListener gameControllerListener : gameControllerListeners) {
                gameControllerListener.pieceDraggedOffBoard(event);
            }
        }
    }
    
    private void notifyIllegalMoveAttempted(AnimationEvent event) {
        synchronized(gameControllerListeners) {
            for(IGameControllerListener gameControllerListener : gameControllerListeners) {
                gameControllerListener.illegalMoveAttempted(event);
            }
        }
    }
    
    private void notifyBeginMoveAnimation(MoveAnimationEvent event) {
        synchronized(gameControllerListeners) {
            for(IGameControllerListener gameControllerListener : gameControllerListeners) {
                gameControllerListener.beginMoveAnimation(event);
            }
        }
    }
    
    private void notifyEndMoveAnimation(MoveAnimationEvent event) {
        synchronized(gameControllerListeners) {
            for(IGameControllerListener gameControllerListener : gameControllerListeners) {
                gameControllerListener.endMoveAnimation(event);
            }
        }
    }

    private void notifyBeginTransitionAnimation(AnimationEvent event) {
        synchronized(gameControllerListeners) {
            for(IGameControllerListener gameControllerListener : gameControllerListeners) {
                gameControllerListener.beginTransitionAnimation(event);
            }
        }
    }
    
    private void notifyEndTransitionAnimation(AnimationEvent event) {
        synchronized(gameControllerListeners) {
            for(IGameControllerListener gameControllerListener : gameControllerListeners) {
                gameControllerListener.endTransitionAnimation(event);
            }
        }
    }
    
    private class InternalGameListener extends GameAdapter {
        
        private RectangularBoard boardBeforeChange = null;
        
        @Override
        public void beforePositionChange(Game game) {
            RectangularBoard board = game.getBoard();
            
            boardBeforeChange = new RectangularBoard(board.getRowCount(), board.getColumnCount());
            for(Square square : board) {
                boardBeforeChange.setPiece(board.getPiece(square), square);
            }
        }

        @Override
        public void afterPositionChange(Game changedGame) {
            final int gamePosition = game.getCurrentPosition();
            if(!isAnimating && (boardControlPosition != gamePosition) && !isMakingInvisibleMove) {
                isAnimating = true;
                game.setLocked(true);
                if(gamePosition == boardControlPosition - 1) {
                    Move move = game.getNextMove();
                    clearAllowedMoves();
                    unplayMove(move, new Runnable() {
                        @Override
                        public void run() {
                            synchronize();
                            boardControlPosition = gamePosition;
                            isAnimating = false;
                            game.setLocked(false);
                            updateCursor();
                        }
                    });
                } else if(gamePosition == boardControlPosition + 1) {
                    Move move = game.getPreviousMove();
                    playMove(move, false, new Runnable() {
                        @Override
                        public void run() {
                            synchronize();
                            boardControlPosition = gamePosition;
                            isAnimating = false;
                            setUpAllowedMoves();
                            game.setLocked(false);
                            updateCursor();
                        }
                    });
                } else {
                    final AnimationEvent event = new AnimationEvent(boardTransitionDuration);
                    notifyBeginTransitionAnimation(event);
                    clearAllowedMoves();
                    boardControl.animate(getTransitionAnimations(), boardTransitionDuration, new Runnable() {
                        @Override
                        public void run() {
                            synchronize();
                            boardControlPosition = gamePosition;
                            isAnimating = false;
                            updateAllowedMoves();
                            afterFinishedPlaying();
                            game.setLocked(false);
                            updateCursor();
                            
                            notifyEndTransitionAnimation(event);
                        }
                    });
                }
                boardBeforeChange = null;
            }
        }

        @Override
        public void ended(GameResult result) {
            clearAllowedMoves();
            if(isDraggingPiece) {
                final BoardPiece<P> draggingPiece = boardControl.getDraggingPiece();
                Point origin = draggingPiece.getLocation();
                Point destination = boardControl.getSquareLocation(boardControl.getDraggingStartSquare());
                int movementDuration = getMovementDuration(origin, destination);
                boardControl.cancelPieceDragAndMovePieceToOriginalSquare(movementDuration, new Runnable() {
                    @Override
                    public void run() {
                        draggingPiece.setAlpha(1.0F);
                        draggingPiece.setShadowIntensity(0.0F);
                        
                        isAnimating = false;
                        isDraggingPiece = false;
                        game.setLocked(false);
                        synchronize();
                        updateCursor();
                    }
                });
            } else {
                updateCursor();
            }
        }
        
        private IBoardAnimation[] getTransitionAnimations() {
            List<IBoardAnimation> animations = new ArrayList<IBoardAnimation>();
            
            RectangularBoard currentBoard = game.getBoard();
            for(Piece oldPiece : boardBeforeChange.getPieces()) {
                Square oldSquare = boardBeforeChange.findPiece(oldPiece);
                Square newSquare = currentBoard.findPiece(oldPiece);
                BoardSquare oldBoardSquare = GameController.getBoardSquare(oldSquare);
                if(newSquare != null) {
                    if(!newSquare.equals(oldSquare)) {
                        BoardSquare destination = GameController.getBoardSquare(newSquare);
                        animations.add(new MoveAnimation(oldBoardSquare, destination));
                    }
                } else {
                    Piece newPiece = currentBoard.getPiece(oldSquare);
                    if(newPiece == null) {
                        animations.add(new FadeOutAnimation(oldBoardSquare));
                    } else {
                        if(boardBeforeChange.findPiece(newPiece) != null) {
                            animations.add(new FadeOutAnimation(oldBoardSquare));
                        } else {
                            animations.add(new CrossfadeAnimation<P>(oldBoardSquare, getPieceType(newPiece)));
                        }
                    }
                }
            }
            
            for(Piece newPiece : currentBoard.getPieces()) {
                boolean isNewPiece = boardBeforeChange.findPiece(newPiece) == null;
                if(isNewPiece) {
                    Square newSquare = currentBoard.findPiece(newPiece);
                    Piece oldPiece = boardBeforeChange.getPiece(newSquare);
                    boolean squareWasEmpty = oldPiece == null;
                    boolean oldPieceMoved = !squareWasEmpty && (currentBoard.findPiece(oldPiece) != null);
                    if(squareWasEmpty || oldPieceMoved) {
                        BoardSquare newBoardSquare = GameController.getBoardSquare(newSquare);
                        animations.add(new FadeInAnimation<P>(newBoardSquare, getPieceType(newPiece)));
                    }
                }
            }
            
            return animations.toArray(new IBoardAnimation[0]);
        }

    }
    
    private class InternalBoardListener extends BoardAdapter<P> {

        @Override
        public void squareLeft(BoardSquareEvent<P> event) {
            currentBoardSquare = null;
            if(!isLocked()) {
                updateAllowedMoves();
                updateCursor();
            }
        }

        @Override
        public void squareEntered(BoardSquareEvent<P> event) {
            currentBoardSquare = event.getSquare();
            if(!isLocked()) {
                updateAllowedMoves();
                updateCursor();
            }
        }

        @Override
        public void squareHovered(BoardSquareEvent<P> event) {
            if(!isLocked() && !isShowingAllowedMoves) {
                if((allowedDestinations != null) && (allowedDestinations.length > 0)) {
                    updateSelectedDestinations();
                } else if((allowedOrigins != null) && (allowedOrigins.length > 0)) {
                    updateSelectedOrigins();
                }
            }
        }

        @Override
        public void squareClicked(SquareClickEvent<P> event) {
            if(!isLocked() && (event.getClickedButton() == MouseButton.LEFT_BUTTON)) {
                final Square origin;
                final Square destination;
                if((allowedDestinations != null) && (allowedDestinations.length > 0)) {
                    origin = GameController.getSquare(event.getSquare());
                    destination = allowedDestinations[selectedDestination];
                } else if((allowedOrigins != null) && (allowedOrigins.length > 0)) {
                    origin = allowedOrigins[selectedOrigin];
                    destination = GameController.getSquare(event.getSquare());
                } else {
                    origin = destination = null;
                }
                
                if(destination != null) {
                    isAnimating = true;
                    game.setLocked(true);
                    BoardSquare boardOrigin = GameController.getBoardSquare(origin);
                    final BoardPiece<P> movingPiece = boardControl.getBoardPiece(boardOrigin);
                    playMove(new Move(origin, destination), true, new Runnable() {
                        @Override
                        public void run() {
                            movingPiece.setHighlight(null);
                            doAfterMoveInput(origin, destination);
                        }
                    });
                }
            }
        }

        @Override
        public void pieceGrabbed(PieceGrabEvent<P> event) {
            if(!isLocked() && (allowedDestinations != null) && (allowedDestinations.length > 0)) {
                isAnimating = true;
                isDraggingPiece = true;
                game.setLocked(true);
                BoardPiece<P> grabbedPiece = event.getGrabbedPiece();
                grabbedPiece.setAlpha(draggingPieceAlpha);
                grabbedPiece.setShadowIntensity(draggingPieceShadowIntensity);
                selectedDestination = -1;
                updateCursor();
                updateSelectedDestinations();
            } else {
                event.allow = false;
            }
        }

        @Override
        public void pieceDragged(PieceDragEvent<P> event) {
            currentBoardSquare = event.getTargetSquare();
            Square currentSquare = GameController.getSquare(currentBoardSquare);
            if((currentBoardSquare != null) && ArrayUtils.contains(allowedDestinations, currentSquare)) {
                selectedDestination = ArrayUtils.indexOf(allowedDestinations, currentSquare);
            } else {
                selectedDestination = -1;
            }
            updateSelectedDestinations();
        }

        @Override
        public void pieceDropped(PieceDragEvent<P> event, IDraggingPieceHandler draggingPieceHandler) {
            final BoardPiece<P> draggedPiece = event.getDraggedPiece();
            final BoardSquare draggingStartSquare = event.getDraggingStartSquare();
            final BoardSquare targetSquare = event.getTargetSquare();

            final Square origin = GameController.getSquare(draggingStartSquare);
            final Square destination = GameController.getSquare(targetSquare);
            
            final Runnable defaultEndMoveCallback = new Runnable() {
                @Override
                public void run() {
                    draggedPiece.setAlpha(1.0F);
                    draggedPiece.setShadowIntensity(0.0F);
                }
            };
            
            isDraggingPiece = false;
            
            if((targetSquare == null) || !ArrayUtils.contains(allowedDestinations, destination)) {
                Point originLocation = draggedPiece.getLocation();
                Point destinationLocation = boardControl.getSquareLocation(draggingStartSquare);
                int movementDuration = getMovementDuration(originLocation, destinationLocation);
                draggingPieceHandler.moveBackToOriginalSquare(movementDuration, new Runnable() {
                    @Override
                    public void run() {
                        defaultEndMoveCallback.run();
                        finishMoving();
                    }
                });
                if(targetSquare == null) {
                    notifyPieceDraggedOffBoard(new AnimationEvent(movementDuration));
                } else {
                    notifyIllegalMoveAttempted(new AnimationEvent(movementDuration));
                }
            } else {
                final Move move = new Move(origin, destination);
                final Runnable endMoveCallback = new Runnable() {
                    @Override
                    public void run() {
                        completePlayingMove(move, true, new Runnable() {
                            @Override
                            public void run() {
                                defaultEndMoveCallback.run();
                                doAfterMoveInput(origin, destination);
                                afterFinishedPlaying();
                            }
                        });
                    }
                };
                Square attackedSquare = game.getAttackedSquare(new Move(origin, destination));
                final BoardSquare attackedBoardSquare = GameController.getBoardSquare(attackedSquare);
                BoardPiece<P> capturedBoardPiece = boardControl.getBoardPiece(attackedBoardSquare);
                if(capturedBoardPiece == null) {
                    final MoveAnimationEvent moveEvent = new MoveAnimationEvent(droppedPieceMovementDuration, move, true);
                    notifyBeginMoveAnimation(moveEvent);
                    draggingPieceHandler.moveToTargetSquare(droppedPieceMovementDuration, new Runnable() {
                        @Override
                        public void run() {
                            notifyEndMoveAnimation(moveEvent);
                            endMoveCallback.run();
                        }
                    });
                } else {
                    if(attackedSquare.equals(destination)) {
                        final MoveAnimationEvent moveEvent = new MoveAnimationEvent(minMovementDuration + pieceFadeDuration, move, true);
                        notifyBeginMoveAnimation(moveEvent);
                        draggingPieceHandler.moveToTargetSquare(minMovementDuration, new MoveCallback() {
                            
                            @Override
                            public void started(final Mover mover) {
                                mover.pause();
                                boardControl.fadeOutPiece(targetSquare, pieceFadeDuration, new FadeCallback() {
                                    @Override
                                    public void ended() {
                                        mover.resume();
                                    }
                                });
                            }

                            @Override
                            public void ended(Mover mover) {
                                notifyEndMoveAnimation(moveEvent);
                                endMoveCallback.run();
                            }
                            
                        });
                    } else {
                        final MoveAnimationEvent moveEvent = new MoveAnimationEvent(droppedPieceMovementDuration + pieceFadeDuration, move, true);
                        notifyBeginMoveAnimation(moveEvent);
                        draggingPieceHandler.moveToTargetSquare(droppedPieceMovementDuration, new Runnable() {
                            @Override
                            public void run() {
                                if(!game.isOver()) {
                                    boardControl.fadeOutPiece(attackedBoardSquare, pieceFadeDuration, new Runnable() {
                                        @Override
                                        public void run() {
                                            notifyEndMoveAnimation(moveEvent);
                                            endMoveCallback.run();
                                        }
                                    });
                                } else {
                                    endMoveCallback.run();
                                }
                            }
                        });
                    }
                }
            }
        }
        
    }

}
