/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.model;

import java.util.ArrayList;
import java.util.Deque;
import java.util.LinkedList;
import java.util.List;

import com.haundrix.games.model.Piece.PieceColor;

/*
 * Typically classes extending from this one act as façades
 * to the business layer of the application that implements the
 * game rules.
 * 
 * Some assumptions were made about the games this class may represent:
 * - every game is played on a rectangular board by two players who alternate moves,
 * - every move is made by moving a piece from an origin square to a destination square.
 * 
 * Implementation note:
 * this class is not thread-safe, as the implementation of part of the API rely on
 * methods such as doGoTo, which temporarily change the internal state of the game.
 * Care must be taken for not to call these methods when there is an ongoing animation
 * being played.
 */
public abstract class Game implements Cloneable {
    
    private final int initialPlyCount;
    private int currentPosition = 0;
    private List<Move> moveHistory = new ArrayList<Move>();
    private Deque<Boolean> insertedNewMoveHistory = new LinkedList<Boolean>();
    
    protected GameResult result;
    
    // this field must hold all the state of the game at a given position
    protected GameState gameState;
    
    private GameInfo info;
    
    private List<IGameListener> listeners = new ArrayList<IGameListener>();
    
    private boolean isLocked = false;
    
    // TODO Clone the supplied game state in order to cut the links to the outside world.
    public Game(GameState gameState, int initialPlyCount) {
        this.gameState = gameState;
        this.initialPlyCount = initialPlyCount;
        
        info = new GameInfo();
        info.setGame(this);
    }
    
    public int getInitialPlyCount() {
        return initialPlyCount;
    }

    // The move numbering starts at 1.
    public int getInitialNextMoveNumber() {
        return (initialPlyCount / 2) + 1;
    }

    public PieceColor getInitialColorToMove() {
        return getColorToMove(initialPlyCount);
    }
    
    public PieceColor getCurrentColorToMove() {
        return getColorToMove(initialPlyCount + currentPosition);
    }
    
    public PieceColor getLastColorToMove() {
        return getColorToMove(initialPlyCount + getPlayedPlyCount());
    }
    
    public GameResult getResult() {
        return result;
    }
    
    public void setResult(GameResult result) {
        if(isOver()) {
            throw new IllegalStateException("Error! The game is already ended.");
        }
        if(result == null) {
            throw new IllegalArgumentException("Error! Result can't be null.");
        }

        this.result = result;
        notifyEnded();
    }
    
    public GameResult getResult(Move move) {
        doMakeMove(move);
        GameResult result = this.result;
        doUnmakePreviousMove();
        
        return result;
    }

    // TODO Return a clone of the game state in order to avoid it being manipulated externally.
    public GameState getState() {
        return gameState;
    }
    
    public GameInfo getInfo() {
        return info;
    }

    public void setInfo(GameInfo info) {
        this.info.setGame(null);
        
        this.info = info;
        this.info.setGame(this);
        
        notifyInfoChanged();
    }

    public void addGameListener(IGameListener gameListener) {
        synchronized(listeners) {
            listeners.add(gameListener);
        };
    }

    public void removeGameListener(IGameListener gameListener) {
        synchronized(listeners) {
            listeners.remove(gameListener);
        }
    }
    
    public boolean isLocked() {
        return isLocked;
    }
    
    public void setLocked(boolean locked) {
        isLocked = locked;
    }
    
    @Override
    public boolean equals(Object otherObject) {
        if((otherObject == null) || !(otherObject instanceof Game)) {
            return false;
        }
        
        Game otherGame = (Game) otherObject;
        
        if(!info.equals(otherGame.getInfo())) {
            return false;
        }
        
        if(initialPlyCount != otherGame.getInitialPlyCount()) {
            return false;
        }
        
        if(moveHistory.size() != otherGame.moveHistory.size()) {
            return false;
        }
        for(int i = 0; i < moveHistory.size(); i++) {
            if(!moveHistory.get(i).equals(otherGame.moveHistory.get(i))) {
                return false;
            }
        }
        
        if(!gameState.equals(otherGame.getState())) {
            return false;
        }
        
        return result == null ? otherGame.getResult() == null : result.equals(otherGame.getResult());
    }

    /*
     * The clone operation does not copy the game info or the game listeners.
     * The result object is properly handled by the default cloning mechanism,
     * since it's immutable by contract.
     */
    @Override
    public Game clone() {
        Game clone;
        try {
            clone = (Game) super.clone();
        } catch(CloneNotSupportedException e) {
            throw new RuntimeException(e);
        }
        
        clone.currentPosition = 0;
        clone.moveHistory = new ArrayList<Move>();
        clone.info = new GameInfo();
        clone.listeners = new ArrayList<IGameListener>();
        clone.isLocked = false;
        
        int currentPosition = this.currentPosition;
        doGoToBegin();
        GameState clonedGameState = gameState.clone();
        clone.gameState = clonedGameState;
        for(Move move : moveHistory) {
            clone.move(move.clone());
        }
        clone.doGoTo(currentPosition);
        doGoTo(currentPosition);
        
        return clone;
    }

    public boolean isAtBegin() {
        return currentPosition == 0;
    }

    public boolean isAtEnd() {
        return currentPosition == moveHistory.size();
    }

    public int getCurrentPosition() {
        return currentPosition;
    }
    
    public int getLastPosition() {
        return moveHistory.size();
    }

    public int getCurrentPlyCount() {
        return currentPosition + initialPlyCount;
    }
    
    public int getPlayedPlyCount() {
        return moveHistory.size();
    }
    
    // The move numbering starts at 1.
    public int getMoveNumber(int position) {
        if((position == 0) && (initialPlyCount == 0)) {
            // the move number is undefined in this situation
            String errorMessage = "Error! Undefined move number for the initial position ";
            errorMessage += "if the initial ply count is zero.";
            throw new IllegalArgumentException(errorMessage);
        }
        
        return (position + initialPlyCount - 1) / 2 + 1;
    }
    
    // The move numbering starts at 1.
    public int getNextMoveNumber() {
        return (currentPosition + initialPlyCount) / 2 + 1;
    }

    public void goToBegin() {
        if(!isLocked) {
            goTo(0);
        }
    }

    public void goToEnd() {
        if(!isLocked) {
            goTo(moveHistory.size());
        }
    }
    
    public void goToPrevious() {
        if(!isLocked) {
            goTo(currentPosition - 1);
        }
    }

    public void goToNext() {
        if(!isLocked) {
            goTo(currentPosition + 1);
        }
    }

    public void goTo(int position) {
        if(!isLocked) {
            if((position < 0) || (position > moveHistory.size())) {
                throw new IllegalArgumentException("Error! Game position out of bounds.");
            }
        
            if(position != currentPosition) {
                notifyBeforePositionChange();
                doGoTo(position);
                notifyAfterPositionChange();
            }
        }
    }

    // Variations are currently not supported.
    // Newly input moves are inserted at the end of the game.
    public void move(Move move) {
        if(isOver()) {
            throw new IllegalStateException("Error! The game is already ended.");
        }

        boolean wasAtEnd = isAtEnd();
        int previousPosition = this.currentPosition;
        
        boolean movedToEnd = wasAtEnd && !isLocked;
        if(movedToEnd) {
            notifyBeforePositionChange();
        }
        
        if(!wasAtEnd) {
            doGoToEnd();
        }
        
        checkLegal(move);
        doMakeMove(move);
        
        if(!movedToEnd) {
            doGoTo(previousPosition);
        }
        
        notifyMoved();
        if(movedToEnd) {
            notifyAfterPositionChange();
        }
        if(isOver()) {
            notifyEnded();
        }
    }

    public Move getPreviousMove() {
        return getMove(currentPosition - 1);
    }
    
    public Move getNextMove() {
        return getMove(currentPosition);
    }
    
    public Move getLastMove() {
        return getMove(getLastPosition() - 1);
    }

    public Move getMove(int moveNumber) {
        if((moveNumber < 0) || (moveNumber >= moveHistory.size())) {
            throw new IllegalArgumentException("Error! Move number out of bounds!");
        }
        
        return moveHistory.get(moveNumber);
    }
    
    public boolean isOver() {
        return result != null;
    }

    public boolean isFinished() {
        return isOver() && isAtEnd();
    }

    // TODO This allows clients to modify the game's internal state!
    //       Change it to return an UnmodifiableBoard wrapper or a clone.
    //       Investigate whether ChessGame.getGameState can also be changed
    //          to return an UnmodifiableChessGameState instead.
    //       Games' internal state should only be changed through move() or navigation APIs. 
    public RectangularBoard getBoard() {
        return gameState.getBoard();
    }

    public abstract boolean isLegal(Move move);
    
    public abstract Square[] getLegalDestinationsFrom(Square origin);
    
    public abstract Square[] getLegalOriginsTo(Square destination);
    
    public Move[] getAllLegalMoves() {
        List<Move> legalMoves = new ArrayList<Move>();
        for(Square origin : getSquaresOccupiedByPiecesOfColor(gameState.getColorToMove())) {
            for(Square destination : getLegalDestinationsFrom(origin)) {
                legalMoves.add(new Move(origin, destination));
            }
        }
        
        return legalMoves.toArray(new Move[0]);
    }
    
    public abstract Square getAttackedSquare(Move move);
    
    public boolean isCapture(Move move) {
        Square attackedSquare = getAttackedSquare(move);
        return getBoard().getPiece(attackedSquare) != null;
    }

    public boolean isCaptureMove(Move move) {
        checkLegal(move);
        return isCapture(move);
    }

    public boolean previousMoveWasCapture() {
        doGoToPrevious();
        Move previousMove = getNextMove();
        boolean previousMoveWasCapture = isCapture(previousMove);
        doGoToNext();
        
        return previousMoveWasCapture;
    }

    public Square getPreviousAttackedSquare() {
        doGoTo(currentPosition - 1);
        Move previousMove = moveHistory.get(currentPosition);
        Square previousAttackedSquare = getAttackedSquare(previousMove);
        doGoTo(currentPosition + 1);
        
        return previousAttackedSquare;
    }

    public String getLastMoveDescription(IMoveNotation notation) {
        int currentPosition = this.currentPosition;
        doGoTo(getLastPosition() - 1);
        Move lastMove = getNextMove();
        String moveDescription = notation.getMoveDescription(lastMove, this);
        doGoTo(currentPosition);
        
        return moveDescription;
    }

    public List<String> getMoveDescriptions(IMoveNotation notation) {
        int previousPosition = currentPosition;
        List<String> moveDescriptions = new ArrayList<String>();
        for(int i = 0; i < moveHistory.size(); i++) {
            doGoTo(i);
            Move move = moveHistory.get(i);
            moveDescriptions.add(notation.getMoveDescription(move, this));
        }
        doGoTo(previousPosition);
        
        return moveDescriptions;
    }
    
    protected final void doMakeMove(Move move) {
        boolean wasAtEnd = isAtEnd();
        
        // If the game is not at the end, the only acceptable 
        // move here is the next move in the move history:
        if(!wasAtEnd && (move != getNextMove())) {
            throw new IllegalArgumentException("Error! Can't make a move if it isn't a new move " +
                    "or the next move in the current position.");
        }
        
        doMove(move);
        currentPosition++;
        
        if(wasAtEnd) {
            moveHistory.add(move);
            result = evaluateGameResult();
        }
        
        insertedNewMoveHistory.addFirst(wasAtEnd);
    }
    
    protected final void doUnmakePreviousMove() {
        doUnmove(getPreviousMove());
        currentPosition--;
        
        if(insertedNewMoveHistory.removeFirst()) {
            moveHistory.remove(moveHistory.size() - 1);
            result = null;
        }
    }

    protected abstract void doMove(Move move);
    
    protected abstract void doUnmove(Move move);

    protected abstract GameResult evaluateGameResult();

    protected void doGoToBegin() {
        doGoTo(0);
    }

    protected void doGoToEnd() {
        doGoTo(moveHistory.size());
    }
    
    protected void doGoToPrevious() {
        doGoTo(currentPosition - 1);
    }

    protected void doGoToNext() {
        doGoTo(currentPosition + 1);
    }

    protected void doGoTo(int position) {
        if(position < currentPosition) {
            while(currentPosition > position) {
                doUnmove(moveHistory.get(currentPosition - 1));
                currentPosition--;
            }
        } else {
            while(currentPosition < position) {
                doMove(moveHistory.get(currentPosition));
                currentPosition++;
            }
        }
    }

    protected void checkLegal(Move move) {
        if(!isLegal(move)) {
            throw new IllegalMoveException();
        }
    }

    protected Square[] getSquaresOccupiedByPiecesOfColor(PieceColor pieceColor) {
        List<Square> squares = new ArrayList<Square>();
        for(Square square : getBoard()) {
            Piece piece = getBoard().getPiece(square);
            if((piece != null) && piece.getColor().equals(pieceColor)) {
                squares.add(square);
            }
        }
        
        return squares.toArray(new Square[0]);
    }

    protected boolean existLegalMove() {
        for(Square square : getSquaresOccupiedByPiecesOfColor(gameState.getColorToMove())) {
            if(getLegalDestinationsFrom(square).length > 0) {
                return true;
            }
        }
        
        return false;
    }

    void notifyInfoChanged() {
        synchronized(listeners) {
            for(IGameListener listener : listeners) {
                listener.infoChanged(info);
            }
        }
    }

    private PieceColor getColorToMove(int plyCount) {
        return plyCount % 2 == 0 ? PieceColor.WHITE : PieceColor.BLACK;
    }

    private void notifyBeforePositionChange() {
        synchronized(listeners) {
            for(IGameListener listener : listeners) {
                listener.beforePositionChange(this);
            }
        }
    }
    
    private void notifyAfterPositionChange() {
        synchronized(listeners) {
            for(IGameListener listener : listeners) {
                listener.afterPositionChange(this);
            }
        }
    }

    private void notifyMoved() {
        synchronized(listeners) {
            for(IGameListener listener : listeners) {
                listener.moved(this);
            }
        }
    }

    private void notifyEnded() {
        synchronized(listeners) {
            for(IGameListener listener : listeners) {
                listener.ended(result);
            }
        }
    }

}
