/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.model;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.haundrix.games.model.Piece.PieceColor;

public class RectangularBoard implements Iterable<Square> {
    
    private SquareColor upperLeftSquareColor = SquareColor.LIGHT;
    
    private Piece[][] pieces;
    
    public RectangularBoard(int rowCount, int columnCount) {
        pieces = new Piece[rowCount][columnCount];
    }

    public RectangularBoard(Piece[][] pieces) {
        int columnCount = pieces[0].length;
        for(int row = 1; row < pieces.length; row++) {
            if(pieces[row].length != columnCount) {
                throw new IllegalArgumentException("The array of pieces must be rectangular.");
            }
        }
        
        this.pieces = pieces;
    }

    public int getRowCount() {
        return pieces.length;
    }

    public int getColumnCount() {
        return pieces[0].length;
    }
    
    public SquareColor getUpperLeftSquareColor() {
        return upperLeftSquareColor;
    }

    public void setUpperLeftSquareColor(SquareColor upperLeftSquareColor) {
        this.upperLeftSquareColor = upperLeftSquareColor;
    }

    public Piece getPiece(Square square) {
        return getPiece(square.getRow(), square.getColumn());
    }

    public Piece getPiece(int row, int column) {
        return pieces[row][column];
    }

    public void setPiece(Piece piece, Square square) {
        setPiece(piece, square.getRow(), square.getColumn());
    }

    public void setPiece(Piece piece, int row, int column) {
        pieces[row][column] = piece;
    }
    
    public Piece[] getPieces() {
        List<Piece> pieces = new ArrayList<Piece>();
        
        for(Square square : this) {
            Piece piece = getPiece(square);
            if(piece != null) {
                pieces.add(piece);
            }
        }
        
        return pieces.toArray(new Piece[0]);
    }
    
    public SquareColor getSquareColor(Square square) {
        return getSquareColor(square.getRow(), square.getColumn());
    }
    
    public SquareColor getSquareColor(int row, int column) {
        return (row + column) % 2 == 0 ? upperLeftSquareColor : upperLeftSquareColor.getOppositeColor();
    }

    @Override
    public boolean equals(Object object) {
        if((object == null) || !(object instanceof RectangularBoard)) {
            return false;
        }
        
        RectangularBoard otherBoard = (RectangularBoard) object;
        if((otherBoard.getRowCount() != getRowCount()) || (otherBoard.getColumnCount() != getColumnCount())) {
            return false;
        }
        
        for(int row = 0; row < getRowCount(); row++) {
            for(int column = 0; column < getColumnCount(); column++) {
                Piece piece = pieces[row][column];
                if(piece == null) {
                    if(otherBoard.getPiece(row, column) != null) {
                        return false;
                    }
                } else {
                    if(!piece.equals(otherBoard.getPiece(row, column))) {
                        return false;
                    }
                }
            }
        }
        
        return true;
    }

    @Override
    public Iterator<Square> iterator() {
        return new Iterator<Square>() {
            
            private int row = 0;
            private int column = 0;

            @Override
            public boolean hasNext() {
                return (row < getRowCount()) && (column < getColumnCount());
            }

            @Override
            public Square next() {
                Square square = new Square(row, column);

                column++;
                if(column >= getColumnCount()) {
                    column = 0;
                    row++;
                }
                
                return square;
            }

            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
            
        };
    }
    
    // Looks for exactly this piece instance on the board.
    public Square findPiece(Piece piece) {
        for(Square square : this) {
            if(getPiece(square) == piece) {
                return square;
            }
        }
        
        return null;
    }

    // Returns the location on board of the piece of the specified class and color.
    // If there is more than one such piece, the location of the first piece found
    // is returned.
    public Square findPiece(Class<? extends Piece> pieceClass, PieceColor pieceColor) {
        for(Square square : this) {
            Piece piece = getPiece(square);
            if((piece != null) && (piece.getClass() == pieceClass) && (piece.getColor().equals(pieceColor))) {
                return square;
            }
        }
        
        return null;
    }

    public Square[] findPieces(Class<? extends Piece> pieceClass, PieceColor pieceColor) {
        List<Square> squares = new ArrayList<Square>();
        
        for(Square square : this) {
            Piece piece = getPiece(square);
            if((piece != null) && (piece.getClass() == pieceClass) && (piece.getColor().equals(pieceColor))) {
                squares.add(square);
            }
        }
        
        return squares.toArray(new Square[0]);
    }

    public Square getSquareWithRow(int row, Square[] squares) {
        checkRowBounds(row);
        checkBounds(squares);
        
        for(Square square : squares) {
            if(square.getRow() == row) {
                return square;
            }
        }
        
        return null;
    }

    public Square getSquareWithColumn(int column, Square[] squares) {
        checkColumnBounds(column);
        checkBounds(squares);
        
        for(Square square : squares) {
            if(square.getColumn() == column) {
                return square;
            }
        }
        
        return null;
    }

    public boolean isNeighbouringSquare(Square origin, Square destination) {
        checkBounds(origin, destination);
        
        if(origin.equals(destination)) {
            return false;
        }
        
        int absRowDifference = Math.abs(origin.getRow() - destination.getRow());
        int absColumnDifference = Math.abs(origin.getColumn() - destination.getColumn());
        return (absRowDifference <= 1) && (absColumnDifference <= 1);
    }
    
    public boolean isColumn(Square origin, Square destination) {
        checkBounds(origin, destination);
        
        if(origin.equals(destination)) {
            return false;
        }

        return origin.getColumn() == destination.getColumn();
    }
    
    public boolean isFreeColumn(Square origin, Square destination) {
        checkBounds(origin, destination);
        
        if(!isColumn(origin, destination)) {
            return false;
        }
        
        int startRow = origin.getRow();
        int endRow = destination.getRow();
        int column = origin.getColumn();
        if(startRow < endRow) {
            for(int row = startRow + 1; row < endRow; row++) {
                if(pieces[row][column] != null) {
                    return false;
                }
            }
        } else {
          for(int row = startRow - 1; row > endRow; row--) {
                if(pieces[row][column] != null) {
                    return false;
                }
            }
        }
        
        return true;
    }
    
    public boolean isRank(Square origin, Square destination) {
        checkBounds(origin, destination);
        
        if(origin.equals(destination)) {
            return false;
        }
        
        return origin.getRow() == destination.getRow();
    }
    
    public boolean isFreeRank(Square origin, Square destination) {
        checkBounds(origin, destination);
        
        if(!isRank(origin, destination)) {
            return false;
        }
        
        int row = origin.getRow();
        int startColumn = origin.getColumn();
        int endColumn = destination.getColumn();
        if(startColumn < endColumn) {
            for(int column = startColumn + 1; column < endColumn; column++) {
                if(pieces[row][column] != null) {
                    return false;
                }
            }
        } else {
            for(int column = startColumn - 1; column > endColumn; column--) {
                if(pieces[row][column] != null) {
                    return false;
                }
            }
        }
        
        return true;
    }
    
    public boolean isDiagonal(Square origin, Square destination) {
        checkBounds(origin, destination);
        
        if(origin.equals(destination)) {
            return false;
        }
        
        int absRowDifference = Math.abs(origin.getRow() - destination.getRow());
        int absColumnDifference = Math.abs(origin.getColumn() - destination.getColumn());
        return absRowDifference == absColumnDifference;
    }

    public boolean isFreeDiagonal(Square origin, Square destination) {
        checkBounds(origin, destination);
        
        if(!isDiagonal(origin, destination)) {
            return false;
        }
        
        int startRow = origin.getRow();
        int endRow = destination.getRow();
        int startColumn = origin.getColumn();
        int endColumn = destination.getColumn();
        int rowIncrement = startRow < endRow ? 1 : -1;
        int columnIncrement = startColumn < endColumn ? 1 : -1;
        int nIntermediateSquares = Math.abs(startRow - endRow) - 1;
        int row = startRow + rowIncrement;
        int column = startColumn + columnIncrement;
        for(int i = 0; i < nIntermediateSquares; i++) {
            if(pieces[row][column] != null) {
                return false;
            }
            row += rowIncrement;
            column += columnIncrement;
        }
        
        return true;
    }

    public boolean isLShapedPath(Square origin, Square destination) {
        checkBounds(origin, destination);
        
        int absRowDifference = Math.abs(origin.getRow() - destination.getRow());
        int absColumnDifference = Math.abs(origin.getColumn() - destination.getColumn());
        boolean condition1 = (absRowDifference == 1) && (absColumnDifference == 2);
        boolean condition2 = (absRowDifference == 2) && (absColumnDifference == 1);
        return condition1 || condition2;
    }
    
    public boolean hasSameRow(Square[] squares) {
        checkBounds(squares);
        
        for(int i = 0; i < squares.length - 1; i++) {
            Square square = squares[i];
            for(int j = i + 1; j < squares.length; j++) {
                if(squares[j].getRow() == square.getRow()) {
                    return true;
                }
            }
        }
        
        return false;
    }

    public boolean hasSameColumn(Square[] squares) {
        checkBounds(squares);
        
        for(int i = 0; i < squares.length - 1; i++) {
            Square square = squares[i];
            for(int j = i + 1; j < squares.length; j++) {
                if(squares[j].getColumn() == square.getColumn()) {
                    return true;
                }
            }
        }
        
        return false;
    }

    public List<Square> getNeighbouringSquares(Square origin) {
        List<Square> squares = new ArrayList<Square>();
        
        int startRow = Math.max(origin.getRow() - 1, 0);
        int endRow = Math.min(origin.getRow() + 1, getRowCount() - 1);
        int startColumn = Math.max(origin.getColumn() - 1, 0);
        int endColumn = Math.min(origin.getColumn() + 1, getColumnCount() - 1);
        for(int row = startRow; row <= endRow; row++) {
            for(int column = startColumn; column <= endColumn; column++) {
                Square candidateSquare = new Square(row, column);
                if(!candidateSquare.equals(origin)) {
                    squares.add(candidateSquare);
                }
            }
        }
        
        return squares;
    }

    public List<Square> getFreeColumnSquaresFrom(Square origin) {
        checkBounds(origin);
        
        List<Square> squares = new ArrayList<Square>();
        squares.addAll(getFreeSquaresFrom(origin, -1, 0));
        squares.addAll(getFreeSquaresFrom(origin, 1, 0));
        
        return squares;
    }

    public List<Square> getFreeRankSquaresFrom(Square origin) {
        checkBounds(origin);
        
        List<Square> squares = new ArrayList<Square>();
        squares.addAll(getFreeSquaresFrom(origin, 0, -1));
        squares.addAll(getFreeSquaresFrom(origin, 0, 1));
        
        return squares;
    }

    public List<Square> getFreeDiagonalSquaresFrom(Square origin) {
        checkBounds(origin);
        
        List<Square> squares = new ArrayList<Square>();
        squares.addAll(getFreeSquaresFrom(origin, -1, -1));
        squares.addAll(getFreeSquaresFrom(origin, -1, 1));
        squares.addAll(getFreeSquaresFrom(origin, 1, -1));
        squares.addAll(getFreeSquaresFrom(origin, 1, 1));
        
        return squares;
    }

    public List<Square> getLShapedPathSquaresFrom(Square origin) {
        List<Square> squares = new ArrayList<Square>();
        addRelativeSquareToList(origin, -2, -1, squares);
        addRelativeSquareToList(origin, -2, 1, squares);
        addRelativeSquareToList(origin, 2, -1, squares);
        addRelativeSquareToList(origin, 2, 1, squares);
        addRelativeSquareToList(origin, -1, -2, squares);
        addRelativeSquareToList(origin, -1, 2, squares);
        addRelativeSquareToList(origin, 1, -2, squares);
        addRelativeSquareToList(origin, 1, 2, squares);
        
        return squares;
    }

    public void checkBounds(Square square1, Square square2) {
        checkBounds(square1);
        checkBounds(square2);
    }
    
    public void checkBounds(Square square) {
        if(square == null) {
            throw new IllegalArgumentException("Square cannot be null.");
        }
        
        checkBounds(square.getRow(), square.getColumn());
    }
    
    public void checkBounds(int row, int column) {
        checkRowBounds(row);
        checkColumnBounds(column);
    }

    private void checkRowBounds(int row) {
        if((row < 0) || (row >= getRowCount())) {
            throw new IllegalArgumentException("Square row out of bounds.");
        }
    }
    
    private void checkColumnBounds(int column) {
        if((column < 0) || (column >= getColumnCount())) {
            throw new IllegalArgumentException("Square column out of bounds.");
        }
    }

    private void checkBounds(Square[] squares) {
        for(Square square : squares) {
            checkBounds(square);
        }
    }

    private List<Square> getFreeSquaresFrom(Square origin, int rowIncrement, int columnIncrement) {
        List<Square> squares = new ArrayList<Square>();
        
        int row = origin.getRow() + rowIncrement;
        int column = origin.getColumn() + columnIncrement;
        while((row >= 0) && (row < getRowCount()) && (column >= 0) && (column < getColumnCount())) {
            squares.add(new Square(row, column));
            if(pieces[row][column] != null) {
                break;
            }
            row += rowIncrement;
            column += columnIncrement;
        }
        
        return squares;
    }

    private void addRelativeSquareToList(Square origin, int rowOffset, int columnOffset,
            List<Square> squares) {
        int row = origin.getRow() + rowOffset;
        if((row < 0) || (row >= getRowCount())) {
            return;
        }
        
        int column = origin.getColumn() + columnOffset;
        if((column < 0) || (column >= getColumnCount())) {
            return;
        }
        
        squares.add(new Square(row, column));
    }

}
