/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.model.playing;

import com.haundrix.games.model.Piece.PieceColor;

public abstract class AbstractClock implements IClock {
    
    private PieceColor colorToMove = PieceColor.WHITE;
    
    private ClockTime clockTime;
    
    private boolean isRunning = false;
    
    private long lastPressTime;

    public AbstractClock(long whiteTime, long blackTime) {
        clockTime = new ClockTime(whiteTime, blackTime);
    }
    
    @Override
    public synchronized PieceColor getColorToMove() {
        return colorToMove;
    }

    @Override
    public synchronized ClockTime getClockTime() {
        if(isRunning) {
            return getUpdatedTime(getPassedTime(), false);
        } else {
            return clockTime;
        }
    }

    @Override
    public synchronized long getCurrentPlayerTime() {
        return getPlayerTime(colorToMove);
    }

    @Override
    public synchronized long getOtherPlayerTime() {
        return getPlayerTime(colorToMove.getOppositeColor());
    }

    @Override
    public synchronized boolean isFlagFallen() {
        ClockTime updatedTime = getClockTime();
        
        return (updatedTime.getWhiteTime() <= 0L) || (updatedTime.getBlackTime() <= 0L);
    }

    @Override
    public long getMinimumTimeTill(long playerTime) {
        ClockTime updatedTime = getClockTime();
        
        return Math.min(updatedTime.getWhiteTime() - playerTime, updatedTime.getBlackTime() - playerTime);
    }

    @Override
    public synchronized long getMinimumTimeTillFlagFalls() {
        return getMinimumTimeTill(0L);
    }

    @Override
    public synchronized boolean isRunning() {
        return isRunning;
    }

    @Override
    public synchronized void start() {
        if(isRunning) {
            throw new IllegalStateException("Error! The clock has already been started.");
        }
        
        lastPressTime = System.currentTimeMillis();
        isRunning = true;
    }

    @Override
    public synchronized void stop() {
        if(!isRunning) {
            throw new IllegalStateException("Error! The clock hasn't been started yet or has already been stopped.");
        }
        
        clockTime = getUpdatedTime(getPassedTime(), false);
        
        isRunning = false;
    }

    @Override
    public synchronized void press() {
        if(!isRunning) {
            throw new IllegalStateException("Error! The clock must be running in order to be pressed.");
        }
        
        long passedTime = getPassedTime();
        lastPressTime = System.currentTimeMillis();
        
        clockTime = getUpdatedTime(passedTime, true);
        
        colorToMove = colorToMove.getOppositeColor();
    }

    /* The default behavior is to deduct the passed time from
     * the time of the current side to move.
     * The contract of this method requires that this behavior is unchanged,
     * unless the value of the isClockJustPressed parameter is true,
     * when subclasses will typically override the default behavior.
     */
    protected ClockTime getUpdatedTime(long passedTime, boolean isClockJustPressed) {
        long whiteTime = clockTime.getWhiteTime();
        long blackTime = clockTime.getBlackTime();
        if(colorToMove == PieceColor.WHITE) {
            whiteTime -= passedTime;
        } else {
            blackTime -= passedTime;
        }
        
        return new ClockTime(whiteTime, blackTime);
    }
    
    private long getPassedTime() {
        return System.currentTimeMillis() - lastPressTime;
    }

    private long getPlayerTime(PieceColor playerColor) {
        return playerColor == PieceColor.WHITE ? getClockTime().getWhiteTime() : getClockTime().getBlackTime();
    }

}
