/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.model.playing;

import com.haundrix.games.model.Piece.PieceColor;

/*
 * Clock that specifies a fixed amount of time for the entire game
 * plus an additional time increment per move (a.k.a. Fischer clock).
 * 
 * Fixed time is given in minutes.
 * Increment is given in seconds.
 */
public class FischerClock extends AbstractClock {

    private final int fixedTime;
    private final int increment;
    
    public FischerClock(int fixedTime, int increment) {
        super(fixedTime * 60L * 1000L, fixedTime * 60L * 1000L);
        
        if(fixedTime <= 0) {
            throw new IllegalArgumentException("Error! Fixed time must be > 0.");
        }
        if(increment < 0) {
            throw new IllegalArgumentException("Error! Increment must be >= 0.");
        }
        
        this.fixedTime = fixedTime;
        this.increment = increment;
    }

    public int getFixedTime() {
        return fixedTime;
    }

    public int getIncrement() {
        return increment;
    }

    @Override
    public long getMaxTimeScale() {
        return getFixedTime() * 60L * 1000L;
    }

    @Override
    protected ClockTime getUpdatedTime(long passedTime, boolean isClockJustPressed) {
        ClockTime updatedTime = super.getUpdatedTime(passedTime, isClockJustPressed);
        
        if(isClockJustPressed) {
            long whiteTime = updatedTime.getWhiteTime();
            long blackTime = updatedTime.getBlackTime();
            if(getColorToMove() == PieceColor.WHITE) {
                whiteTime += increment * 1000L;
            } else {
                blackTime += increment * 1000L;
            }
            
            updatedTime = new ClockTime(whiteTime, blackTime);
        }
        
        return updatedTime;
    }

}
