/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.model.playing;

import com.haundrix.games.model.Piece.PieceColor;

/*
 * Clock that specifies a fixed number of moves in a fixed amount of time.
 * Total time is given in minutes.
 */
public class MovesPerTimeClock extends AbstractClock {

    private final int movesCount;
    private final int totalTime;
    
    private int whiteMovesCount = 0;
    private int blackMovesCount = 0;
    
    public MovesPerTimeClock(int movesCount, int totalTime) {
        super(totalTime * 60L * 1000L, totalTime * 60L * 1000L);
        
        if(movesCount <= 0) {
            throw new IllegalArgumentException("Error! Moves count must be > 0.");
        }
        if(totalTime <= 0) {
            throw new IllegalArgumentException("Error! Total time must be > 0.");
        }
        
        this.movesCount = movesCount;
        this.totalTime = totalTime;
    }

    public int getMovesCount() {
        return movesCount;
    }

    public int getTotalTime() {
        return totalTime;
    }

    @Override
    public long getMaxTimeScale() {
        return getTotalTime() * 60L * 1000L;
    }

    @Override
    protected ClockTime getUpdatedTime(long passedTime, boolean isClockJustPressed) {
        ClockTime updatedTime = super.getUpdatedTime(passedTime, isClockJustPressed);
        
        if(isClockJustPressed) {
            long whiteTime = updatedTime.getWhiteTime();
            long blackTime = updatedTime.getBlackTime();
            if(getColorToMove() == PieceColor.WHITE) {
                if(++whiteMovesCount % movesCount == 0) {
                    whiteTime += totalTime * 60L * 1000L;
                }
            } else {
                if(++blackMovesCount % movesCount == 0) {
                    blackTime += totalTime * 60L * 1000L;
                }
            }
            
            updatedTime = new ClockTime(whiteTime, blackTime);
        }
        
        return updatedTime;
    }

}
