/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp;

import java.io.File;
import java.io.IOException;
import java.lang.Thread.UncaughtExceptionHandler;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

import org.apache.commons.io.FileUtils;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.preferences.InstanceScope;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.swt.graphics.Point;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.jdesktop.animation.timing.interpolation.Evaluator;
import org.osgi.framework.BundleContext;
import org.osgi.service.prefs.BackingStoreException;

import com.haundrix.animation.SwtPointEvaluator;
import com.haundrix.games.chess.controller.IChessSoundTheme;
import com.haundrix.utils.IExceptionHandler;
import com.haundrix.utils.SoundPlayer;

/**
 * The activator class controls the plugin life cycle.
 */
public class HaundrixPlugin extends AbstractUIPlugin {

    public static final String ID = "com.haundrix.games";

    private static HaundrixPlugin plugin;
    
    private IExceptionHandler defaultExceptionHandler;

    private SoundPlayer soundPlayer;
    private IChessSoundTheme chessSoundTheme;
    
    @Override
    public void start(BundleContext context) throws Exception {
        super.start(context);
        
        HaundrixPlugin.plugin = this;
        
        Thread.setDefaultUncaughtExceptionHandler(new UncaughtExceptionHandler() {
            @Override
            public void uncaughtException(Thread thread, Throwable exception) {
                HaundrixLog.logError("Exception in thread \"" + thread.getName() + "\".", exception);
            }
        });
        
        Evaluator.register(Point.class, SwtPointEvaluator.class);
        
        // clears the plugin state area
        File stateDir = getStateLocation().toFile();
        FileUtils.cleanDirectory(stateDir);
        
        defaultExceptionHandler = new IExceptionHandler() {
            @Override
            public void handle(Exception exception) {
                synchronized(this) {
                    HaundrixLog.logError("Exception in thread \"" + Thread.currentThread().getName() + "\".", exception);
                }
            }
        };
        soundPlayer = new SoundPlayer(defaultExceptionHandler);
        chessSoundTheme = new PluginChessSoundTheme();
    }

    @Override
    public void stop(BundleContext context) throws Exception {
        chessSoundTheme = null;
        soundPlayer.dispose();
        soundPlayer = null;
        defaultExceptionHandler = null;
        
        Evaluator.deregister(Point.class);
        
        Thread.setDefaultUncaughtExceptionHandler(null);
        
        HaundrixPlugin.plugin = null;
        
        super.stop(context);
    }

    public static HaundrixPlugin getInstance() {
        return HaundrixPlugin.plugin;
    }
    
    public static ImageDescriptor getImageDescriptor(String path) {
        return imageDescriptorFromPlugin(ID, path);
    }
    
    public File getFile(String fileName) throws URISyntaxException, IOException {
        URL url = FileLocator.find(getBundle(), new Path(fileName), null);
        URL fileURL = FileLocator.toFileURL(url);
        URI uri = new URI(fileURL.getProtocol(), fileURL.getFile(), null);
        return new File(uri);
    }
    
    public IExceptionHandler getDefaultExceptionHandler() {
        return defaultExceptionHandler;
    }

    public SoundPlayer getSoundPlayer() {
        return soundPlayer;
    }
    
    public IChessSoundTheme getChessSoundTheme() {
        return chessSoundTheme;
    }
    
    public void savePreferences() {
        try {
            new InstanceScope().getNode(ID).flush();
        } catch(BackingStoreException e) {
            HaundrixLog.logError("Error storing preference of Haundrix plugin.", e);
        }
    }

}
