/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;

import javax.sound.sampled.UnsupportedAudioFileException;

import com.haundrix.games.chess.controller.IChessSoundTheme;
import com.haundrix.utils.SoundFile;
import com.haundrix.utils.SoundUtils;

class PluginChessSoundTheme implements IChessSoundTheme {

    private static final String SOUND_DIR = "rsc/sounds/";
    
    private static final String BLACK_MOVE = SOUND_DIR + "blackMove.ogg";
    private static final String CAPTURE = SOUND_DIR + "capture.ogg";
    private static final String CHECK = SOUND_DIR + "check.ogg";
    private static final String DRAW = SOUND_DIR + "draw.ogg";
    private static final String ILLEGAL_MOVE_LONG = SOUND_DIR + "illegalMoveLong.ogg";
    private static final String ILLEGAL_MOVE_SHORT = SOUND_DIR + "illegalMoveShort.ogg";
    private static final String KNIGHT_ATTACK = SOUND_DIR + "knightAttack.ogg";
    private static final String KNIGHT_CAPTURE = SOUND_DIR + "knightCapture.ogg";
    private static final String NEW_GAME = SOUND_DIR + "newGame.ogg";
    private static final String PROMOTION = SOUND_DIR + "promotion.ogg";
    private static final String TIMEOUT = SOUND_DIR + "timeout.ogg";
    private static final String TIME_TROUBLE = SOUND_DIR + "timeTrouble.ogg";
    private static final String TRANSITION = SOUND_DIR + "transition.ogg";
    private static final String WHITE_MOVE = SOUND_DIR + "whiteMove.ogg";
    private static final String WIN = SOUND_DIR + "win.ogg";
    
    private static final String SOUNDS_SCRATCH_AREA = "sounds";

    private static final int ILLEGAL_MOVE_DURATION_THRESHOLD = 700;

    private SoundFile blackMoveSound;
    private SoundFile captureSound;
    private SoundFile checkSound;
    private SoundFile drawSound;
    private SoundFile illegalMoveLongSound;
    private SoundFile illegalMoveShortSound;
    private SoundFile knightAttackSound;
    private SoundFile knightCaptureSound;
    private SoundFile newGameSound;
    private SoundFile promotionSound;
    private SoundFile timeoutSound;
    private SoundFile timeTroubleSound;
    private SoundFile transitionSound;
    private SoundFile whiteMoveSound;
    private SoundFile winSound;

    private File soundsScratchDir;

    public PluginChessSoundTheme() {
        File pluginStateArea = HaundrixPlugin.getInstance().getStateLocation().toFile();
        soundsScratchDir = new File(pluginStateArea, SOUNDS_SCRATCH_AREA);
        if(!soundsScratchDir.exists()) {
            soundsScratchDir.mkdir();
        }
        
        try {
            blackMoveSound = getSoundFile(BLACK_MOVE);
            captureSound = getSoundFile(CAPTURE, 0L, 379L);
            checkSound = getSoundFile(CHECK);
            drawSound = getSoundFile(DRAW);
            illegalMoveLongSound = getSoundFile(ILLEGAL_MOVE_LONG);
            illegalMoveShortSound = getSoundFile(ILLEGAL_MOVE_SHORT);
            knightAttackSound = getSoundFile(KNIGHT_ATTACK);
            knightCaptureSound = getSoundFile(KNIGHT_CAPTURE);
            newGameSound = getSoundFile(NEW_GAME);
            promotionSound = getSoundFile(PROMOTION);
            timeoutSound = getSoundFile(TIMEOUT);
            timeTroubleSound = getSoundFile(TIME_TROUBLE);
            transitionSound = getSoundFile(TRANSITION);
            whiteMoveSound = getSoundFile(WHITE_MOVE, 50L, 430L);
            winSound = getSoundFile(WIN);
        } catch(Exception e) {
            HaundrixLog.logError("Error loading chess sound theme.", e);
        }
    }
    
    @Override
    public SoundFile getBlackMoveSound() {
        return blackMoveSound;
    }

    @Override
    public SoundFile getCaptureSound() {
        return captureSound;
    }

    @Override
    public SoundFile getCheckSound() {
        return checkSound;
    }

    @Override
    public SoundFile getDrawSound() {
        return drawSound;
    }

    @Override
    public SoundFile getIllegalMoveLongSound() {
        return illegalMoveLongSound;
    }

    @Override
    public SoundFile getIllegalMoveShortSound() {
        return illegalMoveShortSound;
    }

    @Override
    public SoundFile getKnightAttackSound() {
        return knightAttackSound;
    }

    @Override
    public SoundFile getKnightCaptureSound() {
        return knightCaptureSound;
    }

    @Override
    public SoundFile getNewGameSound() {
        return newGameSound;
    }

    @Override
    public SoundFile getPromotionSound() {
        return promotionSound;
    }

    @Override
    public SoundFile getTimeoutSound() {
        return timeoutSound;
    }

    @Override
    public SoundFile getTimeTroubleSound() {
        return timeTroubleSound;
    }

    @Override
    public SoundFile getTransitionSound() {
        return transitionSound;
    }

    @Override
    public SoundFile getWhiteMoveSound() {
        return whiteMoveSound;
    }

    @Override
    public SoundFile getWinSound() {
        return winSound;
    }

    @Override
    public int getShortIllegalMoveDurationThreshold() {
        return ILLEGAL_MOVE_DURATION_THRESHOLD;
    }
    
    private SoundFile getSoundFile(String fileName) throws URISyntaxException,
            IOException, UnsupportedAudioFileException {
        return getSoundFile(fileName, 0L, 0L);
    }

    private SoundFile getSoundFile(String fileName, long startSilence, long endSilence) throws URISyntaxException,
            IOException, UnsupportedAudioFileException {
        File file = HaundrixPlugin.getInstance().getFile(fileName);
        long duration = SoundUtils.getSoundDuration(file, soundsScratchDir);
        return new SoundFile(file, duration, startSilence, endSilence);
    }

}
