/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.editors;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.commands.Command;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.action.IStatusLineManager;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.commands.ICommandService;
import org.eclipse.ui.handlers.RegistryToggleState;
import org.eclipse.ui.part.EditorPart;
import org.eclipse.ui.services.IEvaluationService;

import com.haundrix.games.chess.controller.ChessGameController;
import com.haundrix.games.chess.controller.ChessGameSoundController;
import com.haundrix.games.chess.model.ChessGame;
import com.haundrix.games.chess.model.ChessGameResult;
import com.haundrix.games.chess.model.ChessGameResultReason;
import com.haundrix.games.chess.pgn.PgnExporter;
import com.haundrix.games.chess.view.ChessPieceType;
import com.haundrix.games.controller.AnimationEvent;
import com.haundrix.games.controller.GameControllerAdapter;
import com.haundrix.games.controller.GameNavigationKeyListener;
import com.haundrix.games.controller.HighlightScheme;
import com.haundrix.games.model.Game;
import com.haundrix.games.model.GameAdapter;
import com.haundrix.games.model.GameInfo;
import com.haundrix.games.model.GameOutcome;
import com.haundrix.games.model.GameResult;
import com.haundrix.games.rcp.HaundrixCommands;
import com.haundrix.games.rcp.HaundrixLog;
import com.haundrix.games.rcp.HaundrixPlugin;
import com.haundrix.games.rcp.HaundrixSharedImages;
import com.haundrix.games.rcp.managers.ColorSchemeManager;
import com.haundrix.games.rcp.managers.HighlightSchemeManager;
import com.haundrix.games.rcp.managers.PieceSetManager;
import com.haundrix.games.rcp.managers.PluginPieceSet;
import com.haundrix.games.rcp.preferences.HaundrixPreferences;
import com.haundrix.games.view.ColorScheme;
import com.haundrix.games.view.RectangularBoardControl;
import com.haundrix.rcp.FileEditorInput;
import com.haundrix.rcp.InvalidEditorInput;
import com.haundrix.rcp.RcpUtils;

/*
 * Notes for clients sub-classing this class:
 * 1. The game must be created and the isDirty flag initialized in init(IEditorSite, IEditorInput).
 * 2. The controls are created in createPartControl(Composite). Subclasses may add additional behavior
 *    after calling super.createPartControl(Composite), such as configuring the game controller.
 */
public abstract class ChessGameEditor extends EditorPart {
    
    protected ChessGame game;
    protected ChessGameController gameController;
    protected ChessGameSoundController soundController;
    
    protected boolean isDirty;

    private IPropertyChangeListener preferencesListener = new InternalPreferencesListener();
    
    private boolean activated = false;

    private Image enabledOwl;
    private Image disabledOwl;
    private boolean isCustomMascot;
    private Image enabledMascot;
    private Image disabledMascot;
    private String statusMessage = null;
    
    private List<IFlipListener> flipListeners = new ArrayList<IFlipListener>();

    @Override
    public void createPartControl(final Composite parent) {
        if(getEditorInput() instanceof InvalidEditorInput) {
            RowLayout rowLayout = new RowLayout();
            rowLayout.type = SWT.HORIZONTAL;
            rowLayout.center = true;
            parent.setLayout(rowLayout);
            
            Image errorImage = getSite().getShell().getDisplay().getSystemImage(SWT.ICON_ERROR);
            Label imageLabel = new Label(parent, SWT.HORIZONTAL);
            imageLabel.setImage(errorImage);
            
            String errorMessage = ((InvalidEditorInput) getEditorInput()).getErrorMessage();
            Label errorMessageLabel = new Label(parent, SWT.HORIZONTAL);
            errorMessageLabel.setText(errorMessage);
        } else {
            parent.setLayout(new FillLayout());
            
            game.addGameListener(new InternalGameListener());
            
            PluginPieceSet pieceSet = getPieceSet();
            
            gameController = new ChessGameController(game, parent, pieceSet);
            gameController.addGameControllerListener(new InternalGameControllerListener());

            HaundrixPlugin haundrixPlugin = HaundrixPlugin.getInstance();
            soundController = new ChessGameSoundController(gameController, haundrixPlugin.getChessSoundTheme(),
                    haundrixPlugin.getSoundPlayer());

            getBoardControl().addKeyListener(new GameNavigationKeyListener(game));
            
            enabledOwl = HaundrixSharedImages.OWL_ENABLED.createImage();
            disabledOwl = HaundrixSharedImages.OWL_DISABLED.createImage();
            isCustomMascot = false;
            setCustomMascot(pieceSet);
            updateColors();
            
            updateGameStatusMessage();
            updateExpressions();

            getPreferenceStore().addPropertyChangeListener(preferencesListener);
            
            RcpUtils.createContextMenu(this, getBoardControl());
            
            activated = true;
        }
    }
    
    @Override
    public void setFocus() {
        if(!(getEditorInput() instanceof InvalidEditorInput)) {
            RectangularBoardControl<ChessPieceType> boardControl = getBoardControl();
            boardControl.setFocus();
            
            ICommandService commandService = (ICommandService) getSite().getService(ICommandService.class); 
            Command flipCommand = commandService.getCommand(HaundrixCommands.FLIP_BOARD.getId());
            flipCommand.getState(RegistryToggleState.STATE_ID).setValue(boardControl.isFlipped());
        }
        updateStatusArea();
    }

    @Override
    public void dispose() {
        if(activated) {
            if(isCustomMascot) {
                enabledMascot.dispose();
                disabledMascot.dispose();
            }
            HaundrixSharedImages.OWL_ENABLED.destroyImage();
            HaundrixSharedImages.OWL_DISABLED.destroyImage();
            getPreferenceStore().removePropertyChangeListener(preferencesListener);
        }
        super.dispose();
    }

    @Override
    public boolean isDirty() {
        return isDirty;
    }

    @Override
    public boolean isSaveAsAllowed() {
        return !(getEditorInput() instanceof InvalidEditorInput);
    }

    @Override
    public void doSave(IProgressMonitor monitor) {
        if(!(getEditorInput() instanceof FileEditorInput)) {
            doSaveAs();
        } else {
            File file = ((FileEditorInput) getEditorInput()).getFile();
            if(file.canWrite()) {
                save(file);
            } else {
                doSaveAs();
            }
        }
    }

    @Override
    public void doSaveAs() {
        Shell shell = getSite().getShell();
        OpenChessGameDialog fileDialog = new OpenChessGameDialog(shell, SWT.SAVE);
        fileDialog.setText("Save As");  // TODO externalize this
        fileDialog.setFileName(getPartName());
        fileDialog.setOverwrite(true);
        String pathName = fileDialog.open();
        
        if(pathName != null) {
            pathName = fileDialog.getFileNameWithExtension(pathName);
            FileEditorInput newEditorInput = new FileEditorInput(pathName);
            IEditorPart openEditor = getSite().getPage().findEditor(newEditorInput);
            if((openEditor != null) && (openEditor != this)) {
                MessageDialog.openError(shell, "Error!", "The selected game file is already open.\n" +
                        "It is necessary to close it before attempting to override its contents.");  // TODO externalize this
                return;
            }
            File newFile = newEditorInput.getFile();
            if(!newFile.exists()) {
                try {
                    newFile.createNewFile();
                } catch(IOException e) {
                    MessageDialog.openError(shell, "Error!", "Error creating the file.\n" +
                        "Please check that you have write permission in the specified directory.");  // TODO externalize this
                    return;
                }
            } else if(!newFile.canWrite()) {
                MessageDialog.openError(shell, "Error!", "The selected file is marked as read-only.");  // TODO externalize this
                return;
            }
            save(newFile);
            onGameSavedToFile(newEditorInput);
        }
    }

    public ChessGame getGame() {
        return game;
    }
    
    public ImageData getBoardImage() {
        return getBoardControl().getBoardImage();
    }

    public boolean isFlipped() {
        return getBoardControl().isFlipped();
    }

    public void flipBoard() {
        getBoardControl().flip();
        
        synchronized(flipListeners) {
            for(IFlipListener flipListener : flipListeners) {
                flipListener.flipped(isFlipped());
            }
        }
    }
    
    public void addFlipListener(IFlipListener flipListener) {
        synchronized(flipListeners) {
            flipListeners.add(flipListener);
        }
    }
    
    public void removeFlipListener(IFlipListener flipListener) {
        synchronized(flipListeners) {
            flipListeners.remove(flipListener);
        }
    }

    @Override
    protected void setInput(IEditorInput input) {
        setPartName(input.getName());
        super.setInput(input);
    }

    @Override
    protected void setInputWithNotify(IEditorInput input) {
        setPartName(input.getName());
        super.setInputWithNotify(input);
    }

    protected void onGameSavedToFile(FileEditorInput fileEditorInput) {
    }

    private RectangularBoardControl<ChessPieceType> getBoardControl() {
        return gameController.getBoardControl();
    }

    private IPreferenceStore getPreferenceStore() {
        return HaundrixPlugin.getInstance().getPreferenceStore();
    }

    private void updateStatusArea(String statusMessage) {
        this.statusMessage = statusMessage;
        if(getEditorSite().getPage().getActiveEditor() == this) {
            updateStatusArea();
        }
    }

    private void updateStatusArea() {
        IStatusLineManager statusLineManager = getEditorSite().getActionBars().getStatusLineManager();
        if(statusMessage == null) {
            statusLineManager.setMessage(disabledMascot, null);
        } else {
            statusLineManager.setMessage(enabledMascot, statusMessage);
        }
    }

    private void setCustomMascot(PluginPieceSet pieceSet) {
        if(isCustomMascot) {
            enabledMascot.dispose();
            disabledMascot.dispose();
        }
        
        isCustomMascot = false;
        enabledMascot = enabledOwl;
        disabledMascot = disabledOwl;
        if(pieceSet.getEnabledMascotImage() != null) {
            isCustomMascot = true;
            Display display = PlatformUI.getWorkbench().getDisplay();
            enabledMascot = pieceSet.getEnabledMascotImage().createImage(display);
            disabledMascot = pieceSet.getDisabledMascotImage().createImage(display);
        }
        
        updateStatusArea();
    }

    private boolean getUsePrefColors() {
        return getPreferenceStore().getBoolean(HaundrixPreferences.CHESS_BOARD_USE_PIECE_SET_PREFERRED_COLORS);
    }

    private PluginPieceSet getPieceSet() {
        String pieceSetId = getPreferenceStore().getString(HaundrixPreferences.CHESS_BOARD_PIECE_SET);
        return PieceSetManager.getInstance().getPieceSet(pieceSetId);
    }
    
    private void updatePieceSet() {
        PluginPieceSet pieceSet = getPieceSet();
        getBoardControl().setPieceSet(pieceSet);
        setCustomMascot(pieceSet);
        if(getUsePrefColors()) {
            updateColors();
        }
    }

    private void updateColors() {
        updateColorScheme();
        updateHighlightScheme();
    }
    
    private void updateColorScheme() {
        String colorSchemeId = getUsePrefColors() ?
                getPieceSet().getPreferredColorSchemeId() :
                getPreferenceStore().getString(HaundrixPreferences.CHESS_BOARD_COLOR_SCHEME);
        ColorScheme colorScheme = ColorSchemeManager.getInstance().getColorScheme(colorSchemeId);
        getBoardControl().setColorScheme(colorScheme);;
    }

    private void updateHighlightScheme() {
        String highlightSchemeId = getUsePrefColors() ?
                getPieceSet().getPreferredHighlightSchemeId() :
                getPreferenceStore().getString(HaundrixPreferences.CHESS_BOARD_HIGHLIGHT_SCHEME);
        HighlightScheme highlightScheme = HighlightSchemeManager.getInstance().getHighlightScheme(highlightSchemeId);
        gameController.setHighlightScheme(highlightScheme);
    }

    private void save(File file) {
        try {
            FileWriter writer = new FileWriter(file);
            PgnExporter pgnExporter = new PgnExporter(writer);
            pgnExporter.exportGame(game);
            writer.close();
        } catch(IOException exception) {
            String errorMessage = "Error exporting game to " + file.getAbsolutePath();
            HaundrixLog.logError(errorMessage, exception);
            MessageDialog.openError(getSite().getShell(), "Error!", errorMessage);
            return;
        }
        
        isDirty = false;
        firePropertyChange(PROP_DIRTY);
    }
    
    private void updateExpressions() {
        IEvaluationService evaluationService = (IEvaluationService) getEditorSite().getService(IEvaluationService.class);
        evaluationService.requestEvaluation(IsGameAtEndAndResultUndefinedPropertyTester.FULL_PROPERTY_NAME);
    }
    
    private void updateGameStatusMessage() {
        if(game.isFinished()) {
            updateFinishedGameStatusMessage();
        } else if(game.isCheck()) {
            updateStatusArea("Check!");  // TODO externalize this
        } else {
            updateStatusArea(null);
        }
    }

    private void updateFinishedGameStatusMessage() {
        ChessGameResult result = game.getResult();
        GameOutcome outcome = result.getOutcome();
        ChessGameResultReason resultReason = result.getResultReason();
        
        if(outcome == GameOutcome.WHITE_WINS) {
            String generalMessage = "White wins the game.";  // TODO externalize this
            
            if(resultReason == ChessGameResultReason.CHECKMATE) {
                updateStatusArea("Checkmate! " + generalMessage);  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.RESIGNATION) {
                updateStatusArea("Black resigns. " + generalMessage);  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.TIME_FORFEIT) {
                updateStatusArea("Black forfeits on time. " + generalMessage);  // TODO externalize this
            } else {
                updateStatusArea(generalMessage);
            }
        } else if(outcome == GameOutcome.BLACK_WINS) {
            String generalMessage = "Black wins the game.";  // TODO externalize this
            
            if(resultReason == ChessGameResultReason.CHECKMATE) {
                updateStatusArea("Checkmate! " + generalMessage);  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.RESIGNATION) {
                updateStatusArea("White resigns. " + generalMessage);  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.TIME_FORFEIT) {
                updateStatusArea("White forfeits on time. " + generalMessage);  // TODO externalize this
            } else {
                updateStatusArea(generalMessage);
            }
        } else {
            String generalMessage = "The game is a draw.";  // TODO externalize this
            
            if(resultReason == ChessGameResultReason.STALEMATE) {
                updateStatusArea("Stalemate. " + generalMessage);  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.DEAD_POSITION) {
                updateStatusArea("Neither side can give checkmate. " + generalMessage);  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.POSITION_REPETITION) {
                updateStatusArea("The same position has appeared three times in the game. " + generalMessage);  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.FIFTY_MOVE_RULE) {
                updateStatusArea("Fifty moves with no piece captures or pawn advances. " + generalMessage);  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.COMMON_AGREEMENT) {
                updateStatusArea("Both players agreeded to a draw.");  // TODO externalize this
            } else {
                updateStatusArea(generalMessage);
            }
        }
    }

    private class InternalGameListener extends GameAdapter {
        
        @Override
        public void infoChanged(GameInfo info) {
            isDirty = true;
            firePropertyChange(PROP_DIRTY);
        }

        @Override
        public void afterPositionChange(Game game) {
            updateGameStatusMessage();
            updateExpressions();
        }

        @Override
        public void moved(Game game) {
            updateGameStatusMessage();
            updateExpressions();
            
            isDirty = true;
            firePropertyChange(PROP_DIRTY);
        }

        @Override
        public void ended(GameResult result) {
            updateFinishedGameStatusMessage();
            updateExpressions();
            
            isDirty = true;
            firePropertyChange(PROP_DIRTY);
        }
        
    }

    private final class InternalGameControllerListener extends GameControllerAdapter {
        
        @Override
        public void pieceDraggedOffBoard(AnimationEvent event) {
            updateStatusArea("Oops...");  // TODO externalize this
        }

        @Override
        public void illegalMoveAttempted(AnimationEvent event) {
            updateStatusArea("Illegal move!");  // TODO externalize this
        }
        
    }

    private class InternalPreferencesListener implements IPropertyChangeListener {

        @Override
        public void propertyChange(PropertyChangeEvent event) {
            String preference = event.getProperty();
            if(preference.equals(HaundrixPreferences.CHESS_BOARD_PIECE_SET)) {
                updatePieceSet();
            } else if(preference.equals(HaundrixPreferences.CHESS_BOARD_USE_PIECE_SET_PREFERRED_COLORS)) {
                updateColors();
            } else if(preference.equals(HaundrixPreferences.CHESS_BOARD_COLOR_SCHEME)) {
                if(!getUsePrefColors()) {
                    updateColorScheme();
                }
            } else if(preference.equals(HaundrixPreferences.CHESS_BOARD_HIGHLIGHT_SCHEME)) {
                if(!getUsePrefColors()) {
                    updateHighlightScheme();
                }
            }
        }

    }

}
