/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.editors;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.ui.actions.CompoundContributionItem;

import com.haundrix.games.rcp.HaundrixPlugin;
import com.haundrix.games.rcp.managers.PieceSetCategory;
import com.haundrix.games.rcp.managers.PieceSetManager;
import com.haundrix.games.rcp.managers.PluginPieceSet;
import com.haundrix.games.rcp.preferences.HaundrixPreferences;
import com.haundrix.utils.NameComparator;

public class DynamicFormatPieceSetContributionItem extends CompoundContributionItem {
    
    private IContributionItem[] contributionItems;
    
    private List<FormatPieceSetAction> actions = new ArrayList<FormatPieceSetAction>();
    private FormatPieceSetAction checkedAction;
    
    private IPropertyChangeListener preferencesListener = new IPropertyChangeListener() {
        @Override
        public void propertyChange(PropertyChangeEvent event) {
            if(event.getProperty().equals(HaundrixPreferences.CHESS_BOARD_PIECE_SET)) {
                updateCheckedAction();
            }
        }
    };

    public DynamicFormatPieceSetContributionItem() {
    }

    public DynamicFormatPieceSetContributionItem(String id) {
        super(id);
    }

    @Override
    public void dispose() {
        getPreferenceStore().removePropertyChangeListener(preferencesListener);
        
        super.dispose();
    }

    @Override
    protected IContributionItem[] getContributionItems() {
        if(contributionItems == null) {
            createContributionItems();
        }
        
        return contributionItems;
    }
    
    private IPreferenceStore getPreferenceStore() {
        return HaundrixPlugin.getInstance().getPreferenceStore();
    }

    private void createContributionItems() {
        PieceSetManager pieceSetManager = PieceSetManager.getInstance();

        PieceSetCategory[] categories = pieceSetManager.getAvailableCategories();
        Arrays.sort(categories, NameComparator.getInstance());
        contributionItems = new IContributionItem[categories.length];
        for(int i = 0; i < categories.length; i++) {
            PieceSetCategory category = categories[i];
            MenuManager categoryMenu = new MenuManager(category.getName());
            
            PluginPieceSet[] pieceSets = pieceSetManager.getAvailablePieceSets(category);
            Arrays.sort(pieceSets, NameComparator.getInstance());
            for(PluginPieceSet pieceSet : pieceSets) {
                FormatPieceSetAction action = new FormatPieceSetAction(pieceSet);
                actions.add(action);
                categoryMenu.add(action);
            }
            
            contributionItems[i] = categoryMenu;
        }
        
        updateCheckedAction();
        getPreferenceStore().addPropertyChangeListener(preferencesListener);
    }
    
    private void updateCheckedAction() {
        if(checkedAction != null) {
            checkedAction.setChecked(false);
        }
        
        String pieceSetId = getPreferenceStore().getString(HaundrixPreferences.CHESS_BOARD_PIECE_SET);
        for(FormatPieceSetAction action : actions) {
            if(action.getPieceSet().getId().equals(pieceSetId)) {
                action.setChecked(true);
                checkedAction = action;
                break;
            }
        }
    }

    private class FormatPieceSetAction extends Action {
        
        private PluginPieceSet pieceSet;

        public FormatPieceSetAction(PluginPieceSet pieceSet) {
            super(pieceSet.getName(), IAction.AS_RADIO_BUTTON);
            
            this.pieceSet = pieceSet;
        }

        public PluginPieceSet getPieceSet() {
            return pieceSet;
        }

        @Override
        public void run() {
            if(isChecked()) {
                getPreferenceStore().setValue(HaundrixPreferences.CHESS_BOARD_PIECE_SET, pieceSet.getId());
                HaundrixPlugin.getInstance().savePreferences();
            }
        }
        
    }

}
