/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.editors;

import java.io.IOException;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.services.IEvaluationService;
import org.joda.time.LocalDate;

import com.haundrix.games.chess.controller.ChessGameControllerAdapter;
import com.haundrix.games.chess.model.ChessGame;
import com.haundrix.games.chess.model.ChessGameResult;
import com.haundrix.games.chess.model.ChessGameResultReason;
import com.haundrix.games.chess.model.playing.ComputerPlayer;
import com.haundrix.games.controller.MoveAnimationEvent;
import com.haundrix.games.model.Game;
import com.haundrix.games.model.GameAdapter;
import com.haundrix.games.model.GameInfo;
import com.haundrix.games.model.GameOutcome;
import com.haundrix.games.model.GameResult;
import com.haundrix.games.model.Move;
import com.haundrix.games.model.Piece.PieceColor;
import com.haundrix.games.model.playing.Engine;
import com.haundrix.games.model.playing.EngineConfig;
import com.haundrix.games.model.playing.HumanPlayer;
import com.haundrix.games.model.playing.IClock;
import com.haundrix.games.model.playing.IGameProxy;
import com.haundrix.games.model.playing.IPlayer;
import com.haundrix.games.rcp.HaundrixLog;
import com.haundrix.games.rcp.managers.EngineManager;
import com.haundrix.games.rcp.managers.PluginEngine;
import com.haundrix.utils.IExceptionHandler;

public class LocalChessGameEditor extends ChessGameEditor {
    
    public static final String ID = "com.haundrix.games.localChessGameEditor";
    
    private static final long MINIMUM_CLOCK_CHECK_INTERVAL = 50L;  // in milliseconds
    
    private volatile Display display;

    private IPlayer whitePlayer;
    private IPlayer blackPlayer;
    
    private boolean isDrawOffered = false;

    private boolean notifiedUserOfGameResult = false;
    
    private long playerTimeThreshold;
    private IPlayer currentPlayer;
    private volatile ScheduledFuture<?> forcePlayerToMoveFuture = null;
    
    private ScheduledExecutorService scheduler;
    
    private volatile boolean isDisposed = false;
    private boolean exceptionOccurred = false;

    public IPlayer getWhitePlayer() {
        return whitePlayer;
    }

    public IPlayer getBlackPlayer() {
        return blackPlayer;
    }
    
    public IClock getClock() {
        return getEditorInput().getClock();
    }

    @Override
    public LocalChessGameEditorInput getEditorInput() {
        return (LocalChessGameEditorInput) super.getEditorInput();
    }

    @Override
    public void init(IEditorSite site, IEditorInput input) throws PartInitException {
        if(!(input instanceof LocalChessGameEditorInput)) {
            throw new PartInitException("Error! Invalid editor input type to LocalChessGameEditor.");
        }
        
        setSite(site);
        setInput(input);
        
        game = new ChessGame(true, true);
        
        isDirty = false;
    }

    @Override
    public void createPartControl(Composite parent) {
        super.createPartControl(parent);
        
        display = parent.getDisplay();
        final IClock clock = getClock();
        
        game.addGameListener(new InternalGameListener());
        
        if(clock != null) {
            gameController.addChessGameControllerListener(new InternalChessGameControllerListener());
        }
        
        try {
            IGameProxy gameProxy = new InternalGameProxy();
            createPlayer(true, gameProxy);   // creates white player
            createPlayer(false, gameProxy);  // creates black player
            currentPlayer = whitePlayer;
            
            GameInfo gameInfo = game.getInfo();
            gameInfo.setEvent("Local game");  // TODO externalize this
            gameInfo.setDate(new LocalDate());
            
            if(clock != null) {
                clock.start();
                calculatePlayerTimeThreshold();
                scheduler = Executors.newScheduledThreadPool(2);
                scheduleClockChecks(clock);
                TimeTroubleMonitor.activate();
            }
            blackPlayer.go();
            whitePlayer.go();
        } catch(Exception e) {
            // TODO Improve error message to specify which engine has failed (white or black player).
            String errorMessage = "Error starting chess engine!";  // TODO externalize this
            HaundrixLog.logError(errorMessage, e);
            MessageDialog.openError(parent.getShell(), "Error!", errorMessage);  // TODO externalize this
            return;
        }
        
        if((whitePlayer instanceof ComputerPlayer) && (blackPlayer instanceof HumanPlayer)) {
            flipBoard();
        }
        
        soundController.playNewGameSound();
    }
    
    @Override
    public void setFocus() {
        super.setFocus();
        
        if(game.isOver() && !notifiedUserOfGameResult) {
            soundController.playGameEndingSound(game.getResult());
            notifyUserOfGameResultWithoutBlocking();
        }
    }

    @Override
    public void dispose() {
        isDisposed = true;
        
        if((getClock() != null) && !exceptionOccurred) {
            TimeTroubleMonitor.deactivate();
        }
        
        if((scheduler != null) && !scheduler.isShutdown()) {
            scheduler.shutdownNow();
        }
        
        disposePlayers();
        
        super.dispose();
    }
    
    public boolean isComputerVsComputerMatch() {
        return (whitePlayer instanceof ComputerPlayer) && (blackPlayer instanceof ComputerPlayer);
    }

    public boolean isComputerToPlay() {
        return currentPlayer instanceof ComputerPlayer;
    }
    
    public void offerDraw() {
        if(mayChangeGame() && !isComputerVsComputerMatch()) {
            offerDraw(getUserColor());
        }
    }

    public void resign() {
        if(mayChangeGame() && !isComputerVsComputerMatch()) {
            PieceColor resigningColor = getUserColor();
            GameOutcome outcome = resigningColor == PieceColor.WHITE ? GameOutcome.BLACK_WINS : GameOutcome.WHITE_WINS;
            soundController.playGameEndingSound(new ChessGameResult(outcome, ChessGameResultReason.RESIGNATION));
            game.resign(resigningColor);
        }
    }

    public void moveNow() {
        if(mayChangeGame()) {
            currentPlayer.moveNow();
        }
    }
    
    private void safeAsyncExec(final Runnable runnable) {
        if(!isDisposed) {
            display.asyncExec(new Runnable() {
                @Override
                public void run() {
                    if(!isDisposed) {
                        runnable.run();
                    }
                }
            });
        }
    }

    private void createPlayer(boolean isWhitePlayer, IGameProxy gameProxy) throws IOException {
        IPlayerConfig playerConfig = isWhitePlayer ? getEditorInput().getWhitePlayerConfig() :
            getEditorInput().getBlackPlayerConfig();
        IClock clock = getEditorInput().getClock();
        
        IPlayer player = null;
        String playerName = "";
        if(playerConfig instanceof HumanPlayerConfig) {
            player = new HumanPlayer();
            playerName = ((HumanPlayerConfig) playerConfig).getName();
        } else if(playerConfig instanceof ComputerPlayerConfig) {
            ComputerPlayerConfig computerPlayerConfig = (ComputerPlayerConfig) playerConfig;
            String engineId = computerPlayerConfig.getEngineId();
            EngineConfig engineConfig = computerPlayerConfig.getEngineConfig();
            
            PluginEngine pluginEngine = EngineManager.getInstance().getEngine(engineId);
            playerName = pluginEngine.getName();
            Engine engine = pluginEngine.getEngine();
            
            PieceColor playerColor = isWhitePlayer ? PieceColor.WHITE : PieceColor.BLACK;
            
            boolean opponentIsComputer = isWhitePlayer ?
                    getEditorInput().getBlackPlayerConfig() instanceof ComputerPlayerConfig :
                    getEditorInput().getWhitePlayerConfig() instanceof ComputerPlayerConfig;
            
            player = new ComputerPlayer(playerColor, clock, engine, engineConfig, opponentIsComputer, gameProxy,
                    new InternalExceptionHandler());
            
            if(isWhitePlayer) {
                gameController.setAllowUserWhiteMoves(false);
            } else {
                gameController.setAllowUserBlackMoves(false);
            }
        }
        
        if(isWhitePlayer) {
            whitePlayer = player;
            game.getInfo().setWhitePlayer(playerName);
        } else {
            blackPlayer = player;
            game.getInfo().setBlackPlayer(playerName);
        }
    }
    
    private void disposePlayers() {
        if((whitePlayer != null) && !whitePlayer.isDisposed()) {
            whitePlayer.dispose();
        }
        
        if((blackPlayer != null) && !blackPlayer.isDisposed()) {
            blackPlayer.dispose();
        }
    }

    private String getPlayerName(PieceColor playerColor) {
        return playerColor == PieceColor.WHITE ? game.getInfo().getWhitePlayer() : game.getInfo().getBlackPlayer();
    }

    private PieceColor getUserColor() {
        if((whitePlayer instanceof HumanPlayer) && (blackPlayer instanceof ComputerPlayer)) {
            return PieceColor.WHITE;
        } else if((whitePlayer instanceof ComputerPlayer) && (blackPlayer instanceof HumanPlayer)) {
            return PieceColor.BLACK;
        } else if((whitePlayer instanceof HumanPlayer) && (blackPlayer instanceof HumanPlayer)) {
            return game.getLastColorToMove();
        } else {
            throw new IllegalStateException("Error! There is no user playing in a computer vs computer game.");
        }
    }
    
    private boolean mayChangeGame() {
        boolean isFlagFalled = getClock() != null ? getClock().isFlagFallen() : false;
        return !game.isOver() && !isFlagFalled;
    }

    private void calculatePlayerTimeThreshold() {
        playerTimeThreshold = currentPlayer.calculateTimeThreshold(getClock().getCurrentPlayerTime());
    }

    private void scheduleClockChecks(IClock clock) {
        long checkTime = Math.max(clock.getMinimumTimeTillFlagFalls() / 2L, MINIMUM_CLOCK_CHECK_INTERVAL);
        
        if(forcePlayerToMoveFuture == null) {
            synchronized(currentPlayer) {
                if(clock.getCurrentPlayerTime() - checkTime <= playerTimeThreshold) {
                    forcePlayerToMoveFuture = scheduler.schedule(new ForcePlayerToMoveTask(currentPlayer),
                            clock.getCurrentPlayerTime() - playerTimeThreshold, TimeUnit.MILLISECONDS);
                }
            }
        }
        
        scheduler.schedule(new ClockCheckTask(clock), checkTime, TimeUnit.MILLISECONDS);
    }

    private void offerDraw(PieceColor drawOfferingPlayerColor) {
        IPlayer otherPlayer = drawOfferingPlayerColor == PieceColor.WHITE ? blackPlayer : whitePlayer;
        if(otherPlayer instanceof HumanPlayer) {
            String drawOfferMessage = getPlayerName(drawOfferingPlayerColor) + " offers a draw.\nAccept?";  // TODO externalize this
            Shell shell = getEditorSite().getShell();
            if(MessageDialog.openQuestion(shell, "Question", drawOfferMessage)) {  // TODO externalize this
                if(mayChangeGame()) {
                    acceptDraw();
                }
            }
        } else {
            isDrawOffered = true;
            ((ComputerPlayer) otherPlayer).offerDraw();
        }
    }

    private void acceptDraw() {
        soundController.playGameEndingSound(new ChessGameResult(GameOutcome.DRAW,
                ChessGameResultReason.COMMON_AGREEMENT));
        game.setDrawByCommonAgreement();
    }

    // This method wraps a call to notifyUserOfGameResult() in
    // order to make sure it won't block the GUI thread for now.
    private void notifyUserOfGameResultWithoutBlocking() {
        final Display display = getEditorSite().getShell().getDisplay();
        new Thread() {
            @Override
            public void run() {
                display.asyncExec(new Runnable() {
                    @Override
                    public void run() {
                        notifyUserOfGameResult();
                    }
                });
            }
        }.start();
    }

    private void notifyUserOfGameResult() {
        ChessGameResult result = game.getResult();
        GameOutcome outcome = result.getOutcome();
        ChessGameResultReason resultReason = result.getResultReason();
        
        String gameResultMessage;
        if(outcome == GameOutcome.WHITE_WINS) {
            if(resultReason == ChessGameResultReason.CHECKMATE) {
                gameResultMessage = "Checkmate!";  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.RESIGNATION) {
                gameResultMessage = "Black resigns.";  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.TIME_FORFEIT) {
                gameResultMessage = "Black forfeits on time.";
            } else {
                gameResultMessage = "White wins the game.";  // TODO externalize this
            }
        } else if(outcome == GameOutcome.BLACK_WINS) {
            if(resultReason == ChessGameResultReason.CHECKMATE) {
                gameResultMessage = "Checkmate!";  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.RESIGNATION) {
                gameResultMessage = "White resigns.";  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.TIME_FORFEIT) {
                gameResultMessage = "White forfeits on time.";
            } else {
                gameResultMessage = "Black wins the game.";  // TODO externalize this
            }
        } else {
            if(resultReason == ChessGameResultReason.STALEMATE) {
                gameResultMessage = "Stalemate!\nThe game is a draw.";  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.DEAD_POSITION) {
                gameResultMessage = "Neither side can give checkmate.\nThe game is a draw.";  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.POSITION_REPETITION) {
                gameResultMessage = "The same position has appeared three times in the game.\nThe game is a draw.";  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.FIFTY_MOVE_RULE) {
                gameResultMessage = "Fifty moves with no piece captures or pawn advances.\nThe game is a draw.";  // TODO externalize this
            } else if(resultReason == ChessGameResultReason.COMMON_AGREEMENT) {
                gameResultMessage = "Both players agreeded to a draw.";  // TODO externalize this
            } else {
                gameResultMessage = "The game is a draw.";
            }
        }
        
        Shell shell = getEditorSite().getShell();
        MessageDialog.openInformation(shell, "Game over!", gameResultMessage);  // TODO externalize this
        
        notifiedUserOfGameResult = true;
    }

    private final class InternalGameProxy implements IGameProxy {
        
        @Override
        public void makeMove(final Move move) {
            safeAsyncExec(new Runnable() {
                @Override
                public void run() {
                    if(mayChangeGame()) {
                        gameController.makeMove(move);
                    }
                }
            });
        }

        @Override
        public void makeDrawOffer(final PieceColor drawOfferingPlayerColor) {
            safeAsyncExec(new Runnable() {
                @Override
                public void run() {
                    if(mayChangeGame()) {
                        offerDraw(drawOfferingPlayerColor);
                    }
                }
            });
        }
        
        @Override
        public void communicateDrawAcceptance() {
            safeAsyncExec(new Runnable() {
                @Override
                public void run() {
                    if(mayChangeGame()) {
                        if(isDrawOffered) {
                            isDrawOffered = false;
                            acceptDraw();
                        }
                    }
                }
            });
        }

        @Override
        public void communicateResignation(final PieceColor resigningPlayerColor) {
            safeAsyncExec(new Runnable() {
                @Override
                public void run() {
                    if(mayChangeGame()) {
                        game.resign(resigningPlayerColor);
                    }
                }
            });
        }

    }

    private final class InternalGameListener extends GameAdapter {

        @Override
        public void afterPositionChange(Game game) {
            updateExpressions();
        }

        @Override
        public void moved(Game game) {
            synchronized(currentPlayer) {
                if((forcePlayerToMoveFuture != null)) {
                    if(!forcePlayerToMoveFuture.isDone()) {
                        forcePlayerToMoveFuture.cancel(true);
                    }
                    forcePlayerToMoveFuture = null;
                }
                
                isDrawOffered = false;
                
                IClock clock = getClock();
                
                IPlayer waitingPlayer = currentPlayer == whitePlayer ? blackPlayer : whitePlayer;
                if(clock != null) {
                    waitingPlayer.setTime(clock.getClockTime());
                }
                waitingPlayer.move(game.getLastMove());
                
                if(clock != null) {
                    clock.press();
                    currentPlayer.setTime(clock.getClockTime());
                }
                
                currentPlayer = waitingPlayer;
                
                if(clock != null) {
                    calculatePlayerTimeThreshold();
                }
                
                updateExpressions();
            }
        }

        @Override
        public void ended(GameResult result) {
            IClock clock = getClock();
            if(clock != null) {
                if(clock.isRunning()) {
                    clock.stop();
                }
                scheduler.shutdownNow();
            }
            
            whitePlayer.setResult(result);
            blackPlayer.setResult(result);
            
            if(getEditorSite().getPage().getActiveEditor() == LocalChessGameEditor.this) {
                notifyUserOfGameResultWithoutBlocking();
            }
            disposePlayers();
            
            updateExpressions();
        }

        private void updateExpressions() {
            IEvaluationService evaluationService = (IEvaluationService) getEditorSite().getService(IEvaluationService.class);
            evaluationService.requestEvaluation(IsComputerToPlayPropertyTester.FULL_PROPERTY_NAME);
        }
        
    }

    private class InternalChessGameControllerListener extends ChessGameControllerAdapter {

        @Override
        public void beginMoveAnimation(MoveAnimationEvent event) {
            if(event.isNewMove()) {
                getClock().stop();
            }
        }

        @Override
        public void endMoveAnimation(MoveAnimationEvent event) {
            if(event.isNewMove() && !game.isCastle(event.getMove())) {
                getClock().start();
            }
        }

        @Override
        public void endCastleRookMoveAnimation(MoveAnimationEvent event) {
            if(event.isNewMove()) {
                getClock().start();
            }
        }

        @Override
        public void beginPromotionMoveAnimation(MoveAnimationEvent event) {
            if(event.isNewMove()) {
                getClock().stop();
            }
        }

        @Override
        public void endPromotionMoveAnimation(MoveAnimationEvent event) {
            if(event.isNewMove()) {
                getClock().start();
            }
        }
        
    }

    private class InternalExceptionHandler implements IExceptionHandler {
        
        @Override
        public void handle(Exception exception) {
            // TODO Improve error message to specify which engine has failed (white or black player).
            String errorMessage = "Error interacting with chess engine!";  // TODO externalize this
            HaundrixLog.logError(errorMessage, exception);
            
            if(getClock() != null) {
                TimeTroubleMonitor.deactivate();
            }
            if((scheduler != null) && !scheduler.isShutdown()) {
                scheduler.shutdownNow();
            }
            gameController.setAllowUserWhiteMoves(false);
            gameController.setAllowUserBlackMoves(false);
            disposePlayers();
            
            exceptionOccurred = true;
        }
        
    }

    private class ClockCheckTask implements Runnable {

        private IClock clock;

        public ClockCheckTask(IClock clock) {
            this.clock = clock;
        }

        @Override
        public void run() {
            if(clock.isFlagFallen()) {
                safeAsyncExec(new Runnable() {
                    @Override
                    public void run() {
                        callTimeFlag();
                    }
                });
            } else {
                scheduleClockChecks(clock);
            }
        }

        // This method will execute in the GUI thread.
        private void callTimeFlag() {
            if(!game.isOver()) {
                // TODO What if both flags have fallen? Is it a draw?
                if(clock.getCurrentPlayerTime() <= 0L) {
                    game.callTimeFlag(clock.getColorToMove());
                } else if(clock.getOtherPlayerTime() <= 0L) {
                    game.callTimeFlag(clock.getColorToMove().getOppositeColor());
                } else {
                    // Some miracle has happened!
                    // Maybe the player in time trouble made a move just
                    // in time and was alloted some extra time on the clock.
                    // Let's start monitoring the clock again:
                    scheduleClockChecks(clock);
                }
            }
        }
    }

    public class ForcePlayerToMoveTask implements Runnable {
        
        private IPlayer currentPlayer;

        public ForcePlayerToMoveTask(IPlayer currentPlayer) {
            this.currentPlayer = currentPlayer;
        }

        @Override
        public void run() {
            safeAsyncExec(new Runnable() {
                @Override
                public void run() {
                    currentPlayer.moveNow();
                }
            });
        }

    }

}
