/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.editors;

import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;

import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;

import com.haundrix.games.chess.view.ChessPieceType;
import com.haundrix.games.model.playing.IClock;
import com.haundrix.games.rcp.HaundrixImages;
import com.haundrix.games.rcp.HaundrixPlugin;
import com.haundrix.games.rcp.managers.EngineManager;
import com.haundrix.games.rcp.managers.PieceSetManager;
import com.haundrix.games.rcp.managers.PluginEngine;
import com.haundrix.games.rcp.managers.PluginPieceSet;
import com.haundrix.games.rcp.preferences.HaundrixPreferences;
import com.haundrix.swt.CenteredDialog;
import com.haundrix.utils.ImageUtils;

class NewLocalGameDialog extends CenteredDialog {
    
    // size of the little piece icon that appears
    // in the bottom right corner of the player image
    private static final int PIECE_ICON_SIZE = 32;
    
    private Image humanWhitePlayerImage;
    private Image humanBlackPlayerImage;
    private Image computerWhitePlayerImage;
    private Image computerBlackPlayerImage;
    private boolean imagesCreated = false;
    
    private Label whitePlayerLabel;
    private Label blackPlayerLabel;
    
    private PlayerConfigControl whitePlayerControl;
    private PlayerConfigControl blackPlayerControl;

    private ClockControl clockControl;
    
    private LocalChessGameEditorInput newGameEditorInput;
    
    public NewLocalGameDialog(Shell parent) {
        super(parent);
    }
    
    public LocalChessGameEditorInput getNewGameEditorInput() {
        openAndRunLoop();
        
        return newGameEditorInput;
    }
    
    @Override
    public void dispose() {
        if(imagesCreated) {
            humanWhitePlayerImage.dispose();
            humanBlackPlayerImage.dispose();
            computerWhitePlayerImage.dispose();
            computerBlackPlayerImage.dispose();
        }

        super.dispose();
    }

    @Override
    protected void createWidgets() {
        createPlayerImages();

        dialog.setLayout(new GridLayout(1, true));
        
        Composite playerImagesPanel = new Composite(dialog, SWT.NONE);
        playerImagesPanel.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
        fillPlayerImagesPanel(playerImagesPanel);
        
        PluginEngine[] engines = EngineManager.getInstance().getEngines();
        
        Group whitePlayerGroup = new Group(dialog, SWT.SHADOW_OUT);
        whitePlayerGroup.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
        whitePlayerGroup.setText("White Player");  // TODO externalize this
        whitePlayerGroup.setLayout(new FillLayout());
        whitePlayerControl = new PlayerConfigControl(whitePlayerGroup, engines, whitePlayerLabel,
                humanWhitePlayerImage, computerWhitePlayerImage);
        
        Group blackPlayerGroup = new Group(dialog, SWT.SHADOW_OUT);
        blackPlayerGroup.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
        blackPlayerGroup.setText("Black Player");  // TODO externalize this
        blackPlayerGroup.setLayout(new FillLayout());
        blackPlayerControl = new PlayerConfigControl(blackPlayerGroup, engines, blackPlayerLabel,
                humanBlackPlayerImage, computerBlackPlayerImage);
        
        Group clockGroup = new Group(dialog, SWT.SHADOW_OUT);
        clockGroup.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
        clockGroup.setText("Time Control");  // TODO externalize this
        clockGroup.setLayout(new FillLayout());
        clockControl = new ClockControl(clockGroup);
        
        Composite buttonsPanel = new Composite(dialog, SWT.NONE);
        buttonsPanel.setLayoutData(new GridData(SWT.END, SWT.CENTER, true, true));
        fillButtonsPanel(buttonsPanel);
    }

    private void createPlayerImages() {
        // retrieves the active piece set
        IPreferenceStore preferenceStore = HaundrixPlugin.getInstance().getPreferenceStore();
        String pieceSetId = preferenceStore.getString(HaundrixPreferences.CHESS_BOARD_PIECE_SET);
        PluginPieceSet pieceSet = PieceSetManager.getInstance().getPieceSet(pieceSetId);
        
        // creates the piece icon images
        BufferedImage whitePieceImage = pieceSet.loadPieceImage(ChessPieceType.WHITE_KNIGHT);
        BufferedImage blackPieceImage = pieceSet.loadPieceImage(ChessPieceType.BLACK_KNIGHT);
        BufferedImage whitePieceIconImage = ImageUtils.fastAndGoodScale(whitePieceImage, PIECE_ICON_SIZE, PIECE_ICON_SIZE);
        BufferedImage blackPieceIconImage = ImageUtils.fastAndGoodScale(blackPieceImage, PIECE_ICON_SIZE, PIECE_ICON_SIZE);

        // creates the player images
        Image swtHumanPlayerImage = HaundrixImages.HUMAN.createImage();
        Image swtComputerPlayerImage = HaundrixImages.COMPUTER.createImage();
        BufferedImage awtHumanPlayerImage = ImageUtils.convertToAwt(swtHumanPlayerImage.getImageData());
        BufferedImage awtComputerPlayerImage = ImageUtils.convertToAwt(swtComputerPlayerImage.getImageData());
        swtHumanPlayerImage.dispose();
        swtComputerPlayerImage.dispose();
        
        Display display = dialog.getDisplay();
        humanWhitePlayerImage = createDecoratedPlayerImage(display, awtHumanPlayerImage, whitePieceIconImage);
        humanBlackPlayerImage = createDecoratedPlayerImage(display, awtHumanPlayerImage, blackPieceIconImage);
        computerWhitePlayerImage = createDecoratedPlayerImage(display, awtComputerPlayerImage, whitePieceIconImage);
        computerBlackPlayerImage = createDecoratedPlayerImage(display, awtComputerPlayerImage, blackPieceIconImage);
        
        imagesCreated = true;
    }

    private Image createDecoratedPlayerImage(Display display, BufferedImage playerImage, BufferedImage pieceIconImage) {
        // does the drawing
        int extraSpace = (int) (PIECE_ICON_SIZE / 3.0);
        BufferedImage awtDecoratedPlayerImage = new BufferedImage(playerImage.getWidth() + extraSpace,
                playerImage.getHeight() + extraSpace, BufferedImage.TYPE_INT_ARGB);
        Graphics2D graphics = awtDecoratedPlayerImage.createGraphics();
        graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        graphics.drawImage(playerImage, 0, 0, null);
        graphics.drawImage(pieceIconImage, playerImage.getWidth() - (PIECE_ICON_SIZE - extraSpace),
                playerImage.getHeight() - (PIECE_ICON_SIZE - extraSpace), null);
        graphics.dispose();
        
        // converts the image to the SWT format
        ImageData swtDecoratedPlayerImageData = ImageUtils.convertToSwt(awtDecoratedPlayerImage);
        return new Image(display, swtDecoratedPlayerImageData);
    }

    private void fillPlayerImagesPanel(Composite playerImagesPanel) {
        playerImagesPanel.setLayout(new GridLayout(2, true));
        
        whitePlayerLabel = new Label(playerImagesPanel, SWT.HORIZONTAL | SWT.LEFT);
        whitePlayerLabel.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, true, false));
        
        blackPlayerLabel = new Label(playerImagesPanel, SWT.HORIZONTAL | SWT.LEFT);
        blackPlayerLabel.setLayoutData(new GridData(SWT.END, SWT.CENTER, true, false));
    }

    private void fillButtonsPanel(Composite buttonsPanel) {
        RowLayout layout = new RowLayout();
        layout.center = true;
        layout.fill = true;
        layout.pack = false;
        layout.spacing = 5;
        buttonsPanel.setLayout(layout);
        
        Button cancelButton = new Button(buttonsPanel, SWT.PUSH);
        cancelButton.setText("&Cancel");  // TODO externalize this
        cancelButton.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                dispose();
            }
        });
        
        Button okButton = new Button(buttonsPanel, SWT.PUSH);
        okButton.setText("&OK");  // TODO externalize this
        okButton.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                IPlayerConfig whitePlayerConfig = whitePlayerControl.getPlayerConfig();
                IPlayerConfig blackPlayerConfig = blackPlayerControl.getPlayerConfig();
                IClock clock = clockControl.getClock();
                newGameEditorInput = new LocalChessGameEditorInput("", whitePlayerConfig, blackPlayerConfig, clock);
                
                dispose();
            }
        });
    }

}
