/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.editors;

import java.util.Arrays;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.RowData;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;

import com.haundrix.games.model.playing.EngineConfig;
import com.haundrix.games.rcp.managers.IEngineOptionsGuiConfigurer;
import com.haundrix.games.rcp.managers.PluginEngine;
import com.haundrix.swt.FlipLayout;
import com.haundrix.utils.NameComparator;

class PlayerConfigControl extends Composite {
    
    private Label playerLabel;
    private Image humanPlayerImage;
    private Image computerPlayerImage;

    private Combo playerTypeCombo;
    
    private Text playerNameText;
    
    private PluginEngine[] engines;
    private Combo engineCombo;
    private IEngineOptionsGuiConfigurer[] engineOptionsConfigurers;
    
    public PlayerConfigControl(Composite parent, PluginEngine[] engines, Label playerLabel,
            Image humanPlayerImage, Image computerPlayerImage) {
        super(parent, SWT.NONE);
        
        this.engines = engines;
        Arrays.sort(this.engines, NameComparator.getInstance());
        
        this.playerLabel = playerLabel;
        this.humanPlayerImage = humanPlayerImage;
        this.computerPlayerImage = computerPlayerImage;
        
        setLayout(new RowLayout(SWT.VERTICAL));
        
        playerTypeCombo = new Combo(this, SWT.DROP_DOWN | SWT.READ_ONLY);
        playerTypeCombo.add("Human");  // TODO externalize this
        playerTypeCombo.add("Computer");  // TODO externalize this
        playerTypeCombo.select(0);
        
        Composite playerConfigurationPanel = new Composite(this, SWT.NONE);
        final FlipLayout flipLayout = new FlipLayout();
        playerConfigurationPanel.setLayout(flipLayout);
        
        Composite page0 = new Composite(playerConfigurationPanel, SWT.NONE);
        fillPage0(page0);
        
        Composite page1 = new Composite(playerConfigurationPanel, SWT.NONE);
        fillPage1(page1);
        
        updatePlayerImage();
        flipLayout.setSelection(0);
        playerTypeCombo.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                updatePlayerImage();
                flipLayout.setSelection(playerTypeCombo.getSelectionIndex());
                repackAll();
            }
        });
    }
    
    public IPlayerConfig getPlayerConfig() {
        if(playerTypeCombo.getSelectionIndex() == 0) {
            return new HumanPlayerConfig(playerNameText.getText().trim());
        } else {
            int engineIndex = engineCombo.getSelectionIndex();
            PluginEngine selectedEngine = engines[engineIndex];
            EngineConfig engineConfig = engineOptionsConfigurers[engineIndex].getConfig();
            return new ComputerPlayerConfig(selectedEngine.getId(), engineConfig);
        }
    }

    private void updatePlayerImage() {
        Image playerImage = playerTypeCombo.getSelectionIndex() == 0 ? humanPlayerImage : computerPlayerImage;
        playerLabel.setImage(playerImage);
    }

    // Attention: make sure you test on Windows if you change this method!
    private void repackAll() {
        getShell().pack(true);
        layout(true, true);
    }

    private void fillPage0(Composite page0) {
        RowLayout layout = new RowLayout();
        layout.center = true;
        layout.spacing = 10;
        page0.setLayout(layout);
        
        Label label = new Label(page0, SWT.HORIZONTAL | SWT.LEFT);
        label.setText("Name:");  // TODO externalize this
        
        playerNameText = new Text(page0, SWT.SINGLE | SWT.LEFT);
        RowData layoutData = new RowData();
        layoutData.width = 150;
        playerNameText.setLayoutData(layoutData);
        playerNameText.setText("Human");  // TODO externalize this
    }
    
    private void fillPage1(Composite page1) {
        page1.setLayout(new RowLayout(SWT.VERTICAL));
        
        Composite engineSelectionPanel = new Composite(page1, SWT.NONE);
        fillEngineSelectionPanel(engineSelectionPanel);
        
        Composite engineOptionsPanel = new Composite(page1, SWT.NONE);
        final FlipLayout flipLayout = new FlipLayout();
        engineOptionsPanel.setLayout(flipLayout);
        
        engineOptionsConfigurers = new IEngineOptionsGuiConfigurer[engines.length];
        for(int i = 0; i < engines.length; i++) {
            try {
                engineOptionsConfigurers[i] = engines[i].getEngineOptionsGuiConfigurerClass().newInstance();
            } catch(InstantiationException e) {
                throw new RuntimeException(e);
            } catch(IllegalAccessException e) {
                throw new RuntimeException(e);
            }
            
            Composite engineOptionsPage = new Composite(engineOptionsPanel, SWT.NONE);
            engineOptionsConfigurers[i].createWidgets(engineOptionsPage);
        }
        
        engineCombo.select(0);
        flipLayout.setSelection(0);
        engineCombo.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                flipLayout.setSelection(engineCombo.getSelectionIndex());
                repackAll();
            }
        });
    }

    private void fillEngineSelectionPanel(Composite engineSelectionPanel) {
        RowLayout layout = new RowLayout();
        layout.center = true;
        layout.spacing = 10;
        engineSelectionPanel.setLayout(layout);
        
        Label label = new Label(engineSelectionPanel, SWT.HORIZONTAL | SWT.LEFT);
        label.setText("Engine:");  // TODO externalize this
        
        engineCombo = new Combo(engineSelectionPanel, SWT.DROP_DOWN |SWT.READ_ONLY);
        for(PluginEngine engine : engines) {
            engineCombo.add(engine.getName());
        }
    }

}
