/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.editors;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.PartInitException;

import com.haundrix.games.chess.model.ChessGame;
import com.haundrix.games.chess.model.ChessGameResult;
import com.haundrix.games.chess.pgn.PgnImporter;
import com.haundrix.games.rcp.HaundrixLog;
import com.haundrix.rcp.FileEditorInput;
import com.haundrix.rcp.InvalidEditorInput;
import com.haundrix.rcp.NullEditorInput;

public class RecordedChessGameEditor extends ChessGameEditor {

    public static final String ID = "com.haundrix.games.recordedChessGameEditor";

    private static final long MAX_FILE_SIZE_FOR_LOGGING = 1000000L;  // 1 MB
    
    @Override
    public void init(IEditorSite site, IEditorInput input) throws PartInitException {
        boolean validInputType = (input instanceof NullEditorInput) || (input instanceof FileEditorInput);
        if((input == null) || !(validInputType)) {
            throw new PartInitException("Error! Invalid editor input type to RecordedChessGameEditor.");
        }
        
        setSite(site);
        if(input instanceof NullEditorInput) {
            setInput(input);
            game = new ChessGame();
            isDirty = false;
        } else {
            File file = ((FileEditorInput) input).getFile();
            
            if(!input.exists()) {
                String errorMessage = "Game file has been deleted from disk.";
                setInput(new InvalidEditorInput(input.getName(), errorMessage));
                isDirty = false;
                return;
            }
            
            FileReader reader = null;
            try {
                reader = new FileReader(file);
                game = new PgnImporter(reader).importGame();
                setInput(input);
                isDirty = false;
            } catch(Exception exception) {
                String errorMessage = "Error opening PGN file from " + file.getAbsolutePath() + ".\n";
                if(file.length() < MAX_FILE_SIZE_FOR_LOGGING) {
                    try {
                        errorMessage += FileUtils.readFileToString(file);
                    } catch(IOException e) {
                        // no problem, we won't log the file's content then
                    }
                }
                HaundrixLog.logError(errorMessage, exception);
                setInput(new InvalidEditorInput(input.getName(), "Invalid PGN file."));  // TODO externalize this
                isDirty = false;
            } finally {
                IOUtils.closeQuietly(reader);
            }
        }
    }

    public void setGameResult(ChessGameResult result) {
        soundController.playGameEndingSound(result);
        game.setResult(result);
    }

    @Override
    protected void onGameSavedToFile(FileEditorInput fileEditorInput) {
        setInputWithNotify(fileEditorInput);
    }

}
