/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.editors;

import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import org.eclipse.ui.IPartListener;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.PlatformUI;

import com.haundrix.games.chess.controller.IChessSoundTheme;
import com.haundrix.games.chess.model.ChessGame;
import com.haundrix.games.chess.model.ChessGameResultReason;
import com.haundrix.games.model.Game;
import com.haundrix.games.model.GameAdapter;
import com.haundrix.games.model.GameResult;
import com.haundrix.games.model.IGameListener;
import com.haundrix.games.model.playing.IClock;
import com.haundrix.games.rcp.HaundrixPlugin;
import com.haundrix.rcp.ActiveEditorListener;
import com.haundrix.rcp.IEditorAware;
import com.haundrix.utils.IExceptionHandler;
import com.haundrix.utils.ISoundLoopHandle;
import com.haundrix.utils.SoundPlayer;

public class TimeTroubleMonitor implements IEditorAware<LocalChessGameEditor> {
    
    public static final long TIME_TROUBLE_THRESHOLD = 20L * 1000L;  // 20s
    
    private static final long MINIMUM_CLOCK_CHECK_INTERVAL = 50L;  // in milliseconds

    private static TimeTroubleMonitor instance;
    
    private static int activationCount = 0;

    private SoundPlayer soundPlayer;
    private IChessSoundTheme soundTheme;
    private IExceptionHandler exceptionHandler;
    
    private IPartListener partListener = new ActiveEditorListener<LocalChessGameEditor>(this);
    
    private ScheduledExecutorService scheduler;
    
    private LocalChessGameEditor activeEditor;
    
    private IGameListener gameListener = new InternalGameListener();
    
    private ScheduledFuture<?> future = null;
    private ISoundLoopHandle soundLoopHandle = null;
    
    private TimeTroubleMonitor() {
        HaundrixPlugin haundrixPlugin = HaundrixPlugin.getInstance();
        soundPlayer = haundrixPlugin.getSoundPlayer();
        soundTheme = haundrixPlugin.getChessSoundTheme();
        exceptionHandler = haundrixPlugin.getDefaultExceptionHandler();
        
        PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().addPartListener(partListener);
        
        scheduler = Executors.newSingleThreadScheduledExecutor();
        
        System.out.println("Time monitor activated.");
    }
    
    public static void activate() {
        if(TimeTroubleMonitor.activationCount == 0) {
            TimeTroubleMonitor.instance = new TimeTroubleMonitor();
        }
        TimeTroubleMonitor.activationCount++;
    }
    
    public static void deactivate() {
        TimeTroubleMonitor.activationCount--;
        if(TimeTroubleMonitor.activationCount == 0) {
            TimeTroubleMonitor.instance.dispose();
            TimeTroubleMonitor.instance = null;
        }
    }
    
    @Override
    public Class<LocalChessGameEditor> getEditorClass() {
        return LocalChessGameEditor.class;
    }

    @Override
    public LocalChessGameEditor getActiveEditor() {
        return activeEditor;
    }

    @Override
    public synchronized void setActiveEditor(LocalChessGameEditor activeEditor) {
        if(soundLoopHandle != null) {
            soundLoopHandle.cancel();
            soundLoopHandle = null;
        }
        
        if(this.activeEditor != null) {
            cancelTimeTroubleCheck();
            getGame().removeGameListener(gameListener);
            this.activeEditor = null;
        }
        
        if((activeEditor != null) && (activeEditor.getClock() != null) && !activeEditor.getGame().isOver()) {
            this.activeEditor = activeEditor;
            getGame().addGameListener(gameListener);
            scheduleTimeTroubleCheck(0L);
        }
    }
    
    private IClock getClock() {
        return activeEditor.getClock();
    }
    
    private ChessGame getGame() {
        return activeEditor.getGame();
    }
    
    private void scheduleTimeTroubleCheck(long delay) {
        future = scheduler.schedule(new TimeTroubleCheckTask(), delay, TimeUnit.MILLISECONDS);
    }
    
    private void cancelTimeTroubleCheck() {
        if(future != null) {
            future.cancel(true);
            future = null;
        }
    }

    private boolean isTimeTrouble() {
        return getClock().getCurrentPlayerTime() <= TIME_TROUBLE_THRESHOLD;
    }

    private void dispose() {
        setActiveEditor(null);
        
        IWorkbench workbench = PlatformUI.getWorkbench();
        if(!workbench.isClosing()) {
            workbench.getActiveWorkbenchWindow().getActivePage().removePartListener(partListener);
        }
        
        scheduler.shutdownNow();
        
        System.out.println("Time monitor deactivated.");
    }

    private class TimeTroubleCheckTask implements Runnable {

        @Override
        public void run() {
            try {
                synchronized(TimeTroubleMonitor.this) {
                    if(Thread.interrupted()) {
                        return;
                    }
                    
                    if(isTimeTrouble()) {
                        soundLoopHandle = soundPlayer.loop(soundTheme.getTimeTroubleSound());
                        future = null;
                    } else {
                        long delay = Math.max(getClock().getMinimumTimeTill(TIME_TROUBLE_THRESHOLD) / 2L,
                                MINIMUM_CLOCK_CHECK_INTERVAL);
                        scheduleTimeTroubleCheck(delay);
                    }
                }
            } catch(Exception e) {
                exceptionHandler.handle(e);
            }
        }

    }

    /* Note that game callback methods are always called in the GUI thread,
     * so even though there is a background thread monitoring for timeout
     * events, the ended(GameResult) method will be scheduled to be run
     * in the GUI thread.
     * Therefore there is no need to synchronize these methods with
     * setActiveEditor(LocalChessGameEditor) above, although we still have
     * to synchronize them against TimeTroubleCheckTask.run().
     */
    private class InternalGameListener extends GameAdapter {

        @Override
        public void moved(Game game) {
            synchronized(TimeTroubleMonitor.this) {
                if((soundLoopHandle != null) && !isTimeTrouble()) {
                    soundLoopHandle.cancel();
                    soundLoopHandle = null;
                    
                    if(!getGame().isOver()) {
                        scheduleTimeTroubleCheck(0L);
                    }
                }
            }
        }
        
        @Override
        public void ended(GameResult result) {
            synchronized(TimeTroubleMonitor.this) {
                if(soundLoopHandle != null) {
                    soundLoopHandle.cancel();
                    soundLoopHandle = null;
                } else {
                    cancelTimeTroubleCheck();
                }
            }

            if(result.getResultReason() == ChessGameResultReason.TIME_FORFEIT) {
                soundPlayer.play(soundTheme.getTimeoutSound());
            }
        }

    }

}
