/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.managers;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.Platform;
import org.eclipse.swt.graphics.RGB;

import com.haundrix.games.rcp.HaundrixLog;
import com.haundrix.graphics.GraphicsUtils;

public class ColorSchemeManager extends AbstractManager {
    
    private static final String COLOR_SCHEMES_EXTENSION_POINT = "com.haundrix.games.colorSchemes";

    private static ColorSchemeManager instance;
    
    private Map<String, PluginColorScheme> colorSchemes = new HashMap<String, PluginColorScheme>();
    
    // Implementation note: we do not want to cause the contributor plugin to get activated here.
    private ColorSchemeManager() {
        IExtensionPoint extensionPoint = Platform.getExtensionRegistry().getExtensionPoint(COLOR_SCHEMES_EXTENSION_POINT);
        for(IExtension extension : extensionPoint.getExtensions()) {
            String pluginName = extension.getContributor().getName();
            for(IConfigurationElement colorSchemeElement : extension.getConfigurationElements()) {
                String id = colorSchemeElement.getAttribute("id");
                if((id == null) || id.trim().equals("")) {
                    String errorMessage = "Error loading color scheme from " + pluginName +
                        " plugin: color scheme id cannot be empty.";
                    HaundrixLog.logError(errorMessage);
                    continue;
                }
                
                try {
                    if(colorSchemes.containsKey(id)) {
                        throw new Exception("Duplicate color scheme id.");
                    }
                    
                    String name = colorSchemeElement.getAttribute("name");
                    checkNonEmpty(name, "Color scheme name can't be empty.");
                    
                    RGB lightSquareColor = GraphicsUtils.decodeColor(colorSchemeElement.getAttribute("lightSquareColor"));
                    RGB darkSquareColor = GraphicsUtils.decodeColor(colorSchemeElement.getAttribute("darkSquareColor"));
                    RGB borderColor = GraphicsUtils.decodeColor(colorSchemeElement.getAttribute("borderColor"));
                    RGB topBackgroundColor = GraphicsUtils.decodeColor(colorSchemeElement.getAttribute("topBackgroundColor"));
                    RGB bottomBackgroundColor = GraphicsUtils.decodeColor(colorSchemeElement.getAttribute("bottomBackgroundColor"));
                    
                    PluginColorScheme colorScheme = new PluginColorScheme(id, name, lightSquareColor,
                        darkSquareColor, borderColor, topBackgroundColor, bottomBackgroundColor);
                    colorSchemes.put(id, colorScheme);
                } catch(Exception e) {
                    String errorMessage = "Error loading color scheme " + id + " contributed by plugin "
                        + pluginName + " .";
                    HaundrixLog.logError(errorMessage, e);
                }
            }
        }
    }
    
    public static ColorSchemeManager getInstance() {
        if(ColorSchemeManager.instance == null) {
            ColorSchemeManager.instance = new ColorSchemeManager();
        }
        
        return ColorSchemeManager.instance;
    }
    
    public PluginColorScheme[] getAvailableColorSchemes() {
        return colorSchemes.values().toArray(new PluginColorScheme[0]);
    }
    
    public PluginColorScheme getColorScheme(String colorSchemeId) {
        return colorSchemes.get(colorSchemeId);
    }
    
}
