/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.managers;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.spi.RegistryContributor;

import com.haundrix.games.model.playing.Engine;
import com.haundrix.games.rcp.HaundrixLog;
import com.haundrix.games.rcp.HaundrixPlugin;
import com.haundrix.utils.ProcessConfig;

// TODO Improve the engine manager to load engines on demand.
// The current implementation loads all engines at once,
// which is not very scalable as all contributing plugins will get activated.
public class EngineManager extends AbstractManager {
    
    private static final String ENGINES_EXTENSION_POINT = "com.haundrix.games.engines";
    
    private static final String ENGINES_SCRATCH_AREA = "engines";
    
    private static final String PLUGIN_PATH_VARIABLE = "$plugin";
    private static final String SCRATCH_DIR_VARIABLE = "$scratchDir";

    private static final Class<? extends Engine> DEFAULT_ENGINE_CLASS = Engine.class;
    private static final Class<? extends IEngineOptionsGuiConfigurer> DEFAULT_ENGINE_OPTIONS_GUI_CONFIGURER_CLASS =
        EngineOptionsGuiConfigurer.class;

    private Map<String, PluginEngine> engines = new HashMap<String, PluginEngine>();

    private static EngineManager instance;
    
    @SuppressWarnings("unchecked")
    private EngineManager() {
        File pluginStateArea = HaundrixPlugin.getInstance().getStateLocation().toFile();
        File enginesScratchArea = new File(pluginStateArea, ENGINES_SCRATCH_AREA);
        if(!enginesScratchArea.exists()) {
            enginesScratchArea.mkdir();
        }
        
        IExtensionPoint extensionPoint = Platform.getExtensionRegistry().getExtensionPoint(ENGINES_EXTENSION_POINT);
        for(IExtension extension : extensionPoint.getExtensions()) {
            RegistryContributor contributor = (RegistryContributor) extension.getContributor();
            String pluginName = contributor.getActualName();
            
            String errorMessage = "Error loading engine from " + pluginName + " plugin: ";
            File pluginDir;
            try {
                pluginDir = FileLocator.getBundleFile(Platform.getBundle(pluginName));
            } catch(IOException e) {
                HaundrixLog.logError(errorMessage + "an exception occurred when retrieving plugin's directory.", e);
                continue;
            }
            if(!pluginDir.isDirectory()) {
                HaundrixLog.logError(errorMessage + "plugin must be deployed as a directory.");
                continue;
            }
            String pluginPath = pluginDir.toString();
            
            for(IConfigurationElement engineElement : extension.getConfigurationElements()) {
                String id = engineElement.getAttribute("id");
                if((id == null) || id.trim().equals("")) {
                    HaundrixLog.logError(errorMessage + "engine id can't be empty.");
                    continue;
                }
                
                try {
                    if(engines.containsKey(id)) {
                        throw new Exception("Duplicate engine id.");
                    }
                    
                    File scratchDir = new File(enginesScratchArea, id);
                    if(!scratchDir.exists()) {
                        scratchDir.mkdir();
                    }
                    String scratchDirPath = scratchDir.toString();

                    String name = engineElement.getAttribute("name");
                    checkNonEmpty(name, "Engine name can't be empty.");
                    
                    String engineClassAttribute = engineElement.getAttribute("class");
                    Class<? extends Engine> engineClass = (engineClassAttribute != null) && !engineClassAttribute.trim().equals("") ?
                            (Class<? extends Engine>) Class.forName(engineClassAttribute.trim()) :
                            DEFAULT_ENGINE_CLASS;
                    Engine engine = engineClass.newInstance();
                    
                    IConfigurationElement processElement = engineElement.getChildren("process")[0];
                    ProcessConfig processConfig = getProcessConfig(processElement, pluginPath, scratchDirPath);
                    engine.setProcessConfig(processConfig);
                    
                    Class<? extends IEngineOptionsGuiConfigurer> guiConfigurerClass =
                        DEFAULT_ENGINE_OPTIONS_GUI_CONFIGURER_CLASS;
                    IConfigurationElement[] guiConfigurerElements = engineElement.getChildren("optionsGuiConfigurer");
                    if(guiConfigurerElements.length > 0) {
                        String guiConfigurerClassName = guiConfigurerElements[0].getAttribute("class");
                        checkNonEmpty(guiConfigurerClassName, "Engine options GUI configurer class name can't be empty.");
                        guiConfigurerClass = (Class<? extends IEngineOptionsGuiConfigurer>) Class.forName(guiConfigurerClassName);
                    }
                    
                    PluginEngine pluginEngine = new PluginEngine(id, name, engine, guiConfigurerClass);
                    
                    engines.put(id, pluginEngine);
                } catch(Exception e) {
                    String message = "Error loading engine " + id + " contributed by " + pluginName + " plugin.";
                    HaundrixLog.logError(message, e);
                }
            }
        }
    }
    
    public static EngineManager getInstance() {
        if(EngineManager.instance == null) {
            EngineManager.instance = new EngineManager();
        }
        
        return EngineManager.instance;
    }
    
    public PluginEngine getEngine(String engineId) {
        return engines.get(engineId);
    }
    
    public PluginEngine[] getEngines() {
        return engines.values().toArray(new PluginEngine[0]);
    }
    
    private String expandString(String pluginPath, String scratchDirPath, String string) {
        String expandedString = string;
        if(string.contains(PLUGIN_PATH_VARIABLE)) {
            expandedString = string.replace(PLUGIN_PATH_VARIABLE, pluginPath);
        }
        if(expandedString.contains(SCRATCH_DIR_VARIABLE)) {
            expandedString = expandedString.replace(SCRATCH_DIR_VARIABLE, scratchDirPath);
        }
        
        return expandedString;
    }

    private ProcessConfig getProcessConfig(IConfigurationElement processElement, String pluginPath,
            String scratchDirPath) {
        String executable = processElement.getAttribute("executable");
        checkNonEmpty(executable, "Executable can't be empty.");
        executable = expandString(pluginPath, scratchDirPath, executable);
        
        String[] parameters;
        IConfigurationElement[] parametersElements = processElement.getChildren("parameters");
        if(parametersElements.length > 0) {
            IConfigurationElement[] parameterElements = parametersElements[0].getChildren("parameter");
            parameters = new String[parameterElements.length];
            for(int i = 0; i < parameterElements.length; i++) {
                parameters[i] = parameterElements[i].getAttribute("value");
                checkNonEmpty(parameters[i], "Process parameter can't be empty.");
                parameters[i] = expandString(pluginPath, scratchDirPath, parameters[i]);
            }
        } else {
            parameters = new String[0];
        }
        
        Map<String, String> environment = new HashMap<String, String>();
        IConfigurationElement[] environmentElements = processElement.getChildren("environment");
        if(environmentElements.length > 0) {
            IConfigurationElement[] variableElements = environmentElements[0].getChildren("variable");
            for(IConfigurationElement variableElement : variableElements) {
                String variableName = variableElement.getAttribute("name");
                checkNonEmpty(variableName, "Process environment variable name can't be empty.");
                
                String variableValue = variableElement.getAttribute("value");
                variableValue = expandString(pluginPath, scratchDirPath, variableValue);
                
                environment.put(variableName, variableValue);
            }
        }
        
        return new ProcessConfig(executable, parameters, environment);
    }

}
