/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.managers;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.Platform;

import com.haundrix.games.rcp.HaundrixLog;
import com.haundrix.rcp.InvalidPluginConfigurationException;

public class FigurineFontManager extends AbstractManager {

    private static final String FIGURINE_FONTS_EXTENSION_POINT = "com.haundrix.games.figurineFonts";

    private static FigurineFontManager instance;
    
    private Map<String, PluginFigurineFont> figurineFonts = new HashMap<String, PluginFigurineFont>();
    
    // Implementation note: we do not want to cause the contributor plugin to get activated here.
    private FigurineFontManager() {
        IExtensionPoint extensionPoint = Platform.getExtensionRegistry().getExtensionPoint(FIGURINE_FONTS_EXTENSION_POINT);
        for(IExtension extension : extensionPoint.getExtensions()) {
            String pluginName = extension.getContributor().getName();
            for(IConfigurationElement figurineFontElement : extension.getConfigurationElements()) {
                String id = figurineFontElement.getAttribute("id");
                if((id == null) || id.trim().equals("")) {
                    String errorMessage = "Error loading figurine font from " + pluginName + " plugin: " +
                        "figurine font id can't be empty.";
                    HaundrixLog.logError(errorMessage);
                    continue;
                }
                
                try {
                    if(figurineFonts.containsKey(id)) {
                        throw new Exception("Duplicate figurine font id.");
                    }
                    
                    String name = figurineFontElement.getAttribute("name");
                    checkNonEmpty(name, "Figurine font name can't be empty.");
                    
                    String fontId = figurineFontElement.getAttribute("font");
                    checkNonEmpty(fontId, "Referenced font id can't be empty.");
                    PluginFont font = FontManager.getInstance().getFont(fontId);
                    if(font == null) {
                        throw new InvalidPluginConfigurationException("Could not find referenced font: " +
                                fontId + ".");
                    }
                    
                    PluginFigurineFont figurineFont = new PluginFigurineFont(id, name, font);
                    figurineFonts.put(id, figurineFont);
                } catch(Exception e) {
                    String errorMessage = "Error loading figurine font " + id + " contributed by " +
                        pluginName + " plugin.";
                    HaundrixLog.logError(errorMessage, e);
                }
            }
        }
    }
    
    public static FigurineFontManager getInstance() {
        if(FigurineFontManager.instance == null) {
            FigurineFontManager.instance = new FigurineFontManager();
        }
        
        return FigurineFontManager.instance;
    }
    
    public PluginFigurineFont[] getAvailableFigurineFonts() {
        return figurineFonts.values().toArray(new PluginFigurineFont[0]);
    }
    
    public PluginFigurineFont getFigurineFont(String figurineFontId) {
        return figurineFonts.get(figurineFontId);
    }
    
}
