/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.managers;

import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.io.IOUtils;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.Platform;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.PlatformUI;
import org.osgi.framework.Bundle;

import com.haundrix.games.rcp.HaundrixLog;
import com.haundrix.games.rcp.HaundrixPlugin;
import com.haundrix.rcp.InvalidPluginConfigurationException;

public class FontManager extends AbstractManager {
    
    private static final String FONTS_EXTENSION_POINT = "com.haundrix.games.fonts";

    private static final String FONTS_SCRATCH_AREA = "fonts";

    private final Display display = PlatformUI.getWorkbench().getDisplay();

    private static FontManager instance;
    
    private Map<String, PluginFont> fonts = new HashMap<String, PluginFont>();
    
    // Implementation note: we do not want to cause the contributor plugin to get activated here.
    private FontManager() {
        File pluginStateArea = HaundrixPlugin.getInstance().getStateLocation().toFile();
        File fontsScratchDir = new File(pluginStateArea, FONTS_SCRATCH_AREA);
        if(!fontsScratchDir.exists()) {
            fontsScratchDir.mkdir();
        }
        
        IExtensionPoint extensionPoint = Platform.getExtensionRegistry().getExtensionPoint(FONTS_EXTENSION_POINT);
        for(IExtension extension : extensionPoint.getExtensions()) {
            String pluginName = extension.getContributor().getName();
            Bundle bundle = Platform.getBundle(pluginName);
            for(IConfigurationElement fontElement : extension.getConfigurationElements()) {
                String id = fontElement.getAttribute("id");
                if((id == null) || id.trim().equals("")) {
                    String errorMessage = "Error loading font from " + pluginName + " plugin: font id can't be null";
                    HaundrixLog.logError(errorMessage);
                    continue;
                }
                
                try {
                    if(fonts.containsKey(id)) {
                        throw new Exception("Duplicate font id.");
                    }
                    
                    String fontName = fontElement.getAttribute("fontName");
                    checkNonEmpty(fontName, "Font name can't be empty.");
                    
                    String normalFontFileLocation = fontElement.getAttribute("normalFontFileLocation");
                    checkNonEmpty(normalFontFileLocation, "Normal font file location can't be empty.");
                    File normalFontTempFile = new File(fontsScratchDir, id + "_regular");
                    loadFont(bundle, normalFontFileLocation, normalFontTempFile);

                    String boldFontFileLocation = fontElement.getAttribute("boldFontFileLocation");
                    if((boldFontFileLocation != null) && !boldFontFileLocation.trim().equals("")) {
                        File boldFontTempFile = new File(fontsScratchDir, id + "_bold");
                        loadFont(bundle, boldFontFileLocation, boldFontTempFile);
                    }
                    
                    String italicFontFileLocation = fontElement.getAttribute("italicFontFileLocation");
                    if((italicFontFileLocation != null) && !italicFontFileLocation.trim().equals("")) {
                        File italicFontTempFile = new File(fontsScratchDir, id + "_italic");
                        loadFont(bundle, italicFontFileLocation, italicFontTempFile);
                    }
                    
                    String boldItalicFontFileLocation = fontElement.getAttribute("boldItalicFontFileLocation");
                    if((boldItalicFontFileLocation != null) && !boldItalicFontFileLocation.trim().equals("")) {
                        File boldItalicFontTempFile = new File(fontsScratchDir, id + "_boldItalic");
                        loadFont(bundle, boldItalicFontFileLocation, boldItalicFontTempFile);
                    }
                    
                    PluginFont font = new PluginFont(id, fontName);
                    fonts.put(id, font);
                } catch(Exception e) {
                    String errorMessage = "Error loading font " + id + " contributed by " + pluginName + " plugin.";
                    HaundrixLog.logError(errorMessage, e);
                }
            }

        }
    }
    
    public static FontManager getInstance() {
        if(FontManager.instance == null) {
            FontManager.instance = new FontManager();
        }
        
        return FontManager.instance;
    }
    
    public PluginFont getFont(String fontId) {
        return fonts.get(fontId);
    }
    
    private void loadFont(Bundle bundle, String fontFileLocation, File fontTempFile) throws Exception {
        URL fontFileUrl = bundle.getEntry(fontFileLocation);
        if(fontFileUrl == null) {
            throw new InvalidPluginConfigurationException("Could not find the specified font file: " +
                    fontFileLocation + ".");
        }

        InputStream fontFileStream = null;
        FileOutputStream fontTempFileStream = null;
        try {
            fontFileStream = fontFileUrl.openStream();
            fontTempFileStream = new FileOutputStream(fontTempFile);
            
            IOUtils.copy(fontFileStream, fontTempFileStream);
        } finally {
            IOUtils.closeQuietly(fontFileStream);
            IOUtils.closeQuietly(fontTempFileStream);
        }
        
        // Actually the return of Device.loadFont on Linux can't be trusted
        // until the following bug is corrected:
        // https://bugs.eclipse.org/bugs/show_bug.cgi?id=271024
        if(!display.loadFont(fontTempFile.getAbsolutePath())) {
            throw new Exception("Error loading font to the system.");
        }
    }

}
