/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.managers;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.Platform;
import org.eclipse.swt.graphics.RGB;

import com.haundrix.games.rcp.HaundrixLog;
import com.haundrix.graphics.GraphicsUtils;
import com.haundrix.graphics.Highlight;

public class HighlightSchemeManager extends AbstractManager {
    
    private static final String HIGHLIGHT_SCHEMES_EXTENSION_POINT = "com.haundrix.games.highlightSchemes";

    private static HighlightSchemeManager instance;
    
    private Map<String, PluginHighlightScheme> highlightSchemes = new HashMap<String, PluginHighlightScheme>();
    
    // Implementation note: we do not want to cause the contributor plugin to get activated here.
    private HighlightSchemeManager() {
        IExtensionPoint extensionPoint = Platform.getExtensionRegistry().getExtensionPoint(HIGHLIGHT_SCHEMES_EXTENSION_POINT);
        for(IExtension extension : extensionPoint.getExtensions()) {
            String pluginName = extension.getContributor().getName();
            for(IConfigurationElement highlightSchemeElement : extension.getConfigurationElements()) {
                String id = highlightSchemeElement.getAttribute("id");
                if((id == null) || id.trim().equals("")) {
                    String errorMessage = "Error loading highlight scheme from " + pluginName +
                        " plugin: highlight scheme id cannot be empty.";
                    HaundrixLog.logError(errorMessage);
                    continue;
                }
                
                try {
                    if(highlightSchemes.containsKey(id)) {
                        throw new Exception("Duplicate highlight scheme id.");
                    }
                    
                    String name = highlightSchemeElement.getAttribute("name");
                    checkNonEmpty(name, "Highlight scheme name can't be empty.");
                    
                    Highlight squareWeakHighlight = loadHighlight(highlightSchemeElement.getChildren("squareWeakHighlight")[0]);
                    Highlight squareStrongHighlight = loadHighlight(highlightSchemeElement.getChildren("squareStrongHighlight")[0]);
                    Highlight movingPieceWeakHighlight = loadHighlight(highlightSchemeElement.getChildren("movingPieceWeakHighlight")[0]);
                    Highlight movingPieceStrongHighlight = loadHighlight(highlightSchemeElement.getChildren("movingPieceStrongHighlight")[0]);
                    Highlight capturingPieceWeakHighlight = loadHighlight(highlightSchemeElement.getChildren("capturingPieceWeakHighlight")[0]);
                    Highlight capturingPieceStrongHighlight = loadHighlight(highlightSchemeElement.getChildren("capturingPieceStrongHighlight")[0]);
                    
                    PluginHighlightScheme highlightScheme = new PluginHighlightScheme(id, name, squareWeakHighlight,
                        squareStrongHighlight, movingPieceWeakHighlight, movingPieceStrongHighlight,
                        capturingPieceWeakHighlight, capturingPieceStrongHighlight);
                    highlightSchemes.put(id, highlightScheme);
                } catch(Exception e) {
                    String errorMessage = "Error loading highlight scheme " + id + " contributed by plugin "
                        + pluginName + " .";
                    HaundrixLog.logError(errorMessage, e);
                }
            }
        }
    }
    
    public static HighlightSchemeManager getInstance() {
        if(HighlightSchemeManager.instance == null) {
            HighlightSchemeManager.instance = new HighlightSchemeManager();
        }
        
        return HighlightSchemeManager.instance;
    }
    
    public PluginHighlightScheme[] getAvailableHighlightSchemes() {
        return highlightSchemes.values().toArray(new PluginHighlightScheme[0]);
    }

    public PluginHighlightScheme getHighlightScheme(String highlightSchemeId) {
        return highlightSchemes.get(highlightSchemeId);
    }
    
    private Highlight loadHighlight(IConfigurationElement config) {
        RGB color = GraphicsUtils.decodeColor(config.getAttribute("color"));
        float intensity = Float.parseFloat(config.getAttribute("intensity"));
        float glowFracThickness = Float.parseFloat(config.getAttribute("glowFracThickness"));
        
        return new Highlight(color, intensity, glowFracThickness);
    }

}
