/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.managers;

import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.resource.ImageDescriptor;
import org.osgi.framework.Bundle;

import com.haundrix.games.chess.view.ChessPieceType;
import com.haundrix.games.rcp.HaundrixLog;
import com.haundrix.games.rcp.HaundrixPlugin;
import com.haundrix.games.view.IPieceImageDescriptor;
import com.haundrix.games.view.UrlPieceImageDescriptor;
import com.haundrix.rcp.InvalidPluginConfigurationException;

// TODO handle other image formats in addition to PNG
public class PieceSetManager extends AbstractManager {
    
    private static final String PIECE_SETS_EXTENSION_POINT = "com.haundrix.games.pieceSets";

    private static final Map<ChessPieceType, String> PIECE_FILE_NAMES = new HashMap<ChessPieceType, String>();
    static {
        PIECE_FILE_NAMES.put(ChessPieceType.BLACK_ROOK, "blackRook.png");
        PIECE_FILE_NAMES.put(ChessPieceType.BLACK_KNIGHT, "blackKnight.png");
        PIECE_FILE_NAMES.put(ChessPieceType.BLACK_BISHOP, "blackBishop.png");
        PIECE_FILE_NAMES.put(ChessPieceType.BLACK_QUEEN, "blackQueen.png");
        PIECE_FILE_NAMES.put(ChessPieceType.BLACK_KING, "blackKing.png");
        PIECE_FILE_NAMES.put(ChessPieceType.BLACK_PAWN, "blackPawn.png");
        PIECE_FILE_NAMES.put(ChessPieceType.WHITE_ROOK, "whiteRook.png");
        PIECE_FILE_NAMES.put(ChessPieceType.WHITE_KNIGHT, "whiteKnight.png");
        PIECE_FILE_NAMES.put(ChessPieceType.WHITE_BISHOP, "whiteBishop.png");
        PIECE_FILE_NAMES.put(ChessPieceType.WHITE_QUEEN, "whiteQueen.png");
        PIECE_FILE_NAMES.put(ChessPieceType.WHITE_KING, "whiteKing.png");
        PIECE_FILE_NAMES.put(ChessPieceType.WHITE_PAWN, "whitePawn.png");
    }
    
    private static PieceSetManager instance;
    
    private Map<String, PluginPieceSet> pieceSets = new HashMap<String, PluginPieceSet>();
    private Map<String, PieceSetCategory> categories = new HashMap<String, PieceSetCategory>();
    
    // Implementation note: we do not want to cause the contributor plugin to get activated here.
    private PieceSetManager() {
        IExtensionPoint extensionPoint = Platform.getExtensionRegistry().getExtensionPoint(PIECE_SETS_EXTENSION_POINT);
        createCategories(extensionPoint);
        createPieceSets(extensionPoint);
    }

    public static PieceSetManager getInstance() {
        if(PieceSetManager.instance == null) {
            PieceSetManager.instance = new PieceSetManager();
        }
        
        return PieceSetManager.instance;
    }

    public PieceSetCategory[] getAvailableCategories() {
        return categories.values().toArray(new PieceSetCategory[0]);
    }
    
    public PluginPieceSet[] getAvailablePieceSets() {
        return pieceSets.values().toArray(new PluginPieceSet[0]);
    }
    
    public PluginPieceSet[] getAvailablePieceSets(PieceSetCategory category) {
        if((category == null) || !categories.containsValue(category)) {
            String errorMessage = "Error! Invalid piece set category.";
            errorMessage += "\nReturning empty array instead.";
            HaundrixLog.logError(errorMessage, new IllegalArgumentException(errorMessage));
            return new PluginPieceSet[0];
        }
        
        List<PluginPieceSet> selectedPieceSets = new ArrayList<PluginPieceSet>();
        for(PluginPieceSet pieceSet : pieceSets.values()) {
            if(pieceSet.getCategory() == category) {
                selectedPieceSets.add(pieceSet);
            }
        }
        
        return selectedPieceSets.toArray(new PluginPieceSet[0]);
    }
    
    public PluginPieceSet getPieceSet(String pieceSetId) {
        return pieceSets.get(pieceSetId);
    }
    
    private void createCategories(IExtensionPoint pieceSetsExtensionPoint) {
        for(IExtension extension : pieceSetsExtensionPoint.getExtensions()) {
            String pluginName = extension.getContributor().getName();
            for(IConfigurationElement config : extension.getConfigurationElements()) {
                if(config.getName().equals("category")) {
                    String id = config.getAttribute("id");
                    if((id == null) || id.trim().equals("")) {
                        String errorMessage = "Error loading piece set category from " + pluginName +
                            "plugin: category id can't be empty.";
                        HaundrixLog.logError(errorMessage);
                        continue;
                    }
                    
                    try {
                        if(categories.containsKey(id)) {
                            throw new Exception("Duplicate piece set category id.");
                        }
                        
                        String name = config.getAttribute("name");
                        checkNonEmpty(name, "Piece set category name can't be empty.");
                        
                        categories.put(id, new PieceSetCategory(id, name));
                    } catch(Exception e) {
                        String errorMessage = "Error loading piece set category " + id + " contributed by " +
                            pluginName + " plugin.";
                        HaundrixLog.logError(errorMessage);
                    }
                }
            }
        }
    }

    private void createPieceSets(IExtensionPoint pieceSetsExtensionPoint) {
        for(IExtension extension : pieceSetsExtensionPoint.getExtensions()) {
            String pluginName = extension.getContributor().getName();
            Bundle bundle = Platform.getBundle(pluginName);
            for(IConfigurationElement config : extension.getConfigurationElements()) {
                if(config.getName().equals("pieceSet")) {
                    String id = config.getAttribute("id");
                    if((id == null) || id.trim().equals("")) {
                        String errorMessage = "Error loading piece set from " + pluginName +
                            " plugin: piece set id can't be empty.";
                        HaundrixLog.logError(errorMessage);
                        continue;
                    }
                    
                    try {
                        if(pieceSets.containsKey(id)) {
                            throw new Exception("Duplicate piece set id.");
                        }
                        
                        String name = config.getAttribute("name");
                        checkNonEmpty(name, "Piece set name can't be empty.");
                        
                        String categoryId = config.getAttribute("category");
                        PieceSetCategory category = loadCategory(pluginName, id, categoryId);
                        
                        String prefColorSchemeIdString = config.getAttribute("preferredColorScheme");
                        String prefColorSchemeId = loadPrefColorSchemeId(pluginName, id, prefColorSchemeIdString);
                        
                        String prefHighlightSchemeIdString = config.getAttribute("preferredHighlightScheme");
                        String prefHighlightSchemeId = loadPrefHighlightSchemeId(pluginName, id, prefHighlightSchemeIdString);
                        
                        ImageDescriptor enabledMascotImage = null;
                        ImageDescriptor disabledMascotImage = null;
                        String enabledMascotImagePath = config.getAttribute("mascotImageEnabled");
                        if((enabledMascotImagePath != null) && !(enabledMascotImagePath.trim().equals(""))) {
                            enabledMascotImage = HaundrixPlugin.getImageDescriptor(enabledMascotImagePath);
                            String disabledMascotImagePath = config.getAttribute("mascotImageDisabled");
                            if((disabledMascotImagePath != null) && !(disabledMascotImagePath.trim().equals(""))) {
                                disabledMascotImage = HaundrixPlugin.getImageDescriptor(disabledMascotImagePath);
                            } else {
                                disabledMascotImage = enabledMascotImage;
                            }
                        }
                        
                        String imageDir = config.getAttribute("imageDir");
                        checkNonEmpty(imageDir, "Image directory can't be empty.");
                        Map<ChessPieceType, IPieceImageDescriptor> pieceImageDescriptors =
                            loadPieceImages(bundle, imageDir);
                        
                        PluginPieceSet pieceSet = new PluginPieceSet(id, name, category, pieceImageDescriptors,
                            prefColorSchemeId, prefHighlightSchemeId, enabledMascotImage, disabledMascotImage);
                        pieceSets.put(id, pieceSet);
                    } catch(Exception e) {
                        String errorMessage = "Error loading piece set " + id + " contributed by " + pluginName +
                            " plugin.";
                        HaundrixLog.logError(errorMessage, e);
                    }
                }
            }
        }
    }

    private PieceSetCategory loadCategory(String pluginName, String pieceSetId, String categoryId) {
        PieceSetCategory category = null;
        if((categoryId != null) && !categoryId.trim().equals("")) {
            category = categories.get(categoryId);
            if(category == null) {
                String warningMessage = "Warning: invalid piece set category " + categoryId;
                warningMessage += " when loading piece set " + pieceSetId;
                warningMessage += " from " + pluginName + " plugin.";
                warningMessage += "\nSetting category to null instead.";
                HaundrixLog.logWarning(warningMessage);
            }
        }
        
        return category;
    }
    
    private String loadPrefColorSchemeId(String pluginName, String pieceSetId, String prefColorSchemeIdString) {
        String prefColorSchemeId = null;
        if((prefColorSchemeIdString != null) && !prefColorSchemeIdString.equals("")) {
            if(ColorSchemeManager.getInstance().getColorScheme(prefColorSchemeIdString) != null) {
                prefColorSchemeId = prefColorSchemeIdString;
            } else {
                String warningMessage = "Warning: invalid preferred color scheme id " + prefColorSchemeIdString;
                warningMessage += " when loading piece set " + pieceSetId;
                warningMessage += " from " + pluginName + " plugin.";
                warningMessage += "\nSetting preferred color scheme id to null instead.";
                HaundrixLog.logWarning(warningMessage);
            }
        }
        
        return prefColorSchemeId;
    }

    private String loadPrefHighlightSchemeId(String pluginName, String pieceSetId, String prefHighlightSchemeIdString) {
        String prefHighlightSchemeId = null;
        if((prefHighlightSchemeIdString != null) && !prefHighlightSchemeIdString.equals("")) {
            if(HighlightSchemeManager.getInstance().getHighlightScheme(prefHighlightSchemeIdString) != null) {
                prefHighlightSchemeId = prefHighlightSchemeIdString;
            } else {
                String warningMessage = "Warning: invalid preferred highlight scheme id " + prefHighlightSchemeIdString;
                warningMessage += " when loading piece set " + pieceSetId;
                warningMessage += " from " + pluginName + " plugin.";
                warningMessage += "\nSetting preferred highlight scheme id to null instead.";
                HaundrixLog.logWarning(warningMessage);
            }
        }
        
        return prefHighlightSchemeId;
    }

    private Map<ChessPieceType, IPieceImageDescriptor> loadPieceImages(Bundle bundle, String imageDir) {
        Map<ChessPieceType, IPieceImageDescriptor> pieceImageDescriptors =
            new HashMap<ChessPieceType, IPieceImageDescriptor>();
        for(ChessPieceType piece : PIECE_FILE_NAMES.keySet()) {
            String imageFile = imageDir + PIECE_FILE_NAMES.get(piece);
            URL imageUrl = bundle.getEntry(imageFile);
            if(imageUrl == null) {
                throw new InvalidPluginConfigurationException("Could not find file " + imageFile + " .");
            }
            
            IPieceImageDescriptor pieceImageDescriptor = new UrlPieceImageDescriptor(imageUrl);
            pieceImageDescriptors.put(piece, pieceImageDescriptor);
        }
        
        return pieceImageDescriptors;
    }

}
