/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.preferences;

import java.util.Arrays;

import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.preference.PreferencePage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPreferencePage;

import com.haundrix.games.chess.controller.ChessGameController;
import com.haundrix.games.chess.controller.ChessGameSoundController;
import com.haundrix.games.chess.model.ChessGame;
import com.haundrix.games.chess.view.ChessPieceType;
import com.haundrix.games.rcp.HaundrixPlugin;
import com.haundrix.games.rcp.managers.ColorSchemeManager;
import com.haundrix.games.rcp.managers.HighlightSchemeManager;
import com.haundrix.games.rcp.managers.PieceSetManager;
import com.haundrix.games.rcp.managers.PluginColorScheme;
import com.haundrix.games.rcp.managers.PluginHighlightScheme;
import com.haundrix.games.rcp.managers.PluginPieceSet;
import com.haundrix.games.view.PieceSet;
import com.haundrix.games.view.RectangularBoardControl;
import com.haundrix.utils.NameComparator;

public class ChessBoardPreferencePage extends PreferencePage implements IWorkbenchPreferencePage {
    
    private PluginPieceSet[] pieceSets;
    private PluginColorScheme[] colorSchemes;
    private PluginHighlightScheme[] highlightSchemes;

    private Combo pieceSetCombo;
    private Button usePrefColorsButton;
    private Combo colorSchemeCombo;
    private Combo highlightSchemeCombo;
    private ChessGameController gameController;

    public ChessBoardPreferencePage() {
    }
    
    @Override
    public void init(IWorkbench workbench) {
    }

    @Override
    protected IPreferenceStore doGetPreferenceStore() {
        return HaundrixPlugin.getInstance().getPreferenceStore();
    }

    @Override
    protected Control createContents(Composite parent) {
        Composite pageBody = new Composite(parent, SWT.NONE);
        pageBody.setLayout(new GridLayout(1, false));
        
        Composite pieceSetPanel = new Composite(pageBody, SWT.NONE);
        pieceSetPanel.setLayoutData(new GridData());
        fillPieceSetPanel(pieceSetPanel);
        
        Group colorsPanel = new Group(pageBody, SWT.NONE);
        colorsPanel.setText("Colors");  // TODO externalize this
        colorsPanel.setLayoutData(new GridData());
        fillColorsPanel(colorsPanel);
        
        Group previewPanel = new Group(pageBody, SWT.NONE);
        previewPanel.setText("Preview");  // TODO externalize this
        previewPanel.setLayoutData(new GridData(GridData.FILL, GridData.FILL, true, true));
        fillPreviewPanel(previewPanel);
        
        loadPreferences(false);
        
        return pageBody;
    }

    @Override
    protected void performDefaults() {
        loadPreferences(true);
        
        super.performDefaults();
    }

    @Override
    public boolean performOk() {
        IPreferenceStore preferences = getPreferenceStore();
        
        preferences.setValue(HaundrixPreferences.CHESS_BOARD_PIECE_SET, getSelectedPieceSet().getId());
        preferences.setValue(HaundrixPreferences.CHESS_BOARD_USE_PIECE_SET_PREFERRED_COLORS, getUsePrefColors());
        preferences.setValue(HaundrixPreferences.CHESS_BOARD_COLOR_SCHEME, getSelectedColorScheme().getId());
        preferences.setValue(HaundrixPreferences.CHESS_BOARD_HIGHLIGHT_SCHEME, getSelectedHighlightScheme().getId());
        
        return true;
    }

    private RectangularBoardControl<ChessPieceType> getBoardControl() {
        return gameController.getBoardControl();
    }
    
    private PluginPieceSet getSelectedPieceSet() {
        return pieceSets[pieceSetCombo.getSelectionIndex()];
    }

    private boolean getUsePrefColors() {
        return usePrefColorsButton.getSelection();
    }

    private PluginColorScheme getSelectedColorScheme() {
        return colorSchemes[colorSchemeCombo.getSelectionIndex()];
    }

    private PluginHighlightScheme getSelectedHighlightScheme() {
        return highlightSchemes[highlightSchemeCombo.getSelectionIndex()];
    }

    private void updateColors(String colorSchemeId, String highlightSchemeId) {
        for(int i = 0; i < colorSchemes.length; i++) {
            if(colorSchemes[i].getId().equals(colorSchemeId)) {
                colorSchemeCombo.select(i);
                getBoardControl().setColorScheme(colorSchemes[i]);
                break;
            }
        }
        for(int i = 0; i < highlightSchemes.length; i++) {
            if(highlightSchemes[i].getId().equals(highlightSchemeId)) {
                highlightSchemeCombo.select(i);
                gameController.setHighlightScheme(highlightSchemes[i]);
                break;
            }
        }
    }
    
    private void updatePreferredColors() {
        if(getUsePrefColors()) {
            PluginPieceSet selectedPieceSet = getSelectedPieceSet();
            updateColors(selectedPieceSet.getPreferredColorSchemeId(),
                selectedPieceSet.getPreferredHighlightSchemeId());
        }
    }

    private void fillPieceSetPanel(Composite pieceSetPanel) {
        pieceSetPanel.setLayout(new GridLayout(2, false));
        
        Label label = new Label(pieceSetPanel, SWT.NONE);
        label.setText("Piece set:");  // TODO externalize this
        label.setLayoutData(new GridData());
        
        pieceSetCombo = new Combo(pieceSetPanel, SWT.DROP_DOWN | SWT.READ_ONLY);
        pieceSetCombo.setLayoutData(new GridData());
        pieceSets = PieceSetManager.getInstance().getAvailablePieceSets();
        Arrays.sort(pieceSets, NameComparator.getInstance());
        for(PluginPieceSet pieceSet : pieceSets) {
            pieceSetCombo.add(pieceSet.getName());
        }
        pieceSetCombo.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                getBoardControl().setPieceSet(getSelectedPieceSet());
                updatePreferredColors();
            }
        });
    }

    private void fillColorsPanel(Composite colorsPanel) {
        colorsPanel.setLayout(new GridLayout(2, false));
        
        usePrefColorsButton = new Button(colorsPanel, SWT.CHECK);
        usePrefColorsButton.setText("Use piece set preferred colors.");  // TODO externalize this
        GridData buttonLayoutData = new GridData();
        buttonLayoutData.horizontalSpan = 2;
        usePrefColorsButton.setLayoutData(buttonLayoutData);
        usePrefColorsButton.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                boolean usePrefColors = getUsePrefColors();
                colorSchemeCombo.setEnabled(!usePrefColors);
                highlightSchemeCombo.setEnabled(!usePrefColors);
                updatePreferredColors();
            }
        });
        
        Label colorSchemeLabel = new Label(colorsPanel, SWT.NONE);
        colorSchemeLabel.setText("Color scheme:");  // TODO externalize this
        colorSchemeLabel.setLayoutData(new GridData());
        
        colorSchemeCombo = new Combo(colorsPanel, SWT.DROP_DOWN | SWT.READ_ONLY);
        colorSchemeCombo.setLayoutData(new GridData());
        colorSchemes = ColorSchemeManager.getInstance().getAvailableColorSchemes();
        Arrays.sort(colorSchemes, NameComparator.getInstance());
        for(PluginColorScheme colorScheme : colorSchemes) {
            colorSchemeCombo.add(colorScheme.getName());
        }
        colorSchemeCombo.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                getBoardControl().setColorScheme(getSelectedColorScheme());
            }
        });
        
        Label highlightSchemeLabel = new Label(colorsPanel, SWT.NONE);
        highlightSchemeLabel.setText("Highlight scheme:");  // TODO externalize this
        highlightSchemeLabel.setLayoutData(new GridData());
        
        highlightSchemeCombo = new Combo(colorsPanel, SWT.DROP_DOWN | SWT.READ_ONLY);
        highlightSchemeCombo.setLayoutData(new GridData());
        highlightSchemes = HighlightSchemeManager.getInstance().getAvailableHighlightSchemes();
        Arrays.sort(highlightSchemes, NameComparator.getInstance());
        for(PluginHighlightScheme highlightScheme : highlightSchemes) {
            highlightSchemeCombo.add(highlightScheme.getName());
        }
        highlightSchemeCombo.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                gameController.setHighlightScheme(getSelectedHighlightScheme());
            }
        });
    }

    private void fillPreviewPanel(Group previewPanel) {
        previewPanel.setLayout(new FillLayout());
        
        String pieceSetId = getPreferenceStore().getString(HaundrixPreferences.CHESS_BOARD_PIECE_SET);
        PieceSet<ChessPieceType> pieceSet = PieceSetManager.getInstance().getPieceSet(pieceSetId);
        gameController = new ChessGameController(new ChessGame(), previewPanel, pieceSet);
        
        HaundrixPlugin haundrixPlugin = HaundrixPlugin.getInstance();
        new ChessGameSoundController(gameController, haundrixPlugin.getChessSoundTheme(), haundrixPlugin.getSoundPlayer());
    }

    private void loadPreferences(boolean useDefaults) {
        IPreferenceStore preferences = getPreferenceStore();
        
        String pieceSetPreference = HaundrixPreferences.CHESS_BOARD_PIECE_SET;
        String pieceSetId = useDefaults ? preferences.getDefaultString(pieceSetPreference) :
            preferences.getString(pieceSetPreference);
        for(int i = 0; i < pieceSets.length; i++) {
            if(pieceSets[i].getId().equals(pieceSetId)) {
                pieceSetCombo.select(i);
                getBoardControl().setPieceSet(pieceSets[i]);
                break;
            }
        }
        
        String usePrefColorsPreference = HaundrixPreferences.CHESS_BOARD_USE_PIECE_SET_PREFERRED_COLORS;
        boolean usePrefColors = useDefaults ? preferences.getDefaultBoolean(usePrefColorsPreference) :
            preferences.getBoolean(usePrefColorsPreference);
        usePrefColorsButton.setSelection(usePrefColors);
        
        String colorSchemeId;
        String highlightSchemeId;
        if(usePrefColors) {
            PluginPieceSet pieceSet = getSelectedPieceSet();
            colorSchemeId = pieceSet.getPreferredColorSchemeId();
            highlightSchemeId = pieceSet.getPreferredHighlightSchemeId();
            colorSchemeCombo.setEnabled(false);
            highlightSchemeCombo.setEnabled(false);
        } else {
            String colorSchemePreference = HaundrixPreferences.CHESS_BOARD_COLOR_SCHEME;
            colorSchemeId = useDefaults ? preferences.getDefaultString(colorSchemePreference) :
                preferences.getString(colorSchemePreference);
            
            String highlightSchemePreference = HaundrixPreferences.CHESS_BOARD_HIGHLIGHT_SCHEME;
            highlightSchemeId = useDefaults ? preferences.getDefaultString(highlightSchemePreference) :
                preferences.getString(highlightSchemePreference);
        }
        updateColors(colorSchemeId, highlightSchemeId);
    }

}
