/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.views;

import org.apache.commons.lang.ArrayUtils;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.ActionContributionItem;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.actions.CompoundContributionItem;

import com.haundrix.games.rcp.HaundrixPlugin;
import com.haundrix.games.rcp.preferences.HaundrixPreferences;

public class DynamicFormatFigurineFontSizeContributionItem extends CompoundContributionItem {
    
    private static final int[] SIZES = {11, 12, 13, 14, 15, 16, 17, 18, 19, 20};
    private static final int INITIAL_CUSTOM_SIZE = 22;
    
    private FormatFontSizeAction[] actions;
    private IContributionItem[] contributionItems;
    private IAction previousSelectedAction;
    private CustomFormatFontSizeAction customSizeAction;
    
    private IPropertyChangeListener preferencesListener = new IPropertyChangeListener() {
        @Override
        public void propertyChange(PropertyChangeEvent event) {
            if(event.getProperty().equals(HaundrixPreferences.GAME_HISTORY_MOVES_FONT_SIZE)) {
                updateSelectedAction();
            }
        }
    };

    public DynamicFormatFigurineFontSizeContributionItem() {
    }

    public DynamicFormatFigurineFontSizeContributionItem(String id) {
        super(id);
    }

    @Override
    public void dispose() {
        IPreferenceStore preferences = HaundrixPlugin.getInstance().getPreferenceStore();
        preferences.removePropertyChangeListener(preferencesListener);
        
        super.dispose();
    }

    @Override
    protected IContributionItem[] getContributionItems() {
        if(contributionItems == null) {
            createContributionItems();
        }
        
        return contributionItems;
    }

    private void createContributionItems() {
        actions = new FormatFontSizeAction[SIZES.length];
        contributionItems = new IContributionItem[SIZES.length + 1];
        
        for(int i = 0; i < SIZES.length; i++) {
            actions[i] = new FormatFontSizeAction(SIZES[i]);
            contributionItems[i] = new ActionContributionItem(actions[i]);
        }
        
        customSizeAction = new CustomFormatFontSizeAction(INITIAL_CUSTOM_SIZE);
        contributionItems[SIZES.length] = new ActionContributionItem(customSizeAction);
        
        updateSelectedAction();
        
        IPreferenceStore preferences = HaundrixPlugin.getInstance().getPreferenceStore();
        preferences.addPropertyChangeListener(preferencesListener);
    }
    
    private void updateSelectedAction() {
        if(previousSelectedAction != null) {
            previousSelectedAction.setChecked(false);
        }
        
        IPreferenceStore preferences = HaundrixPlugin.getInstance().getPreferenceStore();
        int fontSize = preferences.getInt(HaundrixPreferences.GAME_HISTORY_MOVES_FONT_SIZE);
        
        int index = ArrayUtils.indexOf(SIZES, fontSize);
        if(index >= 0) {
            customSizeAction.setChecked(false);
            actions[index].setChecked(true);
            previousSelectedAction = actions[index];
        } else {
            customSizeAction.setFontSize(fontSize);
            customSizeAction.setChecked(true);
            previousSelectedAction = customSizeAction;
        }

    }

    private class FormatFontSizeAction extends Action {

        protected int fontSize;
        
        public FormatFontSizeAction(int fontSize) {
            super(Integer.toString(fontSize), IAction.AS_RADIO_BUTTON);
            
            this.fontSize = fontSize;
        }

        @Override
        public void run() {
            if(isChecked()) {
                IPreferenceStore preferences = HaundrixPlugin.getInstance().getPreferenceStore();
                preferences.setValue(HaundrixPreferences.GAME_HISTORY_MOVES_FONT_SIZE, fontSize);
                HaundrixPlugin.getInstance().savePreferences();
            }
        }
        
    }
    
    private class CustomFormatFontSizeAction extends FormatFontSizeAction {

        public CustomFormatFontSizeAction(int fontSize) {
            super(fontSize);
            
            updateLabel();
        }

        public void setFontSize(int fontSize) {
            this.fontSize = fontSize;
            updateLabel();
        }

        @Override
        public void run() {
            if(isChecked()) {
                Shell shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
                FontSizeDialog fontSizeDialog = new FontSizeDialog(shell, fontSize);
                fontSizeDialog.setText("New Size");  // TODO externalize this
                Integer newFontSize = fontSizeDialog.getFontSize();
                if(newFontSize != null) {
                    setFontSize(newFontSize);
                    super.run();
                } else {
                    updateSelectedAction();
                }
            }
        }
        
        private void updateLabel() {
            setText("Other (" + fontSize + ") ...");
        }
        
    }
    
}
