/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.views;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import org.apache.commons.lang.StringUtils;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IPartListener;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.part.ViewPart;

import com.haundrix.games.chess.model.playing.ComputerPlayer;
import com.haundrix.games.model.Piece.PieceColor;
import com.haundrix.games.model.playing.EngineLog;
import com.haundrix.games.model.playing.IPlayer;
import com.haundrix.games.model.playing.EngineLog.ILogListener;
import com.haundrix.games.rcp.editors.LocalChessGameEditor;
import com.haundrix.games.rcp.managers.FontManager;
import com.haundrix.rcp.ActiveEditorListener;
import com.haundrix.rcp.IEditorAware;

public class EngineView extends ViewPart implements IEditorAware<LocalChessGameEditor> {
    
    public static final String ID = "com.haundrix.games.engineView";
    
    public static final String FONT_ID = "com.haundrix.games.fonts.dejaVuSansMono";
    private static final int FONT_SIZE = 9;
    
    private Text text;

    private PieceColor engineColor;
    private EngineLog engineLog;
    private InternalLogListener logListener;
    
    private LocalChessGameEditor localGameEditor;
    private IPartListener partListener = new ActiveEditorListener<LocalChessGameEditor>(this);
    
    private SelectEngineColorControl engineColorControl;

    public PieceColor getEngineColor() {
        return engineColor;
    }

    public void setEngineColor(PieceColor engineColor) {
        if(engineColor != this.engineColor) {
            this.engineColor = engineColor;
            
            updateEngineLog();
        }
    }

    @Override
    public void createPartControl(Composite parent) {
        parent.setLayout(new FillLayout());
        
        text = new Text(parent, SWT.MULTI | SWT.READ_ONLY | SWT.V_SCROLL);
        com.haundrix.games.view.Font font = FontManager.getInstance().getFont(FONT_ID);
        final Font swtFont = new Font(parent.getDisplay(), font.getFontName(), FONT_SIZE, SWT.NORMAL);
        text.setFont(swtFont);
        text.addDisposeListener(new DisposeListener() {
            @Override
            public void widgetDisposed(DisposeEvent e) {
                swtFont.dispose();
            }
        });
        
        IWorkbenchPage page = getViewSite().getPage();
        page.addPartListener(partListener);
        IEditorPart activeEditor = page.getActiveEditor();
        engineColor = PieceColor.WHITE;
        if(activeEditor instanceof LocalChessGameEditor) {
            setActiveEditor((LocalChessGameEditor) activeEditor);
        }
    }

    @Override
    public void setFocus() {
    }

    @Override
    public void dispose() {
        if(this.engineLog != null) {
            this.engineLog.removeListener(logListener);
            logListener.dispose();
        }
        
        getViewSite().getPage().removePartListener(partListener);
        
        super.dispose();
    }

    @Override
    public Class<LocalChessGameEditor> getEditorClass() {
        return LocalChessGameEditor.class;
    }

    @Override
    public LocalChessGameEditor getActiveEditor() {
        return localGameEditor;
    }

    @Override
    public void setActiveEditor(LocalChessGameEditor localGameEditor) {
        this.localGameEditor = localGameEditor;
        
        if(this.localGameEditor == null) {
            engineColor = PieceColor.WHITE;
        } else if(this.localGameEditor.getWhitePlayer() instanceof ComputerPlayer) {
            engineColor = PieceColor.WHITE;
        } else if(this.localGameEditor.getBlackPlayer() instanceof ComputerPlayer) {
            engineColor = PieceColor.BLACK;
        } else {
            engineColor = PieceColor.WHITE;
        }

        if(engineColorControl != null) {
            engineColorControl.updateEngineColor();
        }

        updateEngineLog();
    }
    
    void setEngineColorControl(SelectEngineColorControl engineColorControl) {
        this.engineColorControl = engineColorControl;
    }

    private void updateEngineLog() {
        if(engineLog != null) {
            engineLog.removeListener(logListener);
            logListener.dispose();
            logListener = null;
        }
        engineLog = null;
        text.setText("");
        
        if(localGameEditor != null) {
            IPlayer player = engineColor == PieceColor.WHITE ? localGameEditor.getWhitePlayer() :
                localGameEditor.getBlackPlayer();
            if(player instanceof ComputerPlayer) {
                engineLog = ((ComputerPlayer) player).getLog();

                logListener = new InternalLogListener();
                text.setRedraw(false);  // this is necessary to avoid the undesirable scrolling effect on Windows
                for(String message : engineLog.addListenerAndGetMessages(logListener)) {
                    text.append(message + text.getLineDelimiter());
                }
                text.setRedraw(true);  // this is necessary to avoid the undesirable scrolling effect on Windows
            }
        }
    }

    /*
     * In order to prevent an undesirable scrolling effect form happening,
     * log messages are not sent to the GUI as soon as they are received.
     * Instead, when a new log message arrives a bulk update is scheduled
     * to occur after a short delay, when all pending messages are sent out
     * to the GUI at once.
     */
    private class InternalLogListener implements ILogListener {
        
        private static final long UPDATE_MESSAGES_DELAY = 250;  // in milliseconds

        private final Display display = getViewSite().getShell().getDisplay();
        private final ScheduledExecutorService scheduler = Executors.newSingleThreadScheduledExecutor();
        
        private List<String> pendingMessages = new ArrayList<String>();
        private volatile ScheduledFuture<?> updatePendingMessagesFuture = null;
        
        @Override
        public synchronized void messageLogged(final String message) {
            if(!scheduler.isShutdown()) {
                pendingMessages.add(message);
                
                if(updatePendingMessagesFuture == null) {
                    updatePendingMessagesFuture = scheduler.schedule(new Runnable() {
                        @Override
                        public void run() {
                            display.asyncExec(new Runnable() {
                                @Override
                                public void run() {
                                    updatePendingMessages();
                                }
                            });
                        }
                    }, UPDATE_MESSAGES_DELAY, TimeUnit.MILLISECONDS);
                }
            }
        }
        
        public synchronized void dispose() {
            scheduler.shutdownNow();
        }
        
        // this method is run in the GUI thread
        private synchronized void updatePendingMessages() {
            if(!scheduler.isShutdown()) {
                text.setRedraw(false);
                for(String message : pendingMessages) {
                    String lineDelimiter = text.getLineDelimiter();
                    if(text.getLineCount() >= EngineLog.MAX_SIZE * 2) {
                        // trims the first half of the log in order to avoid memory problems
                        String currentText = text.getText();
                        int index = StringUtils.ordinalIndexOf(currentText, lineDelimiter, EngineLog.MAX_SIZE);
                        String newText = currentText.substring(index + lineDelimiter.length(),
                                currentText.length());
                        text.setText(newText);
                    }
                    text.append(message + lineDelimiter);
                }
                pendingMessages.clear();
                text.setRedraw(true);
            }
            
            updatePendingMessagesFuture = null;
        }

    }
    
}
