/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.views;

import java.util.List;

import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.part.ViewPart;

import com.haundrix.games.chess.model.ChessAlgebraicMoveNotation;
import com.haundrix.games.chess.model.ChessGame;
import com.haundrix.games.controller.GameNavigationKeyListener;
import com.haundrix.games.model.Game;
import com.haundrix.games.model.GameAdapter;
import com.haundrix.games.model.GameOutcome;
import com.haundrix.games.model.GameResult;
import com.haundrix.games.model.IGameListener;
import com.haundrix.games.model.IMoveNotation;
import com.haundrix.games.model.Piece.PieceColor;
import com.haundrix.games.rcp.HaundrixPlugin;
import com.haundrix.games.rcp.editors.ChessGameEditor;
import com.haundrix.games.rcp.managers.FigurineFontManager;
import com.haundrix.games.rcp.preferences.HaundrixPreferences;
import com.haundrix.rcp.ActiveEditorListener;
import com.haundrix.rcp.IEditorAware;
import com.haundrix.swt.ITableDecoratorListener;
import com.haundrix.swt.TableCell;
import com.haundrix.swt.TableColumnDecorator;
import com.haundrix.swt.TableDecorator;

public class GameHistoryView extends ViewPart implements IEditorAware<ChessGameEditor> {
    
    public static final String ID = "com.haundrix.games.gameHistoryView";
    
    private static final String UNDEFINED_LABEL = "*";
    private static final String WHITE_WINS_LABEL = "1 - 0";
    private static final String BLACK_WINS_LABEL = "0 - 1";
    private static final String DRAW_LABEL = "½ - ½";

    private final IMoveNotation notation = new ChessAlgebraicMoveNotation();

    private ChessGameEditor activeEditor;
    
    private ActiveEditorListener<ChessGameEditor> activeEditorListener = new ActiveEditorListener<ChessGameEditor>(this);
    private GameNavigationKeyListener gameNavigationKeyListener = new GameNavigationKeyListener();
    private IGameListener gameListener = new InternalGameListener();
    private IPropertyChangeListener preferencesListener = new InternalPreferencesListener();

    private NavigationBar navigationBar;
    private TableDecorator tableDecorator;
    private Label resultLabel;

    private Font tableHeaderFont;
    private Font movesFont;
    private Font selectedMoveFont;

    private int maxDigits;

    private boolean selfMoving = false;

    @Override
    public void createPartControl(Composite parent) {
        GridLayout layout = new GridLayout(1, false);
        layout.marginWidth = layout.marginHeight = 0;
        layout.horizontalSpacing = layout.verticalSpacing = 0;
        parent.setLayout(layout);
        
        navigationBar = new NavigationBar(parent);
        navigationBar.setLayoutData(new GridData(SWT.CENTER, SWT.FILL, true, false));
        
        tableDecorator = new TableDecorator(parent, true);
        tableDecorator.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
        createMoveTable();
        
        resultLabel = new Label(parent, SWT.HORIZONTAL | SWT.CENTER);
        resultLabel.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
        resultLabel.setBackground(parent.getDisplay().getSystemColor(SWT.COLOR_WHITE));
        
        updateTableHeaderFont();
        updateMovesAndResultFont();

        IWorkbenchPage page = getViewSite().getPage();
        page.addPartListener(activeEditorListener);
        IEditorPart activeEditor = page.getActiveEditor();
        if(activeEditorListener.isValidEditor(activeEditor)) {
            setActiveEditor((ChessGameEditor) activeEditor);
        }
        
        HaundrixPlugin.getInstance().getPreferenceStore().addPropertyChangeListener(preferencesListener);
    }

    @Override
    public void setFocus() {
    }

    @Override
    public void dispose() {
        if(activeEditor != null) {
            getGame().removeGameListener(gameListener);
        }
        getViewSite().getPage().removePartListener(activeEditorListener);
        HaundrixPlugin.getInstance().getPreferenceStore().removePropertyChangeListener(preferencesListener);

        tableHeaderFont.dispose();
        movesFont.dispose();
        selectedMoveFont.dispose();
        
        super.dispose();
    }

    @Override
    public Class<ChessGameEditor> getEditorClass() {
        return ChessGameEditor.class;
    }

    @Override
    public ChessGameEditor getActiveEditor() {
        return activeEditor;
    }

    @Override
    public void setActiveEditor(ChessGameEditor activeEditor) {
        if(this.activeEditor != activeEditor) {
            if(this.activeEditor != null) {
                getGame().removeGameListener(gameListener);
            }
            gameNavigationKeyListener.setGame(null);
            navigationBar.setGame(null);
            
            this.activeEditor = activeEditor;
            
            // loads the move history from the game
            tableDecorator.removeAll();
            resultLabel.setText(UNDEFINED_LABEL);
            if(this.activeEditor != null) {
                gameNavigationKeyListener.setGame(getGame());
                navigationBar.setGame(getGame());
                getGame().addGameListener(gameListener);
                
                tableDecorator.getTable().setRedraw(false);
                List<String> moveDescriptions = getGame().getMoveDescriptions(notation);
                if(getGame().getInitialColorToMove() == PieceColor.BLACK) {
                    addMove("...", 0);
                }
                for(int i = 0; i < moveDescriptions.size(); i++) {
                    addMove(moveDescriptions.get(i), i + 1);
                }
                tableDecorator.getTable().setRedraw(true);
                tableDecorator.pack();
              
                // The line below is to fix a weird bug on Linux:
                // When opening the first game and the application was started without
                // any open game from a previous session, the move table widget is
                // created totally messed up: with a horizontal scroll bar (columns
                // with incorrect widths) and not scrolled down to make the current
                // selection visible.
                // For some mysterious reason, packing the table twice fixes the problem.
                tableDecorator.pack();
                
                updateResult();

                maxDigits = getMaxDigits();

                selectCurrentMove();
            }
        }
    }
    
    private ChessGame getGame() {
        return activeEditor.getGame();
    }

    private void createMoveTable() {
        final Table table = tableDecorator.getTable();
        table.setHeaderVisible(true);
        
        tableDecorator.addListener(new ITableDecoratorListener() {
            @Override
            public void cellSelectionChanged(TableCell previousSelectedCell, TableCell currentSelectedCell) {
                if(getGame().isLocked()) {
                    // can't navigate the game, rolls back the selection
                    tableDecorator.setSelectedCell(previousSelectedCell);
                    return;
                }
                
                // checks whether the user clicked on a valid cell
                int currentSelectedRow = currentSelectedCell.getRow();
                int currentSelectedColumn = currentSelectedCell.getColumn();
                boolean isMoveNumberColumn = currentSelectedColumn == 0;
                int lastMoveRow = getGame().getResult() != null ? table.getItemCount() - 2 :
                    table.getItemCount() - 1;
                boolean isCellWithoutBlackMove = (currentSelectedRow == lastMoveRow) &&
                    (currentSelectedColumn == 2) && (table.getItem(currentSelectedRow).getText(2).equals(""));
                if(isMoveNumberColumn || isCellWithoutBlackMove) {
                    currentSelectedCell = new TableCell(currentSelectedRow, 1);
                    tableDecorator.setSelectedCell(currentSelectedCell);
                }

                selfMoving = true;
                getGame().goTo(getPosition(currentSelectedCell));
                selfMoving = false;
                
                updateTableSelection(previousSelectedCell, currentSelectedCell);
            }
        });
        tableDecorator.addKeyListener(gameNavigationKeyListener);
        
        TableColumnDecorator column1Decorator = new TableColumnDecorator(tableDecorator, SWT.RIGHT);
        column1Decorator.setAutoExpandeable(false);
        TableColumn column1 = column1Decorator.getTableColumn();
        column1.setText("#");  // TODO externalize this
        column1.setResizable(false);
        
        TableColumnDecorator column2Decorator = new TableColumnDecorator(tableDecorator, SWT.CENTER);
        column2Decorator.setAutoExpandeable(true);
        TableColumn column2 = column2Decorator.getTableColumn();
        column2.setText("White");  // TODO externalize this
        column2.setResizable(false);
        
        TableColumnDecorator column3Decorator = new TableColumnDecorator(tableDecorator, SWT.CENTER);
        column3Decorator.setAutoExpandeable(true);
        TableColumn column3 = column3Decorator.getTableColumn();
        column3.setText("Black");  // TODO externalize this
        column3.setResizable(false);
    }
    
    private int getPosition(TableCell currentSelectedCell) {
        int cellCount = currentSelectedCell.getRow() * 2 + 1;
        if(currentSelectedCell.getColumn() == 2) {
            cellCount++;
        }
        
        int position = cellCount;
        if(getGame().getInitialColorToMove() == PieceColor.BLACK) {
            position--;
        }
        
        return position;
    }

    private TableCell getTableCell(int gamePosition) {
        int cellCount = gamePosition;
        if(getGame().getInitialColorToMove() == PieceColor.BLACK) {
            cellCount++;
        }
        
        if(cellCount == 0) {
            return null;
        } else {
            int moveRow = (cellCount - 1) / 2;
            int moveColumn = (cellCount - 1) % 2 == 0 ? 1 : 2;
            return new TableCell(moveRow, moveColumn);
        }
    }

    private void updateTableSelection(TableCell previousSelectedCell, TableCell currentSelectedCell) {
        Table table = tableDecorator.getTable();
        
        if(previousSelectedCell != null) {
            TableItem previousSelectedItem = table.getItem(previousSelectedCell.getRow());
            previousSelectedItem.setFont(0, movesFont);
            previousSelectedItem.setFont(previousSelectedCell.getColumn(), movesFont);
        }

        if(currentSelectedCell != null) {
            TableItem currentSelectedItem = table.getItem(currentSelectedCell.getRow());
            currentSelectedItem.setFont(0, selectedMoveFont);
            currentSelectedItem.setFont(currentSelectedCell.getColumn(), selectedMoveFont);
        }
        
        updateResultFont();
    }

    private void selectCurrentMove() {
        if(activeEditor != null) {
            TableCell currentPositionCell = getTableCell(getGame().getCurrentPosition());
            TableCell previousSelectedCell = tableDecorator.getSelectedCell();
            tableDecorator.setSelectedCell(currentPositionCell);
            updateTableSelection(previousSelectedCell, currentPositionCell);
            
            // ensures that the selected move is visible on the table
            Table table = tableDecorator.getTable();
            if(getGame().getCurrentPosition() == 0) {
                table.setTopIndex(0);
            } else {
                TableItem currentItem = table.getItem(currentPositionCell.getRow());
                table.showItem(currentItem);
            }
        }
    }

    private void updateTableHeaderFont() {
        Font previousTableHeaderFont = tableHeaderFont;
        
        IPreferenceStore preferences = HaundrixPlugin.getInstance().getPreferenceStore();
        int tableHeaderFontSize = preferences.getInt(HaundrixPreferences.GAME_HISTORY_HEADER_FONT_SIZE);
        
        Table table = tableDecorator.getTable();
        Display display = table.getDisplay();
        FontData[] fontDatas = table.getFont().getFontData();
        for(FontData fontData : fontDatas) {
            fontData.setHeight(tableHeaderFontSize);
        }
        tableHeaderFont = new Font(display, fontDatas);
        table.setFont(tableHeaderFont);
        tableDecorator.pack();
        
        if(previousTableHeaderFont != null) {
            previousTableHeaderFont.dispose();
        }
    }

    private void updateMovesAndResultFont() {
        Font previousMoveFont = movesFont;
        Font previousSelectedMoveFont = selectedMoveFont;
        
        IPreferenceStore preferences = HaundrixPlugin.getInstance().getPreferenceStore();
        String figurineFontId = preferences.getString(HaundrixPreferences.GAME_HISTORY_MOVES_FIGURINE_FONT);
        String figurineFontName = FigurineFontManager.getInstance().getFigurineFont(figurineFontId).getFontName();
        int movesFontSize = preferences.getInt(HaundrixPreferences.GAME_HISTORY_MOVES_FONT_SIZE);
        
        Table table = tableDecorator.getTable();
        Display display = table.getDisplay();
        movesFont = new Font(display, figurineFontName, movesFontSize, SWT.NORMAL);
        selectedMoveFont = new Font(display, figurineFontName, movesFontSize, SWT.BOLD);

        for(int row = 0; row < table.getItemCount(); row++) {
            TableItem tableItem = table.getItem(row);
            tableItem.setFont(movesFont);
            for(int column = 0; column < table.getColumnCount(); column++) {
                tableItem.setFont(column, movesFont);
            }
        }
        tableDecorator.pack();
        
        updateResultFont();
        
        selectCurrentMove();

        if(previousMoveFont != null) {
            previousMoveFont.dispose();
            previousSelectedMoveFont.dispose();
        }
    }

    private void updateResultFont() {
        Font resultFont = (activeEditor != null) && getGame().isFinished() ? selectedMoveFont : movesFont;
        resultLabel.setFont(resultFont);
    }

    private void addMove(String moveDescription, int gamePosition) {
        TableCell newMoveCell = getTableCell(gamePosition);
        Table table = tableDecorator.getTable();
        if(newMoveCell.getColumn() == 1) {
            TableItem newMoveItem = new TableItem(table, SWT.NONE);
            newMoveItem.setFont(movesFont);
            int moveNumber = getGame().getMoveNumber(gamePosition);
            newMoveItem.setText(0, moveNumber + ".");
            newMoveItem.setText(1, moveDescription);
        } else {
            TableItem newMoveItem = table.getItem(newMoveCell.getRow());
            newMoveItem.setText(2, moveDescription);
        }
    }

    private void updateResult() {
        GameResult result = getGame().getResult();
        if(result != null) {
            GameOutcome outcome = result.getOutcome();
            if(outcome == GameOutcome.WHITE_WINS) {
                resultLabel.setText(WHITE_WINS_LABEL);
            } else if(outcome == GameOutcome.BLACK_WINS) {
                resultLabel.setText(BLACK_WINS_LABEL);
            } else {
                resultLabel.setText(DRAW_LABEL);
            }
        }
    }
    
    private int getMaxDigits() {
        int lastPosition = getGame().getLastPosition();
        return lastPosition == 0 ? 0 : Integer.toString(getGame().getMoveNumber(lastPosition)).length();
    }
    
    private class InternalGameListener extends GameAdapter {

        @Override
        public void afterPositionChange(Game boardGame) {
            if(!selfMoving) {
                selectCurrentMove();
            }
        }

        @Override
        public void moved(Game boardGame) {
            String moveDescription = getGame().getLastMoveDescription(notation);
            int gamePosition = getGame().getLastPosition();
            addMove(moveDescription, gamePosition);
            if(getMaxDigits() > maxDigits) {
                tableDecorator.pack();
                maxDigits = getMaxDigits();
            } else {
                if(getGame().getLastColorToMove() == PieceColor.BLACK) {
                    tableDecorator.repackIfScrollbarAppeared();
                }
            }
        }

        @Override
        public void ended(GameResult result) {
            updateResult();
            updateResultFont();
            tableDecorator.repackIfScrollbarAppeared();
        }
        
    }

    private class InternalPreferencesListener implements IPropertyChangeListener {

        @Override
        public void propertyChange(PropertyChangeEvent event) {
            String changedProperty = event.getProperty();
            if(changedProperty.equals(HaundrixPreferences.GAME_HISTORY_HEADER_FONT_SIZE)) {
                updateTableHeaderFont();
            } else if((changedProperty.equals(HaundrixPreferences.GAME_HISTORY_MOVES_FIGURINE_FONT)) ||
                changedProperty.equals(HaundrixPreferences.GAME_HISTORY_MOVES_FONT_SIZE)) {
                updateMovesAndResultFont();
            }
        }

    }

}
