/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.views;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;

import com.haundrix.games.model.Game;
import com.haundrix.games.model.GameAdapter;
import com.haundrix.games.model.IGameListener;
import com.haundrix.games.rcp.HaundrixImages;

class NavigationBar extends Composite {

    private Image goToBeginImage;
    private Image goToLongPreviousImage;
    private Image goToPreviousImage;
    private Image goToNextImage;
    private Image goToLongNextImage;
    private Image goToEndImage;
    
    private Button goToBegin;
    private Button goToLongPrevious;
    private Button goToPrevious;
    private Button goToNext;
    private Button goToLongNext;
    private Button goToEnd;
    
    private Game game;
    
    private int longStepSize = 10;
    
    private IGameListener gameListener = new GameAdapter() {

        @Override
        public void afterPositionChange(Game game) {
            updateButtons();
        }

        @Override
        public void moved(Game game) {
            updateButtons();
        }
        
    };
    
    public NavigationBar(Composite parent) {
        super(parent, SWT.NONE);
        
        RowLayout layout = new RowLayout(SWT.HORIZONTAL);
        layout.marginTop = layout.marginBottom = 0;
        layout.marginLeft = layout.marginRight = 0;
        layout.spacing = 0;
        setLayout(layout);
        
        goToBeginImage = HaundrixImages.GO_TO_BEGIN.createImage();
        goToLongPreviousImage = HaundrixImages.GO_TO_LONG_PREVIOUS.createImage();
        goToPreviousImage = HaundrixImages.GO_TO_PREVIOUS.createImage();
        goToNextImage = HaundrixImages.GO_TO_NEXT.createImage();
        goToLongNextImage = HaundrixImages.GO_TO_LONG_NEXT.createImage();
        goToEndImage = HaundrixImages.GO_TO_END.createImage();
        
        goToBegin = new Button(this, SWT.PUSH);
        goToBegin.setEnabled(false);
        goToBegin.setImage(goToBeginImage);
        goToBegin.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                game.goToBegin();
            }
        });
        
        goToLongPrevious = new Button(this, SWT.PUSH);
        goToLongPrevious.setEnabled(false);
        goToLongPrevious.setImage(goToLongPreviousImage);
        goToLongPrevious.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                int newPosition = game.getCurrentPosition() - longStepSize;
                newPosition = Math.max(0, newPosition);
                game.goTo(newPosition);
            }
        });
        
        goToPrevious = new Button(this, SWT.PUSH);
        goToPrevious.setEnabled(false);
        goToPrevious.setImage(goToPreviousImage);
        goToPrevious.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                int newPosition = game.getCurrentPosition() - 1;
                game.goTo(newPosition);
            }
        });
        
        goToNext = new Button(this, SWT.PUSH);
        goToNext.setEnabled(false);
        goToNext.setImage(goToNextImage);
        goToNext.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                int newPosition = game.getCurrentPosition() + 1;
                game.goTo(newPosition);
            }
        });
        
        goToLongNext = new Button(this, SWT.PUSH);
        goToLongNext.setEnabled(false);
        goToLongNext.setImage(goToLongNextImage);
        goToLongNext.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                int newPosition = game.getCurrentPosition() + longStepSize;
                newPosition = Math.min(newPosition, game.getLastPosition());
                game.goTo(newPosition);
            }
        });
        
        goToEnd = new Button(this, SWT.PUSH);
        goToEnd.setEnabled(false);
        goToEnd.setImage(goToEndImage);
        goToEnd.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                game.goToEnd();
            }
        });

        addDisposeListener(new DisposeListener() {
            @Override
            public void widgetDisposed(DisposeEvent e) {
                goToBeginImage.dispose();
                goToLongPreviousImage.dispose();
                goToPreviousImage.dispose();
                goToNextImage.dispose();
                goToLongNextImage.dispose();
                goToEndImage.dispose();
                
                if(game != null) {
                    game.removeGameListener(gameListener);
                }
            }
        });
    }
    
    public void setGame(Game game) {
        if(this.game != null) {
            this.game.removeGameListener(gameListener);
        }
        
        this.game = game;
        if(this.game != null) {
            this.game.addGameListener(gameListener);
        }
        
        updateButtons();
    }

    public int getLongStepSize() {
        return longStepSize;
    }

    public void setLongStepSize(int longStepSize) {
        this.longStepSize = longStepSize;
    }

    private void updateButtons() {
        boolean allowBackMoves = (game != null) && !game.isAtBegin();
        goToBegin.setEnabled(allowBackMoves);
        goToLongPrevious.setEnabled(allowBackMoves);
        goToPrevious.setEnabled(allowBackMoves);
        
        boolean allowForwardMoves = (game != null) && !game.isAtEnd();
        goToNext.setEnabled(allowForwardMoves);
        goToLongNext.setEnabled(allowForwardMoves);
        goToEnd.setEnabled(allowForwardMoves);
    }

}
