/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.views;

import java.awt.image.BufferedImage;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.ProgressBar;

import com.haundrix.games.rcp.editors.TimeTroubleMonitor;
import com.haundrix.utils.ImageUtils;

class PlayerInfoControl extends Composite {
    
    private static final long SECOND = 1000L;
    private static final long MINUTE = 60L * SECOND;
    private static final long HOUR = 60L * MINUTE;
    private static final long TENTHS = 100L;

    private static final int TIME_BAR_MAX_VAL = 1000;

    private static final int PLAYER_ICON_IMAGE_SIZE = 28;
    
    private Label playerImage;
    private Label playerName;
    
    private ProgressBar timeBar;
    
    private Label playerIconImage;
    private Label playerMajorTime;
    
    private Composite minorTimeValuePanel;
    private Label playerMinorTime;
    
    public PlayerInfoControl(Composite parent, boolean playerImageOnTop) {
        super(parent, SWT.NONE);
        
        GridLayout gridLayout = new GridLayout(1, true);
        gridLayout.marginWidth = 0;
        gridLayout.marginHeight = 0;
        gridLayout.verticalSpacing = playerImageOnTop ? 7 : 3;
        setLayout(gridLayout);
        
        if(playerImageOnTop) {
            createPlayerInfoControls();
        }
        createPlayerTimeControls();
        if(!playerImageOnTop) {
            createPlayerInfoControls();
        }
        
        alignMinorTimeBaseline();
    }
    
    public Image getPlayerImage() {
        return playerImage.getImage();
    }
    
    public void setPlayerImage(Image image) {
        if(image != getPlayerImage()) {
            playerImage.setImage(image);
        }
    }
    
    public String getPlayerName() {
        return playerName.getText();
    }
    
    public void setPlayerName(String name) {
        if(!name.equals(getPlayerName())) {
            playerName.setText(name);
        }
    }
    
    public Font getPlayerNameFont() {
        return playerName.getFont();
    }
    
    public void setPlayerNameFont(Font font) {
        if(font != getPlayerNameFont()) {
            playerName.setFont(font);
        }
    }
    
    public Color getPlayerNameFontColor() {
        return playerName.getForeground();
    }
    
    public void setPlayerNameFontColor(Color color) {
        if(color != getPlayerNameFontColor()) {
            playerName.setForeground(color);
        }
    }
    
    public float getTimeFraction() {
        return ((float) timeBar.getSelection()) / ((float) TIME_BAR_MAX_VAL);
    }
    
    public void setTimeFraction(float timeFraction) {
        timeFraction = Math.max(timeFraction, 0.0F);
        
        int newSelection = (int) (timeFraction * TIME_BAR_MAX_VAL);
        if(newSelection != timeBar.getSelection()) {
            timeBar.setSelection(newSelection);
        }
    }
    
    public void setPlayerIconImage(BufferedImage iconImage) {
        Image oldIconImage = playerIconImage.getImage();
        
        BufferedImage awtIconImage = ImageUtils.fastAndGoodScale(iconImage, PLAYER_ICON_IMAGE_SIZE,
                PLAYER_ICON_IMAGE_SIZE);
        ImageData swtIconImage = ImageUtils.convertToSwt(awtIconImage);
        playerIconImage.setImage(new Image(getDisplay(), swtIconImage));
        
        if(oldIconImage != null) {
            oldIconImage.dispose();
        }
    }
    
    public void setPlayerTime(long time) {
        time = Math.max(time, 0L);
        
        long tempTime = time;
        
        int hour = (int) (tempTime / HOUR);
        tempTime = (int) (tempTime % HOUR);
        
        int minute = (int) (tempTime / MINUTE);
        tempTime = (int) (tempTime % MINUTE);
        
        int second = (int) (tempTime / SECOND);
        tempTime = (int) (tempTime % SECOND);
        
        if(time > TimeTroubleMonitor.TIME_TROUBLE_THRESHOLD) {
            String majorTimeText = String.format("%d:%02d:%02d ", hour, minute, second);
            if(!majorTimeText.equals(playerMajorTime.getText())) {
                playerMajorTime.setText(majorTimeText);
            }
            
            minorTimeValuePanel.setVisible(false);
        } else {
            // if time is smaller than TIME_TROUBLE_THRESHOLD, displays tenths of seconds
            int tenths = (int) (tempTime / TENTHS);
            
            String majorTimeText = String.format("%d:%02d:%02d.", hour, minute, second);
            if(!majorTimeText.equals(playerMajorTime.getText())) {
                playerMajorTime.setText(majorTimeText);
            }
            
            String minorTimeText = "" + tenths;
            if(!minorTimeText.equals(playerMinorTime.getText())) {
                playerMinorTime.setText(minorTimeText);
            }
            minorTimeValuePanel.setVisible(true);
        }
    }
    
    public Font getPlayerMajorTimeFont() {
        return playerMajorTime.getFont();
    }
    
    public void setPlayerMajorTimeFont(Font font) {
        if(font != getPlayerMajorTimeFont()) {
            playerMajorTime.setFont(font);
            alignMinorTimeBaseline();
        }
    }
    
    public Font getPlayerMinorTimeFont() {
        return playerMinorTime.getFont();
    }
    
    public void setPlayerMinorTimeFont(Font font) {
        if(font != getPlayerMinorTimeFont()) {
            playerMinorTime.setFont(font);
            alignMinorTimeBaseline();
        }
    }
    
    public Color getPlayerMajorTimeFontColor() {
        return playerMajorTime.getForeground();
    }
    
    public void setPlayerMajorTimeFontColor(Color color) {
        if(color != getPlayerMajorTimeFontColor()) {
            playerMajorTime.setForeground(color);
        }
    }

    public Color getPlayerMinorTimeFontColor() {
        return playerMinorTime.getForeground();
    }
    
    public void setPlayerMinorTimeFontColor(Color color) {
        if(color != getPlayerMinorTimeFontColor()) {
            playerMinorTime.setForeground(color);
        }
    }

    private void createPlayerInfoControls() {
        Composite infoPanel = new Composite(this, SWT.NONE);
        infoPanel.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
        GridLayout layout = new GridLayout(1, true);
        layout.marginWidth = 0;
        layout.marginHeight = 0;
        layout.verticalSpacing = 1;
        infoPanel.setLayout(layout);
        
        playerImage = new Label(infoPanel, SWT.NONE);
        playerImage.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));
        
        // Labels with the SWT.CENTER and SWT.WRAP styles are not rendered correctly
        // on recent Linux distros with GTK 2.18 due to this bug:
        // https://bugs.eclipse.org/bugs/show_bug.cgi?id=299492
        playerName = new Label(infoPanel, SWT.HORIZONTAL | SWT.CENTER | SWT.WRAP);
        playerName.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));
    }

    private void createPlayerTimeControls() {
        Composite timePanel = new Composite(this, SWT.NONE);
        timePanel.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
        GridLayout layout = new GridLayout(1, true);
        layout.marginWidth = 0;
        layout.marginHeight = 0;
        layout.verticalSpacing = 0;
        timePanel.setLayout(layout);
        
        timeBar = new ProgressBar(timePanel, SWT.HORIZONTAL);
        timeBar.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
        timeBar.setMinimum(0);
        timeBar.setMaximum(TIME_BAR_MAX_VAL);
        
        Composite timeDetailsPanel = new Composite(timePanel, SWT.NONE);
        timeDetailsPanel.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
        fillTimeDetailsPanel(timeDetailsPanel);
    }

    private void fillTimeDetailsPanel(Composite timeDetailsPanel) {
        GridLayout layout = new GridLayout(2, false);
        layout.marginWidth = 0;
        layout.marginHeight = 0;
        timeDetailsPanel.setLayout(layout);
        
        playerIconImage = new Label(timeDetailsPanel, SWT.NONE);
        playerIconImage.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, false, false));
        
        Composite timeValuePanel = new Composite(timeDetailsPanel, SWT.NONE);
        timeValuePanel.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));
        fillTimeValuePanel(timeValuePanel);
    }

    private void fillTimeValuePanel(Composite timeValuePanel) {
        GridLayout layout = new GridLayout(2, false);
        layout.marginWidth = 0;
        layout.marginHeight = 0;
        layout.horizontalSpacing = 0;
        timeValuePanel.setLayout(layout);
        
        playerMajorTime = new Label(timeValuePanel, SWT.HORIZONTAL);
        playerMajorTime.setLayoutData(new GridData(SWT.CENTER, SWT.END, false, false));
        
        minorTimeValuePanel = new Composite(timeValuePanel, SWT.NONE);
        minorTimeValuePanel.setLayoutData(new GridData(SWT.CENTER, SWT.END, false, false));
        
        fillMinorTimeValuePanel();
    }

    private void fillMinorTimeValuePanel() {
        RowLayout layout = new RowLayout(SWT.HORIZONTAL);
        layout.marginTop = 0;
        layout.marginRight = 0;
        layout.marginBottom = 0;
        layout.marginLeft = 0;
        minorTimeValuePanel.setLayout(layout);
        
        playerMinorTime = new Label(minorTimeValuePanel, SWT.HORIZONTAL);
    }

    private void alignMinorTimeBaseline() {
        GC gc = new GC(getDisplay());
        gc.setAdvanced(true);
        gc.setTextAntialias(SWT.ON);
        
        gc.setFont(getPlayerMajorTimeFont());
        int majorTimeDescent = gc.getFontMetrics().getDescent();
        
        gc.setFont(getPlayerMinorTimeFont());
        int minorTimeDescent = gc.getFontMetrics().getDescent();
        
        gc.dispose();
        
        RowLayout layout = (RowLayout) minorTimeValuePanel.getLayout();
        layout.marginBottom = majorTimeDescent - minorTimeDescent + 1;
    }

}
