/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.rcp.views;

import java.awt.image.BufferedImage;

import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.part.ViewPart;
import org.jdesktop.animation.timing.TimingEventListener;
import org.jdesktop.animation.timing.TimingSource;

import com.haundrix.animation.SwtTimingSource;
import com.haundrix.games.chess.view.ChessPieceType;
import com.haundrix.games.model.Game;
import com.haundrix.games.model.GameAdapter;
import com.haundrix.games.model.GameInfo;
import com.haundrix.games.model.IGameListener;
import com.haundrix.games.model.Piece.PieceColor;
import com.haundrix.games.model.playing.ClockTime;
import com.haundrix.games.model.playing.HumanPlayer;
import com.haundrix.games.model.playing.IClock;
import com.haundrix.games.rcp.HaundrixImages;
import com.haundrix.games.rcp.HaundrixPlugin;
import com.haundrix.games.rcp.editors.ChessGameEditor;
import com.haundrix.games.rcp.editors.IFlipListener;
import com.haundrix.games.rcp.editors.LocalChessGameEditor;
import com.haundrix.games.rcp.managers.ColorSchemeManager;
import com.haundrix.games.rcp.managers.FontManager;
import com.haundrix.games.rcp.managers.PieceSetManager;
import com.haundrix.games.rcp.managers.PluginFont;
import com.haundrix.games.rcp.managers.PluginPieceSet;
import com.haundrix.games.rcp.preferences.HaundrixPreferences;
import com.haundrix.games.view.ColorScheme;
import com.haundrix.rcp.ActiveEditorListener;
import com.haundrix.rcp.IEditorAware;
import com.haundrix.swt.GradientComposite;
import com.haundrix.utils.ImageUtils;

public class TimeView extends ViewPart implements IEditorAware<ChessGameEditor> {
    
    public static final String ID = "com.haundrix.games.timeView";
    
    private static final int PLAYER_IMAGE_SIZE = 64;
    
    private static final String UNKNOWN_PLAYER_NAME = "N. N.";  // TODO externalize this
    private static final String PLAYER_NAME_FONT_ID = "com.haundrix.games.fonts.baveuse";
    private static final int PLAYER_NAME_FONT_SIZE = 14;
    
    private static final String PLAYER_TIME_FONT_ID = "com.haundrix.games.fonts.haundrixMono";
    private static final int PLAYER_MAJOR_TIME_FONT_SIZE = 17;
    private static final int PLAYER_MINOR_TIME_FONT_SIZE = 10;
    
    private static final int TIME_UPDATE_INTERVAL = 50;  // in milliseconds
    
    private Image whiteKingImage;
    private Image blackKingImage;
    private Image humanImage;
    private Image computerImage;
    
    private Font playerNameFont;
    private Color playerNameFontColor;
    
    private Font currentPlayerMajorTimeFont;
    private Font currentPlayerMinorTimeFont;
    private Font otherPlayerMajorTimeFont;
    private Font otherPlayerMinorTimeFont;
    private Color whitePlayerTimeFontColor;
    private Color blackPlayerTimeFontColor;
    
    private GradientComposite backgroundPanel;
    private Color topBackgroundColor;
    private Color bottomBackgroundColor;

    private PlayerInfoControl topPlayerControl;
    private PlayerInfoControl bottomPlayerControl;
    
    private ChessGameEditor chessEditor;

    private InternalPreferencesListener preferencesListener = new InternalPreferencesListener();
    private ActiveEditorListener<ChessGameEditor> activeEditorListener = new ActiveEditorListener<ChessGameEditor>(this);
    private IGameListener gameListener = new InternalGameListener();
    private IFlipListener flipListener = new InternalFlipListener();
    
    private SwtTimingSource swtTimer;

    private boolean isInitialized = false;

    @Override
    public void createPartControl(Composite parent) {
        createKingImages(true);
        humanImage = HaundrixImages.HUMAN.createImage();
        computerImage = HaundrixImages.COMPUTER.createImage();
        
        Display display = parent.getDisplay();
        FontManager fontManager = FontManager.getInstance();
        
        PluginFont pluginPlayerNameFont = fontManager.getFont(PLAYER_NAME_FONT_ID);
        playerNameFont = new Font(display, pluginPlayerNameFont.getFontName(), PLAYER_NAME_FONT_SIZE, SWT.NORMAL);
        playerNameFontColor = new Color(display, 5, 96, 145);
        
        PluginFont pluginPlayerTimeFont = fontManager.getFont(PLAYER_TIME_FONT_ID);
        currentPlayerMajorTimeFont = new Font(display, pluginPlayerTimeFont.getFontName(),
                PLAYER_MAJOR_TIME_FONT_SIZE, SWT.ITALIC);
        currentPlayerMinorTimeFont = new Font(display, pluginPlayerTimeFont.getFontName(),
                PLAYER_MINOR_TIME_FONT_SIZE, SWT.ITALIC);
        otherPlayerMajorTimeFont = new Font(display, pluginPlayerTimeFont.getFontName(),
                PLAYER_MAJOR_TIME_FONT_SIZE, SWT.NORMAL);
        otherPlayerMinorTimeFont = new Font(display, pluginPlayerTimeFont.getFontName(),
                PLAYER_MINOR_TIME_FONT_SIZE, SWT.NORMAL);
        whitePlayerTimeFontColor = display.getSystemColor(SWT.COLOR_WHITE);
        blackPlayerTimeFontColor = display.getSystemColor(SWT.COLOR_BLACK);
        
        parent.setLayout(new FillLayout());
        
        backgroundPanel = new GradientComposite(parent);
        backgroundPanel.setLayout(new GridLayout(1, true));
        
        topPlayerControl = new PlayerInfoControl(backgroundPanel, true);
        topPlayerControl.setLayoutData(new GridData(SWT.FILL, SWT.BEGINNING, true, true));
        topPlayerControl.setPlayerNameFont(playerNameFont);
        topPlayerControl.setPlayerNameFontColor(playerNameFontColor);
        
        bottomPlayerControl = new PlayerInfoControl(backgroundPanel, false);
        bottomPlayerControl.setLayoutData(new GridData(SWT.FILL, SWT.END, true, true));
        bottomPlayerControl.setPlayerNameFont(playerNameFont);
        bottomPlayerControl.setPlayerNameFontColor(playerNameFontColor);
        
        swtTimer = new SwtTimingSource(parent.getDisplay(), TIME_UPDATE_INTERVAL);
        swtTimer.addEventListener(new TimingEventListener() {
            @Override
            public void timingSourceEvent(TimingSource timingSource) {
                updateTime();
            }
        });

        IWorkbenchPage page = getViewSite().getPage();
        
        IEditorPart activeEditor = page.getActiveEditor();
        ChessGameEditor activeChessEditor = activeEditorListener.isValidEditor(activeEditor) ?
                (ChessGameEditor) activeEditor : null;
        setActiveEditor(activeChessEditor);

        updateBackground();
        
        getPreferenceStore().addPropertyChangeListener(preferencesListener);
        page.addPartListener(activeEditorListener);
        
        isInitialized = true;
    }

    @Override
    public void setFocus() {
    }

    @Override
    public void dispose() {
        if(isInitialized) {
            whiteKingImage.dispose();
            blackKingImage.dispose();
            humanImage.dispose();
            computerImage.dispose();
            
            playerNameFont.dispose();
            playerNameFontColor.dispose();
            
            currentPlayerMajorTimeFont.dispose();
            currentPlayerMinorTimeFont.dispose();
            otherPlayerMajorTimeFont.dispose();
            otherPlayerMinorTimeFont.dispose();
            
            getPreferenceStore().removePropertyChangeListener(preferencesListener);
            getViewSite().getPage().removePartListener(activeEditorListener);
            if(chessEditor != null) {
                chessEditor.getGame().removeGameListener(gameListener);
                chessEditor.removeFlipListener(flipListener);
                
                if(isTimedGame()) {
                    swtTimer.stop();
                }
            }
        }
        
        super.dispose();
    }

    @Override
    public Class<ChessGameEditor> getEditorClass() {
        return ChessGameEditor.class;
    }

    @Override
    public ChessGameEditor getActiveEditor() {
        return chessEditor;
    }

    @Override
    public void setActiveEditor(ChessGameEditor activeChessEditor) {
        if(this.chessEditor != null) {
            this.chessEditor.getGame().removeGameListener(gameListener);
            this.chessEditor.removeFlipListener(flipListener);
            
            if(isTimedGame()) {
                swtTimer.stop();
            }
        }
        
        this.chessEditor = activeChessEditor;
        if(this.chessEditor != null) {
            this.chessEditor.getGame().addGameListener(gameListener);
            this.chessEditor.addFlipListener(flipListener);
            
            if(isTimedGame()) {
                swtTimer.start();
            }
        }
        
        updateContent();
    }
    
    private IPreferenceStore getPreferenceStore() {
        return HaundrixPlugin.getInstance().getPreferenceStore();
    }
    
    private boolean getUsePrefColors() {
        return getPreferenceStore().getBoolean(HaundrixPreferences.CHESS_BOARD_USE_PIECE_SET_PREFERRED_COLORS);
    }

    private PluginPieceSet getCurrentPieceSet() {
        String pieceSetId = getPreferenceStore().getString(HaundrixPreferences.CHESS_BOARD_PIECE_SET);
        return PieceSetManager.getInstance().getPieceSet(pieceSetId);
    }

    private void createKingImages(boolean firstCall) {
        if(!firstCall) {
            Image topPlayerImage = topPlayerControl.getPlayerImage();
            if((topPlayerImage == whiteKingImage) || (topPlayerImage == blackKingImage)) {
                topPlayerControl.setPlayerImage(null);
            }
            
            Image bottomPlayerImage = bottomPlayerControl.getPlayerImage();
            if((bottomPlayerImage == whiteKingImage) || (bottomPlayerImage == blackKingImage)) {
                bottomPlayerControl.setPlayerImage(null);
            }
            
            whiteKingImage.dispose();
            blackKingImage.dispose();
        }
        
        PluginPieceSet pieceSet = getCurrentPieceSet();
        Display display = getViewSite().getShell().getDisplay();
        
        BufferedImage awtWhiteKingImage = pieceSet.loadPieceImage(ChessPieceType.WHITE_KING);
        awtWhiteKingImage = ImageUtils.fastAndGoodScale(awtWhiteKingImage, PLAYER_IMAGE_SIZE, PLAYER_IMAGE_SIZE);
        ImageData swtWhiteKingImage = ImageUtils.convertToSwt(awtWhiteKingImage);
        whiteKingImage = new Image(display, swtWhiteKingImage);
        
        BufferedImage awtBlackKingImage = pieceSet.loadPieceImage(ChessPieceType.BLACK_KING);
        awtBlackKingImage = ImageUtils.fastAndGoodScale(awtBlackKingImage, PLAYER_IMAGE_SIZE, PLAYER_IMAGE_SIZE);
        ImageData swtBlackKingImage = ImageUtils.convertToSwt(awtBlackKingImage);
        blackKingImage = new Image(display, swtBlackKingImage);
    }

    private boolean isTimedGame() {
        return (chessEditor != null) &&
            (chessEditor instanceof LocalChessGameEditor) &&
            ((LocalChessGameEditor) chessEditor).getClock() != null;
    }

    private void updateContent() {
        Image whitePlayerImage = whiteKingImage;
        String whitePlayerName = UNKNOWN_PLAYER_NAME;
        Image blackPlayerImage = blackKingImage;
        String blackPlayerName = UNKNOWN_PLAYER_NAME;
        boolean isFlipped = false;
        if(chessEditor != null) {
            GameInfo gameInfo = chessEditor.getGame().getInfo();
            whitePlayerName = gameInfo.getWhitePlayer();
            if((whitePlayerName == null) || whitePlayerName.trim().equals("")) {
                whitePlayerName = UNKNOWN_PLAYER_NAME;
            }
            blackPlayerName = gameInfo.getBlackPlayer();
            if((blackPlayerName == null) || blackPlayerName.trim().equals("")) {
                blackPlayerName = UNKNOWN_PLAYER_NAME;
            }
            
            isFlipped = chessEditor.isFlipped();
            
            if(chessEditor instanceof LocalChessGameEditor) {
                LocalChessGameEditor localGameEditor = (LocalChessGameEditor) chessEditor;
                whitePlayerImage = localGameEditor.getWhitePlayer() instanceof HumanPlayer ? humanImage : computerImage;
                blackPlayerImage = localGameEditor.getBlackPlayer() instanceof HumanPlayer ? humanImage : computerImage;
            }
        }
        
        PluginPieceSet pieceSet = getCurrentPieceSet();
        BufferedImage whitePlayerIconImage = pieceSet.loadPieceImage(ChessPieceType.WHITE_KNIGHT);
        BufferedImage blackPlayerIconImage = pieceSet.loadPieceImage(ChessPieceType.BLACK_KNIGHT);

        PlayerInfoControl whitePlayerControl = isFlipped ? topPlayerControl : bottomPlayerControl;
        PlayerInfoControl blackPlayerControl = isFlipped ? bottomPlayerControl : topPlayerControl;
        
        whitePlayerControl.setPlayerImage(whitePlayerImage);
        whitePlayerControl.setPlayerName(whitePlayerName);
        whitePlayerControl.setPlayerIconImage(whitePlayerIconImage);
        whitePlayerControl.setPlayerMajorTimeFontColor(whitePlayerTimeFontColor);
        whitePlayerControl.setPlayerMinorTimeFontColor(whitePlayerTimeFontColor);
        
        blackPlayerControl.setPlayerImage(blackPlayerImage);
        blackPlayerControl.setPlayerName(blackPlayerName);
        blackPlayerControl.setPlayerIconImage(blackPlayerIconImage);
        blackPlayerControl.setPlayerMajorTimeFontColor(blackPlayerTimeFontColor);
        blackPlayerControl.setPlayerMinorTimeFontColor(blackPlayerTimeFontColor);
        
        updateTime();
        
        // TODO Investigate this as a possible bug with SWT.
        // This shouldn't be necessary as it should be handled automatically by SWT,
        // but somehow there seems to be a bug lurking around (at least on Linux).
        // Symptom:
        // When the player's name does not fit in a single line only the first line
        // is shown on the Label. Resizing the view causes its content to be repacked
        // and all lines to be shown correctly.
        backgroundPanel.layout(true, true);
    }

    private void updateTime() {
        if(!isTimedGame()) {
            topPlayerControl.setTimeFraction(1.0F);
            topPlayerControl.setPlayerTime(0L);
            topPlayerControl.setPlayerMajorTimeFont(otherPlayerMajorTimeFont);
            topPlayerControl.setPlayerMinorTimeFont(otherPlayerMinorTimeFont);

            bottomPlayerControl.setTimeFraction(1.0F);
            bottomPlayerControl.setPlayerTime(0L);
            bottomPlayerControl.setPlayerMajorTimeFont(otherPlayerMajorTimeFont);
            bottomPlayerControl.setPlayerMinorTimeFont(otherPlayerMinorTimeFont);

            return;
        }
        
        LocalChessGameEditor localGameEditor = (LocalChessGameEditor) chessEditor;
        IClock clock = localGameEditor.getClock();
        ClockTime clockTime = clock.getClockTime();
        
        PlayerInfoControl whitePlayerControl = localGameEditor.isFlipped() ? topPlayerControl : bottomPlayerControl;
        PlayerInfoControl blackPlayerControl = localGameEditor.isFlipped() ? bottomPlayerControl : topPlayerControl;
        
        float whiteTimeFraction = (float) (((double) clockTime.getWhiteTime()) / ((double) clock.getMaxTimeScale()));
        whiteTimeFraction = Math.min(whiteTimeFraction, 1.0F);
        whitePlayerControl.setTimeFraction(whiteTimeFraction);
        whitePlayerControl.setPlayerTime(clockTime.getWhiteTime());
        Font whitePlayerMajorTimeFont;
        Font whitePlayerMinorTimeFont;
        if(clock.getColorToMove() == PieceColor.WHITE) {
            whitePlayerMajorTimeFont = currentPlayerMajorTimeFont;
            whitePlayerMinorTimeFont = currentPlayerMinorTimeFont;
        } else {
            whitePlayerMajorTimeFont = otherPlayerMajorTimeFont;
            whitePlayerMinorTimeFont = otherPlayerMinorTimeFont;
        }
        whitePlayerControl.setPlayerMajorTimeFont(whitePlayerMajorTimeFont);
        whitePlayerControl.setPlayerMinorTimeFont(whitePlayerMinorTimeFont);
        
        float blackTimeFraction = (float) (((double) clockTime.getBlackTime()) / ((double) clock.getMaxTimeScale()));
        blackTimeFraction = Math.min(blackTimeFraction, 1.0F);
        blackPlayerControl.setTimeFraction(blackTimeFraction);
        blackPlayerControl.setPlayerTime(clockTime.getBlackTime());
        Font blackPlayerMajorTimeFont;
        Font blackPlayerMinorTimeFont;
        if(clock.getColorToMove() == PieceColor.BLACK) {
            blackPlayerMajorTimeFont = currentPlayerMajorTimeFont;
            blackPlayerMinorTimeFont = currentPlayerMinorTimeFont;
        } else {
            blackPlayerMajorTimeFont = otherPlayerMajorTimeFont;
            blackPlayerMinorTimeFont = otherPlayerMinorTimeFont;
        }
        blackPlayerControl.setPlayerMajorTimeFont(blackPlayerMajorTimeFont);
        blackPlayerControl.setPlayerMinorTimeFont(blackPlayerMinorTimeFont);
    }

    private void updateBackground() {
        String colorSchemeId;
        if(getUsePrefColors()) {
            String pieceSetId = getPreferenceStore().getString(HaundrixPreferences.CHESS_BOARD_PIECE_SET);
            colorSchemeId = PieceSetManager.getInstance().getPieceSet(pieceSetId).getPreferredColorSchemeId();
        } else {
            colorSchemeId = getPreferenceStore().getString(HaundrixPreferences.CHESS_BOARD_COLOR_SCHEME);
        }
        ColorScheme colorScheme = ColorSchemeManager.getInstance().getColorScheme(colorSchemeId);
        
        Color oldTopBackgroundColor = topBackgroundColor;
        Color oldBottomBackgroundColor = bottomBackgroundColor;
        topBackgroundColor = new Color(backgroundPanel.getDisplay(), colorScheme.getTopBackgroundColor());
        bottomBackgroundColor = new Color(backgroundPanel.getDisplay(), colorScheme.getBottomBackgroundColor());
        backgroundPanel.setTopBackgroundColor(topBackgroundColor);
        backgroundPanel.setBottomBackgroundColor(bottomBackgroundColor);
        if(oldTopBackgroundColor != null) {
            oldTopBackgroundColor.dispose();
        }
        if(oldBottomBackgroundColor != null) {
            oldBottomBackgroundColor.dispose();
        }
    }

    private final class InternalPreferencesListener implements IPropertyChangeListener {
        
        @Override
        public void propertyChange(PropertyChangeEvent event) {
            String changedProperty = event.getProperty();
            
            if(changedProperty.equals(HaundrixPreferences.CHESS_BOARD_PIECE_SET)) {
                createKingImages(false);
                updateContent();
            }
            
            boolean usePrefColors = getUsePrefColors();
            boolean updateBackground =
                changedProperty.equals(HaundrixPreferences.CHESS_BOARD_USE_PIECE_SET_PREFERRED_COLORS) ||
                (changedProperty.equals(HaundrixPreferences.CHESS_BOARD_PIECE_SET) && usePrefColors) ||
                (changedProperty.equals(HaundrixPreferences.CHESS_BOARD_COLOR_SCHEME) && !usePrefColors);
            if(updateBackground) {
                updateBackground();
            }
        }
        
    }

    public class InternalGameListener extends GameAdapter implements IGameListener {

        @Override
        public void infoChanged(GameInfo info) {
            updateContent();
        }

        @Override
        public void moved(Game game) {
            updateTime();
        }

    }

    public class InternalFlipListener implements IFlipListener {

        @Override
        public void flipped(boolean isFlipped) {
            updateContent();
        }

    }

}
