/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.view;

import java.awt.AlphaComposite;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Display;
import org.jdesktop.animation.timing.TimingTarget;
import org.jdesktop.animation.timing.TimingTargetAdapter;
import org.jdesktop.animation.timing.interpolation.PropertySetter;

import com.haundrix.animation.CompositeFadeCallback;
import com.haundrix.animation.ConstantAccelerationTimingTarget;
import com.haundrix.animation.DisposableAwareTimingTarget;
import com.haundrix.animation.FadeCallback;
import com.haundrix.animation.Highlighter;
import com.haundrix.animation.IFadable;
import com.haundrix.animation.IHighlightable;
import com.haundrix.animation.IMovable;
import com.haundrix.animation.StrechedSpringInterpolator;
import com.haundrix.animation.SwtAnimator;
import com.haundrix.graphics.Highlight;
import com.haundrix.graphics.HighlightRenderer;
import com.haundrix.graphics.ShadowRenderer;
import com.haundrix.utils.IDisposeListener;
import com.haundrix.utils.IDisposable;
import com.haundrix.utils.ImageUtils;

public class BoardPiece<P extends IPieceType> implements IDisposable, IMovable, IFadable, IHighlightable {

    private static final float CROSSFADE_ACCELERATION = 0.3F;
    private static final float CROSSFADE_DECELERATION = 0.2F;
    
    private Display display;
    private P piece;
    private PieceSet<P> pieceSet;
    private int size;
    private Point location;
    
    private IBoardTransform coordinateTransform;
    
    private BufferedImage originalImage;
    private BufferedImage resizedImage;
    private BufferedImage highlightBackgroundImage;
    private BufferedImage highlightedAndShadowedImage;
    
    private boolean isCrossfading = false;
    private P targetPiece;
    private float crossfadeFraction;  // 0 shows just the original piece, 1 shows just the target piece
    private BufferedImage originalTargetImage;
    private BufferedImage resizedTargetImage;
    private BufferedImage highlightBackgroundTargetImage;
    private BufferedImage highlightedAndShadowedTargetImage;
    
    private Image finalImage;

    private Highlight highlight = null;
    private float shadowIntensity = 0.0F;
    private float alpha = 1.0F;
    
    private List<IDisposeListener> disposeListeners = new ArrayList<IDisposeListener>();
    private List<IBoardPieceListener<P>> boardPieceListeners = new ArrayList<IBoardPieceListener<P>>();
    
    private boolean isDisposed = false;

    public BoardPiece(Display display, P piece, PieceSet<P> pieceSet, int size, Point location) {
        this.display = display;
        this.piece = piece;
        this.pieceSet = pieceSet;
        this.size = size;
        this.location = location;
        
        loadPieceImages();
    }

    public P getPiece() {
        return piece;
    }
    
    public void setPiece(P piece) {
        if(piece != this.piece) {
            this.piece = piece;
            loadPieceImages();
        }
    }

    public void crossfadeTo(P piece, int duration) {
        crossfadeTo(piece, duration, new FadeCallback());
    }

    public void crossfadeTo(P piece, int duration, Runnable endFadeCallback) {
        crossfadeTo(piece, duration, new FadeCallback(endFadeCallback));
    }
    
    public void crossfadeTo(P piece, int duration, FadeCallback fadeCallback) {
        if(piece != this.piece) {
            final FadeCallback usedFadeCallback = prepareCrossfade(piece, fadeCallback);
            
            PropertySetter propertySetter = new PropertySetter(this, "crossfadeFraction", 1.0F);
            final SwtAnimator animator = new SwtAnimator(display, duration, propertySetter);
            
            final IDisposeListener disposeListener = new IDisposeListener() {
                @Override
                public void disposed(IDisposable disposable) {
                    animator.cancel();
                    usedFadeCallback.fadableDisposed(disposable);
                }
            };
            addDisposeListener(disposeListener);
            
            animator.addTarget(new TimingTargetAdapter() {

                @Override
                public void begin() {
                    usedFadeCallback.started();
                }

                @Override
                public void end() {
                    removeDisposeListener(disposeListener);
                    usedFadeCallback.ended();
                }
                
            });
            
            animator.setAcceleration(CROSSFADE_ACCELERATION);
            animator.setDeceleration(CROSSFADE_DECELERATION);
            animator.start();
        }
    }
    
    public FadeCallback prepareCrossfade(P piece, FadeCallback fadeCallback) {
        FadeCallback usedFadeCallback = null;
        
        if(piece != this.piece) {
            isCrossfading = true;
            targetPiece = piece;
            crossfadeFraction = 0.0F;
            originalTargetImage = pieceSet.loadPieceImage(piece);
            resizedTargetImage = getResizedImage(originalTargetImage);
            highlightBackgroundTargetImage = highlight == null ? null :
                new HighlightRenderer(highlight).getHighlightBackgroundImage(resizedTargetImage);
            highlightedAndShadowedTargetImage = getHighlightedAndShadowedImage(resizedTargetImage,
                    highlightBackgroundTargetImage);
            
            FadeCallback endFadeCallback = new FadeCallback() {
                @Override
                public void ended() {
                    BoardPiece.this.piece = targetPiece;
                    originalImage = originalTargetImage;
                    resizedImage = resizedTargetImage;
                    highlightBackgroundImage = highlightBackgroundTargetImage;
                    highlightedAndShadowedImage = highlightedAndShadowedTargetImage;

                    isCrossfading = false;
                    targetPiece = null;
                    crossfadeFraction = 0.0F;
                    originalTargetImage = null;
                    resizedTargetImage = null;
                    highlightBackgroundTargetImage = null;
                    highlightedAndShadowedTargetImage = null;
                }
            };
            
            usedFadeCallback = fadeCallback == null ? endFadeCallback :
                new CompositeFadeCallback(endFadeCallback, fadeCallback);
        }

        return usedFadeCallback;
    }
    
    public TimingTarget createCrossfadeTimingTarget(P newPiece) {
        PropertySetter propertySetter = new PropertySetter(this, "crossfadeFraction", 1.0F);
        TimingTarget timingTarget = new ConstantAccelerationTimingTarget(CROSSFADE_ACCELERATION, CROSSFADE_DECELERATION,
                propertySetter);
        return new DisposableAwareTimingTarget(this, timingTarget);
    }
    
    public PieceSet<P> getPieceSet() {
        return pieceSet;
    }

    public void setPieceSet(PieceSet<P> pieceSet) {
        this.pieceSet = pieceSet;
        loadPieceImages();
    }

    public int getSize() {
        return size;
    }
    
    public void setSize(int size) {
        if(size != this.size) {
            this.size = size;
            resizeImages();
        }
    }

    @Override
    public Point getLocation() {
        return location;
    }

    @Override
    public void setLocation(Point location) {
        Point previousLocation = this.location;
        this.location = coordinateTransform == null ? location : coordinateTransform.transform(location);
        BoardPieceMovedEvent<P> event = new BoardPieceMovedEvent<P>(this, previousLocation, this.location);
        synchronized(boardPieceListeners) {
            for(IBoardPieceListener<P> listener : boardPieceListeners) {
                listener.boardPieceMoved(event);
            }
        }
    }

    public IBoardTransform getCoordinateTransform() {
        return coordinateTransform;
    }

    public void setCoordinateTransform(IBoardTransform coordinateTransform) {
        this.coordinateTransform = this.coordinateTransform == null ? coordinateTransform :
            new CompositeBoardTransform(this.coordinateTransform, coordinateTransform);
        location = coordinateTransform.transform(location);
    }
    
    public void clearCoordinateTransform() {
        coordinateTransform = null;
    }

    @Override
    public Highlight getHighlight() {
        return highlight;
    }

    @Override
    public void setHighlight(Highlight highlight) {
        boolean settingToNull = (highlight == null) && (this.highlight != null);
        boolean changingHighlight = (highlight != null) && !highlight.equals(this.highlight); 
        if(settingToNull || changingHighlight) {
            this.highlight = highlight;
            highlightAndShadowImages(true);
        }
    }
    
    @Override
    public float getHighlightIntensity() {
        return highlight == null ? 0.0F : highlight.getIntensity();
    }

    @Override
    public void setHighlightIntensity(float highlightIntensity) {
        if(highlight != null) {
            highlight = highlight.derive(highlightIntensity);
            highlightAndShadowImages(false);
        }
    }

    public void highlightTo(Highlight highlight, int duration) {
        Highlighter highlighter = new Highlighter(display, this, highlight, duration);
        highlighter.setAcceleration(0.2F);
        highlighter.setDeceleration(0.2F);
        highlighter.highlight();
    }
    
    public void pulse(Highlight highlight, int duration, int nPulses) {
        Highlighter highlighter = new Highlighter(display, this, highlight, duration);
        highlighter.setInterpolator(StrechedSpringInterpolator.getInstance());
        highlighter.pulse(nPulses);
    }

    public float getShadowIntensity() {
        return shadowIntensity;
    }

    public void setShadowIntensity(float shadowIntensity) {
        if((shadowIntensity < 0.0F) || (shadowIntensity > 1.0F)) {
            throw new IllegalArgumentException("Shadow intensity must be a number within 0 and 1.");
        }
        
        if(shadowIntensity != this.shadowIntensity) {
            this.shadowIntensity = shadowIntensity;
            highlightAndShadowImages(false);
        }
    }

    @Override
    public float getAlpha() {
        return alpha;
    }

    @Override
    public void setAlpha(float alpha) {
        if((alpha < 0.0F) || (alpha > 1.0F)) {
            throw new IllegalArgumentException("Alpha value must be a number within 0 and 1.");
        }
        
        if(alpha != this.alpha) {
            this.alpha = alpha;
            calculateFinalImage();
        }
    }

    // This method is not part of the public API!
    // It is just called by reflection from the Timing Framework API.
    public float getCrossfadeFraction() {
        return crossfadeFraction;
    }

    // This method is not part of the public API!
    // It is just called by reflection from the Timing Framework API.
    public void setCrossfadeFraction(float crossfadeFraction) {
        this.crossfadeFraction = crossfadeFraction;
        calculateFinalImage();
    }

    public Image getImage() {
        return finalImage;
    }

    public Rectangle getBounds() {
        Rectangle imageBounds = finalImage.getBounds();
        return new Rectangle(location.x, location.y, imageBounds.width, imageBounds.height);
    }
    
    public void addBoardPieceListener(IBoardPieceListener<P> boardPieceListener) {
        synchronized(boardPieceListeners) {
            if(!boardPieceListeners.contains(boardPieceListener)) {
                boardPieceListeners.add(boardPieceListener);
            }
        }
    }
    
    public void removeBoardPieceListener(IBoardPieceListener<P> boardPieceListener) {
        synchronized(boardPieceListeners) {
            boardPieceListeners.remove(boardPieceListener);
        }
    }
    
    @Override
    public void addDisposeListener(IDisposeListener disposeListener) {
        synchronized(disposeListeners) {
            if(!disposeListeners.contains(disposeListener)) {
                disposeListeners.add(disposeListener);
            }
        }
    }

    @Override
    public void removeDisposeListener(IDisposeListener disposeListener) {
        synchronized(disposeListeners) {
            disposeListeners.remove(disposeListener);
        }
    }

    @Override
    public boolean isDisposed() {
        return isDisposed;
    }

    @Override
    public void dispose() {
        if(!isDisposed) {
            synchronized(disposeListeners) {
                for(IDisposeListener disposeListener : disposeListeners) {
                    disposeListener.disposed(this);
                }
            }
    
            finalImage.dispose();
            isDisposed = true;
        }
    }
    
    private void loadPieceImages() {
        originalImage = pieceSet.loadPieceImage(piece);
        if(isCrossfading) {
            originalTargetImage = pieceSet.loadPieceImage(targetPiece);
        }
        
        resizeImages();
    }
    
    private void resizeImages() {
        resizedImage = getResizedImage(originalImage);
        if(isCrossfading) {
            resizedTargetImage = getResizedImage(originalTargetImage);
        }
        
        highlightAndShadowImages(true);
    }
    
    private BufferedImage getResizedImage(BufferedImage image) {
        return ImageUtils.fastAndGoodScale(image, size, size);
    }
    
    private void highlightAndShadowImages(boolean rerenderHighlightBackground) {
        if((highlight != null) && rerenderHighlightBackground) {
            highlightBackgroundImage = new HighlightRenderer(highlight).getHighlightBackgroundImage(resizedImage);
        }
        highlightedAndShadowedImage = getHighlightedAndShadowedImage(resizedImage, highlightBackgroundImage);
        if(isCrossfading) {
            if((highlight != null) && rerenderHighlightBackground) {
                highlightBackgroundTargetImage = new HighlightRenderer(highlight).getHighlightBackgroundImage(resizedTargetImage);
            }
            highlightedAndShadowedTargetImage = getHighlightedAndShadowedImage(resizedTargetImage, highlightBackgroundTargetImage);
        }
        
        calculateFinalImage();
    }
    
    private BufferedImage getHighlightedAndShadowedImage(BufferedImage image, BufferedImage highlightBackgroundImage) {
        BufferedImage highlightedImage = highlight == null ? image :
                HighlightRenderer.getHighlightedImage(image, highlightBackgroundImage, highlight.getIntensity());
        return new ShadowRenderer(shadowIntensity).getShadowedImage(highlightedImage);
    }

    private void calculateFinalImage() {
        Rectangle previousBounds = new Rectangle(0, 0, 0, 0);
        if(finalImage != null) {
            previousBounds = getBounds();
            finalImage.dispose();
        }
        
        BufferedImage awtFinalImage;
        if((alpha == 1.0F) && (!isCrossfading)) {
            awtFinalImage = highlightedAndShadowedImage;
        } else {
            int width = highlightedAndShadowedImage.getWidth();
            int height = highlightedAndShadowedImage.getHeight();
            int imageType = highlightedAndShadowedImage.getType();
            if(isCrossfading) {
                if(highlightedAndShadowedTargetImage.getWidth() > width) {
                    width = highlightedAndShadowedTargetImage.getWidth();
                }
                if(highlightedAndShadowedTargetImage.getHeight() > height) {
                    height = highlightedAndShadowedTargetImage.getHeight();
                }
            }
            awtFinalImage = new BufferedImage(width, height, imageType);

            Graphics2D graphics = awtFinalImage.createGraphics();
            graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            float sourceAlpha = isCrossfading ? alpha * (1.0F - crossfadeFraction) : alpha;
            graphics.setComposite(AlphaComposite.SrcOver.derive(sourceAlpha));
            graphics.drawImage(highlightedAndShadowedImage, 0, 0, null);
            if(isCrossfading) {
                graphics.setComposite(AlphaComposite.SrcOver.derive(alpha * crossfadeFraction));
                graphics.drawImage(highlightedAndShadowedTargetImage, 0, 0, null);
            }
            graphics.dispose();
        }
        
        ImageData imageData = ImageUtils.convertToSwt(awtFinalImage);
        finalImage = new Image(display, imageData);

        triggerChangedEvent(previousBounds);
    }

    private void triggerChangedEvent(Rectangle bounds) {
        BoardPieceChangedEvent<P> event = new BoardPieceChangedEvent<P>(this, bounds);
        synchronized(boardPieceListeners) {
            for(IBoardPieceListener<P> listener : boardPieceListeners) {
                listener.boardPieceChanged(event);
            }
        }
    }

}
