/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.games.view;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.MouseTrackAdapter;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.Pattern;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.jdesktop.animation.timing.TimingTarget;
import org.jdesktop.animation.timing.TimingTargetAdapter;
import org.jdesktop.animation.timing.interpolation.Interpolator;

import com.haundrix.animation.CompositeFadeCallback;
import com.haundrix.animation.CompositeMoveCallback;
import com.haundrix.animation.FadeCallback;
import com.haundrix.animation.Fader;
import com.haundrix.animation.MoveCallback;
import com.haundrix.animation.Mover;
import com.haundrix.animation.SwtAnimator;
import com.haundrix.games.view.SquareClickEvent.MouseButton;
import com.haundrix.graphics.Highlight;
import com.haundrix.graphics.IBorder;
import com.haundrix.graphics.IBorderDescriptor;
import com.haundrix.graphics.SimpleBorder;
import com.haundrix.utils.IDisposeListener;
import com.haundrix.utils.IDisposable;

public class RectangularBoardControl<P extends IPieceType> extends Canvas implements Iterable<BoardSquare> {
    
    private static final float FADE_IN_ACCELERATION = 0.2F;
    private static final float FADE_IN_DECELERATION = 0.3F;

    private static final float FADE_OUT_ACCELERATION = 0.3F;
    private static final float FADE_OUT_DECELERATION = 0.2F;

    private static final int DEFAULT_MOVE_DURATION = 300;
    private static final int DEFAULT_FADE_DURATION = 500;
    
    private Color topBackgroundColor;
    private Color bottomBackgroundColor;
    private Pattern backgroundPattern;
    
    private IBorder border;
    private double borderFracThickness = 0.1;
    
    private int columnCount;
    private int rowCount;
    private Color lightColor;
    private Color darkColor;
    
    private PieceSet<P> pieceSet; 
    
    private boolean isFlipped = false;

    private int preferredSquareSize = 40;

    private int leftMargin;
    private int topMargin;
    private int squareSize = preferredSquareSize;
    
    private BoardPiece<P>[][] staticPieces;
    private IBoardPieceListener<P> boardPieceListener = new InternalBoardPieceListener();
    private IDisposeListener disposePieceListener = new InternalDisposePieceListener();
    private Highlight[][] squareHighlights;

    private BoardSquare previousPointedSquare = null;
    
    private boolean isMouseDown = false;
    private MouseButton pressedButton = null;
    private Point clickedLocation = null;
    private boolean firstMove = true;
    
    private List<BoardPiece<P>> movingPieces = new ArrayList<BoardPiece<P>>();
    
    private boolean isDraggingPiece = false;
    private boolean isHandlingDraggingPiece = false;
    private BoardSquare draggingStartSquare = null;
    private BoardPiece<P> draggingPiece = null;
    private IDisposeListener draggingPieceDisposeListener = new IDisposeListener() {
        @Override
        public void disposed(IDisposable disposable) {
            // just in case the dragging piece gets disposed on the way!
            clearDraggingStatus();
        }
    };
    
    private List<IBoardListener<P>> listeners = new ArrayList<IBoardListener<P>>();

    @SuppressWarnings("unchecked")
    public RectangularBoardControl(Composite parent, int columnCount, int rowCount, PieceSet<P> pieceSet) {
        super(parent, SWT.NO_BACKGROUND | SWT.DOUBLE_BUFFERED);
        
        topBackgroundColor = new Color(getDisplay(), 197, 218, 247);
        bottomBackgroundColor = new Color(getDisplay(), 159, 181, 205);
        
        RGB borderColor = new RGB(255, 140, 0);
        border = new SimpleBorder(getDisplay(), getBounds(), 10, borderColor);
        
        this.columnCount = columnCount;
        this.rowCount = rowCount;
        lightColor = new Color(getDisplay(), 255, 206, 158);
        darkColor = new Color(getDisplay(), 209, 139, 71);
        
        this.pieceSet = pieceSet;
        
        staticPieces = new BoardPiece[rowCount][columnCount];
        squareHighlights = new Highlight[rowCount][columnCount];
        
        addPaintListener(new PaintListener() {
            @Override
            public void paintControl(PaintEvent e) {
                onPaintControl(e);
            }
        });
        
        addControlListener(new ControlAdapter() {
            @Override
            public void controlResized(ControlEvent e) {
                onControlResized(e);
            }
        });
        
        addMouseListener(new MouseAdapter() {
            @Override
            public void mouseDown(MouseEvent e) {
                onMouseDown(e);
            }

            @Override
            public void mouseUp(MouseEvent e) {
                onMouseUp(e);
            }
        });

        addMouseMoveListener(new MouseMoveListener() {
            @Override
            public void mouseMove(MouseEvent e) {
                onMouseMove(e);
            }
        });
        
        addMouseTrackListener(new MouseTrackAdapter() {
            @Override
            public void mouseHover(MouseEvent e) {
                onMouseHover(e);
            }
        });
        
        addDisposeListener(new DisposeListener() {
            @Override
            public void widgetDisposed(DisposeEvent e) {
                onDispose();
            }
        });
    }

    public int getColumnCount() {
        return columnCount;
    }

    public int getRowCount() {
        return rowCount;
    }
    
    public int getPreferredSquareSize() {
        return preferredSquareSize;
    }

    public void setPreferredSquareSize(int preferredSquareSize) {
        this.preferredSquareSize = preferredSquareSize;
    }

    public RGB getLightColor() {
        return lightColor.getRGB();
    }
    
    public void setLightColor(RGB lightColor) {
        if(!lightColor.equals(this.lightColor.getRGB())) {
            this.lightColor.dispose();
            this.lightColor = new Color(getDisplay(), lightColor);
            Rectangle bounds = getBoardBounds();
            redraw(bounds.x, bounds.y, bounds.width, bounds.height, true);
        }
    }
    
    public RGB getDarkColor() {
        return darkColor.getRGB();
    }
    
    public void setDarkColor(RGB darkColor) {
        if(!darkColor.equals(this.darkColor.getRGB())) {
            this.darkColor.dispose();
            this.darkColor = new Color(getDisplay(), darkColor);
            Rectangle bounds = getBoardBounds();
            redraw(bounds.x, bounds.y, bounds.width, bounds.height, true);
        }
    }
    
    public double getBorderFracThickness() {
        return borderFracThickness;
    }
    
    public void setBorderFracThickness(double borderFracThickness) {
        this.borderFracThickness = borderFracThickness;
        resize();
        redraw();
    }
    
    public IBorder getBorder() {
        return border;
    }

    public void setBorder(IBorderDescriptor border) {
        int thickness = this.border.getThickness();
        this.border.dispose();
        this.border = border.createBorder(getDisplay(), getBoardBounds());
        this.border.setThickness(thickness);
        redraw();
    }

    public RGB getTopBackgroundColor() {
        return topBackgroundColor.getRGB();
    }
    
    public void setTopBackgroundColor(RGB topBackgroundColor) {
        if(!topBackgroundColor.equals(this.topBackgroundColor.getRGB())) {
            this.topBackgroundColor.dispose();
            this.topBackgroundColor = new Color(getDisplay(), topBackgroundColor);
            calculateBackgroundPattern();
            redraw();
        }
    }
    
    public RGB getBottomBackgroundColor() {
        return bottomBackgroundColor.getRGB();
    }
    
    public void setBottomBackgroundColor(RGB bottomBackgroundColor) {
        if(!bottomBackgroundColor.equals(this.bottomBackgroundColor.getRGB())) {
            this.bottomBackgroundColor.dispose();
            this.bottomBackgroundColor = new Color(getDisplay(), bottomBackgroundColor);
            calculateBackgroundPattern();
            redraw();
        }
    }
    
    public ColorScheme getColorScheme() {
        return new ColorScheme(lightColor.getRGB(), darkColor.getRGB(), border.getColor(),
            topBackgroundColor.getRGB(), bottomBackgroundColor.getRGB());
    }
    
    public void setColorScheme(ColorScheme colorScheme) {
        setLightColor(colorScheme.getLightSquareColor());
        setDarkColor(colorScheme.getDarkSquareColor());
        if(!colorScheme.getBorderColor().equals(border.getColor())) {
            border.setColor(colorScheme.getBorderColor());
            redraw();
        }
        setTopBackgroundColor(colorScheme.getTopBackgroundColor());
        setBottomBackgroundColor(colorScheme.getBottomBackgroundColor());
    }
    
    public PieceSet<P> getPieceSet() {
        return pieceSet;
    }
    
    public void setPieceSet(PieceSet<P> pieceSet) {
        this.pieceSet = pieceSet;
        for(BoardPiece<P> staticPiece : getStaticPieces()) {
            staticPiece.setPieceSet(pieceSet);
        }
        for(BoardPiece<P> movingPiece : movingPieces) {
            movingPiece.setPieceSet(pieceSet);
        }
        if(isDraggingPiece) {
            draggingPiece.setPieceSet(pieceSet);
        }
    }
    
    public Highlight getSquareHighlight(BoardSquare square) {
        return squareHighlights[square.getRow()][square.getColumn()];
    }
    
    public void setSquareHighlight(BoardSquare square, Highlight highlight) {
        Highlight currentHighlight = squareHighlights[square.getRow()][square.getColumn()];
        boolean settingToNull = (highlight == null) && (currentHighlight != null);
        boolean changingHighlight = (highlight != null) && !highlight.equals(currentHighlight); 
        if(settingToNull || changingHighlight) {
            squareHighlights[square.getRow()][square.getColumn()] = highlight;
            Point squareLocation = getSquareLocation(square);
            redraw(squareLocation.x, squareLocation.y, squareSize, squareSize, true);
        }
    }
    
    public P getPiece(BoardSquare square) {
        return getPiece(square.getRow(), square.getColumn());
    }
    
    public P getPiece(int row, int column) {
        BoardPiece<P> staticPiece = staticPieces[row][column];
        return staticPiece == null ? null : staticPiece.getPiece();
    }
    
    public void setPiece(P piece, BoardSquare square) {
        setPiece(piece, square.getRow(), square.getColumn(), null);
    }
    
    public void setPiece(P piece, int row, int column) {
        setPiece(piece, row, column, null);
    }
    
    /*
     * Setting the piece to null removes any existing piece from the specified location.
     */
    public void setPiece(P piece, int row, int column, IPieceConfigurator<P> pieceConfigurator) {
        BoardPiece<P> staticPiece = staticPieces[row][column];
        if(staticPiece != null) {
            staticPiece.dispose();
            staticPieces[row][column] = null;
        }
        
        if(piece != null) {
            Point pieceLocation = getSquareLocation(row, column);
            BoardPiece<P> boardPiece = new BoardPiece<P>(getDisplay(), piece, pieceSet, squareSize, pieceLocation);
            if(pieceConfigurator != null) {
                pieceConfigurator.configure(boardPiece);
            }
            boardPiece.addBoardPieceListener(boardPieceListener);
            boardPiece.addDisposeListener(disposePieceListener);
            staticPieces[row][column] = boardPiece;
            Rectangle pieceBounds = boardPiece.getBounds();
            redraw(pieceBounds.x, pieceBounds.y, pieceBounds.width, pieceBounds.height, true);
        } else {
            staticPieces[row][column] = null;
        }
    }
    
    public BoardPiece<P> getBoardPiece(BoardSquare square) {
        return getBoardPiece(square.getRow(), square.getColumn());
    }

    public BoardPiece<P> getBoardPiece(int row, int column) {
        return staticPieces[row][column];
    }
    
    @Override
    public Iterator<BoardSquare> iterator() {
        return new Iterator<BoardSquare>() {
            
            private int row = 0;
            private int column = 0;

            @Override
            public boolean hasNext() {
                return (row < rowCount) && (column < columnCount);
            }

            @Override
            public BoardSquare next() {
                BoardSquare boardSquare = new BoardSquare(row, column);

                column++;
                if(column >= columnCount) {
                    column = 0;
                    row++;
                }
                
                return boardSquare;
            }

            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
            
        };
    }

    public int getSquareSize() {
        return squareSize;
    }

    public Point getSquareLocation(BoardSquare square) {
        int columnOffset = !isFlipped ? square.getColumn() : columnCount - 1 - square.getColumn();
        int rowOffset = !isFlipped ? square.getRow() : rowCount - 1 - square.getRow();
        
        int x = leftMargin + columnOffset * squareSize;
        int y = topMargin + rowOffset * squareSize;
        
        return new Point(x, y);
    }
    
    public boolean isFlipped() {
        return isFlipped;
    }
    
    public void flip() {
        isFlipped = !isFlipped;
        
        for(BoardSquare square : this) {
            BoardPiece<P> piece = getBoardPiece(square);
            if(piece != null) {
                piece.setLocation(getSquareLocation(square));
            }
        }
        
        Point topLeft = new Point(leftMargin, topMargin);
        Point bottomRight = new Point(leftMargin + columnCount * squareSize, topMargin + rowCount * squareSize);
        IBoardTransform coordinateTransform = new FlipBoardTransform(topLeft, bottomRight, squareSize);
        for(BoardPiece<P> movingPiece : movingPieces) {
            movingPiece.setCoordinateTransform(coordinateTransform);
        }
        if(isHandlingDraggingPiece) {
            draggingPiece.setCoordinateTransform(coordinateTransform);
        }
        
        if(previousPointedSquare != null) {
            notifySquareLeft(previousPointedSquare);
            int newRow = rowCount - 1 - previousPointedSquare.getRow();
            int newColumn = columnCount - 1 - previousPointedSquare.getColumn();
            BoardSquare newSquare = new BoardSquare(newRow, newColumn);
            notifySquareEntered(newSquare);
            previousPointedSquare = newSquare;
        }
        
        Rectangle bounds = getBoardBounds();
        redraw(bounds.x, bounds.y, bounds.width, bounds.height, true);
    }

    public ImageData getBoardImage() {
        // draws an image with the entire control
        Image controlImage = new Image(getDisplay(), getSize().x, getSize().y);
        GC controlImageGc = new GC(controlImage);
        paintControl(controlImageGc);
        controlImageGc.dispose();
        
        // copies just the board and margins to a smaller image
        int borderThickness = border.getThickness();
        int width = getBoardBounds().width + 2 * borderThickness;
        int height = getBoardBounds().height + 2 * borderThickness;
        Image boardImage = new Image(getDisplay(), width, height);
        GC boardImageGc = new GC(boardImage);
        boardImageGc.drawImage(controlImage, leftMargin - borderThickness, topMargin - borderThickness, width, height,
                0, 0, width, height);
        boardImageGc.dispose();
        
        ImageData boardImageData = boardImage.getImageData();
        boardImage.dispose();
        controlImage.dispose();
        
        return boardImageData;
    }

    public void fadeInPiece(P piece, BoardSquare square, int fadeDuration) {
        fadeInPiece(piece, square, fadeDuration, null, new FadeCallback());
    }
    
    public void fadeInPiece(P piece, BoardSquare square, int fadeDuration,
            IPieceConfigurator<P> pieceConfigurator) {
        fadeInPiece(piece, square, fadeDuration, pieceConfigurator, new FadeCallback());
    }

    public void fadeInPiece(P piece, BoardSquare square, int fadeDuration,
            IPieceConfigurator<P> pieceConfigurator, Runnable endFadeCallback) {
        fadeInPiece(piece, square, fadeDuration, pieceConfigurator, new FadeCallback(endFadeCallback));
    }
    
    public void fadeInPiece(P piece, BoardSquare square, int fadeDuration, IPieceConfigurator<P> pieceConfigurator,
            FadeCallback fadeCallback) {
        if(piece != null) {
            BoardPiece<P> newPiece = preparePieceFadeIn(piece, square, pieceConfigurator);
            
            Fader fader = new Fader(getDisplay(), newPiece, 1.0F, fadeDuration, fadeCallback);
            fader.setAcceleration(FADE_IN_ACCELERATION);
            fader.setDeceleration(FADE_IN_DECELERATION);
            fader.fade();
        }
    }

    public void fadeOutPiece(BoardSquare square, int fadeDuration) {
        fadeOutPiece(square, fadeDuration, new FadeCallback());
    }

    public void fadeOutPiece(BoardSquare square, int fadeDuration, Runnable endFadeCallback) {
        fadeOutPiece(square, fadeDuration, new FadeCallback(endFadeCallback));
    }
    
    public void fadeOutPiece(BoardSquare square, int fadeDuration, FadeCallback fadeCallback) {
        final BoardPiece<P> boardPiece = getBoardPiece(square);
        if(boardPiece != null) {
            FadeCallback usedFadeCallback = preparePieceFadeOut(square, fadeCallback);
            
            Fader fader = new Fader(getDisplay(), boardPiece, 0.0F, fadeDuration, usedFadeCallback);
            fader.setAcceleration(FADE_OUT_ACCELERATION);
            fader.setDeceleration(FADE_OUT_DECELERATION);
            fader.fade();
        }
    }

    public void movePiece(BoardSquare origin, final BoardSquare destination, int movementDuration) {
        movePiece(origin, destination, movementDuration, null, null);
    }
    
    public void movePiece(BoardSquare origin, final BoardSquare destination, int movementDuration,
            Interpolator interpolator) {
        movePiece(origin, destination, movementDuration, interpolator, null);
    }

    public void movePiece(BoardSquare origin, final BoardSquare destination, int movementDuration,
            Runnable endMoveCallback) {
        movePiece(origin, destination, movementDuration, null, new MoveCallback(endMoveCallback));
    }

    public void movePiece(BoardSquare origin, final BoardSquare destination, int movementDuration,
            MoveCallback moveCallback) {
        movePiece(origin, destination, movementDuration, null, moveCallback);
    }
    
    public void movePiece(BoardSquare origin, final BoardSquare destination, int movementDuration,
            Interpolator interpolator, MoveCallback moveCallback) {
        BoardPiece<P> movingPiece = getBoardPiece(origin);
        if((movingPiece != null) && (!origin.equals(destination))) {
            MoveCallback usedMoveCallback = beginPreparationOfStaticPieceMovement(origin, destination, moveCallback);
            
            movePiece(movingPiece, destination, movementDuration, interpolator, usedMoveCallback);
        }
    }
    
    @SuppressWarnings("unchecked")
    public void animate(IBoardAnimation[] animations, int duration, final Runnable endAnimationCallback) {
        List<TimingTarget> timingTargets = new ArrayList<TimingTarget>();
        
        // Important: move animations must be prepared before fade in animations.
        
        final List<MoveCallback> moveCallbacks = new ArrayList<MoveCallback>();
        for(IBoardAnimation animation : animations) {
            if(animation instanceof MoveAnimation) {
                prepareMoveAnimation((MoveAnimation) animation, moveCallbacks, timingTargets);
            }
        }
        
        final List<FadeCallback> fadeCallbacks = new ArrayList<FadeCallback>();
        for(IBoardAnimation animation : animations) {
            if(animation instanceof FadeInAnimation) {
                prepareFadeInAnimation((FadeInAnimation<P>) animation, timingTargets);
            } else if(animation instanceof FadeOutAnimation) {
                prepareFadeOutAnimation((FadeOutAnimation) animation, fadeCallbacks, timingTargets);
            } else if(animation instanceof CrossfadeAnimation) {
                prepareCrossfadeAnimation((CrossfadeAnimation<P>) animation, fadeCallbacks, timingTargets);
            }
        }
        
        final SwtAnimator animator = new SwtAnimator(getDisplay(), duration);

        final DisposeListener disposeListener = new DisposeListener() {
            @Override
            public void widgetDisposed(DisposeEvent e) {
                animator.cancel();
            }
        };
        addDisposeListener(disposeListener);
        
        TimingTarget cleanupTimingTarget = new TimingTargetAdapter() {
            @Override
            public void end() {
                // Important: fade out callbacks must be called before move callbacks.
                for(FadeCallback fadeCallback : fadeCallbacks) {
                    fadeCallback.ended();
                }
                for(MoveCallback moveCallback : moveCallbacks) {
                    moveCallback.ended(null);
                }
                endAnimationCallback.run();
                
                removeDisposeListener(disposeListener);
            }
        };
        
        for(TimingTarget timingTarget : timingTargets) {
            animator.addTarget(timingTarget);
        }
        animator.addTarget(cleanupTimingTarget);
        
        animator.start();
    }

    public boolean isDraggingPiece() {
        return isDraggingPiece;
    }
    
    public boolean isHandlingDraggingPiece() {
        return isHandlingDraggingPiece;
    }
    
    public BoardSquare getDraggingStartSquare() {
        return draggingStartSquare;
    }
    
    public BoardPiece<P> getDraggingPiece() {
        return draggingPiece;
    }

    /*
     * The dragging piece is simply disposed of from the board and its corresponding
     * piece drag and piece drop events will no longer be fired.
     */
    public void cancelPieceDrag() {
        if(isDraggingPiece) {
            isHandlingDraggingPiece = true;
            draggingPiece.dispose();
        }
    }
    
    /*
     * The dragging piece is moved back to its original location and its corresponding
     * piece drag and piece drop events will no longer be fired.
     */
    public void cancelPieceDragAndMovePieceToOriginalSquare(int movementDuration, Runnable endMoveCallback) {
        if(isDraggingPiece) {
            isHandlingDraggingPiece = true;
            new InternalDraggingPieceHandler(null).moveBackToOriginalSquare(movementDuration, endMoveCallback);
        }
    }
    
    /*
     * The dragging piece is moved back to the specified location and its corresponding
     * piece drag and piece drop events will no longer be fired.
     */
    public void cancelPieceDragAndMovePieceTo(int movementDuration, BoardSquare targetSquare, Runnable endMoveCallback) {
        if(isDraggingPiece) {
            isHandlingDraggingPiece = true;
            new InternalDraggingPieceHandler(targetSquare).moveToTargetSquare(movementDuration, endMoveCallback);
        }
    }
    
    /*
     * The dragging piece is faded out and its corresponding
     * piece drag and piece drop events will no longer be fired.
     */
    public void cancelPieceDragAndFadeOutPiece(int fadeDuration, Runnable endFadeCallback) {
        if(isDraggingPiece) {
            isHandlingDraggingPiece = true;
            new InternalDraggingPieceHandler(null).fadeOut(fadeDuration, endFadeCallback);
        }
    }

    public void addBoardListener(IBoardListener<P> listener) {
        synchronized(listeners) {
            if(!listeners.contains(listener)) {
                listeners.add(listener);
            }
        }
    }
    
    public void removeBoardListener(IBoardListener<P> listener) {
        synchronized(listeners) {
            listeners.remove(listener);
        }
    }

    @Override
    public Point computeSize(int wHint, int hHint, boolean changed) {
        int squareSize;
        if((wHint == SWT.DEFAULT) && (hHint != SWT.DEFAULT)) {
            squareSize = getMaxSquareHeight(hHint);
        } else if((wHint != SWT.DEFAULT) && (hHint == SWT.DEFAULT)) {
            squareSize = getMaxSquareWidth(wHint);
        } else if((wHint != SWT.DEFAULT) && (hHint != SWT.DEFAULT)) {
            squareSize = getMaxSquareSize(wHint, hHint);
        } else {
            squareSize = preferredSquareSize;
        }
        
        int borderThickness = (int) (squareSize * borderFracThickness);
        int width = squareSize * columnCount + 2 * borderThickness;
        int height = squareSize * rowCount + 2 * borderThickness;
        return new Point(width, height);
    }
    
    private MouseButton getClickedButton(MouseEvent e) {
        if(e.button == 1) {
            return MouseButton.LEFT_BUTTON;
        } else if(e.button == 3) {
            return MouseButton.RIGHT_BUTTON;
        } else {
            return null;
        }
    }

    private int getMaxSquareWidth(int width) {
        return (int) (width / (columnCount + 2.0 * borderFracThickness));
    }
    
    private int getMaxSquareHeight(int height) {
        return (int) (height / (rowCount + 2.0 * borderFracThickness));
    }
    
    private int getMaxSquareSize(int width, int height) {
        int maxSquareWidth = getMaxSquareWidth(width);
        int maxSquareHeight = getMaxSquareHeight(height);
        return Math.min(maxSquareWidth, maxSquareHeight);
    }

    private void calculateBackgroundPattern() {
        if(backgroundPattern != null) {
            backgroundPattern.dispose();
        }
        backgroundPattern = new Pattern(getDisplay(), 0.0F, 0.0F, 0.0F, getClientArea().height,
                topBackgroundColor, bottomBackgroundColor);
    }
    
    private Rectangle getBoardBounds() {
        return new Rectangle(leftMargin, topMargin, squareSize * columnCount, squareSize * rowCount);
    }
    
    private BoardSquare getSquare(int x, int y) {
        if((x < leftMargin) || (y < topMargin)) {
            return null;
        }

        int row = (y - topMargin) / squareSize;
        int column = (x - leftMargin) / squareSize;

        if((row >= rowCount) || (column >= columnCount)) {
            return null;
        }
        
        if(isFlipped) {
            row = rowCount - 1 - row;
            column = columnCount - 1 - column;
        }

        return new BoardSquare(row, column);
    }

    private Point getSquareLocation(int row, int column) {
        return getSquareLocation(new BoardSquare(row, column));
    }

    private BoardSquare getPieceSquare(BoardPiece<P> boardPiece) {
        for(BoardSquare boardSquare : this) {
            if(getBoardPiece(boardSquare) == boardPiece) {
                return boardSquare;
            }
        }

        return null;
    }

    private List<BoardPiece<P>> getStaticPieces() {
        List<BoardPiece<P>> staticPieces = new ArrayList<BoardPiece<P>>();
        for(BoardSquare boardSquare : this) {
            BoardPiece<P> staticPiece = getBoardPiece(boardSquare);
            if(staticPiece != null) {
                staticPieces.add(staticPiece);
            }
        }
        
        return staticPieces;
    }
    
    private BoardPiece<P> preparePieceFadeIn(P piece, BoardSquare square, IPieceConfigurator<P> pieceConfigurator) {
        BoardPiece<P> newPiece = null;
        
        if(piece != null) {
            BoardPiece<P> existingPiece = getBoardPiece(square);
            if(existingPiece != null) {
                existingPiece.dispose();
                staticPieces[square.getRow()][square.getColumn()] = null;
            }
            
            Point newPieceLocation = getSquareLocation(square);
            newPiece = new BoardPiece<P>(getDisplay(), piece, pieceSet, squareSize, newPieceLocation);
            if(pieceConfigurator != null) {
                pieceConfigurator.configure(newPiece);
            }
            newPiece.setAlpha(0.0F);
            newPiece.addBoardPieceListener(boardPieceListener);
            newPiece.addDisposeListener(disposePieceListener);
            staticPieces[square.getRow()][square.getColumn()] = newPiece;
            Rectangle pieceBounds = newPiece.getBounds();
            redraw(pieceBounds.x, pieceBounds.y, pieceBounds.width, pieceBounds.height, true);
        }
        
        return newPiece;
    }

    private FadeCallback preparePieceFadeOut(BoardSquare square, FadeCallback fadeCallback) {
        FadeCallback usedFadeCallback = null;
        
        final BoardPiece<P> boardPiece = getBoardPiece(square);
        if(boardPiece != null) {
            FadeCallback removePieceCallback = new FadeCallback() {
                @Override
                public void ended() {
                    BoardSquare pieceSquare = getPieceSquare(boardPiece);
                    boardPiece.dispose();
                    if(pieceSquare != null) {
                        staticPieces[pieceSquare.getRow()][pieceSquare.getColumn()] = null;
                    }
                }
            };
            usedFadeCallback = fadeCallback == null ? removePieceCallback :
                new CompositeFadeCallback(removePieceCallback, fadeCallback);
        }
        
        return usedFadeCallback;
    }

    private MoveCallback beginPreparationOfStaticPieceMovement(BoardSquare origin, BoardSquare destination,
            MoveCallback moveCallback) {
        MoveCallback usedMoveCallback = null;
        
        final BoardPiece<P> boardPiece = getBoardPiece(origin);
        if((boardPiece != null) && (!origin.equals(destination))) {
            MoveCallback finishMoveCallback = new MoveCallback() {

                @Override
                public void movableDisposed(Mover mover, IDisposable disposable) {
                    movingPieces.remove(boardPiece);
                }
                
                @Override
                public void ended(Mover mover) {
                    movingPieces.remove(boardPiece);
                }

            };
            usedMoveCallback = moveCallback == null ?
                    finishMoveCallback :
                    new CompositeMoveCallback(finishMoveCallback, moveCallback);
            staticPieces[origin.getRow()][origin.getColumn()] = null;
            movingPieces.add(boardPiece);
        }
        
        return usedMoveCallback;
    }

    private MoveCallback finishPreparationOfPieceMovement(final BoardPiece<P> movingPiece, final BoardSquare destination,
            MoveCallback moveCallback) {
        MoveCallback completeMoveCallback = new MoveCallback() {
            @Override
            public void ended(Mover mover) {
                BoardPiece<P> targetPiece = staticPieces[destination.getRow()][destination.getColumn()];
                if(targetPiece != null) {
                    targetPiece.dispose();
                }

                staticPieces[destination.getRow()][destination.getColumn()] = movingPiece;
                movingPiece.clearCoordinateTransform();
                Rectangle pieceBounds = movingPiece.getBounds();
                redraw(pieceBounds.x, pieceBounds.y, pieceBounds.width, pieceBounds.height, true);
            }
        };
        
        return moveCallback == null ? completeMoveCallback : new CompositeMoveCallback(completeMoveCallback, moveCallback);
    }

    private void prepareMoveAnimation(MoveAnimation moveAnimation, List<MoveCallback> moveCallbacks,
            List<TimingTarget> timingTargets) {
        BoardSquare origin = moveAnimation.getOrigin();
        BoardSquare destination = moveAnimation.getDestination();
        BoardPiece<P> movingPiece = getBoardPiece(origin);
        if((movingPiece != null) && (!origin.equals(destination))) {
            MoveCallback moveCallback = beginPreparationOfStaticPieceMovement(origin, destination, null);
            MoveCallback usedMoveCallback = finishPreparationOfPieceMovement(movingPiece, destination, moveCallback);
            moveCallbacks.add(usedMoveCallback);
            
            TimingTarget timingTarget = Mover.createTimingTarget(movingPiece, getSquareLocation(destination));
            timingTargets.add(timingTarget);
        }
    }
    
    private void prepareFadeInAnimation(FadeInAnimation<P> fadeInAnimation, List<TimingTarget> timingTargets) {
        BoardPiece<P> fadingInPiece = preparePieceFadeIn(fadeInAnimation.getPiece(),
                fadeInAnimation.getSquare(), null);
        if(fadingInPiece != null) {
            TimingTarget timingTarget = Fader.createTimingTarget(fadingInPiece, 1.0F, FADE_IN_ACCELERATION,
                    FADE_IN_DECELERATION);
            timingTargets.add(timingTarget);
        }
    }

    private void prepareFadeOutAnimation(FadeOutAnimation fadeOutAnimation, List<FadeCallback> fadeCallbacks,
            List<TimingTarget> timingTargets) {
        FadeCallback fadeOutCallback = preparePieceFadeOut(fadeOutAnimation.getSquare(), null);
        if(fadeOutCallback != null) {
            fadeCallbacks.add(fadeOutCallback);
            
            BoardPiece<P> fadingOutPiece = getBoardPiece(fadeOutAnimation.getSquare());
            TimingTarget timingTarget = Fader.createTimingTarget(fadingOutPiece, 0.0F, FADE_OUT_ACCELERATION,
                    FADE_OUT_DECELERATION);
            timingTargets.add(timingTarget);
        }
    }

    private void prepareCrossfadeAnimation(CrossfadeAnimation<P> crossfadeAnimation, List<FadeCallback> fadeCallbacks,
            List<TimingTarget> timingTargets) {
        BoardPiece<P> crossfadingPiece = getBoardPiece(crossfadeAnimation.getSquare());
        P newPiece = crossfadeAnimation.getNewPiece();
        if((crossfadingPiece != null) && (newPiece != crossfadingPiece.getPiece())) {
            FadeCallback crossfadeCallback = crossfadingPiece.prepareCrossfade(newPiece, null);
            fadeCallbacks.add(crossfadeCallback);
            
            TimingTarget timingTarget = crossfadingPiece.createCrossfadeTimingTarget(newPiece);
            timingTargets.add(timingTarget);
        }
    }

    private void clearDraggingStatus() {
        isDraggingPiece = false;
        isHandlingDraggingPiece = false;
        draggingStartSquare = null;
        
        draggingPiece.clearCoordinateTransform();
        draggingPiece = null;
    }

    private void resize() {
        int width = getSize().x;
        int height = getSize().y;
        
        if((width == 0) && (height == 0)) {
            // There's no need to treat this special case,
            // as usually the size will be reset to a meaningful value shortly.
            // Let's just ignore this resize.
            return;
        }

        int originalSquareSize = squareSize;
        int originalLeftMargin = leftMargin;
        int originalTopMargin = topMargin;
        
        squareSize = getMaxSquareSize(width, height);
        leftMargin = (width - squareSize * columnCount) / 2;
        topMargin = (height - squareSize * rowCount) / 2;
        
        calculateBackgroundPattern();
        
        int borderThickness = (int) (borderFracThickness * squareSize);
        border.setThickness(borderThickness);
        border.setInset(getBoardBounds());
        
        for(BoardSquare boardSquare : this) {
            BoardPiece<P> boardPiece = getBoardPiece(boardSquare);
            if(boardPiece != null) {
                boardPiece.setSize(squareSize);
                boardPiece.setLocation(getSquareLocation(boardSquare));
            }
        }
        
        IBoardTransform coordinateTransform = new ResizeBoardTransform(originalTopMargin, originalLeftMargin,
                originalSquareSize, topMargin, leftMargin, squareSize);
        for(BoardPiece<P> movingPiece : movingPieces) {
            movingPiece.setSize(squareSize);
            movingPiece.setCoordinateTransform(coordinateTransform);
        }
        
        if(draggingPiece != null) {
            draggingPiece.setSize(squareSize);
            
            if(isHandlingDraggingPiece) {
                draggingPiece.setCoordinateTransform(coordinateTransform);
            } else {
                // re-centers the dragging piece on the cursor location
                Point location = draggingPiece.getLocation();
                int newX = location.x + (originalSquareSize - squareSize) / 2;
                int newY = location.y + (originalSquareSize - squareSize) / 2;
                draggingPiece.setLocation(new Point(newX, newY));
            }
        }
    }

    private void paintControl(GC gc) {
        gc.setAdvanced(true);
        gc.setAntialias(SWT.ON);
        
        Rectangle clipping = gc.getClipping();
        Rectangle boardArea = getBoardBounds();
        Rectangle clippedBoardArea = clipping.intersection(boardArea);
        boolean clippedBoardAreaIsEmpty = (clippedBoardArea == null) || (clippedBoardArea.width == 0) ||
            (clippedBoardArea.height == 0); 
        
        if(clippedBoardAreaIsEmpty || !clippedBoardArea.intersection(clipping).equals(clipping)) {
            // draws the background
            gc.setBackgroundPattern(backgroundPattern);
            gc.fillRectangle(clipping);
            
            // draws the border
            border.draw(gc);
        }
        
        // draws the board with the pieces
        if(!clippedBoardAreaIsEmpty) {
            BoardSquare topLeftSquare = getSquare(clippedBoardArea.x, clippedBoardArea.y);
            BoardSquare bottomRightSquare = getSquare(clippedBoardArea.x + clippedBoardArea.width - 1,
                    clippedBoardArea.y + clippedBoardArea.height - 1);
            BoardSquare startSquare = !isFlipped ? topLeftSquare : bottomRightSquare;
            BoardSquare endSquare = !isFlipped ? bottomRightSquare : topLeftSquare;
            for(int column = startSquare.getColumn(); column <= endSquare.getColumn(); column++) {
                for(int row = startSquare.getRow(); row <= endSquare.getRow(); row++) {
                    Color squareColor = (column + row) % 2 == 0 ?
                            lightColor : darkColor;
                    gc.setBackground(squareColor);
                    Point squareLocation = getSquareLocation(row, column);
                    gc.fillRectangle(squareLocation.x, squareLocation.y, squareSize, squareSize);
                    Highlight squareHighlight = squareHighlights[row][column];
                    if(squareHighlight != null) {
                        Color highlightColor = new Color(getDisplay(), squareHighlight.getColor());
                        gc.setBackground(highlightColor);
                        int previousAlpha = gc.getAlpha();
                        gc.setAlpha((int) (squareHighlight.getIntensity() * 255));
                        gc.fillRectangle(squareLocation.x, squareLocation.y, squareSize, squareSize);
                        gc.setAlpha(previousAlpha);
                        highlightColor.dispose();
                    }
                    BoardPiece<P> boardPiece = staticPieces[row][column];
                    if(boardPiece != null) {
                        Image pieceImage = boardPiece.getImage();
                        gc.drawImage(pieceImage, squareLocation.x, squareLocation.y);
                    }
                }
            }
        }
        
        // draws the moving pieces
        for(BoardPiece<P> movingPiece : movingPieces) {
            gc.drawImage(movingPiece.getImage(), movingPiece.getLocation().x, movingPiece.getLocation().y);
        }

        // draws the dragging piece
        if(isDraggingPiece) {
            gc.drawImage(draggingPiece.getImage(), draggingPiece.getLocation().x, draggingPiece.getLocation().y);
        }
    }
    
    private void movePiece(final BoardPiece<P> movingPiece, final BoardSquare destination,
            int movementDuration, Interpolator interpolator, MoveCallback moveCallback) {
        MoveCallback usedMoveCallback = finishPreparationOfPieceMovement(movingPiece, destination, moveCallback);
        
        Mover mover = new Mover(getDisplay(), movingPiece, getSquareLocation(destination), movementDuration,
                usedMoveCallback);
        if(interpolator != null) {
            mover.setInterpolator(interpolator);
        }
        mover.move();
    }
    
    private void notifyBoardLeft() {
        synchronized(listeners) {
            for(IBoardListener<P> listener : listeners) {
                listener.boardLeft();
            }
        }
    }

    private void notifyBoardEnter() {
        synchronized(listeners) {
            for(IBoardListener<P> listener : listeners) {
                listener.boardEntered();
            }
        }
    }

    private void notifySquareLeft(BoardSquare leftSquare) {
        BoardPiece<P> piece = staticPieces[leftSquare.getRow()][leftSquare.getColumn()];
        BoardSquareEvent<P> event = new BoardSquareEvent<P>(leftSquare, piece);
        synchronized(listeners) {
            for(IBoardListener<P> listener : listeners) {
                listener.squareLeft(event);
            }
        }
    }

    private void notifySquareEntered(BoardSquare enteredSquare) {
        BoardPiece<P> piece = staticPieces[enteredSquare.getRow()][enteredSquare.getColumn()];
        BoardSquareEvent<P> event = new BoardSquareEvent<P>(enteredSquare, piece);
        synchronized(listeners) {
            for(IBoardListener<P> listener : listeners) {
                listener.squareEntered(event);
            }
        }
    }
    
    private void notifySquareHovered(BoardSquare hoveredSquare) {
        if(hoveredSquare != null) {
            BoardPiece<P> piece = staticPieces[hoveredSquare.getRow()][hoveredSquare.getColumn()];
            BoardSquareEvent<P> event = new BoardSquareEvent<P>(hoveredSquare, piece);
            synchronized(listeners) {
                for(IBoardListener<P> listener : listeners) {
                    listener.squareHovered(event);
                }
            }
        }
    }

    private void notifySquareClicked(BoardSquare clickedSquare) {
        BoardPiece<P> piece = staticPieces[clickedSquare.getRow()][clickedSquare.getColumn()];
        SquareClickEvent<P> event = new SquareClickEvent<P>(clickedSquare, piece, pressedButton);
        synchronized(listeners) {
            for(IBoardListener<P> listener : listeners) {
                listener.squareClicked(event);
            }
        }
    }

    private void notifyPieceGrabbed(PieceGrabEvent<P> event) {
        synchronized(listeners) {
            for(IBoardListener<P> listener : listeners) {
                listener.pieceGrabbed(event);
            }
        }
    }

    private void notifyPieceDragged(BoardSquare hoveredSquare) {
        BoardPiece<P> hoveredPiece = hoveredSquare != null ?
                staticPieces[hoveredSquare.getRow()][hoveredSquare.getColumn()] :
                null;
        PieceDragEvent<P> event = new PieceDragEvent<P>(draggingPiece, draggingStartSquare,
                hoveredSquare, hoveredPiece);
        synchronized(listeners) {
            for(IBoardListener<P> listener : listeners) {
                listener.pieceDragged(event);
            }
        }
    }

    private void notifyPieceDropped(BoardSquare targetSquare, InternalDraggingPieceHandler draggingPieceHandler) {
        BoardPiece<P> targetPiece = targetSquare != null ?
                staticPieces[targetSquare.getRow()][targetSquare.getColumn()] :
                null;
        PieceDragEvent<P> event = new PieceDragEvent<P>(draggingPiece, draggingStartSquare, targetSquare,
                targetPiece);
        synchronized(listeners) {
            for(IBoardListener<P> listener : listeners) {
                listener.pieceDropped(event, draggingPieceHandler);
            }
        }
    }

    private void onControlResized(ControlEvent e) {
        resize();
    }

    private void onPaintControl(PaintEvent e) {
        paintControl(e.gc);
    }

    private void onMouseDown(MouseEvent e) {
        // right button clicks are only handled if there isn't a context menu attached to the control
        boolean isValidRightButton = (e.button == 3) && (getMenu() == null);
        
        boolean isValidButton = (e.button == 1) || isValidRightButton;
        if(!isMouseDown && isValidButton && (e.count == 1)) {
            BoardSquare square = getSquare(e.x, e.y);
            if(square != null) {
                isMouseDown = true;
                pressedButton = getClickedButton(e);
                clickedLocation = new Point(e.x, e.y);
            }
        }
    }

    private void onMouseUp(MouseEvent e) {
        MouseButton releasedButton = getClickedButton(e);
        if(isMouseDown && (releasedButton == pressedButton)) {
            if(isDraggingPiece && !isHandlingDraggingPiece) {
                isHandlingDraggingPiece = true;
                BoardSquare targetSquare = getSquare(draggingPiece.getLocation().x + squareSize / 2,
                        draggingPiece.getLocation().y + squareSize / 2);
                InternalDraggingPieceHandler draggingPieceHandler = new InternalDraggingPieceHandler(targetSquare);
                notifyPieceDropped(targetSquare, draggingPieceHandler);
                if(!draggingPieceHandler.called) {
                    if(targetSquare == null) {
                        draggingPieceHandler.fadeOut(DEFAULT_FADE_DURATION);
                    } else {
                        draggingPieceHandler.moveToTargetSquare(DEFAULT_MOVE_DURATION);
                    }
                }
            } else {
                Point currentLocation = new Point(e.x, e.y);
                if(currentLocation.equals(clickedLocation)) {
                    notifySquareClicked(getSquare(e.x, e.y));
                }
            }
            isMouseDown = false;
            pressedButton = null;
            clickedLocation = null;
            firstMove = true;
        }
    }

    private void onMouseMove(MouseEvent e) {
        BoardSquare hoveredSquare = getSquare(e.x, e.y);
        
        if(isMouseDown && (pressedButton == MouseButton.LEFT_BUTTON)) {
            // does not allow dragging a new piece if there's still a piece being dragged
            if(firstMove && !isDraggingPiece) {
                BoardSquare clickedSquare = getSquare(clickedLocation.x, clickedLocation.y);
                BoardPiece<P> boardPiece = staticPieces[clickedSquare.getRow()][clickedSquare.getColumn()];
                if(boardPiece != null) {
                    PieceGrabEvent<P> event = new PieceGrabEvent<P>(boardPiece, clickedSquare);
                    notifyPieceGrabbed(event);
                    if(event.allow) {
                        isDraggingPiece = true;
                        draggingStartSquare = clickedSquare;
                        draggingPiece = boardPiece;
                        draggingPiece.addDisposeListener(draggingPieceDisposeListener);
                        staticPieces[clickedSquare.getRow()][clickedSquare.getColumn()] = null;
                    }
                }
            }
            firstMove = false;
            if(isDraggingPiece && !isHandlingDraggingPiece) {
                draggingPiece.setLocation(new Point(e.x - squareSize / 2, e.y - squareSize / 2));
                notifyPieceDragged(hoveredSquare);
            }
        }
        
        // if we didn't trigger the piece dragged event, then we must notify
        // clients of the changed square through the standard move events
        if(!isDraggingPiece || isHandlingDraggingPiece) {
            // the move events are triggered in this sequence: 
            // square left > board left > board entered > square entered
            
            if((previousPointedSquare != null) && ((hoveredSquare == null) || (!hoveredSquare.equals(previousPointedSquare)))) {
                notifySquareLeft(previousPointedSquare);
            }
            
            if((hoveredSquare == null) && (previousPointedSquare != null)) {
                notifyBoardLeft();
            } else if((previousPointedSquare == null) && (hoveredSquare != null)) {
                notifyBoardEnter();
            }
            
            if((hoveredSquare != null) && ((previousPointedSquare == null) || (!previousPointedSquare.equals(hoveredSquare)))) {
                notifySquareEntered(hoveredSquare);
            }
            
            previousPointedSquare = hoveredSquare;
        }
    }

    private void onMouseHover(MouseEvent e) {
        // TODO what if there's a button pressed?
        notifySquareHovered(getSquare(e.x, e.y));
    }

    private void onDispose() {
        for(BoardPiece<P> staticPiece : getStaticPieces()) {
            staticPiece.dispose();
        }
        
        // We need to copy the moving pieces to a temporary list in order to dispose them,
        // since as they are automatically removed from the movingPieces list as they are
        // being disposed, that would cause a concurrent modification exception if we tried
        // to iterate directly over the original list here.
        List<BoardPiece<P>> theMovingPieces = new ArrayList<BoardPiece<P>>();
        theMovingPieces.addAll(movingPieces);
        for(BoardPiece<P> movingPiece : theMovingPieces) {
            movingPiece.dispose();
        }
        
        darkColor.dispose();
        lightColor.dispose();
        
        border.dispose();
        
        backgroundPattern.dispose();
        bottomBackgroundColor.dispose();
        topBackgroundColor.dispose();
    }
    
    private class InternalBoardPieceListener implements IBoardPieceListener<P> {

        @Override
        public void boardPieceChanged(BoardPieceChangedEvent<P> event) {
            Rectangle previousPieceBounds = event.getPreviousPieceBounds();
            redraw(previousPieceBounds.x, previousPieceBounds.y, previousPieceBounds.width,
                    previousPieceBounds.height, true);
            Rectangle pieceBounds = event.getPiece().getBounds();
            redraw(pieceBounds.x, pieceBounds.y, pieceBounds.width, pieceBounds.height, true);
        }

        @Override
        public void boardPieceMoved(BoardPieceMovedEvent<P> event) {
            Point previousLocation = event.getPreviousLocation();
            Rectangle pieceBounds = event.getPiece().getBounds();
            redraw(previousLocation.x, previousLocation.y, pieceBounds.width, pieceBounds.height, true);
            redraw(pieceBounds.x, pieceBounds.y, pieceBounds.width, pieceBounds.height, true);
        }

    }
    
    private class InternalDisposePieceListener implements IDisposeListener {

        @Override
        public void disposed(IDisposable disposable) {
            Rectangle pieceBounds = ((BoardPiece<?>) disposable).getBounds();
            redraw(pieceBounds.x, pieceBounds.y, pieceBounds.width, pieceBounds.height, true);
        }
        
    }

    private class InternalDraggingPieceHandler implements IDraggingPieceHandler {
        
        private boolean called = false;
        
        private BoardSquare targetSquare;
        
        public InternalDraggingPieceHandler(BoardSquare targetSquare) {
            this.targetSquare = targetSquare;
        }

        @Override
        public void moveToTargetSquare(int duration) {
            if(!called) {
                if(targetSquare == null) {
                    throw new IllegalStateException("Cannot move piece to target square because it was released outside the board area.");
                }
                
                moveTo(targetSquare, duration);
            }
        }
        
        @Override
        public void moveToTargetSquare(int duration, Interpolator interpolator) {
            if(!called) {
                if(targetSquare == null) {
                    throw new IllegalStateException("Cannot move piece to target square because it was released outside the board area.");
                }
                
                moveTo(targetSquare, duration, interpolator);
            }
        }

        @Override
        public void moveToTargetSquare(int duration, Runnable endMoveCallback) {
            if(!called) {
                if(targetSquare == null) {
                    throw new IllegalStateException("Cannot move piece to target square because it was released outside the board area.");
                }
                
                moveTo(targetSquare, duration, endMoveCallback);
            }
        }

        @Override
        public void moveToTargetSquare(int duration, MoveCallback moveCallback) {
            if(!called) {
                if(targetSquare == null) {
                    throw new IllegalStateException("Cannot move piece to target square because it was released outside the board area.");
                }
                
                moveTo(targetSquare, duration, moveCallback);
            }
        }
        
        @Override
        public void moveToTargetSquare(int duration, Interpolator interpolator, MoveCallback moveCallback) {
            if(!called) {
                if(targetSquare == null) {
                    throw new IllegalStateException("Cannot move piece to target square because it was released outside the board area.");
                }
                
                moveTo(targetSquare, duration, interpolator, moveCallback);
            }
        }

        @Override
        public void moveBackToOriginalSquare(int duration) {
            if(!called) {
                moveTo(draggingStartSquare, duration);
            }
        }

        @Override
        public void moveBackToOriginalSquare(int duration, Interpolator interpolator) {
            if(!called) {
                moveTo(draggingStartSquare, duration, interpolator);
            }
        }

        @Override
        public void moveBackToOriginalSquare(int duration, Runnable endMoveCallback) {
            if(!called) {
                moveTo(draggingStartSquare, duration, endMoveCallback);
            }
        }

        @Override
        public void moveBackToOriginalSquare(int duration, MoveCallback moveCallback) {
            if(!called) {
                moveTo(draggingStartSquare, duration, moveCallback);
            }
        }

        @Override
        public void moveBackToOriginalSquare(int duration, Interpolator interpolator, MoveCallback moveCallback) {
            if(!called) {
                moveTo(draggingStartSquare, duration, interpolator, moveCallback);
            }
        }

        @Override
        public void moveTo(BoardSquare square, int duration) {
            if(!called) {
                moveTo(square, duration, null, null);
            }
        }

        @Override
        public void moveTo(BoardSquare square, int duration, Interpolator interpolator) {
            if(!called) {
                moveTo(square, duration, interpolator, null);
            }
        }
    
        @Override
        public void moveTo(BoardSquare square, int duration, Runnable endMoveCallback) {
            if(!called) {
                moveTo(square, duration, null, new MoveCallback(endMoveCallback));
            }
        }

        @Override
        public void moveTo(BoardSquare square, int duration, MoveCallback moveCallback) {
            if(!called) {
                moveTo(square, duration, null, moveCallback);
            }
        }

        @Override
        public void moveTo(BoardSquare square, int duration, Interpolator interpolator,
                MoveCallback moveCallback) {
            if(!called) {
                draggingPiece.removeDisposeListener(draggingPieceDisposeListener);
                
                MoveCallback finishMoveCallback = new MoveCallback() {

                    @Override
                    public void movableDisposed(Mover mover, IDisposable disposable) {
                        clearDraggingStatus();
                    }
                    
                    @Override
                    public void ended(Mover mover) {
                        clearDraggingStatus();
                    }

                };
                MoveCallback usedMoveCallback = moveCallback == null ?
                        finishMoveCallback :
                        new CompositeMoveCallback(finishMoveCallback, moveCallback);
                movePiece(draggingPiece, square, duration, interpolator, usedMoveCallback);
                
                called = true;
            }
        }

        @Override
        public void fadeOut(int duration) {
            if(!called) {
                fadeOut(duration, new FadeCallback());
            }
        }

        @Override
        public void fadeOut(int duration, Runnable endFadeCallback) {
            if(!called) {
                fadeOut(duration, new FadeCallback(endFadeCallback));
            }
        }

        @Override
        public void fadeOut(int duration, FadeCallback fadeCallback) {
            if(!called) {
                draggingPiece.removeDisposeListener(draggingPieceDisposeListener);
                
                FadeCallback clearDraggingStatusCallback = new FadeCallback() {

                    @Override
                    public void fadableDisposed(IDisposable disposable) {
                        clearDraggingStatus();
                    }
                    
                    @Override
                    public void ended() {
                        clearDraggingStatus();
                    }

                };
                FadeCallback usedFadeCallback = fadeCallback == null ?
                        clearDraggingStatusCallback :
                        new CompositeFadeCallback(clearDraggingStatusCallback, fadeCallback);
                
                Fader fader = new Fader(getDisplay(), draggingPiece, 0.0F, duration, usedFadeCallback);
                fader.setAcceleration(0.3F);
                fader.setDeceleration(0.2F);
                fader.fade();
                
                called = true;
            }
        }

    }

}
