/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.graphics;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;

import org.eclipse.swt.graphics.RGB;

public class HighlightRenderer {

    private Highlight highlight;

    public HighlightRenderer(Highlight highlight) {
        if(highlight == null) {
            throw new IllegalArgumentException("Highlight cannot be null.");
        }
        
        this.highlight = highlight;
    }

    public Highlight getHighlight() {
        return highlight;
    }
    
    public BufferedImage getHighlightedImage(BufferedImage image) {
        BufferedImage highlightBackground = getHighlightBackgroundImage(image);
        return HighlightRenderer.getHighlightedImage(image, highlightBackground, highlight.getIntensity());
    }
    
    public static BufferedImage getHighlightedImage(BufferedImage image, BufferedImage highlightBackground,
            float highlightIntensity) {
        if(image.getWidth() != highlightBackground.getWidth()) {
            throw new IllegalArgumentException("The image to be highlighted and the highlight background must have the same widths.");
        }
        if(image.getHeight() != highlightBackground.getHeight()) {
            throw new IllegalArgumentException("The image to be highlighted and the highlight background must have the same heights.");
        }
        if((highlightIntensity < 0.0F) || (highlightIntensity > 1.0F)) {
            throw new IllegalArgumentException("The highlight intensity must be a number within 0 and 1.");
        }

        BufferedImage highlightedImage = new BufferedImage(image.getWidth(), image.getHeight(), image.getType());
        Graphics2D graphics = highlightedImage.createGraphics();
        graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        graphics.drawImage(image, 0, 0, null);
        graphics.setComposite(AlphaComposite.SrcOver.derive(highlightIntensity));
        graphics.drawImage(highlightBackground, 0, 0, null);
        graphics.dispose();
        
        return highlightedImage;
    }
    
    public BufferedImage getHighlightBackgroundImage(BufferedImage image) {
        int width = image.getWidth();
        int height = image.getHeight();
        int glowThickness = (int) ((width + height) / 2 * highlight.getGlowFracThickness());
        
        // creates a slightly larger image in order to account for border effects in the blur operation
        int newWidth = width + 2 * glowThickness;
        int newHeight = height + 2 * glowThickness;
        BufferedImage scratchImage = new BufferedImage(newWidth, newHeight, image.getType());
        Graphics2D graphics = scratchImage.createGraphics();
        graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        graphics.drawImage(image, glowThickness, glowThickness, null);
        graphics.setComposite(AlphaComposite.SrcIn);
        RGB highlightColor = highlight.getColor();
        graphics.setColor(new Color(highlightColor.red, highlightColor.green, highlightColor.blue));
        graphics.fillRect(0, 0, newWidth, newHeight);
        graphics.dispose();

        if(glowThickness > 0) {
            BufferedImage blurredImage = new BlurRenderer(glowThickness).getBlurredImage(scratchImage);
            
            // copies the blur image to an image of the same size as the original image
            BufferedImage finalImage = new BufferedImage(width, height, image.getType());
            graphics = finalImage.createGraphics();
            graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            graphics.drawImage(blurredImage, 0, 0, width - 1, height - 1, glowThickness, glowThickness,
                    newWidth - glowThickness - 1, newHeight - glowThickness - 1, null);
            graphics.drawImage(scratchImage, 0, 0, width - 1, height - 1, glowThickness, glowThickness,
                    newWidth - glowThickness - 1, newHeight - glowThickness - 1, null);
            graphics.dispose();
            
            return finalImage;
        } else {
            return scratchImage;
        }
    }
    
}
