/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.graphics;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;

public class ShadowRenderer {

    private float shadowIntensity;

    public ShadowRenderer(float shadowIntensity) {
        if((shadowIntensity < 0.0F) || (shadowIntensity > 1.0F)) {
            throw new IllegalArgumentException("Shadow intensity must be a number within 0 and 1.");
        }

        this.shadowIntensity = shadowIntensity;
    }

    public float getShadowIntensity() {
        return shadowIntensity;
    }
    
    public BufferedImage getShadowedImage(BufferedImage image) {
        if(shadowIntensity != 0.0F) {
            // we trim out the lower transparent area of the image in order to closely match the shadow
            int width = image.getWidth();
            int newHeight = ShadowRenderer.getLastNonTransparentRow(image) + 1;
            BufferedImage shadowedImage = new BufferedImage(width, 2 * newHeight, image.getType());
            Graphics2D graphics = shadowedImage.createGraphics();
            graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

            // draws the main image
            graphics.drawImage(image, 0, 0, width - 1, newHeight - 1, 0, 0, width - 1, newHeight - 1, null);
            
            // draws the shadow
            graphics.scale(1.0, -1.0);
            int startY = -2 * newHeight + 1;
            int endY = -newHeight;
            graphics.drawImage(image, 0, startY, width - 1, endY, 0, 0, width - 1, newHeight - 1, null);
            graphics.setComposite(AlphaComposite.SrcIn.derive(shadowIntensity));
            graphics.setColor(Color.BLACK);
            graphics.fillRect(0, startY, width, newHeight);
            
            graphics.dispose();
            
            return shadowedImage;
        } else {
            return image;
        }
    }

    private static int getLastNonTransparentRow(BufferedImage image) {
        int width = image.getWidth();
        int height = image.getHeight();
        int[] rgbArray = new int[width];
        int lastNonTransparentRow = height - 1;
        outer : while(lastNonTransparentRow >=0) {
            image.getRGB(0, lastNonTransparentRow, width, 1, rgbArray, 0, width);
            for(int column = 0; column < width; column++) {
                int alpha = rgbArray[column] >> 24;
                if(alpha != 0) {
                    break outer;
                }
            }
            lastNonTransparentRow--;
        }
        return lastNonTransparentRow;
    }
    
}
