/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.swt;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;

// This class performs custom painting of table items in order
// to support drawing centered text along a whole row.
//
// It also avoids some limitations of native table painting on the Windows platform:
// 1. On Windows the row heights are determined from the table's font only.
//    So if a table row is set a different font than its parent table then it
//      will not have its height correctly calculated.
// 2. On Windows the first column will always be aligned to the left, even
//    if it was created with the SWT.CENTER or SWT.RIGHT style bits.
public class CustomTableItemDrawingListener implements Listener {

    private static final int MARGIN_WIDTH = 6;
    private static final String CENTER_TEXT = "centerText";
    
    private Table table;

    public CustomTableItemDrawingListener(Table table) {
        this.table = table;
        
        this.table.addListener(SWT.MeasureItem, this);
        this.table.addListener(SWT.EraseItem, this);
        this.table.addListener(SWT.PaintItem, this);
    }
    
    public String getCenterText(TableItem tableItem) {
        return (String) tableItem.getData(CENTER_TEXT);
    }
    
    public void setCenterText(TableItem tableItem, String text) {
        tableItem.setData(CENTER_TEXT, text);
    }

    @Override
    public void handleEvent(Event event) {
        switch(event.type) {
        case SWT.MeasureItem :
            handleMeasureItem(event);
            break;
        case SWT.EraseItem :
            handleEraseItem(event);
            break;
        case SWT.PaintItem :
            handlePaintItem(event);
            break;
        }
    }

    private boolean drawCenterText(Event event) {
        TableItem tableItem = (TableItem) event.item;
        return getCenterText(tableItem) != null;
    }

    private void handleMeasureItem(Event event) {
        if(SwtUtils.isWindows()) {
            String text = ((TableItem) event.item).getText(event.index);
            GC gc = event.gc;
            gc.setTextAntialias(SWT.ON);
            Point textSize = gc.textExtent(text, SWT.DRAW_TRANSPARENT);
            event.width = textSize.x + MARGIN_WIDTH;
            event.height = gc.getFontMetrics().getHeight() + 6;
        }
    }

    private void handleEraseItem(Event event) {
        if(drawCenterText(event) || SwtUtils.isWindows()) {
            event.detail &= ~SWT.FOREGROUND;
        }
    }

    private void handlePaintItem(Event event) {
        if(drawCenterText(event)) {
            GC gc = event.gc;
            gc.setTextAntialias(SWT.ON);

            String text = getCenterText((TableItem) event.item);
            Point textSize = gc.textExtent(text, SWT.DRAW_TRANSPARENT);

            int totalColumnsWidth = 0;
            for(TableColumn column : table.getColumns()) {
                totalColumnsWidth += column.getWidth();
            }
            int textOffset = (totalColumnsWidth - textSize.x) / 2;
            for(int i = 0; i < event.index; i++) {
                textOffset -= table.getColumn(i).getWidth();
            }
            int x = event.x + textOffset;
            
            int y = event.y + (event.height - textSize.y) / 2;
            
            gc.drawString(text, x, y, true);
        } else {
            if(SwtUtils.isWindows()) {
                GC gc = event.gc;
                gc.setTextAntialias(SWT.ON);
                
                String text = ((TableItem) event.item).getText(event.index);
                Point textSize = gc.textExtent(text, SWT.DRAW_TRANSPARENT);
                
                int x;
                int columnStyle = table.getColumn(event.index).getStyle();
                int width = table.getColumn(event.index).getWidth();
                if((columnStyle & SWT.CENTER) == SWT.CENTER) {
                    x = event.x + (width - textSize.x) / 2;
                } else if((columnStyle & SWT.RIGHT) == SWT.RIGHT) {
                    x = event.x + width - textSize.x - MARGIN_WIDTH;
                } else {
                    x = event.x;
                }
                
                int y = event.y + (event.height - textSize.y) / 2;
                
                gc.drawString(text, x, y, true);
            }
        }
    }

}
