/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.swt;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;

/*
 * This table is a wrapper for Table objects in order
 * to extend them with some additional functionality:
 * 
 * 1. Implementing auto-sizable columns, which grab the extra table space as it resizes.
 * 2. Permitting selection of a single cell.
 * 3. Allowing to draw text along a whole row.
 * 
 * Functionality 3 is broken on recent Linux distributions that ship with GTK 2.18 (e.g., Ubuntu 9.10),
 * as the table widgets will always draw grid lines separating the table columns.
 * Unfortunately, the SWT custom table drawing mechanism does not allow overriding this behavior:
 * https://bugs.eclipse.org/bugs/show_bug.cgi?id=299405 .
 */
public class TableDecorator {

    private Color cellBackground;
    private Color cellForeground;
    private Color selectedCellBackground;
    private Color selectedCellForeground;
    
    private Table table;
    private CustomTableItemDrawingListener customTableItemDrawingListener;
    
    private List<TableColumnDecorator> columnDecorators = new ArrayList<TableColumnDecorator>();
    
    private TableCell selectedCell = null;
    
    private boolean hasRepackedAfterScrollbarAppeared = false;
    
    private List<ITableDecoratorListener> listeners = new ArrayList<ITableDecoratorListener>();
    
    public TableDecorator(Composite parent, int style) {
        createTable(parent, style | SWT.H_SCROLL | SWT.V_SCROLL);
    }

    public TableDecorator(Composite parent, boolean allowCellSelection) {
        createTable(parent, SWT.SINGLE | SWT.FULL_SELECTION | SWT.H_SCROLL | SWT.V_SCROLL);
        
        if(allowCellSelection) {
            Display display = table.getDisplay();
            cellBackground = display.getSystemColor(SWT.COLOR_LIST_BACKGROUND);
            cellForeground = display.getSystemColor(SWT.COLOR_LIST_FOREGROUND);
            selectedCellBackground = display.getSystemColor(SWT.COLOR_LIST_SELECTION);
            selectedCellForeground = display.getSystemColor(SWT.COLOR_LIST_SELECTION_TEXT);
            
            table.addMouseListener(new MouseAdapter() {

                @Override
                public void mouseDown(MouseEvent e) {
                    onMouseDown(e);
                }

                @Override
                public void mouseDoubleClick(MouseEvent e) {
                    table.deselectAll();
                }
                
            });
            
            // We don't allow the table to receive focus in order to avoid drawing
            // the focus rectangle around the selected item.
            // A more elegant solution would be to add another listener to the SWT.EraseItem
            // here (see the createTable method below) and disable the SWT.SELECTED bit, but
            // that doesn't work due to this bug:
            // https://bugs.eclipse.org/bugs/show_bug.cgi?id=271032
            // Once this bug is corrected we may follow this approach and re-allow the table to
            // receive keyboard focus, also removing the addKeyListener method below.
            table.addFocusListener(new FocusAdapter() {
                @Override
                public void focusGained(FocusEvent e) {
                    table.getParent().forceFocus();
                }
            });
        }
    }

    public Table getTable() {
        return table;
    }
    
    public void setLayoutData(Object layoutData) {
        table.getParent().setLayoutData(layoutData);
    }
    
    public void addKeyListener(KeyListener keyListener) {
        table.getParent().addKeyListener(keyListener);
    }
    
    public TableCell getSelectedCell() {
        return selectedCell;
    }

    public void setSelectedCell(TableCell cell) {
        if(selectedCell != null) {
            TableItem selectedItem = table.getItem(selectedCell.getRow());
            selectedItem.setBackground(selectedCell.getColumn(), cellBackground);
            selectedItem.setForeground(selectedCell.getColumn(), cellForeground);
        }

        if(cell != null) {
            TableItem clickedItem = table.getItem(cell.getRow());
            clickedItem.setBackground(cell.getColumn(), selectedCellBackground);
            clickedItem.setForeground(cell.getColumn(), selectedCellForeground);
        }
        
        selectedCell = cell;
    }
    
    public void removeAll() {
        table.removeAll();
        selectedCell = null;
        hasRepackedAfterScrollbarAppeared = false;
    }

    public void pack() {
        Table table = getTable();
        
        Rectangle availableArea = table.getParent().getClientArea();
        int vBarWidth = table.getVerticalBar().getSize().x;
        int trimmingsWidth = table.computeTrim(0, 0, 0, 0).width - vBarWidth;
        int columnsWidth = availableArea.width - trimmingsWidth;

        // to be on the safe side, we assume an uncertainty of itemHeight in the calculation of preferredHeight
        if(availableArea.height < getPreferredHeight() + table.getItemHeight()) {
            // subtract the scrollbar width from the total column
            // width if a vertical scrollbar will be required
            columnsWidth -= vBarWidth;
        }
        
        table.setRedraw(false);
        int nFixedColumns = 0;
        int fixedColumnsWidth = 0;
        for(TableColumnDecorator columnDecorator : columnDecorators) {
            if(!columnDecorator.isAutoExpandeable()) {
                nFixedColumns++;
                TableColumn tableColumn = columnDecorator.getTableColumn();
                tableColumn.pack();
                fixedColumnsWidth += tableColumn.getWidth();
            }
        }
        if(fixedColumnsWidth < columnsWidth) {
            int remainingColumnsWidth = columnsWidth - fixedColumnsWidth;
            int expandeableColumnWidth = remainingColumnsWidth / (columnDecorators.size() - nFixedColumns);
            for(TableColumnDecorator columnDecorator : columnDecorators) {
                if(columnDecorator.isAutoExpandeable()) {
                    columnDecorator.getTableColumn().setWidth(expandeableColumnWidth);
                }
            }
        }  // TODO handle the case when fixedColumnsWidth >= columnsWidth
        table.setRedraw(true);
        
        // to be on the safe side, we assume an uncertainty of itemHeight in the calculation of preferredHeight
        hasRepackedAfterScrollbarAppeared = getPreferredHeight() - table.getItemHeight() > availableArea.height;
    }
    
    public void repackIfScrollbarAppeared() {
        if(!hasRepackedAfterScrollbarAppeared) {
            Rectangle availableArea = table.getParent().getClientArea();
            // to be on the safe side, we assume an uncertainty of itemHeight in the calculation of preferredHeight
            if(getPreferredHeight() + table.getItemHeight() > availableArea.height) {
                pack();
            }
        }
    }

    public String getCenterText(TableItem tableItem) {
        return customTableItemDrawingListener.getCenterText(tableItem);
    }
    
    public void setCenterText(TableItem tableItem, String text) {
        customTableItemDrawingListener.setCenterText(tableItem, text);
    }

    public void addListener(ITableDecoratorListener listener) {
        synchronized(listeners) {
            listeners.add(listener);
        }
    }
    
    public void removeListener(ITableDecoratorListener listener) {
        synchronized(listeners) {
            listeners.remove(listener);
        }
    }
    
    void addColumnDecorator(TableColumnDecorator columnDecorator) {
        columnDecorators.add(columnDecorator);
    }
    
    private void createTable(Composite parent, int style) {
        Composite intermediate = new Composite(parent, SWT.NONE);
        intermediate.setLayout(new FillLayout());
        table = new Table(intermediate, style);
        
        customTableItemDrawingListener = new CustomTableItemDrawingListener(table);
        
        table.addControlListener(new ControlAdapter() {
            @Override
            public void controlResized(ControlEvent e) {
                pack();
            }
        });
    }
    
    private int getPreferredHeight() {
        // This seems to be more accurate than table.computeSize(SWT.DEFAULT, SWT.DEFAULT).y,
        // which yields values a little larger than necessary.
        return table.getHeaderHeight() + table.getItemCount() * table.getItemHeight() +
            table.computeTrim(0, 0, 0, 0).height;
    }

    private void notifyCellSelectionChanged(TableCell previousSelectedCell, TableCell currentSelectedCell) {
        synchronized(listeners) {
            for(ITableDecoratorListener listener : listeners) {
                listener.cellSelectionChanged(previousSelectedCell, currentSelectedCell);
            }
        }
    }
    
    private TableCell getClickedCell(Point clickLocation) {
        TableItem clickedItem = table.getItem(clickLocation);
        if(clickedItem == null) {
            return null;
        }
        
        for(int column = 0; column < table.getColumnCount(); column++) {
            if(clickedItem.getBounds(column).contains(clickLocation)) {
                return new TableCell(table.indexOf(clickedItem), column);
            }
        }
        
        return null;
    }

    // This method gets executed just if the table decorator supports cell selection (see constructor).
    private void onMouseDown(MouseEvent e) {
        table.deselectAll();
        
        TableCell clickedCell = getClickedCell(new Point(e.x, e.y));
        if((clickedCell == null) || (clickedCell.equals(selectedCell))) {
            return;
        }
        
        TableCell previousSelectedCell = selectedCell;
        setSelectedCell(clickedCell);
        notifyCellSelectionChanged(previousSelectedCell, selectedCell);
    }

}
