/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.utils;

import java.awt.AlphaComposite;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Transparency;
import java.awt.image.BufferedImage;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.PaletteData;
import org.eclipse.swt.graphics.RGB;

public class ImageUtils {

    // The returned image will always have a direct color model,
    // even if the AWT image uses an indexed palette.
    public static ImageData convertToSwt(BufferedImage awtImage) {
        // retrieves the individual pixels from the AWT image
        int width = awtImage.getWidth();
        int height = awtImage.getHeight();
        int[] rgbArray = new int[width * height];
        int imageType = awtImage.getType();
        if((imageType == BufferedImage.TYPE_INT_RGB) || (imageType == BufferedImage.TYPE_INT_ARGB)) {
            // no color conversions when the image raster already stores the data as RGB integers
            awtImage.getRaster().getDataElements(0, 0, width, height, rgbArray);
        } else {
            awtImage.getRGB(0, 0, width, height, rgbArray, 0, width);
        }

        // copies them into the individual pixels of the SWT image
        ImageData imageData = new ImageData(width, height, 24, new PaletteData(0xFF0000, 0xFF00, 0xFF));
        byte[] data = imageData.data;
        int nWastedBytes = imageData.bytesPerLine - 3 * width;
        int pixelIndex = 0;
        int dataIndex = 0;
        for(int row = 0; row < height; row++) {
            for(int column = 0; column < width; column++) {
                int rgb = rgbArray[pixelIndex++];
                data[dataIndex] = (byte) ((rgb >> 16) & 0xFF);
                data[dataIndex + 1] = (byte) ((rgb >> 8) & 0xFF);
                data[dataIndex + 2] = (byte) (rgb & 0xFF);
                dataIndex += 3;
            }
            dataIndex += nWastedBytes;
        }
        if((awtImage.getTransparency() != Transparency.OPAQUE)) {
            byte[] alphaData = new byte[width * height];
            imageData.alphaData = alphaData;
            for(int i = 0; i < alphaData.length; i++) {
                alphaData[i] = (byte) ((rgbArray[i] >> 24) & 0xFF);
            }
        }

        return imageData;
    }

    // The returned image will always have a direct color model,
    // even if the SWT image uses an indexed palette.
    public static BufferedImage convertToAwt(ImageData swtImage) {
        int width = swtImage.width;
        int height = swtImage.height;
        boolean isOpaque = swtImage.getTransparencyType() == SWT.TRANSPARENCY_NONE;
        int imageType = isOpaque ? BufferedImage.TYPE_INT_RGB : BufferedImage.TYPE_INT_ARGB;
        BufferedImage awtImage = new BufferedImage(width, height, imageType);
        
        int[] pixels = new int[width * height];
        swtImage.getPixels(0, 0, pixels.length, pixels, 0);
        int[] rgbArray = new int[width * height];
        PaletteData palette = swtImage.palette;
        for(int i = 0; i < rgbArray.length; i++) {
            RGB rgb = palette.getRGB(pixels[i]);
            rgbArray[i] = (rgb.red << 16) | (rgb.green << 8) | rgb.blue;
            if(!isOpaque) {
                int alpha;
                if(swtImage.getTransparencyType() == SWT.TRANSPARENCY_ALPHA) {
                    alpha = swtImage.alpha != -1 ? swtImage.alpha : swtImage.alphaData[i] & 0xFF;
                } else if(swtImage.getTransparencyType() == SWT.TRANSPARENCY_PIXEL) {
                    alpha = pixels[i] == swtImage.transparentPixel ? 0 : 255;
                } else {
                    // The only transparency type that isn't handled for the moment is SWT.TRANSPARENCY_MASK.
                    alpha = 255;
                }
                rgbArray[i] |= alpha << 24;
            }
        }
        awtImage.setRGB(0, 0, width, height, rgbArray, 0, width);
        
        return awtImage;
    }

    // Note: the returned image will have an integer representation
    //       (either BufferedImage.TYPE_INT_RGB or BufferedImage.TYPE_INT_ARGB).
    public static BufferedImage fastAndGoodScale(BufferedImage originalImage, int targetWidth, int targetHeight) {
        int width = originalImage.getWidth();
        int height = originalImage.getHeight();
        int newWidth = targetWidth >= width ?
                        // performs the upscaling in a single step
                        targetWidth :
                        // performs the first progressive downscaling step
                        Math.max(width / 2, targetWidth);
        int newHeight = targetHeight >= height ?
                        // performs the upscaling in a single step
                        targetHeight :
                        // performs the first progressive downscaling step
                        Math.max(height / 2, targetHeight);
        int imageType = (originalImage.getTransparency() == Transparency.OPAQUE) ?
                BufferedImage.TYPE_INT_RGB : BufferedImage.TYPE_INT_ARGB;
        BufferedImage scratchImage = new BufferedImage(newWidth, newHeight, imageType);
        Graphics2D graphics = scratchImage.createGraphics();
        graphics.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
        graphics.setComposite(AlphaComposite.Src);
        graphics.drawImage(originalImage, 0, 0, newWidth, newHeight, null);
        width = newWidth;
        height = newHeight;

        // progressively downscales the image
        while((width > targetWidth) || (height > targetHeight)) {
            newWidth = Math.max(width / 2, targetWidth);
            newHeight = Math.max(height / 2, targetHeight);
            graphics.drawImage(scratchImage, 0, 0, newWidth, newHeight, 0, 0, width, height, null);
            width = newWidth;
            height = newHeight;
        }
        graphics.dispose();

        // if the final image has a different dimension than requested, returns
        // a new image with the correct size
        BufferedImage resultImage = scratchImage;
        if((resultImage.getWidth() != targetWidth) || (resultImage.getHeight() != targetHeight)) {
            resultImage = new BufferedImage(targetWidth, targetHeight, imageType);
            Graphics2D resultGraphics = resultImage.createGraphics();
            resultGraphics.drawImage(scratchImage, 0, 0, targetWidth, targetHeight, 0, 0, targetWidth, targetHeight, null);
            resultGraphics.dispose();
        }
        return resultImage;
    }

}
