/*
Copyright (C) 2009, 2010 Alex Ignácio da Silva

This file is part of Haundrix Games.

Haundrix Games is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License version 3 as published by the
Free Software Foundation, with additional permission under section 7 to combine
it with independent works licensed under selected Free/Libre Open Source Software licenses.
Please see the full text of the license that is distributed along with this file.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.
*/
package com.haundrix.utils;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.BooleanControl;
import javax.sound.sampled.FloatControl;
import javax.sound.sampled.SourceDataLine;

import org.apache.commons.io.IOUtils;

public class SoundPlayer {

    private static final int THREAD_POOL_SIZE = 6;

    private static final int EXTERNAL_BUFFER_SIZE = 128000;
    
    // TODO Allow configuring the sound volume.
    private static final float VOLUME = 0.7F;  // must be between 0 and 1
    
    private boolean isMuted = false;
    
    private IExceptionHandler exceptionHandler;
    
    private ScheduledExecutorService scheduler;
    
    private List<SoundLooper> soundLoopers = new ArrayList<SoundLooper>();
    
    public SoundPlayer(IExceptionHandler exceptionHandler) {
        this.exceptionHandler = exceptionHandler;
        
        scheduler = Executors.newScheduledThreadPool(THREAD_POOL_SIZE);
    }
    
    public boolean isMuted() {
        return isMuted;
    }

    public void setMuted(boolean isMuted) {
        this.isMuted = isMuted;
        
        synchronized(soundLoopers) {
            for(SoundLooper soundLooper : soundLoopers) {
                soundLooper.setMuted(this.isMuted);
            }
        }
    }

    public void play(final File soundFile) {
        if(!isMuted) {
            scheduler.schedule(new Runnable() {
                @Override
                public void run() {
                    doPlay(soundFile);
                }
            }, 0L, TimeUnit.MILLISECONDS);
        }
    }
    
    public void play(SoundFile soundFile) {
        if(!isMuted) {
            playAtBegin(soundFile, 0L);
        }
    }
    
    public void playAtBegin(final SoundFile soundFile, long beginInterval) {
        if(!isMuted) {
            scheduler.schedule(new Runnable() {
                @Override
                public void run() {
                    doPlay(soundFile.getFile());
                }
            }, beginInterval - soundFile.getStartSilence(), TimeUnit.MILLISECONDS);
        }
    }
    
    public void playAtFinish(final SoundFile soundFile, long finishInterval) {
        if(!isMuted) {
            scheduler.schedule(new Runnable() {
                @Override
                public void run() {
                    doPlay(soundFile.getFile());
                }
            }, finishInterval - soundFile.getDuration() + soundFile.getEndSilence(), TimeUnit.MILLISECONDS);
        }
    }
    
    public ISoundLoopHandle loop(File soundFile) {
        final SoundLooper soundLooper = new SoundLooper(soundFile);
        soundLooper.setMuted(isMuted);
        synchronized(soundLoopers) {
            soundLoopers.add(soundLooper);
        }
        
        scheduler.schedule(new Runnable() {
            @Override
            public void run() {
                soundLooper.loop();
            }
        }, 0L, TimeUnit.MILLISECONDS);
        
        return soundLooper;
    }
    
    public ISoundLoopHandle loop(SoundFile soundFile) {
        return loop(soundFile.getFile());
    }
    
    public void dispose() {
        scheduler.shutdownNow();
        
        synchronized(soundLoopers) {
            List<ISoundLoopHandle> loopers = new ArrayList<ISoundLoopHandle>(soundLoopers);
            for(ISoundLoopHandle looper : loopers) {
                looper.cancel();
            }
        }
    }
    
    // This method is run in a background thread.
    private void doPlay(File soundFile) {
        AudioInputStream audioStream = null;
        AudioInputStream convertedStream = null;
        SourceDataLine line = null;
        try {
            audioStream = AudioSystem.getAudioInputStream(soundFile);
            AudioFormat format = audioStream.getFormat();
            AudioFormat newFormat = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, format.getSampleRate(),
                    16, format.getChannels(), format.getChannels() * 2, format.getSampleRate(), false);
            convertedStream = AudioSystem.getAudioInputStream(newFormat, audioStream);
            
            line = AudioSystem.getSourceDataLine(newFormat);
            line.open(newFormat);
            line.start();

            SoundPlayer.setDefaultLineVolume(line);
            
            byte[] data = new byte[EXTERNAL_BUFFER_SIZE];
            int bytesRead;
            while((bytesRead = convertedStream.read(data)) != -1) {
                line.write(data, 0, bytesRead);
            }
            
            line.drain();
        } catch(Exception e) {
            exceptionHandler.handle(e);
        } finally {
            if(line != null) {
                if(line.isRunning()) {
                    line.stop();
                }
                if(line.isOpen()) {
                    line.close();
                }
            }
            IOUtils.closeQuietly(convertedStream);
            IOUtils.closeQuietly(audioStream);
        }
    }
    
    private static void setDefaultLineVolume(SourceDataLine line) {
        if(line.isControlSupported(FloatControl.Type.VOLUME)) {
            FloatControl volumeControl = (FloatControl) line.getControl(FloatControl.Type.VOLUME);
            float minVolume = volumeControl.getMinimum();
            float maxVolume = volumeControl.getMaximum();
            float volume = minVolume + (maxVolume - minVolume) * VOLUME;
            volumeControl.setValue(volume);
        }
    }

    private class SoundLooper implements ISoundLoopHandle {

        private File soundFile;
        private AudioFormat newFormat;
        
        private SourceDataLine line;
        
        private volatile boolean isCanceled = false;
        
        public SoundLooper(File soundFile) {
            this.soundFile = soundFile;
            
            try {
                AudioFormat format = AudioSystem.getAudioFileFormat(soundFile).getFormat();
                newFormat = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, format.getSampleRate(),
                        16, format.getChannels(), format.getChannels() * 2, format.getSampleRate(), false);
                line = AudioSystem.getSourceDataLine(newFormat);
                line.open(newFormat);
                line.start();

                SoundPlayer.setDefaultLineVolume(line);
            } catch(Exception e) {
                cancel();
                exceptionHandler.handle(e);
            }
        }
        
        public void setMuted(boolean isMuted) {
            if(line.isControlSupported(BooleanControl.Type.MUTE)) {
                BooleanControl muteControl = (BooleanControl) line.getControl(BooleanControl.Type.MUTE);
                muteControl.setValue(isMuted);
            } else if(line.isControlSupported(FloatControl.Type.VOLUME)) {
                if(isMuted) {
                    FloatControl volumeControl = (FloatControl) line.getControl(FloatControl.Type.VOLUME);
                    volumeControl.setValue(0.0F);
                } else {
                    SoundPlayer.setDefaultLineVolume(line);
                }
            } else {
                // without support for mute or volume controls there isn't much we can do...
            }
        }

        public void loop() {
            AudioInputStream audioStream = null;
            AudioInputStream convertedStream = null;
            try {
                while(!isCanceled) {
                    audioStream = AudioSystem.getAudioInputStream(soundFile);
                    convertedStream = AudioSystem.getAudioInputStream(newFormat, audioStream);
                    
                    byte[] data = new byte[EXTERNAL_BUFFER_SIZE];
                    int bytesRead;
                    while(((bytesRead = convertedStream.read(data)) != -1) && !isCanceled) {
                        line.write(data, 0, bytesRead);
                    }
                    
                    convertedStream.close();
                    convertedStream = null;
                    audioStream.close();
                    audioStream = null;
                }
            } catch(Exception e) {
                IOUtils.closeQuietly(convertedStream);
                IOUtils.closeQuietly(audioStream);
                cancel();
                exceptionHandler.handle(e);
            }
        }

        @Override
        public synchronized void cancel() {
            if(!isCanceled) {
                isCanceled = true;
                if(line != null) {
                    if(line.isRunning()) {
                        line.stop();
                    }
                    if(line.isOpen()) {
                        line.close();
                    }
                }
                
                synchronized(soundLoopers) {
                    soundLoopers.remove(this);
                }
            }
        }
        
    }
    
}
